Package: crlmm
Type: Package
Title: Genotype Calling (CRLMM) and Copy Number Analysis tool for
        Affymetrix SNP 5.0 and 6.0 and Illumina arrays
Version: 1.70.0
Author: Benilton S Carvalho, Robert Scharpf, Matt Ritchie, Ingo
        Ruczinski, Rafael A Irizarry
Maintainer: Benilton S Carvalho <benilton@unicamp.br>,
 Robert Scharpf <rscharpf@jhsph.edu>, Matt Ritchie
 <mritchie@wehi.EDU.AU>
Description: Faster implementation of CRLMM specific to SNP 5.0 and 6.0
        arrays, as well as a copy number tool specific to 5.0, 6.0, and
        Illumina platforms.
License: Artistic-2.0
Depends: R (>= 2.14.0), oligoClasses (>= 1.21.12), preprocessCore (>=
        1.17.7)
LinkingTo: preprocessCore (>= 1.17.7)
Imports: methods, Biobase (>= 2.15.4), BiocGenerics, affyio (>=
        1.23.2), illuminaio, ellipse, mvtnorm, splines, stats, utils,
        lattice, ff, foreach, RcppEigen (>= 0.3.1.2.1), matrixStats,
        VGAM, parallel, graphics, limma, beanplot
Suggests: hapmapsnp6, genomewidesnp6Crlmm (>= 1.0.7), snpStats, RUnit
Collate: AllGenerics.R AllClasses.R methods-AssayData.R methods-CNSet.R
        methods-CNSetLM.R methods-eSet.R methods-SnpSuperSet.R
        methods-PredictionRegion.R cnrma-functions.R cnset-accessors.R
        crlmm-functions.R crlmmGT2.R crlmm-illumina.R krlmm.R plot.R
        snprma-functions.R utils.R zzz.R test_crlmm_package.R
LazyLoad: yes
biocViews: Microarray, Preprocessing, SNP, CopyNumberVariation
RoxygenNote: 7.1.1
git_url: https://git.bioconductor.org/packages/crlmm
git_branch: RELEASE_3_23
git_last_commit: 416d6f5
git_last_commit_date: 2026-04-28
Repository: Bioconductor 3.23
Date/Publication: 2026-04-28
NeedsCompilation: yes
Packaged: 2026-04-28 21:30:10 UTC; biocbuild

useDynLib("crlmm", .registration=TRUE)

importClassesFrom(Biobase, AssayData, eSet)

##importClassesFrom(methods, ANY, character, formula, integer, list,
##                  matrix, oldClass)
##importMethodsFrom(methods, initialize, show)
import(methods)
##importFrom(methods, setOldClass)
##importFrom(methods, as, callNextMethod, is, new, validObject)
##importFrom(methods, callNextMethod, new, validObject, coerce)

importFrom(RcppEigen, fastLmPure)
## importClassesFrom(oligoClasses, CNSet, CNSetLM, ff_matrix,
##                   ff_or_matrix, oligoSnpSet)
importClassesFrom(oligoClasses, CNSet, oligoSnpSet, ff_or_matrix, GenomeAnnotatedDataFrame)
##setOldClass(ff_or_matrix)
import(matrixStats)

importMethodsFrom(Biobase, annotatedDataFrameFrom, annotation,
                  AnnotatedDataFrame, assayData, experimentData, featureData,
                  "featureData<-", featureNames, "featureNames<-",
                  pData, "pData<-", phenoData, "phenoData<-",
                  protocolData, "protocolData<-",
                  sampleNames, snpCall, "snpCall<-",
                  snpCallProbability, "snpCallProbability<-",
                  storageMode, "storageMode<-", varLabels)

importMethodsFrom(BiocGenerics, cbind, colnames, Filter, get,
                  intersect, lapply, ncol, NCOL, nrow, NROW, order,
                  paste, pmax, pmin, rbind, rownames, sapply, setdiff,
                  table, union, unique)

##importMethodsFrom(genefilter, show)

importMethodsFrom(oligoClasses, A, "A<-", B, batch, batchNames,
                  batchStatistics, "batchStatistics<-", calls,
                  chromosome, close, confs, flags, isSnp, mean, nu,
                  open, phi, "sampleNames<-")

importFrom(affyio, read.celfile, read.celfile.header)

importFrom(illuminaio, readIDAT)

importFrom(Biobase, assayDataElement, assayDataElementReplace,
           assayDataNew, copyEnv)

importFrom(ellipse, ellipse)

##importFrom(genefilter, rowSds)

importFrom(lattice, lpolygon, panel.grid, panel.number, panel.xyplot,
           xyplot)

##importFrom(methods, as, callNextMethod, is, new, validObject)

importFrom(mvtnorm, rmvnorm)

importFrom(oligoClasses, celfileDate, chromosome2integer, i2p,
           initializeBigMatrix, initializeBigVector, integerMatrix,
           isPackageLoaded, ldPath, ocLapply, ocProbesets, ocSamples,
           parStatus, splitIndicesByLength, splitIndicesByNode, 
           AssayDataList, genomeBuild, celfileName)

importFrom(preprocessCore, normalize.quantiles, normalize.quantiles.determine.target,
           normalize.quantiles.use.target, subColSummarizeMedian)

importFrom(stats, coef, cov, dnorm, kmeans, lm, mad, median, quantile,
           sd)

importFrom(utils, packageDescription, setTxtProgressBar,
           txtProgressBar)

importFrom(graphics, plot, smoothScatter, abline)

importFrom(beanplot, beanplot)

## foreach
import(foreach)

importFrom(VGAM, vglm, multinomial, coefficients)

importFrom(parallel, makeCluster, detectCores, parRapply, stopCluster, clusterEvalQ, parCapply)

importFrom(limma, loessFit)

importFrom(ff, delete)

##----------------------------------------------------------------------------
## export
##----------------------------------------------------------------------------
exportClasses(PredictionRegion)
exportMethods(CA, CB,
	      A, B, corr, nuA, nuB, phiA, phiB,
	      predictionRegion, posteriorProbability,
	      tau2, Ns, medians, mads,
	      xyplot, calculateRBaf, OligoSetList,
	      BafLrrSetList)
export(crlmm,
       crlmmIllumina,
       constructAffyCNSet,
       genotype,
       genotypeAffy,
       readIdatFiles,
       readGenCallOutput,
       snprma,
       snprma2,
       cnrmaAffy,
       snprmaAffy,
       crlmm2,
       genotype2, genotypeLD,
       genotypeAffy,
       genotype.Illumina,
       plotSamples, plotSNPs,
       crlmmCopynumber2, crlmmCopynumberLD, crlmmCopynumber)
export(genotypes, totalCopynumber, rawCopynumber, xyplot)
export(ABpanel, validCEL, celDates, validCdfNames)

##export(constructOligoSetListFrom, constructBafLrrSetListFrom)

crlmmGT2 <- function(A, B, SNR, mixtureParams, cdfName, row.names=NULL,
                     col.names=NULL, probs=c(1/3, 1/3, 1/3), DF=6,
                     SNRMin=5, recallMin=10, recallRegMin=1000,
                     gender=NULL, desctrucitve=FALSE, verbose=TRUE,
                     returnParams=FALSE, badSNP=.7,
		     callsGt,
		     callsPr){
	pkgname <- getCrlmmAnnotationName(cdfName)
	stopifnot(require(pkgname, character.only=TRUE, quietly=!verbose))
	open(SNR)
	open(A)
	open(B)
	open(mixtureParams)
	## expect objects to be ff
	keepIndex <- which( SNR[] > SNRMin)
	if(length(keepIndex)==0) stop("No arrays above quality threshold!")
	loader("preprocStuff.rda", .crlmmPkgEnv, pkgname)
	gns <- getVarInEnv("gns", .crlmmPkgEnv)
	if(is.null(rownames(A))){
		stopifnot(nrow(A) == length(gns))
		index <- seq_len(nrow(A))
	} else {
		index <- match(gns, rownames(A))
	}
	snpBatches <- splitIndicesByLength(index, ocProbesets())
	NR <- length(unlist(snpBatches))
	if(verbose) message("Calling ", NR, " SNPs for recalibration... ")
	NC <- ncol(A)
	##
	if(verbose) message("Loading annotations.")
	obj1 <- loader("genotypeStuff.rda", .crlmmPkgEnv, pkgname)
	obj2 <- loader("mixtureStuff.rda", .crlmmPkgEnv, pkgname)
	##
	## this is toget rid of the 'no visible binding' notes
	## variable definitions
	XIndex <- getVarInEnv("XIndex")
	autosomeIndex <- getVarInEnv("autosomeIndex")
	YIndex <- getVarInEnv("YIndex")
	SMEDIAN <- getVarInEnv("SMEDIAN")
	theKnots <- getVarInEnv("theKnots")
	regionInfo <- getVarInEnv("regionInfo")
	params <- getVarInEnv("params")
	rm(list=c(obj1, obj2), envir=.crlmmPkgEnv)
	rm(obj1, obj2)
	## use lexical scope
	imputeGender <- function(XIndex, YIndex){
		if(length(YIndex) > 0){
			a <- log2(as.matrix(A[XIndex,,drop=FALSE]))
			b <- log2(as.matrix(B[XIndex,,drop=FALSE]))
			meds.X <- (apply(a+b, 2, median))/2
			## Y
			a <- log2(as.matrix(A[YIndex,,drop=FALSE]))
			b <- log2(as.matrix(B[YIndex,,drop=FALSE]))
			meds.Y <- (apply(a+b, 2, median))/2
			R <- meds.X - meds.Y
			if(sum(SNR[] > SNRMin) == 1){
				gender <- ifelse(R[SNR[] > SNRMin] > 0.5, 2L, 1L)
			} else{
				gender <- kmeans(R, c(min(R[SNR[]>SNRMin]), max(R[SNR[]>SNRMin])))[["cluster"]]
			}
		} else {
			XMedian <- apply(log2(as.matrix(A[XIndex,,drop=FALSE]))+log2(as.matrix(B[XIndex,, drop=FALSE])), 2, median)/2
			if(sum(SNR > SNRMin) == 1){
				gender <- which.min(c(abs(XMedian-8.9), abs(XMedian-9.5)))
			} else{
				gender <- kmeans(XMedian, c(min(XMedian[SNR[]>SNRMin]), max(XMedian[SNR[]>SNRMin])))[["cluster"]]
			}
		}
		return(gender)
	}
	if(is.null(gender)){
		if(ocProbesets() < length(XIndex)){
			if(verbose) message("Using ", ocProbesets(), " SNPs on chrom X and Y to assign gender.")
			XIndex2 <- sample(XIndex, ocProbesets(), replace=FALSE)
		} else XIndex2 <- XIndex
		if(ocProbesets() < length(YIndex)){
			YIndex2 <- sample(YIndex, ocProbesets(), replace=FALSE)
		} else YIndex2 <- YIndex
		message("Imputing gender")
		gender <- imputeGender(XIndex=XIndex2, YIndex=YIndex2)
		##cnSet$gender[,] <- gender
	}
	Indexes <- list(autosomeIndex, XIndex, YIndex)
	cIndexes <- list(keepIndex,
			 keepIndex[which(gender[keepIndex]==2)],
			 keepIndex[which(gender[keepIndex]==1)])
	## call C
	fIndex <- which(gender==2)
	mIndex <- which(gender==1)
	## different here
	## use gtypeCallerR in batches
	batchSize <- ocProbesets()
	open(A)
	open(B)
	open(mixtureParams)
	process1 <- function(idxBatch){
		snps <- snpBatches[[idxBatch]]
		##rSnps <- range(snps)
		last <- (idxBatch-1)*batchSize
		IndexesBatch <- list(autosomeIndex[autosomeIndex %in% snps]-last,
				     XIndex[XIndex %in% snps]-last,
				     YIndex[YIndex %in% snps]-last)
		IndexesBatch <- lapply(IndexesBatch, as.integer)
		tmpA <- as.matrix(A[snps,])
		tmpB <- as.matrix(B[snps,])
		tmp <- gtypeCallerR(tmpA, tmpB, fIndex, mIndex,
				    params[["centers"]][snps,],
				    params[["scales"]][snps,],
				    params[["N"]][snps,],
				    IndexesBatch, cIndexes,
				    sapply(IndexesBatch, length),
				    sapply(cIndexes, length), SMEDIAN,
				    theKnots, mixtureParams[], DF, probs, 0.025)
		tmp
	}
	## Lexical scope
	gc(verbose=FALSE)
	newparamsBatch <- ocLapply(seq(along=snpBatches), process1, neededPkgs="crlmm")
	gc(verbose=FALSE)
	if(verbose) message("finished process1")
	newparams <- vector("list", 3)
	names(newparams) <- c("centers", "scales", "N")
	newparams[["centers"]] <- do.call("rbind", lapply(newparamsBatch, "[[", 1))
	newparams[["scales"]] <- do.call("rbind", lapply(newparamsBatch, "[[", 2))
	newparams[["N"]] <- do.call("rbind", lapply(newparamsBatch, "[[", 3))
	rm(newparamsBatch); gc(verbose=FALSE)
	if(verbose) message("Done.")
	if(verbose) message("Estimating recalibration parameters.")
	d <- newparams[["centers"]] - params$centers
	##
	##regression
	Index <- intersect(which(pmin(newparams[["N"]][, 1],
				      newparams[["N"]][, 2],
				      newparams[["N"]][, 3]) > recallMin &
				 !apply(regionInfo, 1, any)),
			   autosomeIndex)
	if(length(Index) < recallRegMin){
		warning("Recalibration not possible. Possible cause: small sample size.")
		newparams <- params
		dev <- vector("numeric", nrow(newparams[["centers"]]))
		SS <- matrix(Inf, 3, 3)
		DD <- 0
	}else{
		data4reg <- as.data.frame(newparams[["centers"]][Index,])
		names(data4reg) <- c("AA", "AB", "BB")
		regParams <- cbind(  coef(lm(AA~AB*BB, data=data4reg)),
				   c(coef(lm(AB~AA+BB, data=data4reg)), 0),
				   coef(lm(BB~AA*AB, data=data4reg)))
		rownames(regParams) <- c("intercept", "X", "Y", "XY")
		rm(data4reg)
		##
		minN <- 3
		newparams[["centers"]][newparams[["N"]] < minN] <- NA
		Index <- setdiff(which(rowSums(is.na(newparams[["centers"]]))==1), YIndex)
		if(verbose) message("Filling out empty centers", appendLF=FALSE)
		for(i in Index){
			if(verbose) if(i%%10000==0) message(".", appendLF=FALSE)
			mu <- newparams[["centers"]][i, ]
			j <- which(is.na(mu))
			newparams[["centers"]][i, j] <- c(1, mu[-j], prod(mu[-j]))%*%regParams[, j]
			rm(mu, j)
		}
		##
		##remaing NAs are made like originals
		if(length(YIndex)>0){
			noMoveIndex <- union(setdiff(which(rowSums(is.na(newparams[["centers"]]))>0), YIndex),
					     YIndex[rowSums(is.na(newparams[["centers"]][YIndex, ])>1)])
		}
		snps2ignore <- which(rowSums(is.na(newparams[["centers"]])) > 0)
		snps2keep <- setdiff(autosomeIndex, snps2ignore)
		rm(snps2ignore)
		newparams[["centers"]][is.na(newparams[["centers"]])] <- params[["centers"]][is.na(newparams[["centers"]])]
		if(verbose) cat("\n")
		##
		if(verbose) message("Calculating and standardizing size of shift... ", appendLF=FALSE)
		GG <- DD <- newparams[["centers"]] - params[["centers"]]
		DD <- sweep(DD, 2, colMeans(DD[autosomeIndex, ]))
		SS <- cov(DD[autosomeIndex, ])
		SSI <- solve(SS)
		dev <- vector("numeric", nrow(DD))
		if(length(YIndex)){
			dev[-YIndex] <- apply(DD[-YIndex, ], 1, function(x) x%*%SSI%*%x)
			dev[-YIndex] <- 1/sqrt( (2*pi)^3*det(SS))*exp(-0.5*dev[-YIndex])
			##Now Y (only two params)
			SSY <- SS[c(1, 3), c(1, 3)]
			SSI <- solve(SSY)
			dev[YIndex] <- apply(DD[YIndex, c(1, 3)], 1, function(x) x%*%SSI%*%x)
			dev[YIndex] <- 1/sqrt( (2*pi)^2*det(SSY))*exp(-0.5*dev[YIndex])
		} else {
			dev=apply(DD,1,function(x) x%*%SSI%*%x)
			dev=1/sqrt( (2*pi)^3*det(SS))*exp(-0.5*dev)
		}
		gc(verbose=FALSE)
	}
	if (verbose) message("OK")
	##
	## BC: must keep SD
	params[-2] <- newparams[-2]
	rm(newparams)
	gc(verbose=FALSE)
	##
	if(verbose) message("Calling ", NR, " SNPs... ", appendLF=FALSE)
	##
	## ###################
	## ## MOVE TO C#######
	##
	## running in batches
	callsGt.present <- !missing(callsGt)
	callsPr.present <- !missing(callsPr)
	overwriteAB <- !callsGt.present & !callsPr.present
	if(!overwriteAB){
		open(callsGt)
		open(callsPr)
	}
	process2 <- function(idxBatch){
		snps <- snpBatches[[idxBatch]]
		tmpA <- as.matrix(A[snps,])
		tmpB <- as.matrix(B[snps,])
		rSnps <- range(snps)
		last <- (idxBatch-1)*batchSize
		IndexesBatch <- list(autosomeIndex[autosomeIndex %in% snps]-last,
				     XIndex[XIndex %in% snps]-last,
				     YIndex[YIndex %in% snps]-last)
		IndexesBatch <- lapply(IndexesBatch, as.integer)
		ImNull <- gtypeCallerR2(tmpA, tmpB, fIndex, mIndex,
					params[["centers"]][snps,],
					params[["scales"]][snps,],
					params[["N"]][snps,],
					IndexesBatch, cIndexes,
					sapply(IndexesBatch, length),
					sapply(cIndexes, length),
					SMEDIAN, theKnots, mixtureParams[],
					DF, probs, 0.025,
					which(regionInfo[snps, 2]),
					which(regionInfo[snps, 1]))
		if(overwriteAB){
			A[snps,] <- tmpA
			B[snps,] <- tmpB
		} else {
			callsGt[snps, ] <- tmpA
			callsPr[snps, ] <- tmpB
		}
	}
	gc(verbose=FALSE)
	ocLapply(seq(along=snpBatches), process2, neededPkgs="crlmm")
	close(A)
	close(B)
	close(mixtureParams)
	if(!overwriteAB){
		close(callsGt)
		close(callsPr)
	}
	gc(verbose=FALSE)
	message("Done with process2")
	##  END MOVE TO C#######
	## ##################
	##
	dev <- dev/(dev+1/383)
	if(!is.null(row.names)){ rownames(A) <- rownames(B) <- names(dev) <- row.names}
	if(!is.null(col.names)){ colnames(A) <- colnames(B) <- col.names}
	##
	if(length(Index) >= recallRegMin){
		tmp4batchQC <- DD[autosomeIndex,]/(params[["N"]][autosomeIndex,]+1)
		tmpSnpQc <- dev[autosomeIndex]
		SS <- cov(tmp4batchQC[tmpSnpQc < badSNP,])
		batchQC <- mean(diag(SS))
	}else{
		SS <- matrix(0, 3, 3)
		batchQC <- Inf
	}
	##
	if(verbose) message("Done.")
	if (returnParams){
		return(list(calls=A, confs=B, SNPQC=dev, batchQC=batchQC, params=params, DD=DD, covDD=SS, gender=gender, pkgname=pkgname))
	}else{
		return(list(calls=A, confs=B, SNPQC=dev, batchQC=batchQC, DD=DD, covDD=SS, gender=gender, pkgname=pkgname))
	}
}

##---------------------------------------------------------------------------
##---------------------------------------------------------------------------
getProtocolData.Affy <- function(filenames){
	scanDates <- data.frame(ScanDate=sapply(filenames, celfileDate))
	rownames(scanDates) <- basename(rownames(scanDates))
	protocoldata <- new("AnnotatedDataFrame",
			    data=scanDates,
			    varMetadata=data.frame(labelDescription=colnames(scanDates),
			                           row.names=colnames(scanDates)))
	return(protocoldata)
}

##setMethod("snpNames", signature(object="character"),
##	  function(object){
##		  nm <- grep("Crlmm", object)
##		  if(length(nm)==0){
##			  pkgname <- paste(object, "Crlmm", sep="")
##		  } else pkgname <- object
##		  loader("preprocStuff.rda", .crlmmPkgEnv, object)
##		  gns <- getVarInEnv("gns")
##		  return(gns)
##	  })


getFeatureData <- function(cdfName, copynumber=FALSE, genome){
	pkgname <- getCrlmmAnnotationName(cdfName)
	if(!require(pkgname, character.only=TRUE)){
		suggCall <- paste("library(", pkgname, ", lib.loc='/Altern/Lib/Loc')", sep="")
		msg <- paste("If", pkgname, "is installed on an alternative location, please load it manually by using", suggCall)
		message(strwrap(msg))
		stop("Package ", pkgname, " could not be found.")
		rm(suggCall, msg)
	}
	loader("preprocStuff.rda", .crlmmPkgEnv, pkgname)
	loader("genotypeStuff.rda", .crlmmPkgEnv, pkgname)
	loader("mixtureStuff.rda", .crlmmPkgEnv, pkgname)
	gns <- getVarInEnv("gns")
	nm <- grep("Crlmm", pkgname)
	if(length(nm)==0){
		pkgname <- paste(pkgname, "Crlmm", sep="")
	}
	path <- system.file("extdata", package=pkgname)
	##multiple.builds <- length(grep("hg19", list.files(path)) > 0)
	snp.file <- list.files(path, pattern="snpProbes_hg")
	if(length(snp.file)==0){
		no.build <- TRUE
		snp.file <- list.files(path, pattern="snpProbes.rda")
	} else {
		no.build <- FALSE
		if(missing(genome)){
			snp.file <- list.files(path, pattern="snpProbes_hg")
			if(length(snp.file) > 1){
				## use hg19
				message("genome build not specified. Using build hg19 for annotation.")
				snp.file <- snp.file[1]
			}
			genome <- gsub(".rda", "", strsplit(snp.file, "snpProbes_")[[1]][[2]])
		} else snp.file <- paste("snpProbes_", genome, ".rda", sep="")
	}
##	if(!multiple.builds){
##		load(file.path(path, "snpProbes.rda"))
##	} else load(file.path(path, paste("snpProbes_", genome, ".rda", sep="")))
	load(file.path(path, snp.file))
	snpProbes <- get("snpProbes")
	## if we use a different build we may throw out a number of snps...
	snpProbes <- snpProbes[rownames(snpProbes) %in% gns, ]
	if(copynumber){
		if(!no.build){
			cn.file <- paste("cnProbes_", genome, ".rda", sep="")
		} else cn.file <- "cnProbes.rda"
		load(file.path(path, cn.file))
		##		if(!multiple.builds){
		##			load(file.path(path, "cnProbes.rda"))
		##		} else load(file.path(path, paste("cnProbes_", genome, ".rda", sep="")))
		cnProbes <- get("cnProbes")
		snpIndex <- seq(along=gns)
		npIndex <- seq(along=rownames(cnProbes)) + max(snpIndex)
		featurenames <- c(gns, rownames(cnProbes))
	} else featurenames <- gns
	fvarlabels=c("chromosome", "position", "isSnp")
	M <- matrix(NA, length(featurenames), 3, dimnames=list(featurenames, fvarlabels))
	index <- match(rownames(snpProbes), rownames(M)) #only snp probes in M get assigned position
	M[index, "position"] <- snpProbes[, grep("pos", colnames(snpProbes))]
	M[index, "chromosome"] <- chromosome2integer(snpProbes[, grep("chr", colnames(snpProbes))])
	##M[index, "isSnp"] <- 1L
	index <- which(is.na(M[, "isSnp"]))
	M[index, "isSnp"] <- 1L
	if(copynumber){
		index <- match(rownames(cnProbes), rownames(M)) #only snp probes in M get assigned position
		M[index, "position"] <- cnProbes[, grep("pos", colnames(cnProbes))]
		M[index, "chromosome"] <- chromosome2integer(cnProbes[, grep("chr", colnames(cnProbes))])
		M[index, "isSnp"] <- 0L
	}
	##A few of the snpProbes do not match -- I think it is chromosome Y.
	if(any(is.na(M[, "chromosome"])))
		M[is.na(M[, "chromosome"]), "isSnp"] <- 1L
	##return(new("AnnotatedDataFrame", data=M))
	new("GenomeAnnotatedDataFrame",
	    position=as.integer(M[, "position"]),
	    chromosome=as.integer(M[, "chromosome"]),
	    isSnp=as.logical(M[, "isSnp"]),
	    row.names=featurenames)
}

getFeatureData.Affy <- getFeatureData

construct <- function(filenames,
		      cdfName,
		      copynumber=TRUE,
		      sns, verbose=TRUE, batch, fns){
	if(!missing(batch)){
		stopifnot(length(batch) == length(sns))
	}
	if(any(is.na(batch))){
		stop("NA's in batch variable")
	}
	if(missing(sns) & missing(filenames)) stop("one of filenames or samplenames (sns) must be provided")
	if(verbose) message("Initializing container for copy number estimation")
	featureData <- getFeatureData(cdfName, copynumber=copynumber)
	if(!missing(fns)){
		warning("subsetting the featureData can cause the snprma object and CNSet object to be misaligned. This problem is fixed in devel as we match the names prior to assigning values from snprma")
		index <- match(fns, featureNames(featureData))
		if(all(is.na(index))) stop("fns not in featureNames")
		featureData <- featureData[index, ]
	}
	nr <- nrow(featureData); nc <- length(sns)
	cnSet <- new("CNSet",
		     alleleA=initializeBigMatrix(name="A", nr, nc),
		     alleleB=initializeBigMatrix(name="B", nr, nc),
		     call=initializeBigMatrix(name="call", nr, nc),
		     callProbability=initializeBigMatrix(name="callPr", nr,nc),
		     annotation=cdfName,
		     batch=batch)
	if(!missing(sns)){
		sampleNames(cnSet) <- sns
		protocolData <- annotatedDataFrameFrom(A(cnSet), byrow=FALSE)
	} else {
		sampleNames(cnSet) <- basename(filenames)
		protocolData <- getProtocolData.Affy(filenames)
	}
	rownames(pData(protocolData)) <- sns
	protocolData(cnSet) <- protocolData
	featureData(cnSet) <- featureData
	featureNames(cnSet) <- featureNames(featureData)
	pd <- data.frame(matrix(NA, nc, 3), row.names=sns)
	colnames(pd)=c("SKW", "SNR", "gender")
	phenoData(cnSet) <- new("AnnotatedDataFrame", data=pd)
	gns <- getVarInEnv("gns")
	stopifnot(identical(gns, featureNames(cnSet)[1:length(gns)]))
	return(cnSet)
}

constructAffyCNSet <- function(filenames, sns, cdfName, batch, verbose=TRUE, genome){
	##stopifnot(!missing(filenames))
	if(missing(sns)) sns <- basename(filenames)
	stopifnot(!missing(batch))
	##---------------------------------------------------------------------------
	##
	## from snprma2.  Goal is to initialize A and B with
	## appropriate dimension for snps+nps
	##
	##---------------------------------------------------------------------------
	if (missing(cdfName))
		cdfName <- read.celfile.header(filenames[1])[["cdfName"]]
	pkgname <- getCrlmmAnnotationName(cdfName)
	stopifnot(require(pkgname, character.only=TRUE, quietly=!verbose))
	if(verbose) message("Loading annotations and mixture model parameters.")
	obj <- loader("preprocStuff.rda", .crlmmPkgEnv, pkgname)
	pnsa <- getVarInEnv("pnsa")
	pnsb <- getVarInEnv("pnsb")
	gns <- getVarInEnv("gns")
	rm(list=obj, envir=.crlmmPkgEnv)
	rm(obj)
	if(verbose) message("Initializing ff objects.")
	##mixtureParams <- initializeBigMatrix("crlmmMixt-", 4, length(filenames), "double")
	SNR <- initializeBigVector("crlmmSNR-", length(filenames), "double")
	SKW <- initializeBigVector("crlmmSKW-", length(filenames), "double")
	genderff <- initializeBigVector("gender", length(filenames), "integer")
	featureData <- getFeatureData(cdfName, copynumber=TRUE, genome=genome)
	nr <- nrow(featureData); nc <- length(sns)
	A <- initializeBigMatrix("crlmmA-", nr, length(filenames), "integer")
	B <- initializeBigMatrix("crlmmB-", nr, length(filenames), "integer")
	call <- initializeBigMatrix("call-", nr, length(filenames), "integer")
	callPr <- initializeBigMatrix("callPr-", nr, length(filenames), "integer")
	mixtureParams <- initializeBigMatrix("crlmmMixt-", 4, length(filenames), "double")
	rownames(A) <- rownames(B) <- featureNames(featureData)
	rownames(call) <- rownames(callPr) <- featureNames(featureData)
	dirNames <- dirname(filenames)
	unames <- unique(dirNames)
	dirNames <- factor(dirNames, levels=unames)
	dd <- split(seq_len(length(filenames)), dirNames)
	datadir <- list(dirname=names(dd), n=as.integer(sapply(dd, length)))
	if(verbose) message("Instantiating CNSet container")
	cnSet <- new("CNSet",
		     alleleA=A,
		     alleleB=B,
		     call=call,
		     callProbability=callPr,
		     featureData=featureData,
		     annotation=cdfName,
		     batch=as.character(batch),
		     genome=genome,
		     datadir=datadir)
	cnSet@mixtureParams <- mixtureParams
	sampleNames(cnSet) <- sns
	protocolData <- getProtocolData.Affy(filenames)
	protocolData$filename <- basename(filenames)
	rownames(pData(protocolData)) <- sns
	protocolData(cnSet) <- protocolData
	cnSet$SKW <- SKW
	cnSet$SNR <- SNR
	cnSet$gender <- genderff
	stopifnot(validObject(cnSet))
	return(cnSet)
}

snprmaAffy <- function(cnSet,
		       mixtureSampleSize=10^5,
		       eps=0.1,
		       seed=1,
		       verbose=TRUE){
	filenames <- celfileName(cnSet)
	if(verbose) message("Preprocessing ", length(filenames), " files.")
	pkgname <- getCrlmmAnnotationName(annotation(cnSet))
	stopifnot(require(pkgname, character.only=TRUE, quietly=!verbose))
	sampleBatches <- splitIndicesByNode(seq_along(filenames))
	mixtureParams <- cnSet@mixtureParams
	obj1 <- loader("preprocStuff.rda", .crlmmPkgEnv, pkgname)
	obj2 <- loader("genotypeStuff.rda", .crlmmPkgEnv, pkgname)
	obj3 <- loader("mixtureStuff.rda", .crlmmPkgEnv, pkgname)
	autosomeIndex <- getVarInEnv("autosomeIndex")
	pnsa <- getVarInEnv("pnsa")
	pnsb <- getVarInEnv("pnsb")
	fid <- getVarInEnv("fid")
	reference <- getVarInEnv("reference")
	aIndex <- getVarInEnv("aIndex")
	bIndex <- getVarInEnv("bIndex")
	SMEDIAN <- getVarInEnv("SMEDIAN")
	theKnots <- getVarInEnv("theKnots")
	gns <- getVarInEnv("gns")
	rm(list=c(obj1, obj2, obj3), envir=.crlmmPkgEnv)
	rm(obj1, obj2, obj3)
	## for mixture
	set.seed(seed)
	idx <- sort(sample(autosomeIndex, mixtureSampleSize))
	##for skewness. no need to do everything
	idx2 <- sample(length(fid), 10^5)
	A <- A(cnSet)
	B <- B(cnSet)
	SKW <- cnSet$SKW; SNR <- cnSet$SNR
	open(A)
	open(B)
	open(SKW)
	open(mixtureParams)
	open(SNR)
	## RS ADDED
	index <- match(gns, rownames(A))
	rsprocessCEL <- function(i){
		for (k in i){
			y <- as.matrix(read.celfile(filenames[k], intensity.means.only=TRUE)[["INTENSITY"]][["MEAN"]][fid])
			x <- log2(y[idx2])
			SKW[k] <- mean((x-mean(x))^3)/(sd(x)^3)
			rm(x)
			y <- normalize.quantiles.use.target(y, target=reference)
			## RS: add index for row assignment
			ya <- intMedianSummaries(y[aIndex, 1, drop=FALSE], pnsa)
			yb <- intMedianSummaries(y[bIndex, 1, drop=FALSE], pnsb)
			A[index, k] <- ya
			B[index, k] <- yb
			rm(y)
			S <- (log2(A[idx,k])+log2(B[idx, k]))/2 - SMEDIAN
			M <- log2(A[idx, k])-log2(B[idx, k])
			tmp <- fitAffySnpMixture56(S, M, theKnots, eps=eps)
			rm(S, M)
			mixtureParams[, k] <- tmp[["coef"]]
			SNR[k] <- tmp[["medF1"]]^2/(tmp[["sigma1"]]^2+tmp[["sigma2"]]^2)
			rm(tmp)
		}
		TRUE
	}
	ocLapply(sampleBatches, rsprocessCEL, neededPkgs="crlmm")
	close(A)
	close(B)
	close(SKW)
	close(mixtureParams)
	close(SNR)
	return()
}

genotype <- function(filenames,
		     cdfName,
		     batch,
		     mixtureSampleSize=10^5,
		     eps=0.1,
		     verbose=TRUE,
		     seed=1,
		     sns,
		     probs=rep(1/3, 3),
		     DF=6,
		     SNRMin=5,
		     recallMin=10,
		     recallRegMin=1000,
		     gender=NULL,
		     returnParams=TRUE,
		     badSNP=0.7,
		     genome=c("hg19", "hg18")){
	if(!isPackageLoaded("ff")) stop("ff package must be loaded")
	if (missing(sns)) sns <- basename(filenames)
	if (any(duplicated(sns))) stop("sample identifiers must be unique")
	genome <- match.arg(genome)
	cnSet <- constructAffyCNSet(filenames=filenames,
				    sns=sns,
				    cdfName=cdfName,
				    batch=batch, verbose=verbose,
				    genome=genome)
	if(!(is(A(cnSet), "ff") || is(A(cnSet), "ffdf"))){
		stop("The ff package is required for this function.")
	}
	ok <- snprmaAffy(cnSet,
			 mixtureSampleSize=mixtureSampleSize,
			 eps=eps,
			 seed=seed,
			 verbose=verbose)
	ok <- cnrmaAffy(cnSet=cnSet,
			seed=seed,
			verbose=verbose)
	stopifnot(ok)
	ok <- genotypeAffy(cnSet=cnSet,
			   SNRMin=SNRMin,
			   recallMin=recallMin,
			   recallRegMin=recallRegMin,
			   gender=gender,
			   badSNP=badSNP,
			   returnParams=returnParams,
			   verbose=verbose)
	return(cnSet)
}

genotypeAffy <- function(cnSet, SNRMin=5, recallMin=10,
			 recallRegMin=1000,
			 gender=NULL, badSNP=0.7, returnParams=TRUE,
			 verbose=TRUE){
	## The passed arguments A and B currently contain the intensities
	## (not calls and call probabilities)
	tmp <- crlmmGT2(A=A(cnSet),
			B=B(cnSet),
			SNR=cnSet$SNR,
			mixtureParams=cnSet@mixtureParams,
			cdfName=annotation(cnSet),
			row.names=NULL,
			col.names=sampleNames(cnSet),
			SNRMin=SNRMin,
			recallMin=recallMin,
			recallRegMin=recallRegMin,
			gender=gender,
			verbose=verbose,
			returnParams=returnParams,
			badSNP=badSNP,
			callsGt=calls(cnSet),
			callsPr=snpCallProbability(cnSet))
	cnSet$gender[] <- tmp$gender
	if(verbose) message("Genotyping finished.")
	return(TRUE)
}

cnrmaAffy <- function(cnSet, seed=1, verbose=TRUE){
	filenames <- celfileName(cnSet)
	np.index <- which(!isSnp(cnSet))
	A <- A(cnSet)
	cdfName <- annotation(cnSet)
	if(verbose) message("Quantile normalizing nonpolymorphic markers")
	##if(missing(cdfName)) stop("must specify cdfName")
	pkgname <- getCrlmmAnnotationName(cdfName)
	require(pkgname, character.only=TRUE) || stop("Package ", pkgname, " not available")
	sampleBatches <- splitIndicesByNode(seq_along(filenames))
	if(verbose) message("Processing nonpolymorphic probes for ", length(filenames), " files.")
	if(pkgname=="genomewidesnp6Crlmm"){
		loader("1m_reference_cn.rda", .crlmmPkgEnv, pkgname)
	}
	if(pkgname=="genomewidesnp5Crlmm"){
		loader("5.0_reference_cn.rda", .crlmmPkgEnv, pkgname)
	}
	reference <- getVarInEnv("reference")
        loader("npProbesFid.rda", .crlmmPkgEnv, pkgname)
	fid <- getVarInEnv("npProbesFid")
	row.names <- featureNames(cnSet)[np.index]
	row.names <- row.names[row.names %in% names(fid)] ##removes chromosome Y
	fid <- fid[match(row.names, names(fid))]
	np.index <- match(row.names, rownames(A))
	gns <- names(fid)
	set.seed(seed)
	## updates A
	open(A)
	processCEL2 <- function(i){
		for (k in i){
			y <- as.matrix(read.celfile(filenames[k], intensity.means.only=TRUE)[["INTENSITY"]][["MEAN"]][fid])
			A[np.index, k] <- as.integer(normalize.quantiles.use.target(y, target=reference))
		}
		return(TRUE)
	}
	ocLapply(sampleBatches, processCEL2, neededPkgs="crlmm")
	close(A)
	TRUE
}



genotype2 <- function(){
	.Defunct(msg="The genotype2 function has been deprecated. The function genotype should be used instead.  genotype will support large data using ff provided that the ff package is loaded.")
}
genotypeLD <- genotype2

rowCovs <- function(x, y, ...){
	notna <- !is.na(x)
	N <- rowSums(notna)
	x <- suppressWarnings(log2(x))
	if(!missing(y)) y <- suppressWarnings(log2(y))
	return(rowSums((x - rowMeans(x, ...)) * (y - rowMeans(y, ...)), ...)/(N-1))
}

rowMAD <- function(x, y, ...){
	##notna <- !is.na(x)
	mad <- 1.4826*rowMedians(abs(x-rowMedians(x, ...)), ...)
	return(mad)
}

shrink <- function(x, Ns, DF.PRIOR){
	DF <- Ns-1
	DF[DF < 1] <- 1
	x.0 <- apply(x, 2, median, na.rm=TRUE)
	x <- (x*DF + x.0*DF.PRIOR)/(DF.PRIOR + DF)
	for(j in 1:ncol(x)) x[is.na(x[, j]), j] <- x.0[j]
	return(x)
}

rowCors <- function(x, y, ...){
	N <- rowSums(!is.na(x))
	x <- suppressWarnings(log2(x))
	y <- suppressWarnings(log2(y))
	sd.x <- rowSds(x, ...)
	sd.y <- rowSds(y, ...)
	covar <- rowSums((x - rowMeans(x, ...)) * (y - rowMeans(y, ...)), ...)/(N-1)
	return(covar/(sd.x*sd.y))
}

## dqrlsWrapper <- function(x, y, wts, tol=1e-7){
## 	n <- NROW(y)
## 	p <- ncol(x)
## 	ny <- NCOL(y)
## 	.Fortran("dqrls", qr=x*wts, n=n, p=p, y=y * wts, ny=ny,
## 		 tol=as.double(tol), coefficients=mat.or.vec(p, ny),
## 		 residuals=y, effects=mat.or.vec(n, ny),
## 		 rank=integer(1L), pivot=1L:p, qraux=double(p),
## 		 work=double(2 * p), PACKAGE="base")[["coefficients"]]
## }


dqrlsWrapper <- function(x, y, wts, ...)
    fastLmPure(X=x*wts, y=y*wts, method=3)[['coefficients']]

fit.wls <- function(NN, sigma, allele, Y, autosome, X){
	Np <- NN
	Np[Np < 1] <- 1
	W <- (sigma/sqrt(Np))^-1
	Ystar <- Y*W
	complete <- which(rowSums(is.na(W)) == 0 & rowSums(is.na(Ystar)) == 0)
	if(missing(allele)) stop("must specify allele")
	if(autosome & missing(X)){
		if(allele == "A") X <- cbind(1, 2:0)
		if(allele == "B") X <- cbind(1, 0:2)
	}
	if(!autosome & missing(X)){
		if(allele == "A") X <- cbind(1, c(1, 0, 2, 1, 0), c(0, 1, 0, 1, 2))
		if(allele == "B") X <- cbind(1, c(0, 1, 0, 1, 2), c(1, 0, 2, 1, 0))
	}
	betahat <- matrix(NA, ncol(X), nrow(Ystar))
	ww <- rep(1, ncol(Ystar))
	for(i in complete){
		betahat[, i] <- dqrlsWrapper(W[i, ] * X, Ystar[i, ], ww)
		##ssr <- sum((Ystar[i, ] - matrix(Xstar[, i], nrow(X), ncol(X)) %*% matrix(betahat[, i], ncol(X), 1))^2)
	}
	return(betahat)
}

shrinkGenotypeSummaries <- function(strata, index.list, object, MIN.OBS, MIN.SAMPLES, DF.PRIOR,
				    verbose, is.lds=TRUE){
	##if(is.lds) {physical <- get("physical"); open(object)}
	open(object)
	if(verbose) message("      Probe stratum ", strata, " of ", length(index.list))
	marker.index <- index.list[[strata]]
	batches <- split(seq_along(batch(object)), as.character(batch(object)))
	batches <- batches[sapply(batches, length) >= MIN.SAMPLES]
	batch.names <- names(batches)
	batch.index <- which(batchNames(object) %in% batch.names)
	N.AA <- as.matrix(N.AA(object)[marker.index, batch.index])
	N.AB <- as.matrix(N.AB(object)[marker.index, batch.index])
	N.BB <- as.matrix(N.BB(object)[marker.index, batch.index])
	medianA.AA <- as.matrix(medianA.AA(object)[marker.index, batch.index])
	medianA.AB <- as.matrix(medianA.AB(object)[marker.index, batch.index])
	medianA.BB <- as.matrix(medianA.BB(object)[marker.index, batch.index])
	medianB.AA <- as.matrix(medianB.AA(object)[marker.index, batch.index])
	medianB.AB <- as.matrix(medianB.AB(object)[marker.index, batch.index])
	medianB.BB <- as.matrix(medianB.BB(object)[marker.index, batch.index])
	madA.AA <- as.matrix(madA.AA(object)[marker.index, batch.index])
	madA.AB <- as.matrix(madA.AB(object)[marker.index, batch.index])
	madA.BB <- as.matrix(madA.BB(object)[marker.index, batch.index])
	madB.AA <- as.matrix(madB.AA(object)[marker.index, batch.index])
	madB.AB <- as.matrix(madB.AB(object)[marker.index, batch.index])
	madB.BB <- as.matrix(madB.BB(object)[marker.index, batch.index])
	medianA <- medianB <- shrink.madB <- shrink.madA <- vector("list", length(batch.names))
	shrink.tau2A.AA <- tau2A.AA <- as.matrix(tau2A.AA(object)[marker.index, batch.index])
	shrink.tau2B.BB <- tau2B.BB <- as.matrix(tau2B.BB(object)[marker.index, batch.index])
	shrink.tau2A.BB <- tau2A.BB <- as.matrix(tau2A.BB(object)[marker.index, batch.index])
	shrink.tau2B.AA <- tau2B.AA <- as.matrix(tau2B.AA(object)[marker.index, batch.index])
	shrink.corrAA <- corrAA <- as.matrix(corrAA(object)[marker.index, batch.index])
	shrink.corrAB <- corrAB <- as.matrix(corrAB(object)[marker.index, batch.index])
	shrink.corrBB <- corrBB <- as.matrix(corrBB(object)[marker.index, batch.index])
	flags <- as.matrix(flags(object)[marker.index, batch.index])
	for(k in seq(along=batches)){
		sample.index <- batches[[k]]
		this.batch <- unique(as.character(batch(object)[sample.index]))
		medianA[[k]] <- cbind(medianA.AA[, k], medianA.AB[, k], medianA.BB[, k])
		medianB[[k]] <- cbind(medianB.AA[, k], medianB.AB[, k], medianB.BB[, k])
		madA <- cbind(madA.AA[, k], madA.AB[, k], madA.BB[, k])
		madB <- cbind(madB.AA[, k], madB.AB[, k], madB.BB[, k])
		NN <- cbind(N.AA[, k], N.AB[, k], N.BB[, k])
		##RS: estimate DF.PRIOR
		shrink.madA[[k]] <- shrink(madA, NN, DF.PRIOR)
		shrink.madB[[k]] <- shrink(madB, NN, DF.PRIOR)
		## an estimate of the background variance is the MAD
		## of the log2(allele A) intensities among subjects with
		## genotypes BB
		shrink.tau2A.BB[, k] <- shrink(tau2A.BB[, k, drop=FALSE], NN[, 3], DF.PRIOR)[, drop=FALSE]
		shrink.tau2B.AA[, k] <- shrink(tau2B.AA[, k, drop=FALSE], NN[, 1], DF.PRIOR)[, drop=FALSE]
		## an estimate of the signal variance is the MAD
		## of the log2(allele A) intensities among subjects with
		## genotypes AA
		shrink.tau2A.AA[, k] <- shrink(tau2A.AA[, k, drop=FALSE], NN[, 1], DF.PRIOR)[, drop=FALSE]
		shrink.tau2B.BB[, k] <- shrink(tau2B.BB[, k, drop=FALSE], NN[, 3], DF.PRIOR)[, drop=FALSE]
		cor.AA <- corrAA[, k, drop=FALSE]
		cor.AB <- corrAB[, k, drop=FALSE]
		cor.BB <- corrBB[, k, drop=FALSE]
		shrink.corrAA[, k] <- shrink(cor.AA, NN[, 1], DF.PRIOR)
		shrink.corrAB[, k] <- shrink(cor.AB, NN[, 2], DF.PRIOR)
		shrink.corrBB[, k] <- shrink(cor.BB, NN[, 3], DF.PRIOR)
		##
		##---------------------------------------------------------------------------
		## SNPs that we'll use for imputing location/scale of unobserved genotypes
		##---------------------------------------------------------------------------
		index.complete <- indexComplete(NN, medianA[[k]], medianB[[k]], MIN.OBS)
		##
		##---------------------------------------------------------------------------
		## Impute sufficient statistics for unobserved genotypes (plate-specific)
		##---------------------------------------------------------------------------
		unobservedAA <- NN[, 1] < MIN.OBS
		unobservedAB <- NN[, 2] < MIN.OBS
		unobservedBB <- NN[, 3] < MIN.OBS
		unobserved.index <- vector("list", 3)
		## AB, BB observed
		unobserved.index[[1]] <- which(unobservedAA & (NN[, 2] >= MIN.OBS & NN[, 3] >= MIN.OBS))
		## AA and BB observed (strange)
		unobserved.index[[2]] <- which(unobservedAB & (NN[, 1] >= MIN.OBS & NN[, 3] >= MIN.OBS))
		## AB and AA observed
		unobserved.index[[3]] <- which(unobservedBB & (NN[, 2] >= MIN.OBS & NN[, 1] >= MIN.OBS))
		res <- imputeCenter(medianA[[k]], medianB[[k]], index.complete, unobserved.index)
		medianA[[k]] <- res[[1]]
		medianB[[k]] <- res[[2]]
		rm(res)
		##the NA's in 'medianA' and 'medianB' are monomorphic if MIN.OBS = 1
		##
		## RS: For Monomorphic SNPs a mixture model may be better
		## RS: Further, we can improve estimation by borrowing strength across batch
		unobserved.index[[1]] <- which(unobservedAA & unobservedAB)
		unobserved.index[[2]] <- which(unobservedBB & unobservedAB)
		unobserved.index[[3]] <- which(unobservedAA & unobservedBB) ## strange
		res <- imputeCentersForMonomorphicSnps(medianA[[k]], medianB[[k]],
						       index.complete,
						       unobserved.index)
		medianA[[k]] <- res[[1]]; medianB[[k]] <- res[[2]]
		rm(res)
		negA <- rowSums(medianA[[k]] < 0) > 0
		negB <- rowSums(medianB[[k]] < 0) > 0
		flags[, k] <- as.integer(rowSums(NN == 0) > 0 | negA | negB)
	}
	flags(object)[marker.index, batch.index] <- flags
	medianA.AA(object)[marker.index, batch.index] <- do.call("cbind", lapply(medianA, function(x) x[, 1]))
	medianA.AB(object)[marker.index, batch.index] <- do.call("cbind", lapply(medianA, function(x) x[, 2]))
	medianA.BB(object)[marker.index, batch.index] <- do.call("cbind", lapply(medianA, function(x) x[, 3]))
	medianB.AA(object)[marker.index, batch.index] <- do.call("cbind", lapply(medianB, function(x) x[, 1]))
	medianB.AB(object)[marker.index, batch.index] <- do.call("cbind", lapply(medianB, function(x) x[, 2]))
	medianB.BB(object)[marker.index, batch.index] <- do.call("cbind", lapply(medianB, function(x) x[, 3]))
	##
	madA.AA(object)[marker.index, batch.index] <- do.call("cbind", lapply(shrink.madA, function(x) x[, 1]))
	madA.AB(object)[marker.index, batch.index] <- do.call("cbind", lapply(shrink.madA, function(x) x[, 2]))
	madA.BB(object)[marker.index, batch.index] <- do.call("cbind", lapply(shrink.madA, function(x) x[, 3]))
	madB.AA(object)[marker.index, batch.index] <- do.call("cbind", lapply(shrink.madB, function(x) x[, 1]))
	madB.AB(object)[marker.index, batch.index] <- do.call("cbind", lapply(shrink.madB, function(x) x[, 2]))
	madB.BB(object)[marker.index, batch.index] <- do.call("cbind", lapply(shrink.madB, function(x) x[, 3]))
	##
	corrAA(object)[marker.index, batch.index] <- shrink.corrAA
	corrAB(object)[marker.index, batch.index] <- shrink.corrAB
	corrBB(object)[marker.index, batch.index] <- shrink.corrBB
	tau2A.AA(object)[marker.index, batch.index] <- shrink.tau2A.AA
	tau2A.BB(object)[marker.index, batch.index] <- shrink.tau2A.BB
	tau2B.AA(object)[marker.index, batch.index] <- shrink.tau2B.AA
	tau2B.BB(object)[marker.index, batch.index] <- shrink.tau2B.BB
	if(is.lds) return(TRUE) else return(object)
}


fit.lm1 <- function(strata,
		    index.list,
		    object,
		    MIN.SAMPLES,
		    THR.NU.PHI,
		    MIN.NU,
		    MIN.PHI,
		    verbose, is.lds,
		    CHR.X, ...){
	open(object$gender)
	if(is.lds) {physical <- get("physical"); open(object)}
	if(verbose) message("      Probe stratum ", strata, " of ", length(index.list))
	snps <- index.list[[strata]]
	batches <- split(seq_along(batch(object)), as.character(batch(object)))
	batches <- batches[sapply(batches, length) >= MIN.SAMPLES]
	##batchnames <- batchNames(object)
	batch.names <- names(batches)
	batch.index <- which(batchNames(object) %in% batch.names)
	if(length(batches) > 1 && "grandMean" %in% batchNames(object))
		batch.index <- c(batch.index, match("grandMean", batchNames(object)))
	N.AA <- as.matrix(N.AA(object)[snps, batch.index])
	N.AB <- as.matrix(N.AB(object)[snps, batch.index])
	N.BB <- as.matrix(N.BB(object)[snps, batch.index])
	medianA.AA <- as.matrix(medianA.AA(object)[snps, batch.index])
	medianA.AB <- as.matrix(medianA.AB(object)[snps, batch.index])
	medianA.BB <- as.matrix(medianA.BB(object)[snps, batch.index])
	medianB.AA <- as.matrix(medianB.AA(object)[snps, batch.index])
	medianB.AB <- as.matrix(medianB.AB(object)[snps, batch.index])
	medianB.BB <- as.matrix(medianB.BB(object)[snps, batch.index])
	madA.AA <- as.matrix(madA.AA(object)[snps, batch.index])
	madA.AB <- as.matrix(madA.AB(object)[snps, batch.index])
	madA.BB <- as.matrix(madA.BB(object)[snps, batch.index])
	madB.AA <- as.matrix(madB.AA(object)[snps, batch.index])
	madB.AB <- as.matrix(madB.AB(object)[snps, batch.index])
	madB.BB <- as.matrix(madB.BB(object)[snps, batch.index])
	tau2A.AA <- as.matrix(tau2A.AA(object)[snps, batch.index])
	tau2B.BB <- as.matrix(tau2B.BB(object)[snps, batch.index])
	tau2A.BB <- as.matrix(tau2A.BB(object)[snps, batch.index])
	tau2B.AA <- as.matrix(tau2B.AA(object)[snps, batch.index])
	corrAA <- as.matrix(corrAA(object)[snps, batch.index])
	corrAB <- as.matrix(corrAB(object)[snps, batch.index])
	corrBB <- as.matrix(corrBB(object)[snps, batch.index])
	nuA <- as.matrix(nuA(object)[snps,  batch.index])
	phiA <- as.matrix(phiA(object)[snps, batch.index])
	nuB <- as.matrix(nuB(object)[snps, batch.index])
	phiB <- as.matrix(phiB(object)[snps, batch.index])
	flags <- as.matrix(flags(object)[snps, batch.index])
	for(k in seq(along=batches)){
		B <- batches[[k]]
		if(length(B) < MIN.SAMPLES) next()
		this.batch <- unique(as.character(batch(object)[B]))
		medianA <- cbind(medianA.AA[, k], medianA.AB[, k], medianA.BB[, k])
		medianB <- cbind(medianB.AA[, k], medianB.AB[, k], medianB.BB[, k])
		madA <- cbind(madA.AA[, k], madA.AB[, k], madA.BB[, k])
		madB <- cbind(madB.AA[, k], madB.AB[, k], madB.BB[, k])
		NN <- cbind(N.AA[, k], N.AB[, k], N.BB[, k])
		## regress on the medians using the standard errors (hence the division by N) as weights
		res <- fit.wls(NN=NN, sigma=madA, allele="A", Y=medianA, autosome=!CHR.X)
		nuA[, k] <- res[1, ]
		phiA[, k] <- res[2, ]
		rm(res)
		res <- fit.wls(NN=NN, sigma=madB, allele="B", Y=medianB, autosome=!CHR.X)##allele="B", Ystar=YB, W=wB, Ns=Ns)
		nuB[, k] <- res[1, ]
		phiB[, k] <- res[2, ]
	}
	##---------------------------------------------------------------------------
	##
	## Grand mean
	##
	##---------------------------------------------------------------------------
##	if(length(batches) > 1 && "grandMean" %in% batchNames(object)){
	##  TODO:  There are NA's in the medianA.AA for the grandMean and 0's in the madA
	##         - both need to be handled prior to estimating a grand intercept and slope
	if(FALSE){
		## then the last column is for the grandMean
		k <- ncol(medianA.AA)
		medianA <- cbind(medianA.AA[, k], medianA.AB[, k], medianA.BB[, k])
		medianB <- cbind(medianB.AA[, k], medianB.AB[, k], medianB.BB[, k])
		madA <- cbind(madA.AA[, k], madA.AB[, k], madA.BB[, k])
		madB <- cbind(madB.AA[, k], madB.AB[, k], madB.BB[, k])
		NN <- cbind(N.AA[, k], N.AB[, k], N.BB[, k])
		index <- which(rowSums(is.na(medianA)) == 0)
		res <- fit.wls(NN=NN[index, ], sigma=madA[index, ], allele="A", Y=medianA[index, ], autosome=!CHR.X)
		nuA[, k] <- res[1, ]
		phiA[, k] <- res[2, ]
		rm(res)
		res <- fit.wls(NN=NN, sigma=madB, allele="B", Y=medianB, autosome=!CHR.X)##allele="B", Ystar=YB, W=wB, Ns=Ns)
		nuB[, k] <- res[1, ]
		phiB[, k] <- res[2, ]
	}
	if(THR.NU.PHI){
		nuA[nuA < MIN.NU] <- MIN.NU
		nuB[nuB < MIN.NU] <- MIN.NU
		phiA[phiA < MIN.PHI] <- MIN.PHI
		phiB[phiB < MIN.PHI] <- MIN.PHI
	}
	nuA(object)[snps, batch.index] <- nuA
	nuB(object)[snps, batch.index] <- nuB
	phiA(object)[snps, batch.index] <- phiA
	phiB(object)[snps, batch.index] <- phiB
	if(is.lds){
		close(object)
		return(TRUE)
	} else{
		return(object)
	}
}

## nonpolymorphic markers
fit.lm2 <- function(strata,
		    index.list,
		    object,
		    MIN.SAMPLES,
		    THR.NU.PHI,
		    MIN.NU,
		    MIN.PHI,
		    verbose, is.lds, CHR.X, ...){
	if(is.lds) {physical <- get("physical"); open(object)}
	open(object$gender)
	if(verbose) message("      Probe stratum ", strata, " of ", length(index.list))
	marker.index <- index.list[[strata]]
	batches <- split(seq_along(batch(object)), as.character(batch(object)))
	batches <- batches[sapply(batches, length) >= MIN.SAMPLES]
	batch.names <- names(batches)
	batch.index <- which(batchNames(object) %in% batch.names)
	##
	ii <- isSnp(object) & chromosome(object) < 23 & !is.na(chromosome(object))
	flags <- as.matrix(flags(object)[ii, batch.index])
	fns <- featureNames(object)[ii]
	fns.noflags <- fns[rowSums(flags, na.rm=T) == 0]
	if(length(fns.noflags) == 0) {
		warning("All features in one of the probeset batches were flagged. The data is processed in probeset batches to reduce RAM, but the error suggests that increasing the size of the batches may help.  For example, ocProbesets(4*ocProbesets()) would increase the current size of the probeset batches by 4.")
		fns.noflags <- featureNames(object)[marker.index] ## allow processing to continue
	}
	index.flag <- match(fns.noflags, featureNames(object))
	snp.index <- sample(index.flag, min(c(length(index.flag), 5000)))
	##
	nuA.np <- as.matrix(nuA(object)[marker.index, batch.index])
	phiA.np <- as.matrix(phiA(object)[marker.index, batch.index])
	tau2A.AA <- as.matrix(tau2A.AA(object)[marker.index, batch.index])
	##
	nuA.snp <- as.matrix(nuA(object)[snp.index, batch.index])
	nuB.snp <- as.matrix(nuB(object)[snp.index, batch.index])
	phiA.snp <- as.matrix(phiA(object)[snp.index, batch.index])
	phiB.snp <- as.matrix(phiB(object)[snp.index, batch.index])
	medianA.AA <- as.matrix(medianA.AA(object)[snp.index, batch.index])
	medianB.BB <- as.matrix(medianB.BB(object)[snp.index, batch.index])
	medianA.AA.np <- as.matrix(medianA.AA(object)[marker.index, batch.index])
	for(k in seq_along(batches)){
		B <- batches[[k]]
		this.batch <- unique(as.character(batch(object)[B]))
		X <- cbind(1, log2(c(medianA.AA[, k], medianB.BB[, k])))
		Y <- log2(c(phiA.snp[, k], phiB.snp[, k]))
		not.missing <- rowSums(is.na(X)) == 0 & !is.na(Y)
		X <- X[not.missing, ]
		Y <- Y[not.missing]
		betahat <- solve(crossprod(X), crossprod(X, Y))
		##crosshyb <- max(median(medianA.AA[, k]) - median(medianA.AA.np[, k]), 0)
		crosshyb <- 0
		X <- cbind(1, log2(medianA.AA.np[, k] + crosshyb))
		logPhiT <- X %*% betahat
		phiA.np[, k] <- 2^(logPhiT)
		nuA.np[, k] <- medianA.AA.np[,k]-2*phiA.np[, k]
##		cA[, k] <- 1/phiA.np[, J] * (A.np - nuA.np[, J])
	}
	if(THR.NU.PHI){
		nuA.np[nuA.np < MIN.NU] <- MIN.NU
		phiA.np[phiA.np < MIN.PHI] <- MIN.PHI
	}
	nuA(object)[marker.index, batch.index] <- nuA.np
	phiA(object)[marker.index, batch.index] <- phiA.np
	if(is.lds) { close(object); return(TRUE)}
	return(object)
}

summarizeMaleXNps <- function(marker.index,
			      batches,
			      object, MIN.SAMPLES){
	nr <- length(marker.index)
	nc <- length(batchNames(object))
	NN.Mlist <- imputed.medianA <- imputed.medianB <- shrink.madA <- shrink.madB <- vector("list", nc)
	open(object$gender)
	gender <- object$gender[]
	open(A(object))
	AA <- as.matrix(A(object)[marker.index, gender==1])
	madA.AA <- medianA.AA <- matrix(NA, nr, nc)
	colnames(madA.AA) <- colnames(medianA.AA) <- batchNames(object)
	numberMenPerBatch <- rep(NA, nc)
	for(k in seq_along(batches)){
		B <- batches[[k]]
		this.batch <- unique(as.character(batch(object)[B]))
		gender <- object$gender[B]
		if(sum(gender==1) < MIN.SAMPLES) next()
		sns.batch <- sampleNames(object)[B]
		##subset GG apppriately
		sns <- colnames(AA)
		J <- sns%in%sns.batch
		numberMenPerBatch[k] <- length(J)
		medianA.AA[, k] <- rowMedians(AA[, J, drop=FALSE], na.rm=TRUE)
		madA.AA[, k] <- rowMAD(AA[, J, drop=FALSE], na.rm=TRUE)
	}
	return(list(medianA.AA=medianA.AA,
		    madA.AA=madA.AA))
}


summarizeXGenotypes <- function(marker.index,
				batches,
				object,
				GT.CONF.THR,
				MIN.OBS,
				MIN.SAMPLES,
				verbose,
				is.lds,
				DF.PRIOR,
				gender="male", ...){
	I <- unlist(batches)
	open(object$gender)
	if(gender == "male"){
		gender.index <- which(object$gender[] == 1)
	} else {
		gender.index <- which(object$gender[] == 2)
	}
	batches <- lapply(batches, function(x, gender.index) intersect(x, gender.index), gender.index)
	batch.names <- names(batches)
	batch.index <- which(batchNames(object) %in% batch.names)
	nr <- length(marker.index)
	nc <- length(batch.index)
	NN.Mlist <- medianA <- medianB <- shrink.madA <- shrink.madB <- vector("list", nc)
	names(NN.Mlist) <- names(medianA) <- names(medianB) <- names(shrink.madA) <- names(shrink.madB) <- batch.names
	open(calls(object))
	open(snpCallProbability(object))
	open(A(object))
	open(B(object))
	GG <- as.matrix(calls(object)[marker.index, ])
	CP <- as.matrix(snpCallProbability(object)[marker.index, ])
	AA <- as.matrix(A(object)[marker.index, ])
	BB <- as.matrix(B(object)[marker.index, ])
	for(k in seq_along(batches)){
		sample.index <- batches[[k]]
		if(length(sample.index) == 0) next()
		this.batch <- unique(as.character(batch(object)[sample.index]))
		##gender <- object$gender[sample.index]
		if(length(sample.index) < MIN.SAMPLES) next()
		sns.batch <- sampleNames(object)[sample.index]
		##subset GG apppriately
		sns <- colnames(GG)
		J <- sns%in%sns.batch
		G <- GG[, J, drop=FALSE]
		xx <- CP[, J, drop=FALSE]
		p <- i2p(xx)
		highConf <- (1-exp(-xx/1000)) > GT.CONF.THR
		G <- G*highConf
		A <- AA[, J, drop=FALSE]
		B <- BB[, J, drop=FALSE]
		G.AA <- G==1
		G.AA[G.AA==FALSE] <- NA
		G.AB <- G==2
		G.AB[G.AB==FALSE] <- NA
		G.BB <- G==3
		G.BB[G.BB==FALSE] <- NA
		N.AA <- rowSums(G.AA, na.rm=TRUE)
		if(gender == "female")
			N.AB <- rowSums(G.AB, na.rm=TRUE)
		N.BB <- rowSums(G.BB, na.rm=TRUE)
		summaryStats <- function(X, INT, FUNS){
			tmp <- matrix(NA, nrow(X), length(FUNS))
			for(j in seq_along(FUNS)){
				FUN <- match.fun(FUNS[j])
				tmp[, j] <- FUN(X*INT, na.rm=TRUE)
			}
			tmp
		}
		statsA.AA <- summaryStats(G.AA, A, FUNS=c("rowMedians", "rowMAD"))
		if(gender == "female")
			statsA.AB <- summaryStats(G.AB, A, FUNS=c("rowMedians", "rowMAD"))
		statsA.BB <- summaryStats(G.BB, A, FUNS=c("rowMedians", "rowMAD"))
		statsB.AA <- summaryStats(G.AA, B, FUNS=c("rowMedians", "rowMAD"))
		if(gender == "female")
			statsB.AB <- summaryStats(G.AB, B, FUNS=c("rowMedians", "rowMAD"))
		statsB.BB <- summaryStats(G.BB, B, FUNS=c("rowMedians", "rowMAD"))
		if(gender=="male"){
			medianA[[k]] <- cbind(statsA.AA[, 1], ##statsA.AB[, 1],
					      statsA.BB[, 1])
			medianB[[k]] <- cbind(statsB.AA[, 1], ##statsB.AB[, 1],
					      statsB.BB[, 1])
			madA <- cbind(statsA.AA[, 2],  ##statsA.AB[, 2],
				      statsA.BB[, 2])
			madB <- cbind(statsB.AA[, 2], ##statsB.AB[, 2],
				      statsB.BB[, 2])
			NN <- cbind(N.AA, N.BB)
			rm(statsA.AA, statsA.BB, statsB.AA, statsB.BB)
		} else {
			medianA[[k]] <- cbind(statsA.AA[, 1], statsA.AB[, 1],
					      statsA.BB[, 1])
			medianB[[k]] <- cbind(statsB.AA[, 1], statsB.AB[, 1],
					      statsB.BB[, 1])
			madA <- cbind(statsA.AA[, 2],  statsA.AB[, 2],
				      statsA.BB[, 2])
			madB <- cbind(statsB.AA[, 2], statsB.AB[, 2],
				      statsB.BB[, 2])
			NN <- cbind(N.AA, N.AB, N.BB)
			rm(statsA.AA, statsA.AB, statsA.BB, statsB.AA, statsB.AB, statsB.BB)
		}
		NN.Mlist[[k]] <- NN
		shrink.madA[[k]] <- shrink(madA, NN, DF.PRIOR)
		shrink.madB[[k]] <- shrink(madB, NN, DF.PRIOR)
		##---------------------------------------------------------------------------
		## SNPs that we'll use for imputing location/scale of unobserved genotypes
		##---------------------------------------------------------------------------
		index.complete <- indexComplete(NN, medianA[[k]], medianB[[k]], MIN.OBS)
		##---------------------------------------------------------------------------
		## Impute sufficient statistics for unobserved genotypes (plate-specific)
		##---------------------------------------------------------------------------
		if(gender=="male"){
			res <- imputeCenterX(medianA[[k]], medianB[[k]], NN, index.complete, MIN.OBS)
		} else {
			unobservedAA <- NN[, 1] < MIN.OBS
			unobservedAB <- NN[, 2] < MIN.OBS
			unobservedBB <- NN[, 3] < MIN.OBS
			unobserved.index <- vector("list", 3)
			## AB, BB observed
			unobserved.index[[1]] <- which(unobservedAA & (NN[, 2] >= MIN.OBS & NN[, 3] >= MIN.OBS))
			## AA and BB observed (strange)
			unobserved.index[[2]] <- which(unobservedAB & (NN[, 1] >= MIN.OBS & NN[, 3] >= MIN.OBS))
			## AB and AA observed
			unobserved.index[[3]] <- which(unobservedBB & (NN[, 2] >= MIN.OBS & NN[, 1] >= MIN.OBS))
			res <- imputeCenter(medianA[[k]], medianB[[k]], index.complete, unobserved.index)
			medianA[[k]] <- res[[1]]
			medianB[[k]] <- res[[2]]

			## RS: For Monomorphic SNPs a mixture model may be better
			## RS: Further, we can improve estimation by borrowing strength across batch
			unobserved.index[[1]] <- which(unobservedAA & unobservedAB)
			unobserved.index[[2]] <- which(unobservedBB & unobservedAB)
			unobserved.index[[3]] <- which(unobservedAA & unobservedBB) ## strange
			res <- imputeCentersForMonomorphicSnps(medianA[[k]], medianB[[k]],
							       index.complete,
							       unobserved.index)
			medianA[[k]] <- res[[1]]; medianB[[k]] <- res[[2]]
		}
		medianA[[k]] <- res[[1]]
		medianB[[k]] <- res[[2]]
	}
	close(calls(object))
	close(snpCallProbability(object))
	close(A(object))
	close(B(object))
	return(list(madA=shrink.madA,
		    madB=shrink.madB,
		    NN=NN.Mlist,
		    medianA=medianA,
		    medianB=medianB))
}

## X chromosome, SNPs
fit.lm3 <- function(strata,
		    index.list,
		    object,
		    SNRMin,
		    MIN.SAMPLES,
		    MIN.OBS,
		    DF.PRIOR,
		    GT.CONF.THR,
		    THR.NU.PHI,
		    MIN.NU,
		    MIN.PHI,
		    verbose, is.lds, CHR.X, ...){
	##if(is.lds) {physical <- get("physical"); open(object)}
	if(verbose) message("      Probe stratum ", strata, " of ", length(index.list))
	open(object$gender)
	gender <- object$gender[]
	enough.males <- sum(gender==1) >= MIN.SAMPLES
	enough.females <- sum(gender==2) >= MIN.SAMPLES
	if(!enough.males & !enough.females){
		message(paste("fewer than", MIN.SAMPLES, "men and women.  Copy number not estimated for CHR X"))
		return(object)
	}
	marker.index <- index.list[[strata]]
	batches <- split(seq_along(batch(object)), as.character(batch(object)))
	batches <- batches[sapply(batches, length) >= MIN.SAMPLES]
	batch.names <- names(batches)
	batch.index <- which(batchNames(object) %in% batch.names)

	nuA <- as.matrix(nuA(object)[marker.index, batch.index])
	nuB <- as.matrix(nuB(object)[marker.index, batch.index])
	phiA <- as.matrix(phiA(object)[marker.index, batch.index])
	phiB <- as.matrix(phiB(object)[marker.index, batch.index])
	phiA2 <- as.matrix(phiPrimeA(object)[marker.index, batch.index])
	phiB2 <- as.matrix(phiPrimeB(object)[marker.index, batch.index])
	if(enough.males){
		res <- summarizeXGenotypes(marker.index=marker.index,
					   batches=batches,
					   object=object,
					   GT.CONF.THR=GT.CONF.THR,
					   MIN.SAMPLES=MIN.SAMPLES,
					   MIN.OBS=MIN.OBS,
					   verbose=verbose,
					   is.lds=is.lds,
					   DF.PRIOR=DF.PRIOR/2,
					   gender="male")
		madA.Mlist <- res[["madA"]]
		madB.Mlist <- res[["madB"]]
		medianA.Mlist <- res[["medianA"]]
		medianB.Mlist <- res[["medianB"]]
		NN.Mlist <- res[["NN"]]
		rm(res)
		## Need N, median, mad
	}
	if(enough.females){
		res <- summarizeXGenotypes(marker.index=marker.index,
						 batches=batches,
						 object=object,
						 GT.CONF.THR=GT.CONF.THR,
						 MIN.SAMPLES=MIN.SAMPLES,
						 MIN.OBS=MIN.OBS,
						 verbose=verbose,
						 is.lds=is.lds,
						 DF.PRIOR=DF.PRIOR/2,
					   gender="female")
		madA.Flist <- res[["madA"]]
		madB.Flist <- res[["madB"]]
		medianA.Flist <- res[["medianA"]]
		medianB.Flist <- res[["medianB"]]
		NN.Flist <- res[["NN"]]
		rm(res)
              }
        enough.men <- enough.women <- FALSE
	for(k in seq_along(batches)){
		sample.index <- batches[[k]]
		if(is.null(sample.index)) next()
		this.batch <- unique(as.character(batch(object)[sample.index]))
		gender <- object$gender[sample.index]
		enough.men <- sum(gender==1) >= MIN.SAMPLES
		enough.women <- sum(gender==2) >= MIN.SAMPLES
		if(!enough.men & !enough.women) {
			if(verbose) message(paste("fewer than", MIN.SAMPLES, "men and women in batch", this.batch, ". CHR X copy number not available. "))
			next()
		}
		if(enough.women){
			madA.F <- madA.Flist[[k]]
			madB.F <- madB.Flist[[k]]
			medianA.F <- medianA.Flist[[k]]
			medianB.F <- medianB.Flist[[k]]
			NN.F <- NN.Flist[[k]]
		}
		if(enough.men){
			madA.M <- madA.Mlist[[k]]
			madB.M <- madB.Mlist[[k]]
			medianA.M <- medianA.Mlist[[k]]
			medianB.M <- medianB.Mlist[[k]]
			NN.M <- NN.Mlist[[k]]
		}
		if(enough.men & enough.women){
			if(any(madA.F == 0)){
				message("Zeros in median absolute deviation.  Not estimating CN for chrX for batch ", names(batches)[k])
				next()
			}
			betas <- fit.wls(cbind(NN.M, NN.F),
					 sigma=cbind(madA.M, madA.F),
					 allele="A",
					 Y=cbind(medianA.M, medianA.F),
					 autosome=FALSE)
			nuA[, k] <- betas[1, ]
			phiA[, k] <- betas[2, ]
			phiA2[, k] <- betas[3, ]
			betas <- fit.wls(cbind(NN.M, NN.F),
					 sigma=cbind(madB.M, madB.F),
					 allele="B",
					 Y=cbind(medianB.M, medianB.F),
					 autosome=FALSE)
			nuB[, k] <- betas[1, ]
			phiB[, k] <- betas[2, ]
			phiB2[, k] <- betas[3, ]
		}
		if(enough.men & !enough.women){
			betas <- fit.wls(NN.M,##[, c(1,3)],
					 sigma=madA.M,##[, c(1,3)],
					 allele="A",
					 Y=medianA.M,##[, c(1,3)],
					 autosome=FALSE,
					 X=cbind(1, c(1, 0)))
			nuA[, k] <- betas[1, ]
			phiA[, k] <- betas[2, ]
			betas <- fit.wls(NN.M,##[, c(1,3)],
					 sigma=madB.M,#[, c(1,3)],
					 allele="B",
					 Y=medianB.M,#[, c(1,3)],
					 autosome=FALSE,
					 X=cbind(1, c(0, 1)))
			nuB[, k] <- betas[1, ]
			phiB[, k] <- betas[2, ]
		}
		if(!enough.men & enough.women){
			if(any(madA.F == 0)){
				message("Zeros in median absolute deviation.  Not estimating CN for chrX for batch ", names(batches)[k])
				next()
			}
			betas <- fit.wls(NN.F,
					 sigma=madA.F,
					 allele="A",
					 Y=medianA.F,
					 autosome=TRUE) ## can just use the usual design matrix for the women-only analysis
			nuA[, k] <- betas[1, ]
			phiA[, k] <- betas[2, ]
			betas <- fit.wls(NN.F,
					 sigma=madB.F,
					 allele="B",
					 Y=medianB.F,
					 autosome=TRUE) ## can just use the usual design matrix for the women-only analysis
			nuB[, k] <- betas[1, ]
			phiB[, k] <- betas[2, ]
		}
	}
	if(THR.NU.PHI){
		nuA[nuA < MIN.NU] <- MIN.NU
		nuB[nuB < MIN.NU] <- MIN.NU
		phiA[phiA < MIN.PHI] <- MIN.PHI
		phiA2[phiA2 < 1] <- 1
		phiB[phiB < MIN.PHI] <- MIN.PHI
		phiB2[phiB2 < 1] <- 1
	}
	nuA(object)[marker.index, batch.index] <- nuA
	nuB(object)[marker.index, batch.index] <- nuB
	phiA(object)[marker.index, batch.index] <- phiA
	phiB(object)[marker.index, batch.index] <- phiB
	phiPrimeA(object)[marker.index, batch.index] <- phiA2
	phiPrimeB(object)[marker.index, batch.index] <- phiB2
	##
	if(enough.women){
		medianA.AA(object)[marker.index, batch.index] <- do.call("cbind", lapply(medianA.Flist, function(x) x[, 1]))
		medianA.AB(object)[marker.index, batch.index] <- do.call("cbind", lapply(medianA.Flist, function(x) x[, 2]))
		medianA.BB(object)[marker.index, batch.index] <- do.call("cbind", lapply(medianA.Flist, function(x) x[, 3]))
		medianB.AA(object)[marker.index, batch.index] <- do.call("cbind", lapply(medianB.Flist, function(x) x[, 1]))
		medianB.AB(object)[marker.index, batch.index] <- do.call("cbind", lapply(medianB.Flist, function(x) x[, 2]))
		medianB.BB(object)[marker.index, batch.index] <- do.call("cbind", lapply(medianB.Flist, function(x) x[, 3]))
	}
	if(is.lds) {close(object); return(TRUE)} else return(object)
}

##nonpolymorphic markers on X
fit.lm4 <- function(strata,
		    index.list,
		    object,
		    MIN.SAMPLES,
		    THR.NU.PHI,
		    MIN.NU,
		    MIN.PHI,
		    verbose, is.lds=TRUE, ...){
	##if(is.lds) {physical <- get("physical"); open(object)}
	## exclude batches that have fewer than MIN.SAMPLES
	open(object$gender)
	gender <- object$gender[]
	enough.males <- sum(gender==1) >= MIN.SAMPLES
	enough.females <- sum(gender==2) >= MIN.SAMPLES
	if(!enough.males & !enough.females){
		message(paste("fewer than", MIN.SAMPLES, "men and women.  Copy number not estimated for CHR X"))
		return(object)
	}
	excludeBatch <- names(table(batch(object)))[table(batch(object)) < MIN.SAMPLES]
	if(length(excludeBatch) > 0){
		bns <- unique(batch(object))
		batch.index <- which(!bns %in% excludeBatch)
		sample.index <- which(!batch(object)%in% excludeBatch)
	} else {
		sample.index <- 1:ncol(object)
		batch.index <- as.integer(as.factor(unique(batch(object))))
	}
	if(verbose) message("      Probe stratum ", strata, " of ", length(index.list))
	marker.index <- index.list[[strata]]
	batches <- split(seq_along(batch(object)), as.character(batch(object)))
	batches <- batches[batch.index]
	nc <- length(batch.index)
	if(enough.males){
		res <- summarizeMaleXNps(marker.index=marker.index,
					 batches=batches,
					 object=object, MIN.SAMPLES=MIN.SAMPLES)
		medianA.AA.M <- res[["medianA.AA"]][, batch.index, drop=FALSE]
		madA.AA.M <- res[["madA.AA"]][, batch.index, drop=FALSE]
	}
	medianA.AA.F <- as.matrix(medianA.AA(object)[marker.index, batch.index, drop=FALSE]) ## median for women
	madA.AA.F <- as.matrix(madA.AA(object)[marker.index, batch.index, drop=FALSE]) ## median for women
	split.gender <- split(gender[sample.index], as.character(batch(object)[sample.index]))
	N.M <- sapply(split.gender, function(x) sum(x==1))
	N.F <- sapply(split.gender, function(x) sum(x==2))
	nuA <- as.matrix(nuA(object)[marker.index, batch.index, drop=FALSE])
	nuB <- as.matrix(nuB(object)[marker.index, batch.index, drop=FALSE])
	phiA <- as.matrix(phiA(object)[marker.index, batch.index, drop=FALSE])
	phiB <- as.matrix(phiB(object)[marker.index, batch.index, drop=FALSE])
	ii <- isSnp(object) & chromosome(object) < 23 & !is.na(chromosome(object))
	fns <- featureNames(object)[ii]
	flags <- as.matrix(flags(object)[ii, batch.index])
	fns.noflags <- fns[rowSums(flags, na.rm=T) == 0]
	if(length(fns.noflags) == 0){
		warning("All features in one of the probeset batches were flagged. The data is processed in probeset batches to reduce RAM, but the error suggests that increasing the size of the batches may help.  For example, ocProbesets(4*ocProbesets()) would increase the current size of the probeset batches by 4.")
		fns.noflags <- featureNames(object)[marker.index] ## allow processing to continue
	}
	index.flag <- match(fns.noflags, featureNames(object))
	snp.index <- sample(index.flag, min(c(length(index.flag), 10000)))
	##
	N.AA <- as.matrix(N.AA(object)[snp.index, batch.index, drop=FALSE])
	N.AB <- as.matrix(N.AB(object)[snp.index, batch.index, drop=FALSE])
	N.BB <- as.matrix(N.BB(object)[snp.index, batch.index, drop=FALSE])
	enoughAA <- rowSums(N.AA < 5) == 0
	enoughAB <- rowSums(N.AB < 5) == 0
	enoughBB <- rowSums(N.BB < 5) == 0
	snp.index <- snp.index[enoughAA & enoughAB & enoughBB]
	if(length(snp.index) < 100){
		message("too few snps pass criteria for estimating parameters for NP markers on chr X")
		return(object)
	}
	nuA.snp <- as.matrix(nuA(object)[snp.index, batch.index, drop=FALSE])
	nuA.snp.notmissing <- rowSums(is.na(nuA.snp)) == 0
	nuA.snp.notnegative <- rowSums(as.matrix(nuA(object)[snp.index, batch.index, drop=FALSE]) < 20) == 0
	snp.index <- snp.index[nuA.snp.notmissing & nuA.snp.notnegative]
	medianA.AA.snp <- as.matrix(medianA.AA(object)[snp.index, batch.index])
	medianB.BB.snp <- as.matrix(medianB.BB(object)[snp.index, batch.index])
	nuA.snp <- as.matrix(nuA(object)[snp.index, batch.index])
	nuB.snp <- as.matrix(nuB(object)[snp.index, batch.index])
	phiA.snp <- as.matrix(phiA(object)[snp.index, batch.index])
	phiB.snp <- as.matrix(phiB(object)[snp.index, batch.index])
##	pseudoAR <- position(object)[snp.index] < 2709520 | (position(object)[snp.index] > 154584237 & position(object)[snp.index] < 154913754)
##	pseudoAR[is.na(pseudoAR)] <- FALSE
	for(k in seq_along(batches)){
		B <- batches[[k]]
		this.batch <- unique(as.character(batch(object)[B]))
		gender <- object$gender[B]
		enough.men <- N.M[[this.batch]] >= MIN.SAMPLES
		enough.women <- N.F[[this.batch]] >= MIN.SAMPLES
		if(!enough.men & !enough.women) {
			if(verbose) message(paste("fewer than", MIN.SAMPLES, "men and women in batch", this.batch, ". CHR X copy number not available. "))
			next()
		}
		tmp <- cbind(medianA.AA.snp[, k], medianB.BB.snp[,k], phiA.snp[, k], phiB.snp[, k])
		tmp <- tmp[rowSums(is.na(tmp) == 0) & rowSums(tmp < 20) == 0, ]
		if(nrow(tmp) < 100){
			stop("too few markers for estimating nonpolymorphic CN on chromosome X")
		}
		X <- cbind(1, log2(c(tmp[, 1], tmp[, 2])))
		Y <- log2(c(tmp[, 3], tmp[, 4]))
		notmissing.index <- which(rowSums(is.na(X)) == 0 & !is.na(Y))
		X <- X[notmissing.index, ]
		Y <- Y[notmissing.index]
		betahat <- solve(crossprod(X), crossprod(X, Y))
		if(enough.men){
			X.men <- cbind(1, log2(medianA.AA.M[, k]))
			Yhat1 <- as.numeric(X.men %*% betahat)
			## put intercept and slope for men in nuB and phiB
			phiB[, k] <- 2^(Yhat1)
			nuB[, k] <- medianA.AA.M[, k] - 1*phiB[, k]
		}
		X.fem <- cbind(1, log2(medianA.AA.F[, k]))
		Yhat2 <- as.numeric(X.fem %*% betahat)
		phiA[, k] <- 2^(Yhat2)
		nuA[, k] <- medianA.AA.F[, k] - 2*phiA[, k]
	}
	if(THR.NU.PHI){
		nuA[nuA < MIN.NU] <- MIN.NU
		phiA[phiA < MIN.PHI] <- MIN.PHI
		nuB[nuB < MIN.NU] <- MIN.NU
		phiB[phiB < MIN.PHI] <- MIN.PHI
	}
	nuA(object)[marker.index, batch.index] <- nuA
	phiA(object)[marker.index, batch.index] <- phiA
	nuB(object)[marker.index, batch.index] <- nuB
	phiB(object)[marker.index, batch.index] <- phiB
	##if(is.lds) {close(object); return(TRUE)} else return(object)
	if(is.lds) {close(object); return(TRUE)} else return(object)
}

whichPlatform <- function(cdfName){
	index <- grep("genomewidesnp", cdfName)
	if(length(index) > 0){
		platform <- "affymetrix"
	} else{
		index <- grep("human", cdfName)
		platform <- "illumina"
	}
	return(platform)
}

cnrma2 <- function(A, filenames, row.names, verbose=TRUE, seed=1, cdfName, sns){

}

imputeCenter <- function(muA, muB, index.complete, index.missing){
	index <- index.missing
	mnA <- muA
	mnB <- muB
	if(length(index.complete) >= 100){
		for(j in 1:3){
			if(length(index[[j]]) == 0) next()
			X <- cbind(1, mnA[index.complete,  -j, drop=FALSE], mnB[index.complete,  -j, drop=FALSE])
			Y <- cbind(mnA[index.complete, j], mnB[index.complete,  j])
			betahat <- solve(crossprod(X), crossprod(X,Y))
			X <- cbind(1, mnA[index[[j]],  -j, drop=FALSE],  mnB[index[[j]],  -j, drop=FALSE])
			mus <- X %*% betahat
			## threshold imputed intensities that are very small
			mus[mus < 64] <- 64
			muA[index[[j]], j] <- mus[, 1]
			muB[index[[j]], j] <- mus[, 2]
		}
	}
	results <- list(muA, muB)
	return(results)
}

indexComplete <- function(NN, medianA, medianB, MIN.OBS){
	Nindex <- which(rowSums(NN > MIN.OBS) == ncol(NN))
	correct.order <- which(medianA[, 1] > medianA[, ncol(medianA)] & medianB[, ncol(medianB)] > medianB[, 1])
	index.complete <- intersect(Nindex, correct.order)
	size <- min(5000, length(index.complete))
	if(size == 5000) index.complete <- sample(index.complete, 5000, replace=TRUE)
##	if(length(index.complete) < 100){
##		stop("fewer than 100 snps pass criteria for imputing unobserved genotype location/scale")
##	}
	return(index.complete)
}

imputeCentersForMonomorphicSnps <- function(medianA, medianB, index.complete, unobserved.index){
	cols <- c(3, 1, 2)
	if(length(index.complete) < 100){
		##message("index.complete less than 100.  No imputation")
		results <- list(medianA=medianA, medianB=medianB)
		return(results)
	}
	for(j in 1:3){
		if(length(unobserved.index[[j]]) == 0) next()
		kk <- cols[j]
		X <- cbind(1, medianA[index.complete, kk], medianB[index.complete, kk])
		Y <- cbind(medianA[index.complete,  -kk],
			   medianB[index.complete,  -kk])
		betahat <- solve(crossprod(X), crossprod(X,Y))
		X <- cbind(1, medianA[unobserved.index[[j]],  kk], medianB[unobserved.index[[j]],  kk])
		mus <- X %*% betahat
		medianA[unobserved.index[[j]], -kk] <- mus[, 1:2]
		medianB[unobserved.index[[j]], -kk] <- mus[, 3:4]
	}
	results <- list(medianA=medianA, medianB=medianB)
	return(results)
}


imputeCenterX <- function(muA, muB, Ns, index.complete, MIN.OBS){
	##index1 <- which(Ns[, 1] == 0 & Ns[, 3] > MIN.OBS)
	index1 <- which(Ns[, 1] == 0 & Ns[, 2] > MIN.OBS)
	if(length(index1) > 0){
		##X <- cbind(1, muA[index.complete, 3], muB[index.complete, 3])
		X <- cbind(1, muA[index.complete, 2], muB[index.complete, 2])
		Y <- cbind(1, muA[index.complete, 1], muB[index.complete, 1])
##		X <- cbind(1, muA[index.complete[[1]], 3], muB[index.complete[[1]], 3])
##		Y <- cbind(1, muA[index.complete[[1]], 1], muB[index.complete[[1]], 1])
		betahat <- solve(crossprod(X), crossprod(X,Y))
		##now with the incomplete SNPs
		##X <- cbind(1, muA[index1, 3], muB[index1, 3])
		X <- cbind(1, muA[index1, 2], muB[index1, 2])
		mus <- X %*% betahat
		muA[index1, 1] <- mus[, 2]
		muB[index1, 1] <- mus[, 3]
	}
	index1 <- which(Ns[, 2] == 0)
	if(length(index1) > 0){
		X <- cbind(1, muA[index.complete, 1], muB[index.complete, 1])
		Y <- cbind(1, muA[index.complete, 2], muB[index.complete, 2])
##		X <- cbind(1, muA[index.complete[[2]], 1], muB[index.complete[[2]], 1])
##		Y <- cbind(1, muA[index.complete[[2]], 3], muB[index.complete[[2]], 3])
		betahat <- solve(crossprod(X), crossprod(X,Y))
		##now with the incomplete SNPs
		X <- cbind(1, muA[index1, 1], muB[index1, 1])
		mus <- X %*% betahat
		muA[index1, 2] <- mus[, 2]
		muB[index1, 2] <- mus[, 3]
	}
	list(muA, muB)
}


## constructors for Illumina platform
constructIlluminaFeatureData <- function(gns, cdfName){
	pkgname <- paste(cdfName, "Crlmm", sep="")
	path <- system.file("extdata", package=pkgname)
	load(file.path(path, "cnProbes.rda"))
	load(file.path(path, "snpProbes.rda"))
	cnProbes$chr <- chromosome2integer(cnProbes$chr)
	cnProbes <- as.matrix(cnProbes)
	snpProbes$chr <- chromosome2integer(snpProbes$chr)
	snpProbes <- as.matrix(snpProbes)
	mapping <- rbind(snpProbes, cnProbes, deparse.level=0)
	mapping <- mapping[match(gns, rownames(mapping)), ]
	isSnp <- 1L-as.integer(gns %in% rownames(cnProbes))
	mapping <- cbind(mapping, isSnp, deparse.level=0)
	stopifnot(identical(rownames(mapping), gns))
	colnames(mapping) <- c("chromosome", "position", "isSnp")
	new("AnnotatedDataFrame",
	    data=data.frame(mapping),
	    varMetadata=data.frame(labelDescription=colnames(mapping)))
}


constructIlluminaAssayData <- function(np, snp, object, storage.mode="environment", nr){
	stopifnot(identical(snp$gns, featureNames(object)))
	gns <- c(featureNames(object), np$gns)
	sns <- np$sns
	np <- np[1:2]
	snp <- snp[1:2]
	stripnames <- function(x) {
		dimnames(x) <- NULL
		x
	}
	np <- lapply(np, stripnames)
	snp <- lapply(snp, stripnames)
	is.ff <- is(snp[[1]], "ff") | is(snp[[1]], "ffdf")
	if(is.ff){
		lapply(snp, open)
		open(calls(object))
		open(snpCallProbability(object))
##		lapply(np, open)
	}
	##tmp <- rbind(as.matrix(snp[[1]]), as.matrix(np[[1]]), deparse.level=0)
	A.snp <- snp[[1]]
	B.snp <- snp[[2]]
	##Why is np not a ff object?
	A.np <- np[[1]]
	B.np <- np[[2]]
	nc <- ncol(object)
	if(is.ff){
		NA.vec <- rep(NA, nrow(A.np))
		AA <- initializeBigMatrix("A", nr, nc, vmode="integer")
		BB <- initializeBigMatrix("B", nr, nc, vmode="integer")
		GG <- initializeBigMatrix("calls", nr, nc, vmode="integer")
		PP <- initializeBigMatrix("confs", nr, nc, vmode="integer")
		for(j in 1:ncol(object)){
			AA[, j] <- c(snp[[1]][, j], np[[1]][, j])
			BB[, j] <- c(snp[[2]][, j], np[[2]][, j])
			GG[, j] <- c(calls(object)[, j], NA.vec)
			PP[, j] <- c(snpCallProbability(object)[, j], NA.vec)

		}
	} else {
		AA <- rbind(snp[[1]], np[[1]], deparse.level=0)
		BB <- rbind(snp[[2]], np[[2]], deparse.level=0)
		gt <- stripnames(calls(object))
		emptyMatrix <- matrix(integer(), nrow(np[[1]]), ncol(A))
		GG <- rbind(gt, emptyMatrix, deparse.level=0)
		pr <- stripnames(snpCallProbability(object))
		PP <- rbind(pr, emptyMatrix, deparse.level=0)
	}
	assayDataNew(storage.mode,
		     alleleA=AA,
		     alleleB=BB,
		     call=GG,
		     callProbability=PP)
}
constructIlluminaCNSet <- function(crlmmResult,
				   path,
				   snpFile,
				   cnFile){
	load(file.path(path, "snpFile.rda"))
	res <- get("res")
	load(file.path(path, "cnFile.rda"))
	cnAB <- get("cnAB")
	fD <- constructIlluminaFeatureData(c(res$gns, cnAB$gns), cdfName="human370v1c")
	##new.order <- order(fD$chromosome, fD$position)
	##fD <- fD[new.order, ]
	aD <- constructIlluminaAssayData(cnAB, res, crlmmResult, nr=nrow(fD))
	##protocolData(crlmmResult)$batch <- vector("integer", ncol(crlmmResult))
	new("CNSet",
	    call=aD[["call"]],
	    callProbability=aD[["callProbability"]],
	    alleleA=aD[["alleleA"]],
	    alleleB=aD[["alleleB"]],
	    phenoData=phenoData(crlmmResult),
	    protocolData=protocolData(crlmmResult),
	    featureData=fD,
	    batch=batch,
	    annotation="human370v1c")

}



ellipseCenters <- function(object, index, allele, batch, log.it=TRUE){
	ubatch <- unique(batch(object))[batch]
	Nu <- nu(object, allele)[index, batch]
	Phi <- phi(object, allele)[index, batch]
	centers <- list(Nu, Nu+Phi, Nu+2*Phi)
	if(log.it)
		centers <- lapply(centers, log2)
	myLabels <- function(allele){
		switch(allele,
		       A=c("BB", "AB", "AA"),
		       B=c("AA", "AB", "BB"),
		       stop("allele must be 'A' or 'B'")
		       )
	}
	nms <- myLabels(allele)
	names(centers) <- nms
	fns <- featureNames(object)[index]
	centers$fns <- fns
	return(centers)
}


shrinkSummary <- function(object,
			  type="SNP",
			  MIN.OBS=1,
			  MIN.SAMPLES=10,
			  DF.PRIOR=50,
			  verbose=TRUE,
			  marker.index){
	stopifnot(type[[1]] == "SNP")
	CHR.X <- FALSE ## this is no longer needed
	is.lds <- is(calls(object), "ffdf") | is(calls(object), "ff_matrix")
	if(is.lds) stopifnot(isPackageLoaded("ff"))
	if(missing(marker.index)){
		batch <- batch(object)
		is.snp <- isSnp(object)
		is.autosome <- chromosome(object) < 23
		is.annotated <- !is.na(chromosome(object))
		is.X <- chromosome(object) == 23
		marker.index <- whichMarkers(type[[1]], is.snp, is.autosome, is.annotated, is.X)
	}
	if(is.lds){
		index.list <- splitIndicesByLength(marker.index, ocProbesets())
##		if(parStatus()){
##			index.list <- splitIndicesByNode(marker.index)
##		} else index.list <- splitIndicesByLength(marker.index, ocProbesets())
		ocLapply(seq(along=index.list),
			 shrinkGenotypeSummaries,
			 index.list=index.list,
			 object=object,
			 verbose=verbose,
			 MIN.OBS=MIN.OBS,
			 MIN.SAMPLES=MIN.SAMPLES,
			 DF.PRIOR=DF.PRIOR,
			 is.lds=is.lds,
			 neededPkgs="crlmm")
	} else {
		object <- shrinkGenotypeSummaries(strata=1,
			      index.list=list(marker.index),
			      object=object,
			      verbose=verbose,
			      MIN.OBS=MIN.OBS,
			      MIN.SAMPLES=MIN.SAMPLES,
			      DF.PRIOR=DF.PRIOR,
			      is.lds=is.lds)
	}
	if(is.lds) return(TRUE) else return(object)
}

genotypeSummary <- function(object,
			    GT.CONF.THR=0.80,
			    type=c("SNP", "NP", "X.SNP", "X.NP"), ##"X.snps", "X.nps"),
			    verbose=TRUE,
			    marker.index){
	if(type == "X.SNP" | type=="X.NP"){
		CHR.X <- TRUE
	} else CHR.X <- FALSE
	is.lds <- is(calls(object), "ffdf") | is(calls(object), "ff_matrix")
	if(is.lds) stopifnot(isPackageLoaded("ff"))
	if(missing(marker.index)){
		batch <- batch(object)
		is.snp <- isSnp(object)
		is.autosome <- chromosome(object) < 23
		is.annotated <- !is.na(chromosome(object))
		is.X <- chromosome(object) == 23
		marker.index <- whichMarkers(type[[1]], is.snp, is.autosome, is.annotated, is.X)
	}
	summaryFxn <- function(type){
		switch(type,
		       SNP="summarizeSnps",
		       NP="summarizeNps",
		       X.NP="summarizeNps")
	}
	myf <- summaryFxn(type[[1]])
	FUN <- get(myf)
	if(is.lds){
		index.list <- splitIndicesByLength(marker.index, ocProbesets())
		ocLapply(seq(along=index.list),
			 FUN,
			 index.list=index.list,
			 object=object,
			 batchSize=ocProbesets(),
			 GT.CONF.THR=GT.CONF.THR,
			 verbose=verbose,
			 CHR.X=CHR.X,
			 neededPkgs="crlmm")
	} else{
		object <- FUN(strata=1,
			      index.list=list(marker.index),
			      object=object,
			      batchSize=ocProbesets(),
			      GT.CONF.THR=GT.CONF.THR,
			      verbose=verbose,
			      CHR.X=CHR.X)
	}
	if(is.lds) return(TRUE) else return(object)
}

whichMarkers <- function(type, is.snp, is.autosome, is.annotated, is.X){
	switch(type,
	       SNP=which(is.snp & is.autosome & is.annotated),
	       NP=which(!is.snp & is.autosome),
	       X.SNP=which(is.snp & is.X),
	       X.NP=which(!is.snp & is.X),
	       stop("'type' must be one of 'SNP', 'NP', 'X.SNP', or 'X.NP'")
	       )
}

summarizeNps <- function(strata, index.list, object, batchSize,
			 GT.CONF.THR, verbose=TRUE, CHR.X=FALSE, ...){
	is.lds <- is(calls(object), "ffdf") | is(calls(object), "ff_matrix")
	if(is.lds) {physical <- get("physical"); open(object)}
	if(verbose) message("      Probe stratum ", strata, " of ", length(index.list))
	index <- index.list[[strata]]
##	if(CHR.X) {
##		sample.index <- which(object$gender==2)
##		batches <- split(sample.index, as.character(batch(object))[sample.index])
##	} else {
##		batches <- split(seq_along(batch(object)), as.character(batch(object)))
##	}
	batches <- split(seq_along(batch(object)), as.character(batch(object)))
	batchnames <- batchNames(object)
	nr <- length(index)
	nc <- length(batchnames)
	N.AA <- medianA.AA <- madA.AA <- tau2A.AA <- matrix(NA, nr, nc)
	is.illumina <- whichPlatform(annotation(object)) == "illumina"
	open(A(object))
	open(object$gender)
	AA <- as.matrix(A(object)[index, ])
	if(is.illumina){
		BB <- as.matrix(B(object)[index, ])
		AVG <- (AA+BB)/2
		A(object)[index, ] <- AVG
		AA <- AVG
		rm(AVG, BB)
	}
	for(k in seq_along(batches)){
		sample.index <- batches[[k]]
		N.AA[, k] <- length(sample.index)
		if(CHR.X){
			gender <- object$gender[sample.index]
			sample.index <- sample.index[gender == 2]
			if(length(sample.index) == 0) next()
		}
		this.batch <- unique(as.character(batch(object)[sample.index]))
		j <- match(this.batch, batchnames)
		I.A <- AA[, sample.index, drop=FALSE]
		medianA.AA[, k] <- rowMedians(I.A, na.rm=TRUE)
		madA.AA[, k] <- rowMAD(I.A, na.rm=TRUE)
		## log2 Transform Intensities
		I.A <- log2(I.A)
		tau2A.AA[, k] <- rowMAD(I.A, na.rm=TRUE)^2
	}
	N.AA(object)[index,] <- N.AA
	medianA.AA(object)[index,] <- medianA.AA
	madA.AA(object)[index, ] <- madA.AA
	tau2A.AA(object)[index, ] <- tau2A.AA
	if(is.lds) return(TRUE) else return(object)
}

summarizeSnps <- function(strata,
			  index.list,
			  object,
			  GT.CONF.THR=0.80,
			  verbose=TRUE,
			  CHR.X=FALSE, ...){
##	if(is.lds) {
##		physical <- get("physical")
##		open(object)
##	}
	is.lds <- is(calls(object), "ffdf") | is(calls(object), "ff_matrix")
	if(is.lds) stopifnot(isPackageLoaded("ff"))
	if(verbose) message("      Probe stratum ", strata, " of ", length(index.list))
	index <- index.list[[strata]]
	batches <- split(seq_along(batch(object)), as.character(batch(object)))
	batchnames <- batchNames(object)
	nr <- length(index)
	nc <- length(batchnames)
	statsA.AA <- statsA.AB <- statsA.BB <- statsB.AA <- statsB.AB <- statsB.BB <- vector("list", nc)
	corrAA <- corrAB <- corrBB <- tau2A.AA <- tau2A.BB <- tau2B.AA <- tau2B.BB <- matrix(NA, nr, nc)
	Ns.AA <- Ns.AB <- Ns.BB <- matrix(NA, nr, nc)
	if(verbose) message("        Begin reading...")
	GG <- as.matrix(calls(object)[index, ])
	CP <- as.matrix(snpCallProbability(object)[index, ])
	AA <- as.matrix(A(object)[index, ])
	BB <- as.matrix(B(object)[index, ])
	FL <- as.matrix(flags(object)[index, ])
	summaryStats <- function(X, INT, FUNS){
		tmp <- matrix(NA, nrow(X), length(FUNS))
		for(j in seq_along(FUNS)){
			FUN <- match.fun(FUNS[j])
			tmp[, j] <- FUN(X*INT, na.rm=TRUE)
		}
		tmp
	}
	if(verbose) message("        Computing summaries...")
	for(k in seq_along(batches)){
		##note that the genotype frequency for AA would include 'A' on chromosome X for men
		sample.index <- batches[[k]]
		this.batch <- unique(as.character(batch(object)[sample.index]))
		j <- match(this.batch, batchnames)
		G <- GG[, sample.index, drop=FALSE]
		xx <- CP[, sample.index, drop=FALSE]
		highConf <- (1-exp(-xx/1000)) > GT.CONF.THR
		## Some markers may have genotype confidences scores that are ALL below the threshold
		## For these markers, provide statistical summaries based on all the samples and flag
		## Provide summaries for these markers and flag to indicate the problem
		## RS: check whether we need the next to lines and, if so, effect downstream
		ii <- which(rowSums(highConf) == 0)
		if(length(ii) > 0) highConf[ii, ] <- TRUE
		G <- G*highConf
		## table(rowSums(G==0))
		##G <- G*highConf*NORM
		A <- AA[, sample.index, drop=FALSE]
		B <- BB[, sample.index, drop=FALSE]
		## this can be time consuming...do only once
		G.AA <- G==1
		G.AA[G.AA==FALSE] <- NA
		G.AB <- G==2
		G.AB[G.AB==FALSE] <- NA
		G.BB <- G==3
		G.BB[G.BB==FALSE] <- NA
		Ns.AA[, k] <- rowSums(G.AA, na.rm=TRUE)
		Ns.AB[, k] <- rowSums(G.AB, na.rm=TRUE)
		Ns.BB[, k] <- rowSums(G.BB, na.rm=TRUE)
		if(CHR.X){
			gender <- object$gender[sample.index]
			sample.index <- sample.index[gender == 2]
			if(length(sample.index) == 0) next()
		}
		stats <- summaryStats(G.AA, A, FUNS=c("rowMedians", "rowMAD"))
		medianA.AA(object)[index, k] <- stats[, 1]
		##missing.index <- which(rowSums(is.na(medianA.AA(object)[index, ,drop=FALSE])) > 0)
		madA.AA(object)[index, k] <- stats[, 2]
		stats <- summaryStats(G.AB, A, FUNS=c("rowMedians", "rowMAD"))
		medianA.AB(object)[index, k] <- stats[, 1]
		madA.AB(object)[index, k] <- stats[, 2]
		stats <- summaryStats(G.BB, A, FUNS=c("rowMedians", "rowMAD"))
		medianA.BB(object)[index, k] <- stats[, 1]
		madA.BB(object)[index, k] <- stats[, 2]
		stats <- summaryStats(G.AA, B, FUNS=c("rowMedians", "rowMAD"))
		medianB.AA(object)[index, k] <- stats[, 1]
		madB.AA(object)[index, k] <- stats[, 2]
		stats <- summaryStats(G.AB, B, FUNS=c("rowMedians", "rowMAD"))
		medianB.AB(object)[index, k] <- stats[, 1]
		madB.AB(object)[index, k] <- stats[, 2]
		stats <- summaryStats(G.BB, B, FUNS=c("rowMedians", "rowMAD"))
		medianB.BB(object)[index, k] <- stats[, 1]
		madB.BB(object)[index, k] <- stats[, 2]
		## log2 Transform Intensities
		A <- log2(A); B <- log2(B)
		tau2A.AA[, k] <- summaryStats(G.AA, A, FUNS="rowMAD")^2
		tau2A.BB[, k] <- summaryStats(G.BB, A, FUNS="rowMAD")^2
		tau2B.AA[, k] <- summaryStats(G.AA, B, FUNS="rowMAD")^2
		tau2B.BB[, k] <- summaryStats(G.BB, B, FUNS="rowMAD")^2
		corrAA[, k] <- rowCors(A*G.AA, B*G.AA, na.rm=TRUE)
		corrAB[, k] <- rowCors(A*G.AB, B*G.AB, na.rm=TRUE)
		corrBB[, k] <- rowCors(A*G.BB, B*G.BB, na.rm=TRUE)
		rm(A, B, G.AA, G.AB, G.BB, xx, highConf, G)
		gc()
	}
	##---------------------------------------------------------------------------
	## grand mean
	##---------------------------------------------------------------------------
	if(FALSE){ ## no implemented
		if(length(batches) > 1 && "grandMean" %in% batchNames(object)){
			k <- match("grandMean", batchNames(object))
			if(verbose) message("        computing grand means...")
			highConf <- (1-exp(-CP/1000)) > GT.CONF.THR
			rm(CP); gc()
			## Some markers may have genotype confidences scores that are ALL below the threshold
			## For these markers, provide statistical summaries based on all the samples and flag
			## Provide summaries for these markers and flag to indicate the problem
			ii <- which(rowSums(highConf) == 0)
			if(length(ii) > 0) highConf[ii, ] <- TRUE
			GG <- GG*highConf
			rm(highConf); gc()
			Ns <- list(Ns.AA, Ns.AB, Ns.BB)
			rm(Ns.AA, Ns.AB, Ns.BB) ; gc()
			I <- list(AA, BB)
			lI <- lapply(I, log2)
			cors <- list(corrAA, corrAB, corrBB)
			rm(corrAA, corrAB, corrBB); gc()
			rm(AA,BB); gc()
			tau2 <- list(AA=list(tau2A.AA,
				     tau2B.AA),
				     AB=list(NULL, NULL),
				     BB=list(tau2A.BB,
				     tau2B.BB))
			rm(tau2A.AA, tau2B.AA, tau2A.BB, tau2B.BB); gc()
			for(i in 1:3){
				G.call <- isCall(GG, call=i)
				for(j in 1:2){
					computeSummary(object,
						       G.call,
						       call=i,
						       I=I[[j]],
						       allele=c("A", "B")[j],
						       Ns=Ns[[i]],
						       call=i,
						       tau2=tau2[[i]][[j]],
						       index=index)
				}
				updateCors(cors[[i]], G.call, lI)
			}
			##I <- lapply(I, log2)
			##AA <- log2(AA)
			##BB <- log2(BB)
			##		tau2A.AA[, k] <- summaryStats(G.AA, AA, FUNS="rowMAD")^2
			##		tau2A.BB[, k] <- summaryStats(G.BB, AA, FUNS="rowMAD")^2
			##tau2B.AA[, k] <- summaryStats(G.AA, BB, FUNS="rowMAD")^2
			##tau2B.BB[, k] <- summaryStats(G.BB, BB, FUNS="rowMAD")^2
			##		corrAA[, k] <- rowCors(AA*G.AA, BB*G.AA, na.rm=TRUE)
			##		corrAB[, k] <- rowCors(AA*G.AB, BB*G.AB, na.rm=TRUE)
			##		corrBB[, k] <- rowCors(AA*G.BB, BB*G.BB, na.rm=TRUE)
			##		rm(AA, BB); gc()
			##
			## TODO:   fill in NAs -- use code from shrinkGenotypeSummaries
			##
			##		rm(GG, CP, AA, BB, FL, stats)
			##		gc()
			##		G.AA <- GG==1
			##		G.AA[G.AA==FALSE] <- NA
			##		G.AB <- GG==2
			##		G.AB[G.AB==FALSE] <- NA
			##		G.BB <- GG==3
			##		G.BB[G.BB==FALSE] <- NA
			##		Ns.AA[, k] <- rowSums(G.AA, na.rm=TRUE)
			##		Ns.AB[, k] <- rowSums(G.AB, na.rm=TRUE)
			##		Ns.BB[, k] <- rowSums(G.BB, na.rm=TRUE)
			##		N.AA(object)[index,] <- Ns.AA
			##		N.AB(object)[index,] <- Ns.AB
			##		N.BB(object)[index,] <- Ns.BB

			## Calculate row medians and MADs
			##		medianA.AA(object)[index, k] <- stats[, 1]
			##		madA.AA(object)[index, k] <- stats[, 2]
			##		stats <- summaryStats(G.AB, AA, FUNS=c("rowMedians", "rowMAD"))
			##		medianA.AB(object)[index, k] <- stats[, 1]
			##		madA.AB(object)[index, k] <- stats[, 2]
			##		stats <- summaryStats(G.BB, AA, FUNS=c("rowMedians", "rowMAD"))
			##		medianA.BB(object)[index, k] <- stats[, 1]
			##		madA.BB(object)[index, k] <- stats[, 2]
			##		stats <- summaryStats(G.AA, BB, FUNS=c("rowMedians", "rowMAD"))
			##		medianB.AA(object)[index, k] <- stats[, 1]
			##		madB.AA(object)[index, k] <- stats[, 2]
			##		stats <- summaryStats(G.AB, BB, FUNS=c("rowMedians", "rowMAD"))
			##		medianB.AB(object)[index, k] <- stats[, 1]
			##		madB.AB(object)[index, k] <- stats[, 2]
			##		stats <- summaryStats(G.BB, BB, FUNS=c("rowMedians", "rowMAD"))
			##		medianB.BB(object)[index, k] <- stats[, 1]
			##		madB.BB(object)[index, k] <- stats[, 2]
			## log2 Transform Intensities
		}
	}
	##	if(verbose) message("        Begin writing...")
	N.AA(object)[index,] <- Ns.AA
	N.AB(object)[index,] <- Ns.AB
	N.BB(object)[index,] <- Ns.BB
	corrAA(object)[index,] <- corrAA
	corrAB(object)[index,] <- corrAB
	corrBB(object)[index,] <- corrBB
	tau2A.AA(object)[index, ] <- tau2A.AA
	tau2A.BB(object)[index, ] <- tau2A.BB
	tau2B.AA(object)[index, ] <- tau2B.AA
	tau2B.BB(object)[index, ] <- tau2B.BB
	if(is.lds) return(TRUE) else return(object)
}

isCall <- function(G, call){
	G.call <- G==call
	G.call[G.call==FALSE] <- NA
	G.call
}

##computeSummary <- function(object, G.call, call, I, allele, Ns, index){
##	k <- match("grandMean", batchNames(object))
##	stats <- summaryStats(G.call, I, FUNS=c("rowMedians", "rowMAD"))
##	Ns[, k] <- rowSums(G.call, na.rm=TRUE)
##	updateStats(stats, Ns, object, call, allele, index)
##	gc()
##	return()
##}

##updateTau <- function(object, tau2, G.call, call, I, allele, index){
##	k <- match("grandMean", batchNames(object))
##	logI <- log2(I)
##	rm(I); gc()
##	tau2[, k] <- summaryStats(G.call, logI, FUNS="rowMAD")^2
##	if(call==1 & allele=="A"){
##		tau2A.AA(object)[index, ] <- tau2
##	}
##	if(call==1 & allele=="B"){
##		tau2B.AA(object)[index, ] <- tau2
##	}
##	if(call==3 & allele=="A"){
##		tau2A.BB(object)[index, ] <- tau2
##	}
##	if(call==3 & allele=="B"){
##		tau2B.BB(object)[index, ] <- tau2
##	}
##	NULL
##}

##updateCors <- function(cors, G.call, I){
##	k <- match("grandMean", batchNames(object))
##	cors[, k] <- rowCors(I[[1]]*G.call, I[[2]]*G.call, na.rm=TRUE)
##	if(call==1){
##		corrAA(object)[index, ] <- cors
##	}
##	if(call==2){
##		corrAB(object)[index, ] <- cors
##	}
##	if(call==3){
##		corrBB(object)[index, ] <- cors
##	}
##}

##updateStats <- function(stats, Ns, object, call, allele, tau2, index){
##	if(call==1){
##		Ns.AA(object)[index, ] <- Ns
##		if(allele=="A"){
##			medianA.AA(object)[index, k] <- stats[, 1]
##			madA.AA(object)[index, k] <- stats[, 2]
##			updateTau(object, tau2, G.call, call, I, allele, index)
##		} else {
##			medianB.AA(object)[index, k] <- stats[, 1]
##			madB.AA(object)[index, k] <- stats[, 2]
##			updateTau(object, tau2, G.call, call, I, allele, index)
##		}
##	}
##	if(call==2){
##		Ns.AB(object)[index, ] <- Ns
##		if(allele=="A"){
##			medianA.AB(object)[index, k] <- stats[, 1]
##			madA.AB(object)[index, k] <- stats[, 2]
##		} else {
##			medianB.AB(object)[index, k] <- stats[, 1]
##			madB.AB(object)[index, k] <- stats[, 2]
##		}
##	}
##	if(call==3){
##		Ns.BB(object)[index, ] <- Ns
##		if(allele=="A"){
##			medianA.BB(object)[index, k] <- stats[, 1]
##			madA.BB(object)[index, k] <- stats[, 2]
##			updateTau(object, tau2, G.call, call, I, allele, index)
##		} else {
##			medianB.BB(object)[index, k] <- stats[, 1]
##			madB.BB(object)[index, k] <- stats[, 2]
##			updateTau(object, tau2, G.call, call, I, allele, index)
##		}
##	}
##}

crlmmCopynumber <- function(object,
			    MIN.SAMPLES=10,
			    SNRMin=5,
			    MIN.OBS=1,
			    DF.PRIOR=50,
			    bias.adj=FALSE,
			    prior.prob=rep(1/4,4),
			    seed=1,
			    verbose=TRUE,
			    GT.CONF.THR=0.80,
			    MIN.NU=2^3,
			    MIN.PHI=2^3,
			    THR.NU.PHI=TRUE,
			    type=c("SNP", "NP", "X.SNP", "X.NP"),
			    fit.linearModel=TRUE){
	typeof <- paste(type, collapse=",")
	stopifnot(typeof %in% c("SNP", "NP", "SNP,NP", "SNP,X.SNP", "SNP,X.NP", "SNP,NP,X.SNP", "SNP,NP,X.SNP,X.NP"))
	if(GT.CONF.THR >= 1 | GT.CONF.THR < 0) stop("GT.CONF.THR must be in [0,1)")
	batch <- batch(object)
	is.snp <- isSnp(object)
	is.autosome <- chromosome(object) < 23
	is.annotated <- !is.na(chromosome(object))
	is.X <- chromosome(object) == 23
	is.lds <- is(calls(object), "ffdf") | is(calls(object), "ff_matrix")
	if(is.lds) stopifnot(isPackageLoaded("ff"))
	samplesPerBatch <- table(as.character(batch(object)))
	open(object)
	if(any(samplesPerBatch < MIN.SAMPLES)){
		tmp <- paste(names(samplesPerBatch)[samplesPerBatch < MIN.SAMPLES], collapse=", ")
		message("The following batches have fewer than ", MIN.SAMPLES, " samples: ",  tmp)
		message("Not estimating copy number parameters for batch ", tmp)
	}
	mylabel <- function(marker.type){
		switch(marker.type,
		       SNP="autosomal SNPs",
		       NP="autosomal nonpolymorphic markers",
		       X.SNP="chromosome X SNPs",
		       X.NP="chromosome X nonpolymorphic markers")
	}
	if(verbose) message("Computing summary statistics of the genotype clusters for each batch")
	I <- which(type %in% c("SNP", "NP", "X.NP"))
	for(j in seq_along(I)){ ## do not do X.SNP.  Do this during fit.lm3
		i <- I[j]
		marker.type <- type[i]
		if(verbose) message(paste("...", mylabel(marker.type)))
		##if(verbose) message(paste("Computing summary statistics for ", mylabel(marker.type), " genotype clusters for each batch")
		marker.index <- whichMarkers(marker.type, is.snp,
					     is.autosome, is.annotated, is.X)
		if(length(marker.index) == 0) next()
		res <- genotypeSummary(object=object,
				       GT.CONF.THR=GT.CONF.THR,
				       type=marker.type,
				       verbose=verbose,
				       marker.index=marker.index)
		if(!is.lds) {object <- res; rm(res); gc()}
	}
	if(verbose) message("Imputing unobserved genotype medians and shrinking the variances (within-batch, across loci) ")##SNPs only
	if("SNP" %in% type){
		marker.index <- whichMarkers("SNP", is.snp,
					     is.autosome, is.annotated, is.X)
		if(length(marker.index) > 0){
			res <- shrinkSummary(object=object,
					     MIN.OBS=MIN.OBS,
					     MIN.SAMPLES=MIN.SAMPLES,
					     DF.PRIOR=DF.PRIOR,
					     type="SNP",
					     verbose=verbose,
					     marker.index=marker.index)
			if(!is.lds) {object <- res; rm(res); gc()}
		}
	}
	if(verbose) message("Estimating parameters for copy number")##SNPs only
	if(fit.linearModel){
		for(i in seq_along(type)){
			marker.type <- type[i]
			message(paste("...", mylabel(marker.type)))
			CHR.X <- ifelse(marker.type %in% c("X.SNP", "X.NP"), TRUE, FALSE)
			marker.index <- whichMarkers(marker.type, is.snp,
						     is.autosome, is.annotated, is.X)
			if(length(marker.index) == 0) next()
			res <- estimateCnParameters(object=object,
						    type=marker.type,
						    SNRMin=SNRMin,
						    DF.PRIOR=DF.PRIOR,
						    GT.CONF.THR=GT.CONF.THR,
						    MIN.SAMPLES=MIN.SAMPLES,
						    MIN.OBS=MIN.OBS,
						    MIN.NU=MIN.NU,
						    MIN.PHI=MIN.PHI,
						    THR.NU.PHI=THR.NU.PHI,
						    verbose=verbose,
						    marker.index=marker.index,
						    is.lds=is.lds,
						    CHR.X=CHR.X)
		}
		close(object)
	}
	if(is.lds) return(TRUE) else return(object)
}

crlmmCopynumber2 <- function(){
	.Defunct(msg="The crlmmCopynumber2 function has been deprecated. The function crlmmCopynumber should be used instead.  crlmmCopynumber will support large data using ff provided that the ff package is loaded.")
}
crlmmCopynumberLD <- crlmmCopynumber2



estimateCnParameters <- function(object,
				 type=c("SNP", "NP", "X.SNP", "X.NP"),
				 verbose=TRUE,
				 SNRMin=5,
				 DF.PRIOR=50,
				 GT.CONF.THR=0.95,
				 MIN.SAMPLES=10,
				 MIN.OBS=1,
				 MIN.NU=8,
				 MIN.PHI=8,
				 THR.NU.PHI=TRUE,
				 marker.index,
				 CHR.X,
				 is.lds){
	if(missing(marker.index)){
		batch <- batch(object)
		is.snp <- isSnp(object)
		is.autosome <- chromosome(object) < 23
		is.annotated <- !is.na(chromosome(object))
		is.X <- chromosome(object) == 23
		is.lds <- is(calls(object), "ffdf") | is(calls(object), "ff_matrix")
		if(is.lds) stopifnot(isPackageLoaded("ff"))
		CHR.X <- ifelse(type[[1]] %in% c("X.SNP", "X.NP"), TRUE, FALSE)
		marker.index <- whichMarkers(type[[1]], is.snp, is.autosome, is.annotated, is.X)
	}
	lmFxn <- function(type){
		switch(type,
		       SNP="fit.lm1",
		       NP="fit.lm2",
		       X.SNP="fit.lm3",
		       X.NP="fit.lm4")
	}
	myfun <- lmFxn(type[[1]])
	FUN <- get(myfun)
	if(is.lds){
		index.list <- splitIndicesByLength(marker.index, ocProbesets())
##		if(parStatus()){
##			index.list <- splitIndicesByNode(marker.index)
##		} else index.list <- splitIndicesByLength(marker.index, ocProbesets())
		ocLapply(seq(along=index.list),
			 FUN,
			 index.list=index.list,
			 marker.index=marker.index,
			 object=object,
			 batchSize=ocProbesets(),
			 SNRMin=SNRMin,
			 MIN.SAMPLES=MIN.SAMPLES,
			 MIN.OBS=MIN.OBS,
			 DF=DF.PRIOR,
			 GT.CONF.THR=GT.CONF.THR,
			 THR.NU.PHI=THR.NU.PHI,
			 MIN.NU=MIN.NU,
			 MIN.PHI=MIN.PHI,
			 verbose=verbose,
			 is.lds=is.lds,
			 CHR.X=CHR.X,
			 neededPkgs="crlmm")
	} else {
		##FUN <- match.fun(FUN)
		object <- FUN(strata=1,
			      index.list=list(marker.index),
			      marker.index=marker.index,
			      object=object,
			      batchSize=ocProbesets(),
			      SNRMin=SNRMin,
			      MIN.SAMPLES=MIN.SAMPLES,
			      MIN.OBS=MIN.OBS,
			      DF.PRIOR=DF.PRIOR,
			      GT.CONF.THR=GT.CONF.THR,
			      THR.NU.PHI=THR.NU.PHI,
			      MIN.NU=MIN.NU,
			      MIN.PHI=MIN.PHI,
			      is.lds=is.lds,
			      CHR.X=CHR.X,
			      verbose=verbose)
	}
	message("finished")
	return(object)
}


imputeAA.AB <- function(index, N.AA, N.AB, N.BB,
			medianA.AA, medianA.AB, medianA.BB,
			medianB.AA, medianB.AB, medianB.BB){
	gt.to.impute <- 1:2
	imputed <- rep(FALSE, length(index))
	for(i in index){
		Ns <- cbind(N.AA[i, ], N.AB[i, ], N.BB[i, ])
		medianA <- cbind(medianA.AA[i, ], medianA.AB[i, ], medianA.BB[i, ])
		medianB <- cbind(medianB.AA[i, ], medianB.AB[i, ], medianB.BB[i, ])
		colnames(medianA) <- colnames(medianB) <- c("AA", "AB", "BB")

		##Find batches with sufficient data
		K <- which(rowSums(Ns < 3) == 0)
		if(length(K) < 1) next() ## nothing we can do.  use information from other snps
		X <- cbind(1, medianA[K, -gt.to.impute, drop=FALSE], medianB[K, -gt.to.impute, drop=FALSE])
		Y <- cbind(medianA[K, gt.to.impute, drop=FALSE], medianB[K, gt.to.impute, drop=FALSE])
		tmp <- tryCatch(betahat <- solve(crossprod(X), crossprod(X,Y)), error=function(e) NULL)
		if(is.null(tmp)) {
			##cat(".");
			next()
		}
		## Get data from observed genotypes in batch with insufficient data for genotypes 'gt.to.impute'
		L <- which(rowSums(Ns < 3) == 2)
		if(length(L) == 0) next()
		Z <- cbind(1, medianA[L, -gt.to.impute, drop=FALSE], medianB[L, -gt.to.impute, drop=FALSE])
		imputedVals <- Z %*% betahat
		medianA.AA[i, L] <- imputedVals[, 1]
		medianA.AB[i, L] <- imputedVals[, 2]
		medianB.AA[i, L] <- imputedVals[, 3]
		medianB.AB[i, L] <- imputedVals[, 4]
		imputed[i] <- TRUE
	}
	return(list(medianA.AA=medianA.AA, medianA.AB=medianA.AB, medianA.BB=medianA.BB,
		    medianB.AA=medianB.AA, medianB.AB=medianB.AB, medianB.BB=medianB.BB,
		    imputed=imputed))
}

imputeAB.BB <- function(index, N.AA, N.AB, N.BB,
			medianA.AA, medianA.AB, medianA.BB,
			medianB.AA, medianB.AB, medianB.BB){
	gt.to.impute <- 2:3
	imputed <- rep(FALSE, length(index))
	for(j in seq_along(index)){
		i <- index[j]
		Ns <- cbind(N.AA[i, ], N.AB[i, ], N.BB[i, ])
		medianA <- cbind(medianA.AA[i, ], medianA.AB[i, ], medianA.BB[i, ])
		medianB <- cbind(medianB.AA[i, ], medianB.AB[i, ], medianB.BB[i, ])
		colnames(medianA) <- colnames(medianB) <- c("AA", "AB", "BB")

		##Find batches with sufficient data
		K <- which(rowSums(Ns < 3) == 0)
		if(length(K) < 1) next() ## nothing we can do.  use information from other snps
		X <- cbind(1, medianA[K, -gt.to.impute, drop=FALSE], medianB[K, -gt.to.impute, drop=FALSE])
		Y <- cbind(medianA[K, gt.to.impute, drop=FALSE], medianB[K, gt.to.impute, drop=FALSE])
		tmp <- tryCatch(betahat <- solve(crossprod(X), crossprod(X,Y)), error=function(e) NULL)
		if(is.null(tmp)) {
##			cat(".");
			next()
		}

		## Get data from observed genotypes in batch with insufficient data for genotypes 'gt.to.impute'
		L <- which(rowSums(Ns < 3) == 2)
		if(length(L) == 0) next()
		Z <- cbind(1, medianA[L, -gt.to.impute, drop=FALSE], medianB[L, -gt.to.impute, drop=FALSE])
		imputedVals <- Z %*% betahat
		medianA.AB[i, L] <- imputedVals[, 1]
		medianA.BB[i, L] <- imputedVals[, 2]
		medianB.AB[i, L] <- imputedVals[, 3]
		medianB.BB[i, L] <- imputedVals[, 4]
		imputed[j] <- TRUE
	}
	return(list(medianA.AA=medianA.AA, medianA.AB=medianA.AB, medianA.BB=medianA.BB,
		    medianB.AA=medianB.AA, medianB.AB=medianB.AB, medianB.BB=medianB.BB,
		    imputed=imputed))
}

imputeAA <- function(index, N.AA, N.AB, N.BB,
		     medianA.AA, medianA.AB, medianA.BB,
		     medianB.AA, medianB.AB, medianB.BB){
	gt.to.impute <- 1
	imputed <- rep(FALSE, length(index))
	for(j in seq_along(index)){
		i <- index[j]
		Ns <- cbind(N.AA[i, ], N.AB[i, ], N.BB[i, ])
		medianA <- cbind(medianA.AA[i, ], medianA.AB[i, ], medianA.BB[i, ])
		medianB <- cbind(medianB.AA[i, ], medianB.AB[i, ], medianB.BB[i, ])
		colnames(medianA) <- colnames(medianB) <- c("AA", "AB", "BB")

		##Find batches with sufficient data
		K <- which(rowSums(Ns < 3) == 0)
		if(length(K) < 1) next() ## nothing we can do.  use information from other snps
		X <- cbind(1, medianA[K, -gt.to.impute, drop=FALSE], medianB[K, -gt.to.impute, drop=FALSE])
		Y <- cbind(medianA[K, gt.to.impute, drop=FALSE], medianB[K, gt.to.impute, drop=FALSE])
		tmp <- tryCatch(betahat <- solve(crossprod(X), crossprod(X,Y)), error=function(e) NULL)
		if(is.null(tmp)) {
			##cat(".");
			next()
		}
		## Get data from observed genotypes in batch with insufficient data for genotypes 'gt.to.impute'
		L <- which(rowSums(Ns < 3) == 1)
		if(length(L) == 0) next()
		Z <- cbind(1, medianA[L, -gt.to.impute, drop=FALSE], medianB[L, -gt.to.impute, drop=FALSE])
		imputedVals <- Z %*% betahat
		medianA.AA[i, L] <- imputedVals[, 1]
		medianB.AA[i, L] <- imputedVals[, 2]
		imputed[j] <- TRUE
	}
	return(list(medianA.AA=medianA.AA, medianA.AB=medianA.AB, medianA.BB=medianA.BB,
		    medianB.AA=medianB.AA, medianB.AB=medianB.AB, medianB.BB=medianB.BB,
		    imputed=imputed))
}

imputeBB <- function(index, N.AA, N.AB, N.BB,
		     medianA.AA, medianA.AB, medianA.BB,
		     medianB.AA, medianB.AB, medianB.BB){
	gt.to.impute <- 3
	imputed <- rep(FALSE, length(index))
	for(j in seq_along(index)){
		i <- index[j]
		Ns <- cbind(N.AA[i, ], N.AB[i, ], N.BB[i, ])
		medianA <- cbind(medianA.AA[i, ], medianA.AB[i, ], medianA.BB[i, ])
		medianB <- cbind(medianB.AA[i, ], medianB.AB[i, ], medianB.BB[i, ])
		colnames(medianA) <- colnames(medianB) <- c("AA", "AB", "BB")

		##Find batches with sufficient data
		K <- which(rowSums(Ns < 3) == 0)
		if(length(K) < 1) next() ## nothing we can do.  use information from other snps
		X <- cbind(1, medianA[K, -gt.to.impute, drop=FALSE], medianB[K, -gt.to.impute, drop=FALSE])
		Y <- cbind(medianA[K, gt.to.impute, drop=FALSE], medianB[K, gt.to.impute, drop=FALSE])
		colnames(Y) <- c("A.BB", "B.BB")
		tmp <- tryCatch(betahat <- solve(crossprod(X), crossprod(X,Y)), error=function(e) NULL)
		if(is.null(tmp)) {
			##cat(".");
			next()
		}
##		else{
##			R <- Y-crossprod(X, betahat)
##			RSS <- t(R)%*%R
##		}
		## Get data from observed genotypes in batch with insufficient data for genotypes 'gt.to.impute'
		L <- which(rowSums(Ns < 3) == 1)
		if(length(L) == 0) next()
		Z <- cbind(1, medianA[L, -gt.to.impute, drop=FALSE], medianB[L, -gt.to.impute, drop=FALSE])
		imputedVals <- Z %*% betahat
		medianA.BB[i, L] <- imputedVals[, 1]
		medianB.BB[i, L] <- imputedVals[, 2]
		imputed[j] <- TRUE
	}
	return(list(medianA.AA=medianA.AA, medianA.AB=medianA.AB, medianA.BB=medianA.BB,
		    medianB.AA=medianB.AA, medianB.AB=medianB.AB, medianB.BB=medianB.BB,
		    imputed=imputed))
}

imputeAcrossBatch <- function(N.AA, N.AB, N.BB,
			      medianA.AA, medianA.AB, medianA.BB,
			      medianB.AA, medianB.AB, medianB.BB){
	N.missing <- (N.AA < 3) + (N.AB < 3) + (N.BB < 3)
	## find all indices in which one or more batches need to have 2 genotypes imputed
	missingAA.AB <- (N.AA < 3) & (N.AB < 3)
	missingAB.BB <- (N.AB < 3) & (N.BB < 3)
	missingAA <- (N.AA < 3) & (N.AB >= 3)
	missingBB <- (N.BB < 3) & (N.AB >= 3)
	index <- list(AA.AB=which(rowSums(missingAA.AB) > 0),
		      AB.BB=which(rowSums(missingAB.BB) > 0),
		      AA=which(rowSums(missingAA) > 0),
		      BB=which(rowSums(missingBB) > 0))
	imputeNone <- which(rowSums(N.missing == 0) > 0)
	## only works if there are batches with complete data
	index <- lapply(index, intersect, y=imputeNone)
	##indices.to.update <- rep(1:4, each=sapply(index, length))
	updated <- vector("list", 4)
	names(updated) <- c("AA.AB", "AB.BB", "AA", "BB")
	if(length(index[["AA.AB"]] > 0)){
		res <- imputeAA.AB(index[["AA.AB"]],
				   N.AA,
				   N.AB,
				   N.BB,
				   medianA.AA,
				   medianA.AB,
				   medianA.BB,
				   medianB.AA, medianB.AB, medianB.BB)
		updated$AA.AB <- res$imputed
	}
	if(length(index[["AB.BB"]] > 0)){
		res <- imputeAB.BB(index[["AB.BB"]],
				   N.AA,
				   N.AB,
				   N.BB,
				   res[["medianA.AA"]],
				   res[["medianA.AB"]],
				   res[["medianA.BB"]],
				   res[["medianB.AA"]],
				   res[["medianB.AB"]],
				   res[["medianB.BB"]])
		updated$AB.BB <- res$imputed
	}
	if(length(index[["AA"]] > 0)){
		res <- imputeAA(index[["AA"]],
				N.AA,
				N.AB,
				N.BB,
				res[["medianA.AA"]],
				res[["medianA.AB"]],
				res[["medianA.BB"]],
				res[["medianB.AA"]], res[["medianB.AB"]], res[["medianB.BB"]])
		updated$AA <- res$imputed
	}
	if(length(index[["BB"]] > 0)){
		res <- imputeBB(index[["BB"]],
				N.AA,
				N.AB,
				N.BB,
				res[["medianA.AA"]],
				res[["medianA.AB"]],
				res[["medianA.BB"]],
				res[["medianB.AA"]],
				res[["medianB.AB"]],
				res[["medianB.BB"]])
		updated$BB <- res$imputed
	}
	updated.indices <- unlist(updated)
	return(list(res, updated))
}


##calculatePosteriorMean <- function(object, type=c("SNP", "NP", "X.SNP", "X.NP"), verbose=TRUE,
##				   prior.prob=c(1/7, 1/7, 3/7, 1/7, 1/7),
##				   CN=0:4, scale.sd=1){
##	stopifnot(type %in% c("SNP", "NP", "X.SNP", "X.NP"))
##	stopifnot(sum(prior.prob)==1)
##	stopifnot(length(CN) == length(prior.prob))
##	batch <- batch(object)
##	is.snp <- isSnp(object)
##	is.autosome <- chromosome(object) < 23
##	is.annotated <- !is.na(chromosome(object))
##	is.X <- chromosome(object) == 23
##	is.lds <- is(calls(object), "ffdf") | is(calls(object), "ff_matrix")
##	if(is.lds) stopifnot(isPackageLoaded("ff"))
##	samplesPerBatch <- table(as.character(batch(object)))
##	if(!"posteriorMean" %in% assayDataElementNames(object)){
##		message("adding <posteriorMean> slot to assayData")
##		pM <- matrix(NA, nrow(object), ncol(object), dimnames=list(featureNames(object), sampleNames(object)))
##		tmp <- assayDataNew(alleleA=A(object),
##				    alleleB=B(object),
##				    call=calls(object),
##				    callProbability=snpCallProbability(object),
##				    posteriorMean=pM)
##		assayData(object) <- tmp
##	}
##	## add new assay data element for posterior probabilities
##	mylabel <- function(marker.type){
##		switch(marker.type,
##		       SNP="autosomal SNPs",
##		       NP="autosomal nonpolymorphic markers",
##		       X.SNP="chromosome X SNPs",
##		       X.NP="chromosome X nonpolymorphic markers")
##	}
##	if(type=="SNP"){
##		if(verbose) message(paste("...", mylabel(type)))
##		marker.index <- whichMarkers("SNP",
##					     is.snp,
##					     is.autosome,
##					     is.annotated,
##					     is.X)
##		marker.list <- splitIndicesByLength(marker.index, ocProbesets())
##		emit <- ocLapply(seq_along(marker.list),
##				 posteriorMean.snp,
##				 object=object,
##				 index.list=marker.list,
##				 verbose=verbose,
##				 prior.prob=prior.prob,
##				 CN=CN,
##				 scale.sd=scale.sd)
##		#for(i in seq_along(marker.list)){
##		#index <- marker.list[[i]]
##
##	#}
##	} else stop("type not available")
##	if(length(emit)==1) emit <- emit[[1]] else {
##		state.names <- dimnames(emit[[1]])[[3]]
##		tmp <- array(NA, dim=c(length(unlist(marker.list)), ncol(object), length(prior.prob)))
##		for(i in seq_along(marker.list)){
##			marker.index <- marker.list[[i]]
##			tmp[marker.index, , ] <- emit[[i]]
##		}
##		emit <- tmp
##		dimnames(emit) <- list(featureNames(object)[unlist(marker.list)],
##				       sampleNames(object),
##				       state.names)
##	}#stop("need to rbind elements of emit list?")
##	##tmp <- do.call("rbind", emit)
##	match.index <- match(rownames(emit), featureNames(object))
##	S <- length(prior.prob)
##	pM <- matrix(0, length(match.index), ncol(object), dimnames=list(featureNames(object)[match.index], sampleNames(object)))
##	for(i in 1:S) pM <- pM + CN[i]*emit[, , i]
##	posteriorMean(object)[match.index, ] <- pM
##	return(object)
##}

.open <- function(object){
		open(B(object))
		open(tau2A.AA(object))
		open(tau2B.BB(object))
		open(tau2A.BB(object))
		open(tau2B.AA(object))
		open(corrAA(object))
		open(corrAB(object))
		open(corrBB(object))
		open(nuA(object))
		open(nuB(object))
		open(phiA(object))
		open(phiB(object))
}
.close <- function(object){
	close(A(object))
	close(B(object))
	close(tau2A.AA(object))
	close(tau2B.BB(object))
	close(tau2A.BB(object))
	close(tau2B.AA(object))
	close(corrAA(object))
	close(corrAB(object))
	close(corrBB(object))
	close(nuA(object))
	close(nuB(object))
	close(phiA(object))
	close(phiB(object))
}

sum.mymatrix <- function(...){
	x <- list(...)
	return(x[[1]] + do.call(sum, x[-1]))
}
numberGenotypes <- function(CT){
	stopifnot(length(CT)==1)
	copynumber <- paste("cn", CT, sep="")
	switch(copynumber,
		cn0=1,
		cn1=2,
		cn2=3,
		cn3=4,
		cn4=4,
		cn5=6, NULL)
}

.posterior <- function(CT,
		       tau2A,
		       tau2B,
		       sig2A,
		       sig2B,
		       nuA,
		       nuB,
		       phiA,
		       phiB,
		       corrAA,
		       corrAB,
		       corrBB,
		       a,
		       b,
		       scale.sd){
	## 5: AAAAA, AAAAB, AAABB, AABBB, ABBBB, BBBBB
	##CN=4
	## AAAA, AAAB, AABB, ABBB, BBBB:  L = 4
	##CN=3
	##  AAA, AAB, ABB, BBB ; L = 4
	## CN=2
	##  AA, AB, BB; L=3
	## CN = 1: A, B; L=2
	## CN = 0: null; L=1
	L <- numberGenotypes(CT)
	stopifnot(!is.null(L))
	f.x.y <- vector("list", L)
	for(i in seq_along(f.x.y)){
		CA <- (0:CT)[i]
		CB <- CT-CA
		A.scale <- sqrt(tau2A*(CA==0) + sig2A*(CA > 0))
		B.scale <- sqrt(tau2B*(CB==0) + sig2B*(CB > 0))
		if(CA == 0 | CB == 0){## possibly homozygous BB and AA, respectively
			A.scale <- A.scale*scale.sd[1]
			B.scale <- B.scale*scale.sd[1]
		} else { ## one or both greater than zero
			if(length(scale.sd) == 1) scale.sd <- rep(scale.sd, 2)
			A.scale <- A.scale*scale.sd[2]
			B.scale <- B.scale*scale.sd[2]
		}
		A.scale <- as.numeric(A.scale)
		B.scale <- as.numeric(B.scale)
		A.scale2 <- A.scale^2
		B.scale2 <- B.scale^2
		if(CA == 0 & CB == 0) rho <- 0
		if(CA == 0 & CB > 0) rho <- corrBB
		if(CA > 0 & CB == 0) rho <- corrAA
		if(CA > 0 & CB > 0) rho <- corrAB
		rho <- as.numeric(rho)
		meanA <- as.numeric(log2(nuA+CA*phiA))
		meanB <- as.numeric(log2(nuB+CB*phiB))
		covs <- rho*A.scale*B.scale
		##x <- a[, J]
		x <- a
		y <- b
		Q.x.y <- 1/(1-rho^2)*((x - meanA)^2/A.scale2 + (y - meanB)^2/B.scale2 - (2*rho*((x - meanA)*(y - meanB)))/(A.scale*B.scale))
		f.x.y[[i]] <- 1/(2*pi*A.scale*B.scale*sqrt(1-rho^2))*exp(-0.5*Q.x.y)
		class(f.x.y[[i]]) <- "mymatrix"
	}
	res <- do.call("sum", f.x.y)
	return(res)
}





genotypeCorrelation <- function(genotype){
	switch(genotype,
	       NULL="NULL",
	       A="AA",
	       B="BB",
	       AA="AA",
	       AB="AB",
	       BB="BB",
	       AAA="AA",
	       AAB="AB",
	       ABB="AB",
	       BBB="BB",
	       AAAA="AA",
	       AAAB="AB",
	       AABB="AB",
	       ABBB="AB",
	       BBBB="BB")
}

posteriorMean.snp <- function(stratum, object, index.list, CN,
			      prior.prob=c(1/7, 1/7, 3/7, 1/7, 1/7), is.lds=TRUE, verbose=TRUE,
			      scale.sd=1){
	if(length(scale.sd) == 1) rep(scale.sd,2)
	if(verbose) message("Probe stratum ", stratum, " of ", length(index.list))
	index <- index.list[[stratum]]
	test <- tryCatch(open(A(object)), error=function(e) NULL)
	if(!is.null(test)) invisible(.open(object))
	a <- log2(as.matrix(A(object)[index, ]))
	b <- log2(as.matrix(B(object)[index, ]))
	NN <- Ns(object, i=index)[, , 1]
	sig2A <- as.matrix(tau2A.AA(object)[index,])
	sig2B <- as.matrix(tau2B.BB(object)[index,])
	tau2A <- as.matrix(tau2A.BB(object)[index,])
	tau2B <- as.matrix(tau2B.AA(object)[index,])
	corrAA <- as.matrix(corrAA(object)[index, ])
	corrAB <- as.matrix(corrAB(object)[index, ])
	corrBB <- as.matrix(corrBB(object)[index, ])
	nuA <- as.matrix(nuA(object)[index, ])
	phiA <- as.matrix(phiA(object)[index, ])
	nuB <- as.matrix(nuB(object)[index, ])
	phiB <- as.matrix(phiB(object)[index, ])
	if(!is.null(test)) invisible(.close(object))
	S <- length(prior.prob)
	emit <- array(NA, dim=c(nrow(a), ncol(a), S))##SNPs x sample x 'truth'
	sample.index <- split(1:ncol(object), batch(object))
	##emit <- vector("list", length(sample.index))
	for(j in seq_along(sample.index)){
		cat("batch ", j, "\n")
		J <- sample.index[[j]]
		probs <- array(NA, dim=c(nrow(a), length(J), S))
		for(k in seq_along(CN)){
			##CT <- CN[k]
			probs[, , k] <- .posterior(CT=CN[k],
						   tau2A=tau2A,
						   tau2B=tau2B,
						   sig2A=sig2A,
						   sig2B=sig2B,
						   nuA=nuA,
						   nuB=nuB,
						   phiA=phiA,
						   phiB=phiB,
						   corrAA=corrAA,
						   corrAB=corrAB,
						   corrBB=corrBB,
						   a=a[, J],
						   b=b[, J],
						   scale.sd=scale.sd)
			##probs[, , k] <- do.call("sum", f.x.y)
			##if none of the states are likely (outlier), assign NA
			##		emit[, , counter] <- f.x.y
			##		counter <- counter+1
		}
		emit[, J, ] <- probs
	}
	for(i in 1:S) emit[, , i] <- emit[, , i]*prior.prob[i]
	total <- matrix(0, nrow(emit), ncol(emit))
	for(i in 1:S) total <- total+emit[, , i]
	## how to handle outliers...
	##  - use priors (posterior mean will likely be near 2).  smoothing needs to take into account the uncertainty
	##  - need uncertainty estimates for posterior means
	for(i in 1:S) emit[, , i] <- emit[, , i]/total
	dimnames(emit) <- list(featureNames(object)[index], sampleNames(object), paste("states", 1:S, sep="_"))
	## for one marker/one sample, the emission probs must sum to 1
	return(emit)
}





## used for testing
##crlmm2.2 <- function(filenames, row.names=TRUE, col.names=TRUE,
##                   probs=c(1/3, 1/3, 1/3), DF=6, SNRMin=5, gender=NULL,
##                   save.it=FALSE, load.it=FALSE, intensityFile,
##                   mixtureSampleSize=10^5, eps=0.1, verbose=TRUE,
##                   cdfName, sns, recallMin=10, recallRegMin=1000,
##                   returnParams=FALSE, badSNP=.7){
##  if ((load.it || save.it) && missing(intensityFile))
##    stop("'intensityFile' is missing, and you chose either load.it or save.it")
##  if (missing(sns)) sns <- basename(filenames)
##  if (!missing(intensityFile))
##    if (load.it & !file.exists(intensityFile)){
##      load.it <- FALSE
##      message("File ", intensityFile, " does not exist.")
##      message("Not loading it, but running SNPRMA from scratch.")
##    }
##  if (!load.it){
##    res <- snprma2(filenames, fitMixture=TRUE,
##                   mixtureSampleSize=mixtureSampleSize, verbose=verbose,
##                   eps=eps, cdfName=cdfName, sns=sns)
##    open(res[["A"]])
##    open(res[["B"]])
##    open(res[["SNR"]])
##    open(res[["mixtureParams"]])
##    if(save.it){
##      t0 <- proc.time()
##      save(res, file=intensityFile)
##      t0 <- proc.time()-t0
##      if (verbose) message("Used ", t0[3], " seconds to save ", intensityFile, ".")
##    }
##  }else{
##    if (verbose) message("Loading ", intensityFile, ".")
##    obj <- load(intensityFile)
##    if (verbose) message("Done.")
##    if (obj != "res")
##      stop("Object in ", intensityFile, " seems to be invalid.")
##  }
##  if(row.names) row.names=res$gns else row.names=NULL
##  if(col.names) col.names=res$sns else col.names=NULL
##  res2 <- crlmmGT2(res[["A"]], res[["B"]], res[["SNR"]],
##                   res[["mixtureParams"]], res[["cdfName"]],
##                   gender=gender, row.names=row.names,
##                   col.names=col.names, recallMin=recallMin,
##                   recallRegMin=1000, SNRMin=SNRMin,
##                   returnParams=returnParams, badSNP=badSNP,
##                   verbose=verbose)
##
##  res2[["SNR"]] <- res[["SNR"]]
##  res2[["SKW"]] <- res[["SKW"]]
##  return(list2SnpSet(res2, returnParams=returnParams))
##}

genotypes <- function(copyNumber, is.snp=TRUE){
	stopifnot(copyNumber %in% 0:4)
	cn <- paste("x", copyNumber, sep="")
	if(is.snp){
		res <- switch(cn,
			      x0="NULL",
			      x1=LETTERS[1:2],
			      x2=c("AA", "AB", "BB"),
			      x3=c("AAA", "AAB", "ABB", "BBB"),
			      x4=c("AAAA", "AAAB", "AABB", "ABBB", "BBBB"))
	} else {
		res <- switch(cn,
			      x0="NULL",
			      x1="A",
			      x2="AA",
			      x3="AAA",
			      x4="AAAA")
	}
	return(res)
}

dbvn <- function(x, mu, Sigma){
	##stopifnot(ncol(x)==2)
	logA <- x[, , 1]
	logB <- x[, , 2]
	sigma2A <- Sigma[,1]
	sigma2B <- Sigma[,3]
	sigmaAB <- sqrt(sigma2A*sigma2B)
	rho <- Sigma[,2]
	muA <- mu[, 1]
	muB <- mu[, 2]
	Q.x.y <- 1/(1-rho^2)*((logA - muA)^2/sigma2A + (logB - muB)^2/sigma2B - (2*rho*((logA - muA)*(logB - muB)))/sigmaAB)
	f.x.y <- 1/(2*pi*sigmaAB*sqrt(1-rho^2))*exp(-0.5*Q.x.y)
}

ABpanel <- function(x, y, predictRegion,
		    copyNumber=0:4,
		    fill,
		    ...,
		    subscripts){
	panel.grid(h=5, v=5)
	if(!missing(fill)){
		panel.xyplot(x, y, fill=fill[subscripts], ...)
	} else {
		panel.xyplot(x, y, ...)
	}
	pn <- panel.number()
	if(!is.null(copyNumber)){
		for(CN in copyNumber){
			gts <- genotypes(CN)
			index <- match(gts, names(predictRegion))
			pr <- predictRegion[index]
			for(i in seq_along(pr)){ ## for each genotype
				## scale?
				pr2 <- pr[[i]]
				mu <- pr2$mu[pn, , , drop=FALSE] ## pn=panel number
				Sigma <- pr2$cov[pn, , ,drop=FALSE]
				for(j in seq_len(dim(mu)[3])){  ## for each batch
					dat.ellipse <- ellipse(x=Sigma[, 2, j],
							       centre=mu[ , , j],
							       scale=c(sqrt(Sigma[ , 1, j]),
							       sqrt(Sigma[, 3, j])))
					lpolygon(dat.ellipse[,1], dat.ellipse[,2], ...)
				}
			}
		}
	}
}

calculateRTheta <- function(object, ##genotype=c("AA", "AB", "BB"),
			    batch.name,
			    feature.index){
	index.np <- which(!(isSnp(object)[feature.index]))
	j <- match(batch.name, batchNames(object))
	if(missing(j)) stop("batch.name did not match in batchNames(object)")
	CHR <- unique(chromosome(object)[feature.index])
	isX <- CHR == "X"
	centers <- getMedians(batchStatistics(object))
	lapply(centers, open)
	centersMatrix <- lapply(centers, function(x, i, j) x[i, j], j=j, i=feature.index)
	lapply(centers, close)
	centersMatrix <- do.call(cbind, centersMatrix)
	centersA <- centersMatrix[, 1:3, drop=FALSE]
	centersB <- centersMatrix[, 4:6, drop=FALSE]
	rm(centers, centersMatrix); gc()
	centersA[centersA < 64] <- 64
	centersB[centersB < 64] <- 64
	theta <- as.matrix(atan2(centersB, centersA) * 2/pi)
	centersA[is.na(centersA)] <- 0
	centersB[is.na(centersB)] <- 0
	if(length(index.np) > 0) theta[index.np, ] <- 1
	##
	r <- centersA+centersB
	J <- which(batch(object)==batch.name)
	open(A(object))
	open(B(object))
	a <- as.matrix(A(object)[feature.index, J, drop=FALSE])
	b <- as.matrix(B(object)[feature.index, J, drop=FALSE])
	close(A(object))
	close(B(object))
	if(length(index.np) > 0) b[index.np, ] <- 0L
	obs.theta <- atan2(b, a)*2/pi
	calculateBandR <- function(o, r, theta, not.na, M){
		lessAA <- o < theta[, 1]
		lessAB <- o < theta[, 2]
		lessBB <- o < theta[, 3]
		ii <- which(lessAA & not.na)
		jj <- which(!lessBB & not.na)
		i <- which(!lessAA & lessAB & not.na)
		j <- which(!lessAB & lessBB & not.na)
		M[i, 1] <- 0.5*(o[i]-theta[i,1])/(theta[i,2]-theta[i,1])
		M[j, 1] <- 0.5*(o[j]-theta[j,2])/(theta[j,3]-theta[j,2]) + 0.5
		M[ii, 1] <- 0
		M[jj, 1] <- 1
		M[i, 2] <- (o[i]-theta[i,1])*(r[i,2]-r[i,1])/(theta[i,2]-theta[i,1]) + r[i, 1]
		M[j, 2] <- (o[j]-theta[j,2])*(r[j,3]-r[j,2])/(theta[j,3]-theta[j,2]) + r[j, 2]
		M[ii, 2] <- r[ii, 1]
		M[jj, 2] <- r[jj, 3]
		return(M)
	}
	not.na <- !is.na(theta[,1])
	r.expected <- bf <- matrix(NA, length(feature.index), ncol(a))
	M <- matrix(NA, length(feature.index), 2)
	for(j in seq_len(ncol(a))){
		tmp <- calculateBandR(obs.theta[, j], r=r, theta=theta, not.na=not.na, M=M)
		bf[, j] <- tmp[,1]
		r.expected[,j] <- tmp[,2]
	}
	bf[bf < 0] <- 0
	bf[bf > 1] <- 1
	if(length(index.np) > 0) r.expected[index.np, ] <- centersA[index.np, 1]
	obs.r <- a+b
	lrr <- log2(obs.r/r.expected)
	dimnames(bf) <- dimnames(lrr) <- list(featureNames(object)[feature.index],
					      sampleNames(object)[J])
	bf <- integerMatrix(bf, 1000)
	lrr <- integerMatrix(lrr, 100)
	return(list(baf=bf, lrr=lrr))
}

setMethod("Ns", signature(object="AssayData"),
	  function(object, ...){
		  batchnames <- sampleNames(object)
		  N.AA <- assayDataElement(object, "N.AA")
		  N.AB <- assayDataElement(object, "N.AB")
		  N.BB <- assayDataElement(object, "N.BB")
		  res <- array(NA, dim=c(nrow(N.AA), 3, ncol(N.AA)))
		  dimnames(res)[[2]] <- c("AA", "AB", "BB")
		  dimnames(res)[[3]] <- batchnames
		  res[, "AA", ] <- N.AA[,]
		  res[, "AB", ] <- N.AB[,]
		  res[, "BB", ] <- N.BB[,]
		  return(res)
	  })
setMethod("corr", signature(object="AssayData"),
	  function(object, ...){
		  batchnames <- sampleNames(object)
		  corrAA <- assayDataElement(object, "corrAA")
		  corrAB <- assayDataElement(object, "corrAB")
		  corrBB <- assayDataElement(object, "corrBB")
		  res <- array(NA, dim=c(nrow(corrAA), 3, ncol(corrAA)))
		  dimnames(res)[[2]] <- c("AA", "AB", "BB")
		  dimnames(res)[[3]] <- batchnames
		  res[, "AA", ] <- corrAA[,]
		  res[, "AB", ] <- corrAB[,]
		  res[, "BB", ] <- corrBB[,]
		  return(res)
	  })

setMethod("medians", signature(object="AssayData"),
	  function(object, ...){
		  batchnames <- sampleNames(object)
		  medianA.AA <- assayDataElement(object, "medianA.AA")
		  medianA.AB <- assayDataElement(object, "medianA.AB")
		  medianA.BB <- assayDataElement(object, "medianA.BB")
		  medianB.AA <- assayDataElement(object, "medianB.AA")
		  medianB.AB <- assayDataElement(object, "medianB.AB")
		  medianB.BB <- assayDataElement(object, "medianB.BB")
		  res <- array(NA, dim=c(nrow(medianA.AA), 2, 3, ncol(medianA.AA)))
		  dimnames(res)[[2]] <- c("A", "B")
		  dimnames(res)[[3]] <- c("AA", "AB", "BB")
		  dimnames(res)[[4]] <- batchnames
		  res[, "A", "AA", ] <- medianA.AA[,]
		  res[, "A", "AB", ] <- medianA.AB[,]
		  res[, "A", "BB", ] <- medianA.BB[,]
		  res[, "B", "AA", ] <- medianB.AA[,]
		  res[, "B", "AB", ] <- medianB.AB[,]
		  res[, "B", "BB", ] <- medianB.BB[,]
		  return(res)
})

getMedians <- function(object){
	medianA.AA <- assayDataElement(object, "medianA.AA")
	medianA.AB <- assayDataElement(object, "medianA.AB")
	medianA.BB <- assayDataElement(object, "medianA.BB")
	medianB.AA <- assayDataElement(object, "medianB.AA")
	medianB.AB <- assayDataElement(object, "medianB.AB")
	medianB.BB <- assayDataElement(object, "medianB.BB")
	list(A.AA=medianA.AA,
	     A.AB=medianA.AB,
	     A.BB=medianA.BB,
	     B.AA=medianB.AA,
	     B.AB=medianB.AB,
	     B.BB=medianB.BB)
}

setMethod("mads", signature(object="AssayData"),
	  function(object, ...){
		  batchnames <- sampleNames(object)
		  madA.AA <- assayDataElement(object, "madA.AA")
		  madA.AB <- assayDataElement(object, "madA.AB")
		  madA.BB <- assayDataElement(object, "madA.BB")
		  madB.AA <- assayDataElement(object, "madB.AA")
		  madB.AB <- assayDataElement(object, "madB.AB")
		  madB.BB <- assayDataElement(object, "madB.BB")
		  res <- array(NA, dim=c(nrow(madA.AA), 2, 3, ncol(madA.AA)))
		  dimnames(res)[[2]] <- c("A", "B")
		  dimnames(res)[[3]] <- c("AA", "AB", "BB")
		  dimnames(res)[[4]] <- batchnames
		  res[, "A", "AA", ] <- madA.AA[,]
		  res[, "A", "AB", ] <- madA.AB[,]
		  res[, "A", "BB", ] <- madA.BB[,]
		  res[, "B", "AA", ] <- madB.AA[,]
		  res[, "B", "AB", ] <- madB.AB[,]
		  res[, "B", "BB", ] <- madB.BB[,]
		  return(res)
})



setMethod("tau2", signature(object="AssayData"),
	  function(object, ...){
		  batchnames <- sampleNames(object)
		  tau2A.AA <- assayDataElement(object, "tau2A.AA")
		  tau2A.BB <- assayDataElement(object, "tau2A.BB")
		  tau2B.AA <- assayDataElement(object, "tau2B.AA")
		  tau2B.BB <- assayDataElement(object, "tau2B.BB")
		  res <- array(NA, dim=c(nrow(tau2A.AA), 2, 2, ncol(tau2A.AA)))
		  dimnames(res)[[2]] <- c("A", "B")
		  dimnames(res)[[3]] <- c("AA", "BB")
		  dimnames(res)[[4]] <- batchnames
		  res[, "A", "AA", ] <- tau2A.AA[,]
		  res[, "A", "BB", ] <- tau2A.BB[,]
		  res[, "B", "AA", ] <- tau2B.AA[,]
		  res[, "B", "BB", ] <- tau2B.BB[,]
		  return(res)
})

# Loading required libraries
THISPKG <- "crlmm"

.onAttach <- function(libname, pkgname) {
	packageStartupMessage("Welcome to crlmm version ", packageDescription(THISPKG, fields="Version"))
}

.onUnload <- function( libpath ){
	library.dynam.unload(THISPKG, libpath)
}
.crlmmPkgEnv <- new.env(parent=emptyenv())

test_crlmm <- function(){
	library(oligoClasses)
	if (require(genomewidesnp6Crlmm) & require(hapmapsnp6)){
		path <- system.file("celFiles", package="hapmapsnp6")
		cels <- list.celfiles(path, full.names=TRUE)
		(crlmmOutput <- crlmm(cels))
	}
}

test_duplicates <- function(){
	library(crlmm);library(RUnit); library(oligoClasses)
	if (require(genomewidesnp6Crlmm) & require(hapmapsnp6)){
		path <- system.file("celFiles", package="hapmapsnp6")
		cels <- list.celfiles(path, full.names=TRUE)
		## Quickly fails because sample identifiers are not unique
		checkException(crlmmOutput <- crlmm(cels[c(1,1, 2)]), silent=FALSE)

	}
	if(FALSE){
		library2(ff)
		datadir <- "/thumper/ctsa/snpmicroarray/illumina/IDATS/370k"
		## read in your samplesheet
		samplesheet <- read.csv(file.path(datadir, "HumanHap370Duo_Sample_Map.csv"), header=TRUE, as.is=TRUE)
		samplesheet <- samplesheet[-c(28:46,61:75,78:79), ]
		arrayNames <- file.path(datadir, unique(samplesheet[, "SentrixPosition"]))
		index <- c(1,1,2,2,3,4)
		any(duplicate(arrayNames[index]))
		arrayNames <- arrayNames[index]
		ss <- samplesheet[index, ]
		arrayInfo <- list(barcode=NULL, position="SentrixPosition")
		cnSet <- genotype.Illumina(sampleSheet=ss,
					   arrayNames=arrayNames,
					   arrayInfoColNames=arrayInfo,
					   cdfName="human370v1c",
					   batch=rep("1", nrow(ss)))
		checkTrue(validObject(cnSet))
	}

}

stupidSplineBasis <- function(x,knots){
  x <- pmin(x,knots[3])
  x <- pmax(x,knots[1])
  cbind(1, x, x^2, pmax(0, (x-knots[2]))^2)
}

changeToCrlmmAnnotationName <- function(x){
  pkgDir <- system.file(package=THISPKG)
  wanted <- paste(tolower(gsub("_", "", x)), "crlmm.stuff", sep=".")
  file.path(pkgDir, "extdata", wanted)
}

getCrlmmAnnotationName <- function(x){
	paste(tolower(gsub("_", "", x)), "Crlmm", sep="")
}

## medianSummaries <- function(mat, grps)
##   .Call("R_subColSummarize_median", mat, grps, PACKAGE = "preprocessCore")

medianSummaries <- function(mat, grps)
.Call("subColSummarizeMedianPP", mat, grps)

intMedianSummaries <- function(mat, grps)
  as.integer(medianSummaries(mat, grps))



## .crlmmPkgEnv is an enviroment that will
## store all the variables used by the pkg.
## it's meant to not overwrite user's variables
## and get rid of the NOTES generated by
## R CMD check

isLoaded <- function(dataset, environ=.crlmmPkgEnv)
	exists(dataset, envir=environ)

getVarInEnv <- function(dataset, environ=.crlmmPkgEnv){
	if (!isLoaded(dataset, environ=environ))
		stop("Variable ", dataset, " not found in supplied environment")
	environ[[dataset]]
}

list2SnpSet <- function(x, returnParams=FALSE){
  pd <- data.frame(SNR=x[["SNR"]][], gender=x[["gender"]],
                   batchQC=rep(x[["batchQC"]], ncol(x[["calls"]])),
                   row.names=colnames(x[["calls"]]))
  pdv <- data.frame(labelDescription=c("Signal-to-noise Ratio",
                      "Gender: Male (1) and Female (2)",
                      "Quality score for batch"),
                    row.names=c("SNR", "gender", "batchQC"))

  recall <- length(x[["DD"]]) > 1
  if (returnParams){
    if (recall){
      fd <- data.frame(SNPQC=x[["SNPQC"]],
                       cAA=x[["params"]][["centers"]][,1],
                       cAB=x[["params"]][["centers"]][,2],
                       cBB=x[["params"]][["centers"]][,3],
                       sAA=x[["params"]][["scales"]][,1],
                       sAB=x[["params"]][["scales"]][,2],
                       sBB=x[["params"]][["scales"]][,3],
                       nAA=x[["params"]][["N"]][,1],
                       nAB=x[["params"]][["N"]][,2],
                       nBB=x[["params"]][["N"]][,3],
                       spAA=x[["DD"]][,1],
                       spAB=x[["DD"]][,2],
                       spBB=x[["DD"]][,3],
                       row.names=rownames(x[["calls"]]))
      fdv <- data.frame(labelDescription=c(
                          "SNP Quality Score",
                          "Center AA", "Center AB", "Center BB",
                          "Scale AA", "Scale AB", "Scale BB",
                          "N AA", "N AB", "N BB",
                          "Shift in parameters AA",
                          "Shift in parameters AB",
                          "Shift in parameters BB"),
                        row.names=c(
                          "SNPQC",
                          "cAA", "cAB", "cBB",
                          "sAA", "sAB", "sBB",
                          "nAA", "nAB", "nBB",
                          "spAA", "spAB", "spBB"))
    }else{
      fd <- data.frame(SNPQC=x[["SNPQC"]],
                       cAA=x[["params"]][["centers"]][,1],
                       cAB=x[["params"]][["centers"]][,2],
                       cBB=x[["params"]][["centers"]][,3],
                       sAA=x[["params"]][["scales"]][,1],
                       sAB=x[["params"]][["scales"]][,2],
                       sBB=x[["params"]][["scales"]][,3],
                       nAA=x[["params"]][["N"]][,1],
                       nAB=x[["params"]][["N"]][,2],
                       nBB=x[["params"]][["N"]][,3],
                       row.names=rownames(x[["calls"]]))
      fdv <- data.frame(labelDescription=c(
                          "SNP Quality Score",
                          "Center AA", "Center AB", "Center BB",
                          "Scale AA", "Scale AB", "Scale BB",
                          "N AA", "N AB", "N BB"),
                        row.names=c(
                          "SNPQC",
                          "cAA", "cAB", "cBB",
                          "sAA", "sAB", "sBB",
                          "nAA", "nAB", "nBB"))
    }
  }else{
    if (recall){
      fd <- data.frame(SNPQC=x[["SNPQC"]],
                       spAA=x[["DD"]][,1],
                       spAB=x[["DD"]][,2],
                       spBB=x[["DD"]][,3],
                       row.names=rownames(x[["calls"]]))
      fdv <- data.frame(labelDescription=c("SNP Quality Score",
                          "Shift in parameters AA",
                          "Shift in parameters AB",
                          "Shift in parameters BB"),
                        row.names=c("SNPQC", "spAA", "spAB", "spBB"))
    }else{
      fd <- data.frame(SNPQC=x[["SNPQC"]],
                       row.names=rownames(x[["calls"]]))
      fdv <- data.frame(labelDescription=c("SNP Quality Score"),
                        row.names=c("SNPQC"))
    }
  }
  new("SnpSet",
      assayData=assayDataNew("lockedEnvironment",
        call=x[["calls"]], callProbability=x[["confs"]]),
      phenoData=new("AnnotatedDataFrame",
        data=pd, varMetadata=pdv),
      featureData=new("AnnotatedDataFrame",
        data=fd, varMetadata=fdv),
      annotation=x[["pkgname"]])
}

loader <- function(theFile, envir, pkgname){
	theFile <- file.path(system.file(package=pkgname),
			     "extdata", theFile)
	if (!file.exists(theFile))
		stop("File ", theFile, " does not exist in ", pkgname)
	load(theFile, envir=envir)
}

celDates <- function(celfiles){
	if(!all(file.exists(celfiles))) stop("1 or more cel file does not exist")
	celdates <- celtimes <- vector("character", length(celfiles))
	for(i in seq(along=celfiles)){
		if(i %% 100 == 0) cat(".")
		tmp <- read.celfile.header(celfiles[i], info="full")$DatHeader
		tmp <- strsplit(tmp, "\ +")
		celdates[i] <- tmp[[1]][6]
		celtimes[i] <- tmp[[1]][7]
	}
	tmp <- paste(celdates, celtimes)
	celdts <- strptime(tmp, "%m/%d/%y %H:%M:%S")
	return(celdts)
}

illuminaCdfNames <- function(){
	c("human1mv1c",# 1M
	  "human370v1c",            # 370CNV
	  "human650v3a",            # 650Y
	  "human610quadv1b",        # 610 quad
	  "human660quadv1a",        # 660 quad
	  "human370quadv3c",        # 370CNV quad
	  "human550v3b",            # 550K
	  "human1mduov3b",          # 1M Duo
	  "humanomni1quadv1b",      # Omni1 quad
	  "humanomni25quadv1b",     # Omni2.5 quad
	  "humanomni258v1a",        # Omni2.5 8 v1 A
          "humanomni258v1p1b",      # Omni2.5 8 v1.1 B
          "humanomni5quadv1b",      # Omni5 quad          
	  "humanomniexpress12v1b",  # Omni express 12
	  "humanimmuno12v1b",       # Immuno chip 12
	  "humancytosnp12v2p1h",    # CytoSNP 12
          "humanexome12v1p2a",      # Exome 12 v1.2 A
          "humanomniexpexome8v1p1b",
          "nopackage")
}

affyCdfNames <- function(){
	c("genomewidesnp6",
	  "genomewidesnp5")
}

validCdfNames <- function(){
	c(affyCdfNames(),
	  illuminaCdfNames())
}

cleancdfname <- function(x) strsplit(x, "Crlmm")[[1]][[1]]

isValidCdfName <- function(cdfName){
	chipList <- validCdfNames()
	match.arg(cleancdfname(cdfName), chipList)
	return(TRUE)
}

isPackageLoaded <- function(pkg){
	stopifnot(is.character(pkg))
	pkg <- paste("package:", pkg, sep="")
	pkg %in% search()
}

paramNames <- function(){
	c("tau2A",
	  "tau2B", "sig2A", "sig2B",
	  "nuA", ##"nuA.se",
	  "nuB",  ##"nuB.se",
	  "phiA",
	  "phiB",
	  "phiPrimeA",
	  "phiPrimeB",
	  ##"phiA.se", "phiB.se",
	  "corrAB",
	  "corrBB",
	  "corrAA")
}

loadObject <- function(filename, load.it){
	fname <- paste(filename, ".rda", sep="")
	if(load.it & file.exists(file.path(ldPath(), fname))){
		message("load.it is TRUE, loading previously saved ff object")
		return(TRUE)
	} else return(FALSE)
}



setMethod("annotatedDataFrameFrom", "ff_matrix", Biobase:::annotatedDataFrameFromMatrix)
setMethod("annotatedDataFrameFrom", "ffdf", Biobase:::annotatedDataFrameFromMatrix)

## Document this...
getBAF <- function(theta, canonicalTheta)
    .Call('normalizeBAF', theta, canonicalTheta)


validCEL <- function(celfiles){
	for(i in seq_along(celfiles)){
		res <- tryCatch(read.celfile(celfiles[i], intensity.means.only=TRUE), error=function(e) NULL)
		if(is.null(res)) {
			msg <- message("Problem reading ", celfiles[i])
			stop(msg)
		}
	}
	return("Successfully read all cel files")
}

setMethod("posteriorMean", signature(object="CNSet"), function(object) assayDataElement(object, "posteriorMean"))
setReplaceMethod("posteriorMean", signature(object="CNSet", value="matrix"), function(object, value) assayDataElementReplace(object, "posteriorMean", value))
##setReplaceMethod("mixtureParams", signature(object="CNSet", value="ANY"), function(object, value) object@mixtureParams <- mixtureParams)

linearParamElementReplace <- function(obj, elt, value) {
    storage.mode <- storageMode(batchStatistics(obj))
    switch(storage.mode,
           "lockedEnvironment" = {
               aData <- copyEnv(batchStatistics(obj))
               if (is.null(value)) rm(list=elt, envir=aData)
               else aData[[elt]] <- value
               Biobase:::assayDataEnvLock(aData)
               batchStatistics(obj) <- aData
           },
           "environment" = {
               if (is.null(value)) rm(list=elt, envir=batchStatistics(obj))
               else batchStatistics(obj)[[elt]] <- value
           },
           list = batchStatistics(obj)[[elt]] <- value)
    obj
}

## parameters
## allele A
##   autosome SNPs
##   autosome NPs
##   chromosome X NPs for women
C1 <- function(object, marker.index, batch.index, sample.index){
	acn <- matrix(NA, nrow=length(marker.index), ncol=length(sample.index))
	for(k in seq_along(batch.index)){
		l <- batch.index[k]
		## calculate cn for all the samples that are in this batch
		jj <- sample.index[as.character(batch(object))[sample.index] == batchNames(object)[l]]
		bg <- nuA(object)[marker.index, l]
		slope <- phiA(object)[marker.index, l]
		I <- as.matrix(A(object)[marker.index, jj])
		acn[, match(jj, sample.index)] <- 1/slope * (I - bg)
	}
	return(as.matrix(acn))
}

## allele B  (treated allele 'A' for chromosome X NPs)
##   autosome SNPs
##   chromosome X for male nonpolymorphic markers
C2 <- function(object, marker.index, batch.index, sample.index, NP.X=FALSE){
	acn <- matrix(NA, nrow=length(marker.index), ncol=length(sample.index))
	for(k in seq_along(batch.index)){
		l <- batch.index[k]
		jj <- sample.index[as.character(batch(object))[sample.index] == batchNames(object)[l]]
		bg <- nuB(object)[marker.index, l]
		slope <- phiB(object)[marker.index, l]
		if(!NP.X){
			I <- as.matrix(B(object)[marker.index, jj])
		} else I <- as.matrix(A(object)[marker.index, jj])
		acn[, match(jj, sample.index)] <- 1/slope * (I - bg)
	}
##	if(length(acn) > 1){
##		acn <- do.call("cbind", acn)
##	} else acn <- acn[[1]]
	return(as.matrix(acn))
}

## Chromosome X SNPs
C3 <- function(object, allele, marker.index, batch.index, sample.index){
##	acn <- vector("list", length(batch.index))
	acn <- matrix(NA, nrow=length(marker.index), ncol=length(sample.index))
	for(k in seq_along(batch.index)){
		l <- batch.index[k]
		##j <- which(as.character(batch(object))[sample.index] == batchNames(object)[l])
		jj <- sample.index[as.character(batch(object))[sample.index] == batchNames(object)[l]]
		##phiA2 and phiB2 are not always estimable  -- need both men and women
		phiA2 <- phiPrimeA(object)[marker.index, l]
		phiB2 <- phiPrimeB(object)[marker.index, l]
		phiA <- phiA(object)[marker.index, l]
		phiB <- phiB(object)[marker.index, l]
		nuA <- nuA(object)[marker.index, l]
		nuB <- nuB(object)[marker.index, l]
		phiA <- phiA(object)[marker.index, l]
		IA <- as.matrix(A(object)[marker.index, jj])
		IB <- as.matrix(B(object)[marker.index, jj])
		## I = nu + acn * phi
		## acn = 1/phi* (I-nu)
		phistar <- phiB2/phiA
		tmp <- (IB - nuB - phistar*IA + phistar*nuA)/phiB
		CB <- tmp/(1-phistar*phiA2/phiB)
		index <- which(is.na(phiB2) | is.na(phiA2))
		if(length(index) > 0){
			cb <- 1/phiB[index] * (IB[index, ] - nuB[index])
			CB[index, ] <- cb
		}
		if(allele == "B"){
			acn[, match(jj, sample.index)] <- CB
			##acn[[k]] <- CB
		}
		if(allele == "A"){
			CA <- (IA-nuA-phiA2*CB)/phiA
			if(length(index) > 0){
				ca <- 1/phiA[index] * (IA[index, ] - nuA[index])
 				CA[index, ] <- ca
			}
			acn[, match(jj, sample.index)] <- CA
		}
	}
##	if(length(acn) > 1){
##		acn <- do.call("cbind", acn)
##	} else acn <- acn[[1]]
	return(as.matrix(acn))
}




ACN <- function(object, allele, i , j){
	if(missing(i) & missing(j)) stop("must specify rows (i) or columns (j)")
	is.ff <- is(calls(object), "ff") | is(calls(object), "ffdf")
	missing.i <- missing(i)
	missing.j <- missing(j)
	if(!missing.i){
		is.ann <- !is.na(chromosome(object)[i])
		is.X <- chromosome(object)[i]==23 & is.ann
		is.auto <- chromosome(object)[i] < 23 & is.ann
		is.snp <- isSnp(object)[i] & is.ann
	} else{
		is.ann <- !is.na(chromosome(object))
		is.X <- chromosome(object)==23 & is.ann
		is.auto <- chromosome(object) < 23 & is.ann
		is.snp <- isSnp(object) & is.ann
		i <- 1:nrow(object)
	}
	## Define batch.index and sample.index
	if(!missing.j) {
		batches <- unique(as.character(batch(object)[j]))
		##batches <- as.character(batch(object)[j])
		batch.index <- match(batches, batchNames(object))
	} else {
		batch.index <- seq_along(batchNames(object))
		j <- 1:ncol(object)
	}
	nr <- length(i)
	nc <- length(j)
	acn <- matrix(NA, nr, nc)
	dimnames(acn) <- list(featureNames(object)[i],
			      sampleNames(object)[j])
	if(allele == "A"){
		if(is.ff){
			open(nuA(object))
			open(phiA(object))
			open(A(object))
		}
		## --
		## 4 types of markers for allele A
		##--
		## 1. autosomal SNPs or autosomal NPs
		if(any(is.auto)){
			auto.index <- which(is.auto)
			marker.index <- i[is.auto]
			acn[auto.index, ] <- C1(object, marker.index, batch.index, j)
		}
		if(any(is.X)){
			##2. CHR X SNPs (men and women)
			if(any(is.snp)){
				if(is.ff) {
					open(phiPrimeA(object))
					open(phiPrimeB(object))
					open(phiB(object))
					open(nuB(object))
					open(B(object))
				}
				marker.index <- i[is.X & is.snp]
				acn.index <- which(is.X & is.snp)
				acn[acn.index, ] <- C3(object, allele="A", marker.index, batch.index, j)
				if(is.ff) {
					close(phiPrimeA(object))
					close(phiPrimeB(object))
					close(phiB(object))
					close(nuB(object))
					close(B(object))
				}
			}
			if(any(!is.snp)){
				marker.index <- i[is.X & !is.snp]
				acn.index <- which(is.X & !is.snp)
				acn[acn.index, ] <- NA
				female.index <- j[object$gender[j] == 2]
				## 3. CHR X NPs: women
				if(length(female.index) > 0){
					female.batch.index <- match(unique(as.character(batch(object))[female.index]), batchNames(object))
					jj <- which(object$gender[j] == 2)
					acn[acn.index, jj] <- C1(object, marker.index, female.batch.index, female.index)
				}
				male.index <- j[object$gender[j] == 1]
				if(length(male.index) > 0){
					if(is.ff){
						open(nuB(object))
						open(phiB(object))
					}
					male.batch.index <- match(unique(as.character(batch(object))[male.index]), batchNames(object))
					jj <- which(object$gender[j] == 1)
					acn[acn.index, jj] <- C2(object, marker.index, male.batch.index, male.index, NP.X=TRUE)
					if(is.ff){
						close(nuB(object))
						close(phiB(object))
					}
				}
			}
		}
		if(is.ff){
			close(nuA(object))
			close(phiA(object))
			close(A(object))
		}
	}
	if(allele == "B"){
		if(is.ff){
			open(nuB(object))
			open(phiB(object))
			open(B(object))
		}
		if(any(!is.snp)){
			acn.index <- which(!is.snp)
			acn[acn.index, ] <- 0
		}
		if(any(is.auto)){
			auto.index <- which(is.auto & is.snp)
			if(length(auto.index) > 0){
				marker.index <- i[auto.index]
				acn[auto.index, ] <- C2(object, marker.index, batch.index, j)
			}
		}
		if(any(is.X)){
			if(is.ff){
				open(phiPrimeA(object))
				open(phiPrimeB(object))
				open(phiA(object))
				open(nuA(object))
				open(A(object))
			}
			marker.index <- i[is.X & is.snp]
			acn.index <- which(is.X & is.snp)
			acn[acn.index, ] <- C3(object, allele="B", marker.index, batch.index, j)
			if(is.ff){
				close(phiPrimeA(object))
				close(phiPrimeB(object))
				close(phiA(object))
				close(nuA(object))
				close(A(object))
			}
			if(any(!is.snp)){
				acn.index <- which(!is.snp)
				marker.index <- i[!is.snp]
				acn[acn.index, ] <- 0
			}
		}
	}
	if(is.ff){
		close(nuB(object))
		close(phiB(object))
		close(B(object))
	}
	return(acn)
}

setMethod("CA", signature=signature(object="CNSet"),
	  function(object, ...){
		  ca <- ACN(object, allele="A", ...)
		  ca[ca < 0] <- 0
		  ca[ca > 5] <- 5
		  return(ca)
	  })
setMethod("CB", signature=signature(object="CNSet"),
	  function(object, ...) {
		  cb <- ACN(object, allele="B", ...)
		  cb[cb < 0] <- 0
		  cb[cb > 5] <- 5
		  return(cb)
	  })

setMethod("totalCopynumber", signature=signature(object="CNSet"),
	  function(object, ...){
		  ca <- CA(object, ...)
		  cb <- CB(object, ...)
		  return(ca+cb)
	  })
rawCopynumber <- totalCopynumber

setMethod("posteriorProbability", signature(object="CNSet"),
	  function(object, predictRegion, copyNumber=0:4, w){
		  if(missing(w)) w <- rep(1/length(copyNumber),length(copyNumber))
		  stopifnot(sum(w)==1)
		  logI <- array(NA, dim=c(nrow(object), ncol(object), 2), dimnames=list(NULL, NULL, LETTERS[1:2]))
		  getIntensity <- function(object){
			  logI[, , 1] <- log2(A(object))
			  logI[, , 2] <- log2(B(object))
			  return(logI)
		  }
		  logI <- getIntensity(object)
		  ##gts <- lapply(as.list(0:4), genotypes)
		  prob <- array(NA, dim=c(nrow(object), ncol(object), length(copyNumber)),
				dimnames=list(NULL,NULL, paste("copynumber",copyNumber, sep="")))
		  bns <- batchNames(object)
		  snp.index <- which(isSnp(object))
		  if(length(snp.index) > 0){
			  for(i in seq_along(copyNumber)){
				  G <- genotypes(copyNumber[i])
				  P <- array(NA, dim=c(length(snp.index), ncol(object), length(G)),
					     dimnames=list(NULL,NULL,G))
				  for(g in seq_along(G)){
					  gt <- G[g]
					  for(j in seq_along(bns)){
						  this.batch <- bns[j]
						  sample.index <- which(batch(object) == this.batch)
						  mu <- predictRegion[[gt]]$mu[snp.index, , this.batch, drop=FALSE]
						  dim(mu) <- dim(mu)[1:2]
						  if(all(is.na(mu))){
							  P[, sample.index, gt] <- NA
						  } else {
							  Sigma <- predictRegion[[gt]]$cov[snp.index, , this.batch, drop=FALSE]
							  dim(Sigma) <- dim(Sigma)[1:2]
							  P[, sample.index, gt] <- dbvn(x=logI[snp.index, sample.index, , drop=FALSE], mu=mu, Sigma=Sigma)
						  }
					  }
				  }
				  ## the marginal probability for the total copy number
				  ##  -- integrate out the probability for the different genotypes
				  for(j in 1:ncol(P)){
					  PP <- P[, j, , drop=FALSE]
					  dim(PP) <- dim(PP)[c(1, 3)]
					  prob[snp.index, j, i] <- rowSums(PP, na.rm=TRUE)
				  }
			  }
		  } ## length(snp.index) > 0
		  ##---------------------------------------------------------------------------
		  ##
		  ##  nonpolymorphic markers
		  ##
		  ##---------------------------------------------------------------------------
		  np.index <- which(!isSnp(object))
		  if(length(np.index) > 0){
			  for(i in seq_along(copyNumber)){
				  G <- genotypes(copyNumber[i], is.snp=FALSE)
				  P <- array(NA, dim=c(length(np.index), ncol(object), length(G)),
					     dimnames=list(NULL,NULL,G))
				  for(g in seq_along(G)){
					  gt <- G[g]
					  for(j in seq_along(bns)){
						  this.batch <- bns[j]
						  sample.index <- which(batch(object) == this.batch)
						  mu <- predictRegion[[gt]]$mu[np.index, 1, this.batch, drop=FALSE]
						  dim(mu) <- dim(mu)[1]
						  if(all(is.na(mu))){
							  P[, sample.index, gt] <- NA
						  } else {
							  Sigma <- predictRegion[[gt]]$cov[np.index, 1, this.batch, drop=FALSE]
							  dim(Sigma) <- dim(Sigma)[1]
							  P[, sample.index, gt] <- dnorm(logI[np.index, sample.index, 1], mean=mu, sd=Sigma)
						  }
					  } ## j in seq_along(bns)
				  } ## g in seq_along(G)
				  ## the marginal probability for the total copy number
				  ##  -- integrate out the probability for the different genotypes
				  prob[np.index, , i] <- P
			  } ## seq_along(copyNumber)
		  } ## length(np.index) > 0
		  ##constant.  Probs across copy number states must
		  ##sum to one.  nc <- apply(prob, c(1,3), sum,
		  ##na.rm=TRUE)
		  wm <- matrix(w, nrow(object), length(copyNumber), byrow=TRUE)
		  nc <- matrix(NA, nrow(object), ncol(object))
		  for(j in seq_len(ncol(object))){
			  pp <- prob[, j, ] * wm
			  nc <- rowSums(pp, na.rm=TRUE)
			  prob[, j, ] <- pp/nc
		  }
		  return(prob)
	  })

setMethod("calculatePosteriorMean", signature(object="CNSet"),
	  function(object, posteriorProb, copyNumber=0:4, ...){
		  stopifnot(dim(posteriorProb)[[3]] == length(copyNumber))
		  pm <- matrix(0, nrow(object), ncol(object))
		  for(i in seq_along(copyNumber)){
			  pm <- pm + posteriorProb[, , i] * copyNumber[i]
		  }
		  return(pm)
	  })

##.bivariateCenter <- function(nu, phi){
##			  ##  lexical scope for mus, CA, CB
##			  if(CA <= 2 & CB <= 2 & (CA+CB) < 4){
##				  mus[,1, ] <- log2(nu[, 1, ] + CA *
##						    phi[, 1, ])
##				  mus[,2, ] <- log2(nu[, 2, ] + CB *
##						    phi[, 2, ])
##			  } else { ## CA > 2
##				  if(CA > 2){
##					  theta <- pi/4*Sigma[,2,]
##					  shiftA <- CA/4*phi[, 1, ] * cos(theta)
##					  shiftB <- CA/4*phi[, 1, ] * sin(theta)
##					  mus[, 1, ] <- log2(nu[, 1, ] + 2 * phi[,1,]+shiftA)
##					  mus[, 2, ] <- log2(nu[, 2, ] + CB *phi[,2,]+shiftB)
##				  }
##				  if(CB > 2){
##					  ## CB > 2
##					  theta <- pi/2-pi/4*Sigma[,2,]
##					  shiftA <- CB/4*phi[, 2, ] * cos(theta)
##					  shiftB <- CB/4*phi[, 2, ] * sin(theta)
##					  mus[, 1, ] <- log2(nu[, 1, ] + CA*phi[,1,]+shiftA)
##					  mus[, 2, ] <- log2(nu[, 2, ]+ 2*phi[,2,]+shiftB)
##				  }
##				  if(CA == 2 & CB == 2){
##					  mus[, 1, ] <- log2(nu[, 1, ] + 1/2*CA*phi[,1,])
##					  mus[, 2, ] <- log2(nu[, 2, ]+ 1/2*CB*phi[,2,])
##				  }
##			  }
##			  mus
##		  }

## for a given copy number, return a named list of bivariate normal prediction regions
##   - elements of list are named by genotype
##   'AAA'  list should have 'mu' and 'Sigma'
##                mu is a R x 2 matrix, R is number of features
##                Sigma is a R x 4 matrix.  Each row can be coerced to a 2 x 2 matrix
##          For nonpolymorphic markers, 2nd column of mu is NA and elements 2-4 of Sigma are NA
##
setMethod("predictionRegion", signature(object="CNSet", copyNumber="integer"),
	  function(object, copyNumber=0:4){
		  ## would it be more efficient to store as an array?
		  ## mu: features x genotype x allele
		  ## Sigma: features x genotype x covariance
		  stopifnot(all(copyNumber %in% 0:4))
		  getNu <- function(object){
			  nu[, 1, ] <- nuA(object)
			  nu[, 2, ] <- nuB(object)
			  nu
		  }
		  getPhi <- function(object){
			  phi[,1,] <- phiA(object)
			  phi[,2,] <- phiB(object)
			  phi
		  }
		  gts <- lapply(as.list(copyNumber), genotypes)
		  nms <- unlist(gts)
		  res <- vector("list", length(nms))
		  ##names(res) <- paste("copyNumber", copyNumber, sep="")
		  names(res) <- nms
		  bnms <- batchNames(object)
		  nu <- array(NA, dim=c(nrow(object), 2, length(bnms)))
		  phi <- array(NA, dim=c(nrow(object), 2, length(bnms)))
		  nu <- getNu(object)
		  phi <- getPhi(object)
		  ##mus <- matrix(NA, nrow(nu), 2, dimnames=list(NULL, LETTERS[1:2]))
		  mus <- array(NA, dim=c(nrow(nu), 2, length(bnms)),
			       dimnames=list(NULL, LETTERS[1:2],
			       bnms))
		  ## Sigma <- matrix(NA, nrow(nu), 3)
		  Sigma <- array(NA, dim=c(nrow(nu), 3, length(bnms)),
				 dimnames=list(NULL, c("varA", "cor", "varB"),
				 bnms))
		  bivariateCenter <- function(nu, phi){
			  ##  lexical scope for mus, CA, CB
			  mus[,1, ] <- log2(nu[, 1, ] + CA *
					    phi[, 1, ])
			  mus[,2, ] <- log2(nu[, 2, ] + CB *
					    phi[, 2, ])
			  return(mus)
		  }
		  np.index <- which(!isSnp(object))
		  for(i in seq_along(copyNumber)){
			  G <- genotypes(copyNumber[i])
			  tmp <- vector("list", length(G))
			  names(tmp) <- G
			  CN <- copyNumber[i]
			  for(g in seq_along(G)){
				  gt <- G[g]
				  CB <- g-1
				  CA <- CN-CB
				  gt.corr <- genotypeCorrelation(gt)
				  nma <- ifelse(CA == 0, "tau2A.BB", "tau2A.AA")
				  nmb <- ifelse(CB == 0, "tau2B.AA", "tau2B.BB")
				  Sigma[, 1, ] <- getVar(object, nma)
				  Sigma[, 3, ] <- getVar(object, nmb)
				  Sigma[, 2, ] <- getCor(object, gt.corr)
				  if(length(np.index) > 0){
					  Sigma[, 1, ] <- getVar(object, "tau2A.AA")
					  Sigma[np.index, 2, ] <- NA
				  }
				  res[[gt]]$mu <- bivariateCenter(nu,phi)
				  ## adjust correlation
##				  if(CA == 0 | CB == 0){
##					  Sigma[,2,] <- 0
##				  }
				  res[[gt]]$cov <- Sigma
			  }
		  }
		  res <- as(res, "PredictionRegion")
		  return(res)
	  })



setMethod("xyplotcrlmm", signature(x="formula", data="CNSet", predictRegion="list"),
	  function(x, data, predictRegion, ...){
		  fns <- featureNames(data)
		  fns <- matrix(fns, nrow(data), ncol(data), byrow=FALSE)
		  fns <- as.character(fns)
		  df <- list(A=as.numeric(log2(A(data))),
			     B=as.numeric(log2(B(data))),
			     gt=as.integer(calls(data)),
			     gt.conf=as.numeric(confs(data)),
			     snpid=fns)#, snp=snpId)
		  df <- as.data.frame(df)
		  df$snpid <- factor(df$snpid, levels=unique(df$snpid), ordered=TRUE)
		  bns <- batchNames(data)
		  predictRegion <- lapply(predictRegion, function(x, bns){
			  batch.index <- match(bns, dimnames(x$mu)[[3]])
			  x$mu <- x$mu[, , batch.index, drop=FALSE]
			  x$cov <- x$cov[, , batch.index, drop=FALSE]
			  return(x)
		  }, bns=bns)
		  ##predictRegion is an argument of ABpanel
		  xyplot(x, df, predictRegion=predictRegion, ...)
	  })

setMethod("xyplot", signature(x="formula", data="CNSet"),
	  function(x, data, ...){
		  if("predictRegion" %in% names(list(...))){
			  xyplotcrlmm(x, data, ...)
		  } else{
			  callNextMethod()
		  }
})

setMethod("calculateRBaf", signature(object="CNSet"),
	  function(object, batch.name, chrom){
		  calculateRBafCNSet(object, batch.name, chrom)
	  })

calculateRBafCNSet <- function(object, batch.name, chrom){
	if(missing(batch.name)) {
		batch.name <- batchNames(object)
		if("grandMean" %in% batch.name)
			batch.name <- batch.name[-length(batch.name)]
	}
	if(missing(chrom)) chrom <- unique(chromosome(object))
	if(!(all(batch.name %in% batchNames(object)))) stop("batch.name must be belong to batchNames(object)")
	chr <- chromosome(object)
	valid.chrs <- chr <= 23 & chr %in% chrom
	index <- which(valid.chrs)
	indexlist <- split(index, chr[index])
	J <- which(batch(object) %in% batch.name)
	sns <- sampleNames(object)[J]
	sampleindex <- split(J, factor(batch(object)[J], levels=unique(batch(object)[J])))
	if(!all(valid.chrs)) warning("Only computing log R ratios and BAFs for autosomes and chr X")
	## if ff package is loaded, these will be ff objects
	chr <- names(indexlist)
	rlist <- blist <- vector("list", length(indexlist))
	path <- ldPath()
	processByChromosome <- function(i, chr, path){
		ldPath(path)
		nr <- length(i)
		##CHR <- names(i)
		bafname <- paste("baf_chr", chr, sep="")
		rname <- paste("lrr_chr", chr, sep="")
		bmatrix <- initializeBigMatrix(bafname, nr=nr, nc=length(sns), vmode="integer")
		rmatrix <- initializeBigMatrix(rname, nr=nr, nc=length(sns), vmode="integer")
		colnames(rmatrix) <- colnames(bmatrix) <- sns
		## put rownames in order of physical position
		ix <- order(position(object)[i])
		i <- i[ix]
		rownames(rmatrix) <- rownames(bmatrix) <- featureNames(object)[i]
		for(j in seq_along(sampleindex)){
			bname <- batch.name[j]
			J <- sampleindex[[j]]
			res <- calculateRTheta(object=object, # crlmm:::
						       batch.name=bname,
						       feature.index=i)
			k <- match(sampleNames(object)[J], sns)
			bmatrix[, k] <- res[["baf"]]
			rmatrix[, k] <- res[["lrr"]]
		}
		list(bmatrix, rmatrix)
	}
	## calcualte R BAF by chromosome
	if(isPackageLoaded("ff")){
		pkgs <- c("oligoClasses", "ff", "Biobase", "crlmm")
	} else pkgs <- c("oligoClasses", "Biobase", "crlmm")
	i <- NULL
	res <- foreach(i=indexlist, chr=names(indexlist), .packages=pkgs) %dopar% {
		processByChromosome(i, chr, path)
	}
	blist <- lapply(res, "[[", 1)
	rlist <- lapply(res, "[[", 2)
	res <- list(baf=blist, lrr=rlist)
	return(res)
}

##setMethod("calculateRBaf", signature(object="CNSet"),
##	  function(object, batch.name){
##		  all.autosomes <- all(chromosome(object) < 23)
##		  if(!all.autosomes){
##			  stop("method currently only defined for chromosomes 1-22")
##		  }
##		  if(missing(batch.name)) batch.name <- batchNames(object)[1]
##		  stopifnot(batch.name %in% batchNames(object))
##		  if(length(batch.name) > 1){
##			  warning("only the first batch in batch.name processed")
##			  batch.name <- batch.name[1]
##		  }
##		  RTheta.aa <- calculateRTheta(object, "AA", batch.name)
##		  RTheta.ab <- calculateRTheta(object, "AB", batch.name)
##		  RTheta.bb <- calculateRTheta(object, "BB", batch.name)
##
##		  J <- which(batch(object) == batch.name)
##
##		  theta.aa <- matrix(RTheta.aa[, "theta"], nrow(object), length(J), byrow=FALSE)
##		  theta.ab <- matrix(RTheta.ab[, "theta"], nrow(object), length(J), byrow=FALSE)
##		  theta.bb <- matrix(RTheta.bb[, "theta"], nrow(object), length(J), byrow=FALSE)
##
##		  a <- A(object)[, J, drop=FALSE]
##		  b <- B(object)[, J, drop=FALSE] ## NA's for b where nonpolymorphic
##		  is.np <- !isSnp(object)
##		  ##b[is.np, ] <- a[is.np, ]
##		  b[is.np, ] <- 0L
##		  dns <- dimnames(a)
##		  dimnames(a) <- dimnames(b) <- NULL
##		  obs.theta <- atan2(b, a)*2/pi
##
##		  lessAA <- obs.theta < theta.aa
##		  lessAB <- obs.theta < theta.ab
##		  lessBB <- obs.theta < theta.bb
##		  grAA <- !lessAA
##		  grAB <- !lessAB
##		  grBB <- !lessBB
##		  not.na <- !is.na(theta.aa)
##		  I1 <- grAA & lessAB & not.na
##		  I2 <- grAB & lessBB & not.na
##
##		  bf <- matrix(NA, nrow(object), ncol(a))
##		  bf[I1] <- 0.5 * ((obs.theta-theta.aa)/(theta.ab-theta.aa))[I1]
##		  bf[I2] <- (.5 * (obs.theta - theta.ab) / (theta.bb - theta.ab))[I2] + 0.5
##		  bf[lessAA] <- 0
##		  bf[grBB] <- 1
##
##		  r.expected <- matrix(NA, nrow(object), ncol(a))
##		  r.aa <- matrix(RTheta.aa[, "R"], nrow(object), length(J), byrow=FALSE)
##		  r.ab <- matrix(RTheta.ab[, "R"], nrow(object), length(J), byrow=FALSE)
##		  r.bb <- matrix(RTheta.bb[, "R"], nrow(object), length(J), byrow=FALSE)
##		  rm(RTheta.aa, RTheta.ab, RTheta.bb); gc()
##		  obs.r <- a+b
##
##		  lessAA <- lessAA & not.na
##		  grBB <- grBB & not.na
##		  tmp <- ((obs.theta - theta.aa) * (r.ab-r.aa)/(theta.ab-theta.aa))[I1] + r.aa[I1]
##		  r.expected[I1] <- tmp
##		  tmp <- ((obs.theta - theta.ab) * (r.bb - r.ab)/(theta.bb-theta.ab))[I2] + r.ab[I2]
##		  r.expected[I2] <- tmp
##		  r.expected[lessAA] <- r.aa[lessAA]
##		  r.expected[grBB] <- r.bb[grBB]
##		  index.np <- which(is.np)
##		  if(length(index.np) > 0){
##			  a.np <- A(object)[index.np, J, drop=FALSE]
##			  meds <- rowMedians(a.np, na.rm=TRUE)
##			  r.expected[index.np, ] <- matrix(meds, length(index.np), ncol(a.np))
##		  }
##		  lrr <- log2(obs.r/r.expected)
##
##		  dimnames(bf) <- dimnames(lrr) <- dns
##		  res <- list(baf=bf,
##			      lrr=lrr)
##		  return(res)
##	  })


setAs("CNSet", "oligoSetList", function(from, to){
	constructOligoSetListFrom(from)
})



setMethod(OligoSetList, "CNSet", function(object,...){
	constructOligoSetListFrom(object, ...)
})
setMethod(BafLrrSetList, "CNSet", function(object,...){
  constructBafLrrSetListFrom(object, ...)
})





constructOligoSetListFrom <- function(object, ...){
  ##row.index <- seq_len(nrow(object))
  ##col.index <- seq_len(ncol(object))
  is.lds <- ifelse(is(calls(object), "ff_matrix") | is(calls(object), "ffdf"), TRUE, FALSE)
  if(is.lds) stopifnot(isPackageLoaded("ff"))
  b.r <- calculateRBaf(object, ...)
  b <- b.r[["baf"]]
  r <- b.r[["lrr"]]
  j <- match(colnames(r[[1]]), sampleNames(object))
  rns <- lapply(r, rownames)
  fDList <- foreach(featureid=rns) %do%{
    featureData(object)[match(featureid, featureNames(object)), ]
  }
  names(fDList) <- sapply(fDList, function(x) chromosome(x)[1])
  gtPlist <- gtlist <- vector("list", length(r))
  for(i in seq_along(r)){
    gtlist[[i]] <- initializeBigMatrix("call", nr=nrow(r[[i]]), nc=length(j), vmode="integer")
    gtPlist[[i]] <- initializeBigMatrix("callPr", nr=nrow(r[[i]]), nc=length(j), vmode="integer")
    featureid <- rownames(r[[i]])
    ix <- match(featureid, featureNames(object))
    rownames(gtPlist[[i]]) <- rownames(gtlist[[i]]) <- featureid
    colnames(gtPlist[[i]]) <- colnames(gtlist[[i]]) <- colnames(r[[i]])
    for(k in seq_along(j)){
      gtlist[[i]][, k] <- calls(object)[ix, j[k]]
      gtPlist[[i]][, k] <- snpCallProbability(object)[ix, j[k]]
    }
  }
  ad <- AssayDataList(baf=b, copyNumber=r, call=gtlist, callProbability=gtPlist)
  object <- new("oligoSetList",
                assayDataList=ad,
                featureDataList=fDList,
                chromosome=names(fDList),
                phenoData=phenoData(object)[j, ],
                annotation=annotation(object),
                genome=genomeBuild(object))
  return(object)
}



constructBafLrrSetListFrom <- function(object, ...){
	is.lds <- ifelse(is(calls(object), "ff_matrix") | is(calls(object), "ffdf"), TRUE, FALSE)
	if(is.lds) stopifnot(isPackageLoaded("ff"))
	b.r <- calculateRBaf(object, ...)
	b <- b.r[["baf"]]
	r <- b.r[["lrr"]]
	j <- match(colnames(r[[1]]), sampleNames(object))
	rns <- lapply(r, rownames)
	featureid <- NULL
	fDList <- foreach(featureid=rns) %do%{
		featureData(object)[match(featureid, featureNames(object)), ]
	}
	names(fDList) <- sapply(fDList, function(x) chromosome(x)[1])
	ad <- AssayDataList(baf=b, lrr=r)
	object <- new("BafLrrSetList",
		      assayDataList=ad,
		      featureDataList=fDList,
		      chromosome=names(fDList),
		      phenoData=phenoData(object)[j, ],
		      annotation=annotation(object),
		      genome=genomeBuild(object))
	return(object)
}

computeBR <- constructBafLrrSetListFrom

setMethod("[", signature(x="PredictionRegion"),
	  function(x, i, j, ..., drop=FALSE){
		  if(missing(i) & missing(j)) return(x)
		  if(!missing(i)){
			  xx <- lapply(x, function(x, i){
				  x$mu <- x$mu[i,,,drop=FALSE]
				  x$cov <- x$cov[i, , , drop=FALSE]
				  return(x)
			  }, i=i)
			  x <- as(xx, "PredictionRegion")
		  }
		  return(x)
	  })

getSigma2 <- function(object){
	phi.a <- phiA(object)
	phi.b <- phiB(object)
	cbind(phi.a, phi.b)
}

getCor <- function(object, gt){
	if(gt =="NULL") return(rep(0, nrow(object)))
	assayDataElement(batchStatistics(object), paste("corr", gt, sep=""))
}

getTau2 <- function(object, gt){
	bs <- batchStatistics(object)
	nma <- paste("tau2A", gt, sep=".")
	nmb <- paste("tau2B", gt, sep=".")
	tau2.a <- assayDataElement(bs, nma)
	tau2.b <- assayDataElement(bs, nmb)
	cbind(tau2.a, tau2.b)
}

getVar <- function(object, nm){
	bs <- batchStatistics(object)
	assayDataElement(bs, nm)
}

setMethod("nuA", signature=signature(object="CNSet"), function(object) nu(object, "A"))
setMethod("nuB", signature=signature(object="CNSet"), function(object) nu(object, "B"))

setMethod("phiA", signature=signature(object="CNSet"), function(object) phi(object, "A"))
setMethod("phiB", signature=signature(object="CNSet"), function(object) phi(object, "B"))
setMethod("phiPrimeA", signature=signature(object="CNSet"), function(object) {
	assayDataElement(batchStatistics(object), "phiPrimeA")
})
setMethod("phiPrimeB", signature=signature(object="CNSet"), function(object) {
	assayDataElement(batchStatistics(object), "phiPrimeB")
})
setMethod("tau2A", signature=signature(object="CNSet"), function(object) tau2(object, "A"))
setMethod("tau2B", signature=signature(object="CNSet"), function(object) tau2(object, "B"))

setMethod("Ns", signature=signature(object="CNSet"),
	   function(object, ...){
		   Ns(batchStatistics(object), ...)
	   })
setMethod("corr", signature=signature(object="CNSet"),
	   function(object, ...){
		   corr(batchStatistics(object), ...)
	   })

setMethod("medians", signature=signature(object="CNSet"),
	   function(object, ...){
		   medians(batchStatistics(object), ...)
	   })
setMethod("mads", signature=signature(object="CNSet"),
	   function(object, ...){
		   mads(batchStatistics(object), ...)
	   })
setMethod("tau2", signature=signature(object="CNSet"),
	   function(object, ...){
		   tau2(batchStatistics(object), ...)
	   })
##---------------------------------------------------------------------------
## Number of samples with Genotype AA, AB, or BB by batch
setMethod("N.AA", signature=signature(object="CNSet"), function(object){
	assayDataElement(batchStatistics(object), "N.AA")
})
setMethod("N.AB", signature=signature(object="CNSet"), function(object){
	assayDataElement(batchStatistics(object), "N.AB")
})
setMethod("N.BB", signature=signature(object="CNSet"), function(object){
	assayDataElement(batchStatistics(object), "N.BB")
})

setReplaceMethod("N.AA", signature=signature(object="CNSet", value="ff_or_matrix"),
	  function(object, value){
		  linearParamElementReplace(object, "N.AA", value)
	  })
setReplaceMethod("N.AB", signature=signature(object="CNSet", value="ff_or_matrix"),
	  function(object, value){
		  linearParamElementReplace(object, "N.AB", value)
	  })
setReplaceMethod("N.BB", signature=signature(object="CNSet", value="ff_or_matrix"),
	  function(object, value){
		  linearParamElementReplace(object, "N.BB", value)
	  })


##---------------------------------------------------------------------------
##  median intensity by genotype cluster for each allele
##should we update the entire matrix rather than one column...
setMethod("medianA.AA", signature=signature(object="CNSet"), function(object) {
	assayDataElement(batchStatistics(object), "medianA.AA")
})
setMethod("medianA.AB", signature=signature(object="CNSet"), function(object) {
	assayDataElement(batchStatistics(object), "medianA.AB")
})
setMethod("medianA.BB", signature=signature(object="CNSet"), function(object) {
	assayDataElement(batchStatistics(object), "medianA.BB")
})
setMethod("medianB.AA", signature=signature(object="CNSet"), function(object) {
	assayDataElement(batchStatistics(object), "medianB.AA")
})
setMethod("medianB.AB", signature=signature(object="CNSet"), function(object) {
	assayDataElement(batchStatistics(object), "medianB.AB")
})
setMethod("medianB.BB", signature=signature(object="CNSet"), function(object) {
	assayDataElement(batchStatistics(object), "medianB.BB")
})
setReplaceMethod("medianA.AA", signature=signature(object="CNSet", value="ff_or_matrix"),
	  function(object, value){
		  linearParamElementReplace(object, "medianA.AA", value)
	  })
setReplaceMethod("medianA.AB", signature=signature(object="CNSet", value="ff_or_matrix"),
	  function(object, value){
		  linearParamElementReplace(object, "medianA.AB", value)
	  })
setReplaceMethod("medianA.BB", signature=signature(object="CNSet", value="ff_or_matrix"),
	  function(object, value){
		  linearParamElementReplace(object, "medianA.BB", value)
	  })
setReplaceMethod("medianB.AA", signature=signature(object="CNSet", value="ff_or_matrix"),
	  function(object, value){
		  linearParamElementReplace(object, "medianB.AA", value)
	  })
setReplaceMethod("medianB.AB", signature=signature(object="CNSet", value="ff_or_matrix"),
	  function(object, value){
		  linearParamElementReplace(object, "medianB.AB", value)
	  })
setReplaceMethod("medianB.BB", signature=signature(object="CNSet", value="ff_or_matrix"),
	  function(object, value){
		  linearParamElementReplace(object, "medianB.BB", value)
	  })

##---------------------------------------------------------------------------
##  mad intensity by genotype cluster for each allele
setMethod("madA.AA", signature=signature(object="CNSet"), function(object) {
	assayDataElement(batchStatistics(object), "madA.AA")
})
setMethod("madA.AB", signature=signature(object="CNSet"), function(object) {
	assayDataElement(batchStatistics(object), "madA.AB")
})
setMethod("madA.BB", signature=signature(object="CNSet"), function(object) {
	assayDataElement(batchStatistics(object), "madA.BB")
})
setMethod("madB.AA", signature=signature(object="CNSet"), function(object) {
	assayDataElement(batchStatistics(object), "madB.AA")
})
setMethod("madB.AB", signature=signature(object="CNSet"), function(object) {
	assayDataElement(batchStatistics(object), "madB.AB")
})
setMethod("madB.BB", signature=signature(object="CNSet"), function(object) {
	assayDataElement(batchStatistics(object), "madB.BB")
})
setReplaceMethod("madA.AA", signature=signature(object="CNSet", value="ff_or_matrix"),
	  function(object, value){
		  linearParamElementReplace(object, "madA.AA", value)
	  })
setReplaceMethod("madA.AB", signature=signature(object="CNSet", value="ff_or_matrix"),
	  function(object, value){
		  linearParamElementReplace(object, "madA.AB", value)
	  })
setReplaceMethod("madA.BB", signature=signature(object="CNSet", value="ff_or_matrix"),
	  function(object, value){
		  linearParamElementReplace(object, "madA.BB", value)
	  })
setReplaceMethod("madB.AA", signature=signature(object="CNSet", value="ff_or_matrix"),
	  function(object, value){
		  linearParamElementReplace(object, "madB.AA", value)
	  })
setReplaceMethod("madB.AB", signature=signature(object="CNSet", value="ff_or_matrix"),
	  function(object, value){
		  linearParamElementReplace(object, "madB.AB", value)
	  })
setReplaceMethod("madB.BB", signature=signature(object="CNSet", value="ff_or_matrix"),
	  function(object, value){
		  linearParamElementReplace(object, "madB.BB", value)
	  })

##---------------------------------------------------------------------------
##  mad of log(intensities) by genotype cluster for each allele
setMethod("tau2A.AA", signature=signature(object="CNSet"), function(object) {
	assayDataElement(batchStatistics(object), "tau2A.AA")
})
setMethod("tau2A.BB", signature=signature(object="CNSet"), function(object) {
	assayDataElement(batchStatistics(object), "tau2A.BB")
})
setMethod("tau2B.AA", signature=signature(object="CNSet"), function(object) {
	assayDataElement(batchStatistics(object), "tau2B.AA")
})
setMethod("tau2B.BB", signature=signature(object="CNSet"), function(object) {
	assayDataElement(batchStatistics(object), "tau2B.BB")
})
setReplaceMethod("tau2A.AA", signature=signature(object="CNSet", value="ff_or_matrix"),
	  function(object, value){
		  linearParamElementReplace(object, "tau2A.AA", value)
	  })
setReplaceMethod("tau2A.BB", signature=signature(object="CNSet", value="ff_or_matrix"),
	  function(object, value){
		  linearParamElementReplace(object, "tau2A.BB", value)
	  })
setReplaceMethod("tau2B.AA", signature=signature(object="CNSet", value="ff_or_matrix"),
	  function(object, value){
		  linearParamElementReplace(object, "tau2B.AA", value)
	  })
setReplaceMethod("tau2B.BB", signature=signature(object="CNSet", value="ff_or_matrix"),
	  function(object, value){
		  linearParamElementReplace(object, "tau2B.BB", value)
	  })

## correlation of log2A and log2B within each genotype cluster
setMethod("corrAA", signature=signature(object="CNSet"), function(object){
	assayDataElement(batchStatistics(object), "corrAA")
  })
setMethod("corrAB", signature=signature(object="CNSet"), function(object){
	assayDataElement(batchStatistics(object), "corrAB")
  })
setMethod("corrBB", signature=signature(object="CNSet"), function(object){
	assayDataElement(batchStatistics(object), "corrBB")
  })
setReplaceMethod("corrAA", signature=signature(object="CNSet", value="ff_or_matrix"),
	  function(object, value){
		  linearParamElementReplace(object, "corrAA", value)
})

setReplaceMethod("corrAB", signature=signature(object="CNSet", value="ff_or_matrix"),
	  function(object, value){
		  linearParamElementReplace(object, "corrAB", value)
})

setReplaceMethod("corrBB", signature=signature(object="CNSet", value="ff_or_matrix"),
	  function(object, value){
		  linearParamElementReplace(object, "corrBB", value)
})

setReplaceMethod("phiPrimeA", signature=signature(object="CNSet", value="ff_or_matrix"),
	  function(object, value){
		  linearParamElementReplace(object, "phiPrimeA", value)
})

setReplaceMethod("phiPrimeB", signature=signature(object="CNSet", value="ff_or_matrix"),
	  function(object, value){
		  linearParamElementReplace(object, "phiPrimeB", value)
})

##setMethod("mad.AA", signature=signature(object="CNSet"), function(object) mads(object, "AA"))
##setMethod("mad.AB", signature=signature(object="CNSet"), function(object) mads(object, "AB"))
##setMethod("mad.BB", signature=signature(object="CNSet"), function(object) mads(object, "BB"))
##

##setReplaceMethod("median.AA", signature=signature(object="CNSet", value="ff_or_matrix"),
##	  function(object, value){
##		  linearParamElementReplace(object, "median.AA", value)
##	  })
##setReplaceMethod("median.AB", signature=signature(object="CNSet", value="ff_or_matrix"),
##	  function(object, value){
##		  linearParamElementReplace(object, "median.AB", value)
##	  })
##setReplaceMethod("median.BB", signature=signature(object="CNSet", value="ff_or_matrix"),
##	  function(object, value){
##		  linearParamElementReplace(object, "median.BB", value)
##	  })
##setReplaceMethod("mad.AA", signature=signature(object="CNSet", value="ff_or_matrix"),
##	  function(object, value){
##		  linearParamElementReplace(object, "mad.AA", value)
##	  })
##setReplaceMethod("mad.AB", signature=signature(object="CNSet", value="ff_or_matrix"),
##	  function(object, value){
##		  linearParamElementReplace(object, "mad.AB", value)
##	  })
##setReplaceMethod("mad.BB", signature=signature(object="CNSet", value="ff_or_matrix"),
##	  function(object, value){
##		  linearParamElementReplace(object, "mad.BB", value)
##	  })

setReplaceMethod("nuA", signature=signature(object="CNSet", value="ff_or_matrix"),
	  function(object, value){
		  linearParamElementReplace(object, "nuA", value)
	  })

setReplaceMethod("nuB", signature=signature(object="CNSet", value="ff_or_matrix"),
	  function(object, value){
		  linearParamElementReplace(object, "nuB", value)
})

setReplaceMethod("phiA", signature=signature(object="CNSet", value="ff_or_matrix"),
	  function(object, value){
		  linearParamElementReplace(object, "phiA", value)
})

setReplaceMethod("phiB", signature=signature(object="CNSet", value="ff_or_matrix"),
	  function(object, value){
		  linearParamElementReplace(object, "phiB", value)
})

setReplaceMethod("tau2A", signature=signature(object="CNSet", value="ff_or_matrix"),
	  function(object, value){
		  linearParamElementReplace(object, "tau2A", value)
})

setReplaceMethod("tau2B", signature=signature(object="CNSet", value="ff_or_matrix"),
	  function(object, value){
		  linearParamElementReplace(object, "tau2B", value)
})



setReplaceMethod("flags", signature=signature(object="CNSet", value="ff_or_matrix"),
		 function(object, value){
			 linearParamElementReplace(object, "flags", value)
})
setReplaceMethod("flags", signature=signature(object="CNSet", value="ff_matrix"),
		 function(object, value){
			 linearParamElementReplace(object, "flags", value)
})

setReplaceMethod("snpCall", "CNSet",
                 function(object, value){
			 assayDataElementReplace(object, "call", value)
		 })

setReplaceMethod("snpCallProbability", "CNSet",
                 function(object, value){
			 assayDataElementReplace(object, "callProbability", value)
		 })

genotypingTest <- function(){
	require(genomewidesnp6Crlmm) & require(hapmapsnp6)
	path <- system.file("celFiles", package="hapmapsnp6")
	cels <- list.celfiles(path, full.names=TRUE)
	ocProbesets(50e3)
	batch <- as.factor(rep("A", length(cels)))
	(cnSet <- genotype(cels, cdfName="genomewidesnp6", batch=batch))
	library(ff)
	ldPath(tempdir())
	(cnSet2 <- genotype(cels, cdfName="genomewidesnp6", batch=batch))
	checkTrue(all.equal(calls(cnSet), calls(cnSet2)[,]))
}

#genotypingTestIllumina <- function(){
#	setwd("/thumper/ctsa/snpmicroarray/illumina/IDATS/370k/")
#	library(crlmm)
#	library(ff)
#	ldPath(tempdir())
#
#	samples <- read.csv("samples370k.csv", as.is=TRUE)
#	RG <- readIdatFiles(sampleSheet=samples,
#			    arrayInfoColNames=list(barcode=NULL, position="SentrixPosition"),
#			    saveDate=TRUE)
#
#	crlmmResult <-  crlmmIllumina(RG=RG,
#				      cdfName="human370v1c",
#				      returnParams=TRUE)
#	checkTrue(is(calls(crlmmResult)[1:5,1], "integer"))
#
#	crlmmResult2 <- crlmmIlluminaV2(sampleSheet=samples,
#					arrayInfoColNames=list(barcode=NULL, position="SentrixPosition"),
#					saveDate=TRUE, cdfName="human370v1c", returnParams=TRUE)
#	checkTrue(identical(calls(crlmmResult)[1:5, ]),
#		  identical(calls(crlmmResult2)[1:5, ]))
#
#	crlmmResult3 <- genotype.Illumina(sampleSheet=samples,
#					  arrayInfoColNames=list(barcode=NULL,
#					  position="SentrixPosition"),
#					  saveDate=TRUE, cdfName="human370v1c",
#					  batch = as.factor(rep(1, nrow(samples))))
#
#	checkTrue(identical(calls(crlmmResult)[1:5, ]),
#		  identical(calls(crlmmResult3)[1:5, ]))
#
#}

# function below works OK provided all .idat files are in the current working directory
# - could add an option to allow files in Illumina directory structure to be handled
# or to use the optional 'Path' column in sampleSheet
# - there is a lot of header information that is currently discarded - could try and store this somewhere in the resulting NChannelSet

readIdatFiles = function(sampleSheet=NULL,
			  arrayNames=NULL,
			  ids=NULL,
			  path=".",
			  arrayInfoColNames=list(barcode="SentrixBarcode_A", position="SentrixPosition_A"),
			  highDensity=FALSE,
			  sep="_",
			  fileExt=list(green="Grn.idat", red="Red.idat"),
			  saveDate=FALSE, verbose=FALSE) {
       verbose <- FALSE
       if(!is.null(arrayNames)) {
               pd = new("AnnotatedDataFrame", data = data.frame(Sample_ID=arrayNames))
       }
       if(!is.null(sampleSheet)) { # get array info from Illumina's sample sheet
	       if(is.null(arrayNames)){
		       if(!is.null(arrayInfoColNames$barcode) && (arrayInfoColNames$barcode %in% colnames(sampleSheet))) {
			       barcode = sampleSheet[,arrayInfoColNames$barcode]
			       arrayNames=barcode
		       }
		       if(!is.null(arrayInfoColNames$position) && (arrayInfoColNames$position %in% colnames(sampleSheet))) {
			       position = sampleSheet[,arrayInfoColNames$position]
			       if(is.null(arrayNames))
				       arrayNames=position
			       else
				       arrayNames = paste(arrayNames, position, sep=sep)
			       if(highDensity) {
				       hdExt = list(A="R01C01", B="R01C02", C="R02C01", D="R02C02")
				       for(i in names(hdExt))
					       arrayNames = sub(paste(sep, i, sep=""), paste(sep, hdExt[[i]], sep=""), arrayNames)
			       }
		       }
	       }
	       pd = new("AnnotatedDataFrame", data = sampleSheet)
	       sampleNames(pd) = make.unique(basename(arrayNames))
       }
       if(is.null(arrayNames)) {
               arrayNames = gsub(paste(sep, fileExt$green, sep=""), "", dir(pattern=fileExt$green, path=path))
               if(!is.null(sampleSheet)) {
                      sampleSheet=NULL
                      cat("Could not find required info in \'sampleSheet\' - ignoring.  Check \'sampleSheet\' and/or \'arrayInfoColNames\'\n")
               }
               pd = new("AnnotatedDataFrame", data = data.frame(Sample_ID=arrayNames))
       }
       narrays = length(arrayNames)
       grnfiles = paste(arrayNames, fileExt$green, sep=sep)
       redfiles = paste(arrayNames, fileExt$red, sep=sep)
       if(length(grnfiles)==0 || length(redfiles)==0)
	       stop("Cannot find .idat files")
       if(length(grnfiles)!=length(redfiles))
	       stop("Cannot find matching .idat files")
       if(path[1] != "." & path[1] != ""){
	       grnidats = file.path(path, grnfiles)
	       redidats = file.path(path, redfiles)
       }  else {
	       message("'path' arg not set.  Assuming files are in local directory, or that complete path is provided in 'arrayNames'")
	       grnidats = grnfiles
	       redidats = redfiles
       }
       if(!all(file.exists(grnidats))) stop("Missing some of the *Grn.idat files")
       if(!all(file.exists(redidats))) stop("Missing some of the *Red.idat files")
       headerInfo = list(nProbes = rep(NA, narrays),
                         Barcode = rep(NA, narrays),
                         ChipType = rep(NA, narrays),
                         Manifest = rep(NA, narrays), # not sure about this one - sometimes blank
                         Position = rep(NA, narrays)) # this may also vary a bit
       dates = list(decode=rep(NA, narrays),
                    scan=rep(NA, narrays))
       ## read in the data
       for(i in seq_along(arrayNames)) {
	       if(verbose) {
	       ## RS
		       ##cat("reading", arrayNames[i], "\t")
		       cat("reading", basename(arrayNames[i]), "\t")
		       cat(paste(sep, fileExt$green, sep=""), "\t")
	       }
	       idsG = idsR = G = R = NULL
	       G = readIDAT(grnidats[i])
	       idsG = rownames(G$Quants)
	       headerInfo$nProbes[i] = G$nSNPsRead
	       headerInfo$Barcode[i] = G$Barcode
	       headerInfo$ChipType[i] = G$ChipType
	       headerInfo$Manifest[i] = G$Unknown$MostlyNull
	       headerInfo$Position[i] = G$Unknowns$MostlyA
               if(headerInfo$nProbes[i]>(headerInfo$nProbes[1]+headerInfo$nProbes[1]*0.04) || headerInfo$nProbes[i]<(headerInfo$nProbes[1]-headerInfo$nProbes[1]*0.04)) {
		       warning("Chips are not of the same type.  Skipping ", basename(grnidats[i]), " and ", basename(redidats[i]))
		       next()
	       }
               if(saveDate) {
                      if(nrow(G$RunInfo)>=2) {
	              dates$decode[i] = G$RunInfo[1, 1]
	              dates$scan[i] = G$RunInfo[2, 1]
                      }
               }
	       if(i==1) {
		       if(is.null(ids) && !is.null(G)){
			       ids = idsG
		       } # else stop("Could not find probe IDs")
		       nprobes = length(ids)
		       narrays = length(arrayNames)
#                       if(cdfName=="nopackage") {
    	               RG = new("NChannelSet",
                                 R = initializeBigMatrix(name = "R", nr = nprobes, nc = narrays, vmode = "integer"),
                                 G = initializeBigMatrix(name = "G", nr = nprobes, nc = narrays, vmode = "integer"),
		                 zero = initializeBigMatrix(name = "zero", nr = nprobes, nc = narrays, vmode = "integer"),
				 annotation=headerInfo$Manifest[1],
				 phenoData=pd, storage.mode="environment")
#                       } else {
#    	               RG = new("NChannelSet",
#                                 R = matrix(0, nprobes, narrays),
#                                 G = matrix(0, nprobes, narrays),
#		                 zero = matrix(1, nprobes, narrays),
#				 annotation=headerInfo$Manifest[1],
#				 phenoData=pd, storage.mode="environment")                           
#                       }
		       featureNames(RG) = ids
                       
		       if(!is.null(sampleSheet) && !is.null(sampleSheet$Sample_ID)){
		            sampleNames(RG) = make.unique(sampleSheet$Sample_ID)
		       } else  sampleNames(RG) = make.unique(basename(arrayNames))
		       gc(verbose=FALSE)
	       }
	       if(length(ids)==length(idsG)) {
		       if(sum(ids==idsG)==nprobes) {
			       RG@assayData$G[,i] = G$Quants[, "Mean"]
			       zeroG = G$Quants[, "NBeads"]==0
		       }
	       } else {
		       indG = match(ids, idsG)
                       nasG = is.na(indG)
		       RG@assayData$G[!nasG,i] = G$Quants[indG[!nasG], "Mean"]
		       zeroG = G$Quants[indG[!nasG], "NBeads"]==0
	       }
	       rm(G)
	       gc(verbose=FALSE)
	       if(verbose) {
                      cat(paste(sep, fileExt$red, sep=""), "\n")
	       }
	       R = readIDAT(redidats[i])
	       idsR = rownames(R$Quants)

	       if(length(ids)==length(idsG)) {
		       if(sum(ids==idsR)==nprobes) {
			       RG@assayData$R[,i] = R$Quants[ ,"Mean"]
		               zeroR = R$Quants[ ,"NBeads"]==0
                               RG@assayData$zero[,i] = zeroG | zeroR
		       }
	       } else {
		       indR = match(ids, idsR)
	               nasR = is.na(indR)
		       RG@assayData$R[!nasR,i] = R$Quants[indR[!nasR], "Mean"]
		       zeroR = R$Quants[indR[!nasR], "NBeads"]==0
                       RG@assayData$zero[!nasR,i] = zeroG | zeroR
	       }
#	       RG@assayData$zero[,i] = zeroG | zeroR
	       rm(R, zeroG, zeroR)
	       gc(verbose=FALSE)
       }
       if(saveDate) {
	       protocolData(RG)[["ScanDate"]] = dates$scan
       }
       storageMode(RG) = "lockedEnvironment"
       RG
}


getNumberOfSNPs = function(afile, path){
    fullfilename = file.path(path, afile)
    headerSection = readLines(fullfilename, n=15)

    headerLine = headerSection[10][1]
    delimiterList = c(",", "\t")

    headers = unlist(strsplit(headerLine, delimiterList[1]))
    if (length(headers)!=1) {
        delimiterIndex = 1
    }
    if (length(headers) == 1) {
        headers = unlist(strsplit(headerLine, delimiterList[2]))
        if (length(headers) != 1) {
            delimiterIndex = 2
        }
        if (length(headers) == 1) {
            stop("Input file ", fullfilename, " is not delimited by either comm(,) or tab(\\t)")
        }
    }

    SNPLine = headerSection[5][1]
    elements = unlist(strsplit(SNPLine, delimiterList[delimiterIndex]))
    numSNP = as.integer(elements[2])
    return(numSNP)
}

checkNumberOfSNPs = function(filenames, path){
    numSNP = getNumberOfSNPs(filenames[1], path)
    if (length(filenames) > 1) {
		for (i in 2:length(filenames)){
			if (getNumberOfSNPs(filenames[i], path) != numSNP){
				return(FALSE)
			}
		}
    }
    return(TRUE)
}


getNumberOfSamples = function(filenames, path, numSNP){
    sampleCount = rep(0, length(filenames))
    for (i in 1:length(filenames)){
    	# number of sample in input file line 7 is not reliable, calculate from number of lines and number of SNPs
    	fullfilename = file.path(path, filenames[i])
	LineCount = .Call("countFileLines", fullfilename)
  	if (((LineCount - 10) %% numSNP) != 0){
           stop("Please check input file: ", fullfilename, " Line count is not a multiple of number of SNPs")
    	}
	sampleCount[i] = LineCount %/% numSNP
    }	
    return(sampleCount)
}

processOneGenCallFile = function(afile, numSNP, numSample,
    colnames=list("SampleID"="Sample ID", "SNPID"="SNP Name", "XRaw"="X Raw", "YRaw"="Y Raw"),
    verbose=FALSE) {
  
    headerSection = readLines(afile, n=15)

    headerLine = headerSection[10][1]
    delimiterList = c(",", "\t")

    headers = unlist(strsplit(headerLine, delimiterList[1]))
    if (length(headers)!=1) {
        delimiterIndex = 1
    }
    if (length(headers) == 1) {
        headers = unlist(strsplit(headerLine, delimiterList[2]))
        if (length(headers) != 1) {
            delimiterIndex = 2
        }
        if (length(headers) == 1) {
            stop("Input file is not delimited by either comm(,) or tab(\\t)")
        }
    }
     
    if(sum(is.na(match(colnames, headers))) != 0)
	stop("Cannot find required columns: ", colnames[is.na(match(colnames, headers))], " in ", file, "\nPlease check whether this data was exported.")
    
    SNPIDPos = which(headers == colnames$SNPID)
    sampleIDPos = which(headers == colnames$SampleID)
    XValuePos = which(headers == colnames$XRaw)
    YValuePos = which(headers == colnames$YRaw)   
        
    if(verbose) {
        message("Number of SNPs in file: ", afile, " is ", numSNP, " and number of samples is ", numSample)
        if (delimiterIndex == 1) message("File is comma-seperated. ")
        if (delimiterIndex == 2) message("File is tab-seperated. ")              
    }

    values = .Call("readGenCallOutputCFunc", afile, numSNP, numSample, SNPIDPos, sampleIDPos, XValuePos, YValuePos, delimiterIndex)
             
    return(values)
}

readGenCallOutput = function(filenames, path=".", cdfName,
    colnames=list("SampleID"="Sample ID", "SNPID"="SNP Name", "XRaw"="X Raw", "YRaw"="Y Raw"),
    type=list("SampleID"="character", "SNPID"="character", "XRaw"="integer", "YRaw"="integer"), verbose=FALSE) {

    if(!identical(names(type), names(colnames)))
       stop("The arguments 'colnames' and 'type' must have consistent names")
    if(missing(cdfName)) stop("must specify cdfName")

    if (!checkNumberOfSNPs(filenames, path)){
       stop("Number of SNPs in each file must be identical to form one output NChannelSet object.")
    }

    if (verbose) message("Checking number of samples and features in each file. ")
    numSNP = getNumberOfSNPs(filenames[1], path)
    sampleCounts = getNumberOfSamples(filenames, path, numSNP)
    numSample = sum(sampleCounts)

    X = initializeBigMatrix(name = "X", nr = numSNP, nc = numSample, vmode = "integer")
    Y = initializeBigMatrix(name = "Y", nr = numSNP, nc = numSample, vmode = "integer")
    zero = initializeBigMatrix(name = "zero", nr = numSNP, nc = numSample, vmode = "integer")
    
    totSampleNames = rep(NA, numSample)
   
    baseIndex = 1
    if (verbose) message("Start processing ", length(filenames), " input file(s)")
    for (i in 1:length(filenames)){
    	fullfilename = file.path(path, filenames[i])
    	valuesThisFile = processOneGenCallFile(fullfilename, numSNP, sampleCounts[i], colnames, verbose)
	
	if (i == 1){
	    totSNPNames = rownames(valuesThisFile$Xvalues)
	} else {
	    # matching on SNP names? now assume they come in order
	}
	maxIndex = baseIndex + sampleCounts[i] - 1
        m=match(totSNPNames, rownames(valuesThisFile$Xvalues))
        valuesThisFile$Xvalues=valuesThisFile$Xvalues[m,]
        valuesThisFile$Yvalues=valuesThisFile$Yvalues[m,]
	X[, baseIndex:maxIndex] = valuesThisFile$Xvalues
	Y[, baseIndex:maxIndex] = valuesThisFile$Yvalues
        zero[, baseIndex:maxIndex] = (X[, baseIndex:maxIndex] == 0) || (Y[, baseIndex:maxIndex] == 0)
	totSampleNames[baseIndex:(baseIndex + sampleCounts[i] - 1)] = colnames(valuesThisFile$Xvalues)
	rm(valuesThisFile)
	baseIndex = baseIndex + sampleCounts[i]
    }

    if(verbose) message("Creating NChannelSet object\n")

    XY = new("NChannelSet", X=X, Y=Y, zero=zero, annotation=cdfName, storage.mode = "environment")
    sampleNames(XY) = totSampleNames
    featureNames(XY) = totSNPNames
    
    if(verbose)
    cat("Done\n")

    XY
}



RGtoXY = function (RG, chipType, anno, verbose = TRUE) 
{
    if (chipType != "nopackage") {
        needToLoad <- !all(sapply(c("addressA", "addressB", "base"), 
            isLoaded))
        if (needToLoad) {
            chipList = c("human1mv1c", "human370v1c", "human650v3a", 
                "human610quadv1b", "human660quadv1a", "human370quadv3c", 
                "human550v3b", "human1mduov3b", "humanomni1quadv1b", 
                "humanomni25quadv1b", "humanomni258v1a", "humanomni258v1p1b", 
                "humanomni5quadv1b", "humanomniexpress12v1b", 
                "humanimmuno12v1b", "humancytosnp12v2p1h", "humanexome12v1p2a", 
                "humanomniexpexome8v1p1b")
            if (missing(chipType)) {
                chipType = match.arg(cleancdfname(annotation(RG)), 
                  chipList)
            }
            else chipType = match.arg(cleancdfname(chipType), 
                chipList)
            pkgname = getCrlmmAnnotationName(chipType)
            if (!require(pkgname, character.only = TRUE, quietly = !verbose)) {
                suggCall = paste("library(", pkgname, ", lib.loc='/Altern/Lib/Loc')", 
                  sep = "")
                msg = paste("If", pkgname, "is installed on an alternative location, please load it manually by using", 
                  suggCall)
                message(strwrap(msg))
                stop("Package ", pkgname, " could not be found.")
                rm(suggCall, msg)
            }
            if (verbose) 
                message("Loading chip annotation information.")
            loader("address.rda", .crlmmPkgEnv, pkgname)
        }
        aids = getVarInEnv("addressA")
        bids = getVarInEnv("addressB")
        snpbase = getVarInEnv("base")
        ids = names(aids)
        nsnps = length(aids)
        narrays = ncol(RG)
        infI = !is.na(bids) & bids != 0
        aord = match(aids, featureNames(RG))
        bord = match(bids, featureNames(RG))
        xyPhenoData = AnnotatedDataFrame(data = RG@phenoData@data, 
            varMetadata = RG@phenoData@varMetadata)
        levels(xyPhenoData@varMetadata$channel) = c("X", "Y", 
            "zero", "_ALL_")
        XY <- new("NChannelSet", assayDataNew(X = initializeBigMatrix(name = "X", 
            nr = nsnps, nc = narrays, vmode = "integer"), Y = initializeBigMatrix(name = "Y", 
            nr = nsnps, nc = narrays, vmode = "integer"), zero = initializeBigMatrix(name = "zero", 
            nr = nsnps, nc = narrays, vmode = "integer")), phenoData = xyPhenoData, 
            protocolData = RG@protocolData, annotation = chipType)
        storageMode(XY) = "environment"
        featureNames(XY) = ids
        sampleNames(XY) = sampleNames(RG)
        rm(list = c("bord", "infI", "aids", "bids", "ids", "snpbase"))
        gc(verbose = FALSE)
        not.na <- !is.na(aord)
        not.na.aord <- aord[not.na]
        r <- as.matrix(assayData(RG)[["R"]][not.na.aord, ])
        XY@assayData$X[not.na, ] <- r
        rm(r)
        gc(verbose = FALSE)
        g <- as.matrix(assayData(RG)[["G"]][not.na.aord, ])
        XY@assayData$Y[not.na, ] <- g
        rm(g)
        gc(verbose = FALSE)
        z <- as.matrix(assayData(RG)[["zero"]][not.na.aord, ])
        XY@assayData$zero[not.na, ] <- z
        rm(z)
        gc(verbose = FALSE)
        rm(RG)
        gc(verbose = FALSE)
    }
    if (chipType == "nopackage" && !is.null(anno)) {
        pkgname = NULL
        if(sum(colnames(anno@data)=="AddressA_ID")!=1){
           
           message("annotation format is wrong, could not find a column with name: AddressA_ID")
         }else{
         anno2=anno@data
           # chr = as.character(anno$Chr)
  #chr[chr=="X"] = 23
  #  chr[chr=="Y"] = 24
  #  chr[chr=="XY"] = 25
#chr[chr=="MT"] = 26
#anno[,"chromosome"]=as.integer(chr)

       aids = anno2[, "AddressA_ID"]
        bids = anno2[, "AddressB_ID"]
        snpbase = anno2[, "SNP"]
        names(aids) = ids = anno2[, "featureNames"]
        nsnps = length(aids)
        narrays = ncol(RG)
        infI = !is.na(bids) & bids != 0
        aord = match(aids, featureNames(RG))
        bord = match(bids,featureNames(RG))
        xyPhenoData = AnnotatedDataFrame(data = RG@phenoData@data, 
            varMetadata = RG@phenoData@varMetadata)
        levels(xyPhenoData@varMetadata$channel) = c("X", "Y", 
            "zero", "_ALL_")
        XY <- new("NChannelSet", assayDataNew(X = initializeBigMatrix(name = "X", 
            nr = nsnps, nc = narrays, vmode = "integer"), Y = initializeBigMatrix(name = "Y", 
            nr = nsnps, nc = narrays, vmode = "integer"), zero = initializeBigMatrix(name = "zero", 
            nr = nsnps, nc = narrays, vmode = "integer")), phenoData = xyPhenoData, 
            protocolData = RG@protocolData)
        storageMode(XY) = "environment"
       # featureNames(XY) = names(aids)
        sampleNames(XY) = sampleNames(RG)
        rm(list = c("bord", "infI", "aids", "bids", "snpbase"))
        gc(verbose = FALSE)
        not.na <- !is.na(aord)
        not.na.aord <- aord[not.na]
        r <- as.matrix(assayData(RG)[["R"]][not.na.aord, ])
        XY@assayData$X[not.na, ] <- r
        rm(r)
        gc(verbose = FALSE)
        g <- as.matrix(assayData(RG)[["G"]][not.na.aord, ])
        XY@assayData$Y[not.na, ] <- g
        rm(g)
        gc(verbose = FALSE)
        z <- as.matrix(assayData(RG)[["zero"]][not.na.aord, ])
        XY@assayData$zero[not.na, ] <- z
        rm(z)
        gc(verbose = FALSE)
        rm(RG)
        gc(verbose = FALSE)
   
    
    
    
         }
      }
           XY 
}


stripNormalize = function(XY, useTarget=TRUE, verbose=TRUE, quantile.method="between") {
        if(quantile.method=="between") {
	  if(useTarget){
		objectsNeeded <- c("stripnum", "reference")
	  } else objectsNeeded <- "stripnum"
	  needToLoad <- !all(sapply(objectsNeeded, isLoaded))
	  if(needToLoad){
		pkgname = getCrlmmAnnotationName(annotation(XY))
		if(!require(pkgname, character.only=TRUE, quietly=!verbose)){
			suggCall = paste("library(", pkgname, ", lib.loc='/Altern/Lib/Loc')", sep="")
			msg = paste("If", pkgname, "is installed on an alternative location, please load it manually by using", suggCall)
			message(strwrap(msg))
			stop("Package ", pkgname, " could not be found.")
			rm(suggCall, msg)
		}
		if(verbose) message("Loading strip and reference normalization information.")
		loader("preprocStuff.rda", .crlmmPkgEnv, pkgname)
	  }
	  stripnum = getVarInEnv("stripnum")
	  if(useTarget)
		targetdist = getVarInEnv("reference")
	  if(verbose){
		message("Quantile normalizing ", ncol(XY), " arrays by ", max(stripnum), " strips.")
		if (getRversion() > '2.7.0') pb = txtProgressBar(min=0, max=max(stripnum), style=3)
	  }
          for(s in 1:max(stripnum)) {
            if(verbose) {
              if (getRversion() > '2.7.0') setTxtProgressBar(pb, s)
              else cat(".")
            }
            sel = stripnum==s
            ##RS: faster to access data directly
            ##subX = as.matrix(exprs(channel(XY, "X"))[sel,])
            ##subY = as.matrix(exprs(channel(XY, "Y"))[sel,])
            subX <- as.matrix(assayData(XY)[["X"]][sel, ])
            subY <- as.matrix(assayData(XY)[["Y"]][sel, ])
            if(useTarget)
              tmp = normalize.quantiles.use.target(cbind(subX, subY), targetdist[[s]])
            else
              tmp = normalize.quantiles(cbind(subX, subY))
            XY@assayData$X[sel,] = matrix(as.integer(tmp[,1:(ncol(tmp)/2)]+16), nrow(tmp), ncol(tmp)/2)
            XY@assayData$Y[sel,] = matrix(as.integer(tmp[,(ncol(tmp)/2+1):ncol(tmp)]+16), nrow(tmp), ncol(tmp)/2)
            rm(subX, subY, tmp, sel)
            gc(verbose=FALSE)
          }
          if(verbose)
            cat("\n")
        }
        if(quantile.method=="within") {  # ignore strip information
	  if(useTarget){
		objectsNeeded <- c("Xref", "Yref")
	  } else objectsNeeded <- ""
	  needToLoad <- !all(sapply(objectsNeeded, isLoaded))
	  if(needToLoad){
		pkgname = getCrlmmAnnotationName(annotation(XY))
		if(!require(pkgname, character.only=TRUE, quietly=!verbose)){
			suggCall = paste("library(", pkgname, ", lib.loc='/Altern/Lib/Loc')", sep="")
			msg = paste("If", pkgname, "is installed on an alternative location, please load it manually by using", suggCall)
			message(strwrap(msg))
			stop("Package ", pkgname, " could not be found.")
			rm(suggCall, msg)
		}
		if(verbose) message("Loading reference normalization information.")
		loader("targetXY.rda", .crlmmPkgEnv, pkgname)
	  }
	  if(useTarget) {
                Xref = getVarInEnv("Xref")
	        Yref = getVarInEnv("Yref")
          } else{
                Xref = normalize.quantiles.determine.target(as.matrix(assayData(XY)[["X"]]))
                gc(verbose=FALSE)
                Yref = normalize.quantiles.determine.target(as.matrix(assayData(XY)[["Y"]]))
                gc(verbose=FALSE)
          }
	  if(verbose){
		message("Quantile normalizing ", ncol(XY), " arrays, one at a time.")
		if (getRversion() > '2.7.0') pb = txtProgressBar(min=0, max=ncol(XY), style=3)
	  }
          for(s in 1:ncol(XY)) {
            if(verbose) {
              if (getRversion() > '2.7.0') setTxtProgressBar(pb, s)
              else cat(".")
            }
            ##RS: faster to access data directly
            ##subX = as.matrix(exprs(channel(XY, "X"))[sel,])
            ##subY = as.matrix(exprs(channel(XY, "Y"))[sel,])
            subX <- as.matrix(assayData(XY)[["X"]][,s])
            subY <- as.matrix(assayData(XY)[["Y"]][,s])
            tmpX = normalize.quantiles.use.target(subX, as.integer(Xref))
            tmpY = normalize.quantiles.use.target(subY, as.integer(Yref))              
            XY@assayData$X[,s] = as.integer(tmpX+16)
            XY@assayData$Y[,s] = as.integer(tmpY+16)
            rm(subX, subY, tmpX, tmpY)
          }
          if(verbose)
            cat("\n")
        }
  XY
}


preprocessInfinium2 = function(XY, mixtureSampleSize=10^5,
				fitMixture=TRUE,
				eps=0.1,
				verbose=TRUE,
				seed=1,
				cdfName,
				sns,
				stripNorm=TRUE,
				useTarget=TRUE,
                                quantile.method="between") { #,
#               outdir=".") {
#				save.it=FALSE,
#				snpFile,
#				cnFile) {
	if(stripNorm)
		XY = stripNormalize(XY, useTarget=useTarget, verbose=verbose, quantile.method=quantile.method)
	## MR: the code below is mostly straight from snprma.R
	if (missing(sns)) sns = sampleNames(XY) #$X
	if(missing(cdfName))
		cdfName = annotation(XY)
	needToLoad <- !all(sapply(c("autosomeIndex",
				    "SMEDIAN",
				    "theKnots",
				    "npProbesFid"), isLoaded))
	if(needToLoad){
		pkgname = getCrlmmAnnotationName(cdfName)
		if(!require(pkgname, character.only=TRUE, quietly=!verbose)){
			suggCall = paste("library(", pkgname, ", lib.loc='/Altern/Lib/Loc')", sep="")
			msg = paste("If", pkgname, "is installed on an alternative location, please load it manually by using", suggCall)
			message(strwrap(msg))
			stop("Package ", pkgname, " could not be found.")
			rm(suggCall, msg)
		}
		if(verbose) message("Loading snp annotation and mixture model parameters.")
		loader("genotypeStuff.rda", .crlmmPkgEnv, pkgname)
                if(fitMixture)
 		  loader("mixtureStuff.rda", .crlmmPkgEnv, pkgname)
		loader("snpProbesFid.rda", .crlmmPkgEnv, pkgname)
		loader("npProbesFid.rda", .crlmmPkgEnv, pkgname)
	}
	autosomeIndex = getVarInEnv("autosomeIndex")
        if(fitMixture) {
          SMEDIAN = getVarInEnv("SMEDIAN")
          theKnots = getVarInEnv("theKnots")
        }
        npIndex = getVarInEnv("npProbesFid")
	snpIndex = getVarInEnv("snpProbesFid")
	narrays = ncol(XY)
	nprobes = length(npIndex)
	if(length(nprobes)>0) {
		## RS: channel creates an expression set.  This is much slower than directly accessing the data
		A <- matrix(as.integer(assayData(XY)[["X"]][npIndex, ]), nprobes, narrays)
		##system.time(A <- matrix(as.integer(exprs(channel(XY, "X"))[npIndex,]), nprobes, narrays))
		B <- matrix(as.integer(assayData(XY)[["Y"]][npIndex, ]), nprobes, narrays)
		##B = matrix(as.integer(exprs(channel(XY, "Y"))[npIndex,]), nprobes, narrays)
		## new lines below - useful to keep track of zeroed out probes
		zero <- matrix(as.integer(assayData(XY)[["zero"]][npIndex, ]), nprobes, narrays)
		##zero = matrix(as.integer(exprs(channel(XY, "zero"))[npIndex,]), nprobes, narrays)
		
		colnames(A) = colnames(B) = colnames(zero) = sns
		rownames(A) = rownames(B) = rownames(zero) = names(npIndex)
		cnAB = list(A=A, B=B, zero=zero, sns=sns, gns=names(npIndex), cdfName=cdfName)
		##      t0 <- proc.time()
		##      save(cnAB, file=cnFile)
		##      t0 <- proc.time()-t0
		##      if(verbose) message("Used ", round(t0[3],1), " seconds to save ", cnFile, ".")
		rm(A, B, zero)
#		print(gc())
		gc(verbose=FALSE)
	}
  # next process snp probes
  nprobes = length(snpIndex)
  ##We will read each cel file, summarize, and run EM one by one
  ##We will save parameters of EM to use later
  mixtureParams = matrix(NA, 4, narrays)
  SNR = rep(NA, narrays)
  SKW = rep(NA, narrays)
  ## This is the sample for the fitting of splines
  ## BC: I like better the idea of the user passing the seed,
  ##     because this might intefere with other analyses
  ##     (like what happened to GCRMA)
  set.seed(seed)
  idx = sort(sample(autosomeIndex, mixtureSampleSize))
  idx2 = sample(nprobes, 10^5)
  ##S will hold (A+B)/2 and M will hold A-B
  ##NOTE: We actually dont need to save S. Only for pics etc...
  ##f is the correction. we save to avoid recomputing

  A = matrix(0, nprobes, narrays)
  B = matrix(0, nprobes, narrays)
  zero = matrix(NA, nprobes, narrays)
  if(verbose && fitMixture){
     message("Calibrating ", narrays, " arrays.")
     if (getRversion() > '2.7.0') pb = txtProgressBar(min=0, max=narrays, style=3)
  }
  for(i in 1:narrays){
	  ## RS: faster to access data directly without using channel method
	  ##A[,i] = as.integer(exprs(channel(XY, "X"))[snpIndex,i])
	  A[, i] <- as.integer(assayData(XY)[["X"]][snpIndex, i])
	  B[, i] <- as.integer(assayData(XY)[["Y"]][snpIndex, i])
	  zero[, i] <- as.integer(assayData(XY)[["zero"]][snpIndex, i])
	  ##B[,i] = as.integer(exprs(channel(XY, "Y"))[snpIndex,i])
	  ##zero[,i] = as.integer(exprs(channel(XY, "zero"))[snpIndex,i])
	  SKW[i] = mean((A[idx2,i]-mean(A[idx2,i]))^3)/(sd(A[idx2,i])^3)
	  if(fitMixture){
		  S = (log2(A[idx,i])+log2(B[idx,i]))/2 - SMEDIAN
		  M = log2(A[idx,i])-log2(B[idx,i])
		  ##we need to test the choice of eps.. it is not the max diff between funcs
		  tmp = fitAffySnpMixture56(S, M, theKnots, eps=eps)
		  mixtureParams[, i] = tmp[["coef"]]
		  SNR[i] = tmp[["medF1"]]^2/(tmp[["sigma1"]]^2+tmp[["sigma2"]]^2)
	          if(verbose) {
		    if (getRversion() > '2.7.0') setTxtProgressBar(pb, i)
		    else cat(".")
                  }
	  }
	  ## run garbage collection every now and then
	  if(i %% 100 == 0) gc(verbose=FALSE);
  }
  if (verbose && fitMixture) {
	  if (getRversion() > '2.7.0') close(pb)
	  else cat("\n")
  }
  if (!fitMixture) SNR = mixtureParams = NA
  ## gns comes from preprocStuff.rda

#  if(save.it & !missing(snpFile)) {
#    t0 <- proc.time()
#    save(res, file=snpFile)
#    t0 <- proc.time()-t0
#    if(verbose) message("Used ", round(t0[3],1), " seconds to save ", snpFile, ".")
#  }

  res = list(A=A, B=B,
             zero=zero, sns=sns, gns=names(snpIndex), SNR=SNR, SKW=SKW,
             mixtureParams=mixtureParams, cdfName=cdfName, cnAB=cnAB)

#  open(res[["A"]])
#  open(res[["B"]])
#  open(res[["zero"]])
#  open(res[["SNR"]])
#  open(res[["mixtureParams"]])

#  if(save.it & !missing(snpFile)) {
#    t0 <- proc.time()
#    save(res, file=snpFile)
#    t0 <- proc.time()-t0
#    if(verbose) message("Used ", round(t0[3],1), " seconds to save ", snpFile, ".")
#  }

#  close(res[["A"]])
#  close(res[["B"]])
#  close(res[["zero"]])
#  close(res[["SNR"]])
#  close(res[["mixtureParams"]])

  return(res)
}

## crlmmIllumina and genotype.Illumina are now the same function
## Use genotype.Illumina instead
#crlmmIllumina <- function(RG, XY, stripNorm=TRUE, useTarget=TRUE,
#			  row.names=TRUE, col.names=TRUE,
#			  probs=c(1/3, 1/3, 1/3), DF=6, SNRMin=5, gender=NULL,
#			  seed=1,
#			  mixtureSampleSize=10^5, eps=0.1, verbose=TRUE,
#			  cdfName, sns, recallMin=10, recallRegMin=1000,
#			  returnParams=FALSE, badSNP=.7) {
#	if(missing(cdfName)) {
#		if(!missing(RG))
#			cdfName = annotation(RG)
#		if(!missing(XY))
#			cdfName = annotation(XY)
#	}
#	if(!isValidCdfName(cdfName))
#		stop("cdfName not valid.  see validCdfNames")
#	if(!missing(RG)) {
#		if(missing(XY))
#			XY = RGtoXY(RG, chipType=cdfName)
#		else
#			stop("Both RG and XY specified - please use one or the other")
#	}
#	if (missing(sns)) sns = sampleNames(XY)
#	res <- preprocessInfinium2(XY, mixtureSampleSize=mixtureSampleSize,
#				   fitMixture=TRUE, verbose=verbose,
#				   seed=seed, eps=eps, cdfName=cdfName, sns=sns,
#				   stripNorm=stripNorm, useTarget=useTarget) #,
#	if(row.names) row.names=res$gns else row.names=NULL
#	if(col.names) col.names=res$sns else col.names=NULL
#	res2 <- crlmmGT(A=res[["A"]],
#			B=res[["B"]],
#			SNR=res[["SNR"]],
#			mixtureParams=res[["mixtureParams"]],
#			cdfName=cdfName,
#			row.names=row.names,
#			col.names=col.names,
#			probs=probs,
#			DF=DF,
#			SNRMin=SNRMin,
#			recallMin=recallMin,
#			recallRegMin=recallRegMin,
#			gender=gender,
#			verbose=verbose,
#			returnParams=returnParams,
#			badSNP=badSNP)
#	res2[["SNR"]] = res[["SNR"]]
#	res2[["SKW"]] = res[["SKW"]]
#	rm(res); gc(verbose=FALSE)
#	return(list2SnpSet(res2, returnParams=returnParams))
#}


## Deprecate crlmmIlluminaV2 function
## MR: Below is a more memory efficient version of crlmmIllumina() which
## reads in the .idats and genotypes in the one function and removes objects
## after they have been used
#crlmmIlluminaV2 = function(sampleSheet=NULL,
#			  arrayNames=NULL,
#			  ids=NULL,
#			  path=".",
#			  arrayInfoColNames=list(barcode="SentrixBarcode_A", position="SentrixPosition_A"),
#			  highDensity=FALSE,
#			  sep="_",
#			  fileExt=list(green="Grn.idat", red="Red.idat"),
#			  saveDate=FALSE,
#			  stripNorm=TRUE,
#			  useTarget=TRUE,
# #                         outdir=".",
#			  row.names=TRUE,
#			  col.names=TRUE,
#			  probs=c(1/3, 1/3, 1/3), DF=6, SNRMin=5, gender=NULL,
#                          seed=1,  # save.it=FALSE, snpFile, cnFile,
#                          mixtureSampleSize=10^5, eps=0.1, verbose=TRUE,
#                          cdfName, sns, recallMin=10, recallRegMin=1000,
#                          returnParams=FALSE, badSNP=.7) {
#
#    if(missing(cdfName)) stop("must specify cdfName")
#    if(!isValidCdfName(cdfName)) stop("cdfName not valid.  see validCdfNames")
#
##    is.lds = ifelse(isPackageLoaded("ff"), TRUE, FALSE)
#    is.lds=FALSE
#    RG = readIdatFiles(sampleSheet=sampleSheet, arrayNames=arrayNames,
#                       ids=ids, path=path, arrayInfoColNames=arrayInfoColNames,
#                       highDensity=highDensity, sep=sep, fileExt=fileExt, saveDate=saveDate)
#
#
#    XY = RGtoXY(RG, chipType=cdfName)
##    if(is.lds) {
##      open(RG@assayData$R); open(RG@assayData$G); open(RG@assayData$zero)
##      delete(RG@assayData$R); delete(RG@assayData$G); delete(RG@assayData$zero)
##    }
#    rm(RG); gc(verbose=FALSE)
#
#    if (missing(sns)) { sns = sampleNames(XY)
#    }
#    res = preprocessInfinium2(XY, mixtureSampleSize=mixtureSampleSize, fitMixture=TRUE, verbose=verbose,
#                               seed=seed, eps=eps, cdfName=cdfName, sns=sns, stripNorm=stripNorm, useTarget=useTarget) #, outdir=outdir) #,
##                               save.it=save.it, snpFile=snpFile, cnFile=cnFile)
#
##    if(is.lds) {
##      open(XY@assayData$X); open(XY@assayData$Y); open(XY@assayData$zero)
##      delete(XY@assayData$X); delete(XY@assayData$Y); delete(XY@assayData$zero)
##    }
#    rm(XY); gc(verbose=FALSE)
#
#    if(row.names) row.names=res$gns else row.names=NULL
#    if(col.names) col.names=res$sns else col.names=NULL
###
###  if(is.lds){
###	  res2 <- crlmmGT2(A=res[["A"]],
###			   B=res[["B"]],
###			   SNR=res[["SNR"]],
###			   mixtureParams=res[["mixtureParams"]],
###			   cdfName=cdfName,
###			   row.names=row.names,
###			   col.names=col.names,
###			   probs=probs,
###			   DF=DF,
###			   SNRMin=SNRMin,
###			   recallMin=recallMin,
###			   recallRegMin=recallRegMin,
###			   gender=gender,
###			   verbose=verbose,
###			   returnParams=returnParams,
###			   badSNP=badSNP)
###  } else {
#	  res2 <- crlmmGT(A=res[["A"]],
#			   B=res[["B"]],
#			   SNR=res[["SNR"]],
#			   mixtureParams=res[["mixtureParams"]],
#			   cdfName=cdfName,
#			   row.names=row.names,
#			   col.names=col.names,
#			   probs=probs,
#			   DF=DF,
#			   SNRMin=SNRMin,
#			   recallMin=recallMin,
#			   recallRegMin=recallRegMin,
#			   gender=gender,
#			   verbose=verbose,
#			   returnParams=returnParams,
#			   badSNP=badSNP)
###  }
#
###    FUN = ifelse(is.lds, "crlmmGT2", "crlmmGT")
###    ## genotyping
###    crlmmGTfxn = function(FUN,...){
###		switch(FUN,
###		       crlmmGT2=crlmmGT2(...),
###		       crlmmGT=crlmmGT(...))
###              }
###    res2 = crlmmGTfxn(FUN,
###                     A=res[["A"]],
###                     B=res[["B"]],
###                     SNR=res[["SNR"]],
###                     mixtureParams=res[["mixtureParams"]],
###                     cdfName=cdfName,
###                     row.names=row.names,
###                     col.names=col.names,
###                     probs=probs,
###                     DF=DF,
###                     SNRMin=SNRMin,
###                     recallMin=recallMin,
###                     recallRegMin=recallRegMin,
###                     gender=gender,
###                     verbose=verbose,
###                     returnParams=returnParams,
###                     badSNP=badSNP)
#
##    if(is.lds) {
##      open(res[["SNR"]]); open(res[["SKW"]])
##    }
#    res2[["SNR"]] = res[["SNR"]]
#    res2[["SKW"]] = res[["SKW"]]
# #  if(is.lds) {
# #    delete(res[["A"]]); delete(res[["B"]])
# #    delete(res[["SKW"]]); delete(res[["SNR"]]); delete(res[["mixtureParams"]])
# #  }
#    rm(res); gc(verbose=FALSE)
#    return(list2SnpSet(res2, returnParams=returnParams))
#}

# Functions analogous to Rob's Affy functions to set up container
getProtocolData.Illumina = function(filenames, sep="_", fileExt="Grn.idat", verbose=FALSE) {
       narrays = length(filenames)

       headerInfo = list(nProbes = rep(NA, narrays),
                         Barcode = rep(NA, narrays),
                         ChipType = rep(NA, narrays),
                         Manifest = rep(NA, narrays),
                         Position = rep(NA, narrays))

       scanDates = data.frame(ScanDate=rep(NA, narrays), DecodeDate=rep(NA, narrays))
       rownames(scanDates) <- make.unique(gsub(paste(sep, fileExt, sep=""), "", filenames))
       ## read in the data
       for(i in seq_along(filenames)) {
               if(verbose)
	               cat("reading", filenames[i], "\n")
	       idsG = G = NULL
	       G = readIDAT(filenames[i])
	       idsG = rownames(G$Quants)
	       headerInfo$nProbes[i] = G$nSNPsRead
	       headerInfo$Barcode[i] = G$Barcode
	       headerInfo$ChipType[i] = G$ChipType
	       headerInfo$Manifest[i] = G$Unknown$MostlyNull
	       headerInfo$Position[i] = G$Unknowns$MostlyA
               if(headerInfo$nProbes[i]>(headerInfo$nProbes[1]+10000) || headerInfo$nProbes[i]<(headerInfo$nProbes[1]-10000)) {
		       warning("Chips are not of the same type.  Skipping ", basename(filenames[i]))
		       next()
	       }
               if(nrow(G$RunInfo)>=2) {
   	           scanDates$ScanDate[i] = G$RunInfo[1, 1]
	           scanDates$DecodeDate[i] = G$RunInfo[2, 1]
               }
	       rm(G)
	       gc(verbose=FALSE)
       }
       protocoldata = new("AnnotatedDataFrame",
			    data=scanDates,
			    varMetadata=data.frame(labelDescription=colnames(scanDates),
			                           row.names=colnames(scanDates)))
       return(protocoldata)
}

getAvailableIlluminaGenomeBuild <- function(path){
	snp.file <- list.files(path, pattern="snpProbes_hg")
	if(length(snp.file) > 1){
		## use hg19
		message("genome build not specified. Using build hg19 for annotation.")
		snp.file <- snp.file[1]
	}
	if(length(snp.file) == 1)
		genome <- gsub(".rda", "", strsplit(snp.file, "snpProbes_")[[1]][[2]])
	if(length(snp.file)==0) genome <- ""
	genome
}


constructInf <- function(sampleSheet=NULL,
			 arrayNames=NULL,
			 path=".",
			 arrayInfoColNames=list(barcode="SentrixBarcode_A", position="SentrixPosition_A"),
			 highDensity=FALSE,
			 sep="_",
			 fileExt=list(green="Grn.idat", red="Red.idat"),
                         XY,
			 cdfName,
                         anno,
                         genome,
			 verbose=FALSE,
			 batch=NULL, #fns,
			 saveDate=TRUE) { #, outdir="."){
        if(is.null(XY)) {
  	  if(!is.null(arrayNames)) {
		pd = new("AnnotatedDataFrame", data = data.frame(Sample_ID=arrayNames))
	  }
	  if(!is.null(sampleSheet)) { # get array info from Illumina's sample sheet
		if(is.null(arrayNames)) {
			if(!is.null(arrayInfoColNames$barcode) && (arrayInfoColNames$barcode %in% colnames(sampleSheet))) {
				barcode = sampleSheet[,arrayInfoColNames$barcode]
				arrayNames=barcode
			}
			if(!is.null(arrayInfoColNames$position) && (arrayInfoColNames$position %in% colnames(sampleSheet))) {
				position = sampleSheet[,arrayInfoColNames$position]
				if(is.null(arrayNames))
					arrayNames=position
				else
					arrayNames = paste(arrayNames, position, sep=sep)
				if(highDensity) {
					hdExt = list(A="R01C01", B="R01C02", C="R02C01", D="R02C02")
					for(i in names(hdExt))
						arrayNames = sub(paste(sep, i, sep=""), paste(sep, hdExt[[i]], sep=""), arrayNames)
				}
			}
		  }
		  pd = new("AnnotatedDataFrame", data = sampleSheet)
		  sampleNames(pd) = make.unique(basename(arrayNames))
	  }
	  if(is.null(arrayNames)) {
		arrayNames = gsub(paste(sep, fileExt$green, sep=""), "", dir(pattern=fileExt$green, path=path))
		if(!is.null(sampleSheet)) {
			sampleSheet=NULL
			cat("Could not find required info in \'sampleSheet\' - ignoring.  Check \'sampleSheet\' and/or \'arrayInfoColNames\'\n")
		}
		pd = new("AnnotatedDataFrame", data = data.frame(Sample_ID=arrayNames))
  	  }
	  narrays = length(arrayNames)

	  if(!is.null(batch)) {
		stopifnot(length(batch) == narrays)
	  }
 	  if(is.null(batch)) {
                batch = rep("batch1", narrays) # assume only one batch stop("Must specify 'batch'")
	  }
          if(is(batch, "factor")) batch = as.character(batch)

	  grnfiles = paste(arrayNames, fileExt$green, sep=sep)
	  redfiles = paste(arrayNames, fileExt$red, sep=sep)
	  if(length(grnfiles)==0 || length(redfiles)==0)
		stop("Cannot find .idat files")
	  if(length(grnfiles)!=length(redfiles))
		stop("Cannot find matching .idat files")
	  if(path[1] != "." & path[1] != ""){
		grnidats = file.path(path, grnfiles)
		redidats = file.path(path, redfiles)
	  }  else {
		message("path arg not set.  Assuming files are in local directory, or that complete path is provided")
		grnidats = grnfiles
		redidats = redfiles
    	  }
	  if(!all(file.exists(grnidats))) stop("Missing some of the *Grn.idat files")
	  if(!all(file.exists(redidats))) stop("Missing some of the *Red.idat files")

	  if(verbose) message("Initializing container for genotyping and copy number estimation")
          if(cdfName!="nopackage") {
	    pkgname <- getCrlmmAnnotationName(cdfName)
	    path <- system.file("extdata", package=pkgname)
	    genome <- getAvailableIlluminaGenomeBuild(path)
	    featureData = getFeatureData(cdfName, copynumber=TRUE, genome=genome)
          } else {
            if(!is.integer(anno$chr)) {
                  chr = as.character(anno$chromosome)
                  chr[chr=="X"] = 23
                  chr[chr=="Y"] = 24
                  chr[chr=="XY"] = 25
            }
            featureData = new("GenomeAnnotatedDataFrame",
                          isSnp=as.logical(anno$isSnp),
                          position=as.integer(anno$position),
                          chromosome=as.integer(chr),
                          row.names=anno$featureNames)
          }
            nr <- nrow(featureData)
            nc <- narrays
	    sns <-  if (!is.null(sampleSheet) && !is.null(sampleSheet$Sample_ID)) {
		make.unique(sampleSheet$Sample_ID)
            } else{
		make.unique(basename(arrayNames))
            } 
	  biga <- initializeBigMatrix(name="A", nr, nc)
	  bigb <- initializeBigMatrix(name="B", nr, nc)
	  bigc <- initializeBigMatrix(name="call", nr, nc)
	  bigd <- initializeBigMatrix(name="callPr", nr,nc)
	  colnames(biga) <- colnames(bigb) <- colnames(bigc) <- colnames(bigd) <- sns
	  cnSet <- new("CNSet",
		     alleleA=biga,
		     alleleB=bigb,
		     call=bigc,
		     callProbability=bigd,
		     annotation=cdfName,
		     featureData=featureData,
		     batch=batch,
		     genome=genome)
##        if (!is.null(sampleSheet) && !is.null(sampleSheet$Sample_ID)) {
##		sampleNames(cnSet) = make.unique(sampleSheet$Sample_ID)
##        } else{
##		sampleNames(cnSet) <- make.unique(basename(arrayNames))
##	}
  	  if(saveDate){
		protocolData = getProtocolData.Illumina(grnidats, sep=sep, fileExt=fileExt$green, verbose=verbose)
	  } else{
		protocolData = annotatedDataFrameFrom(A(cnSet), byrow=FALSE)
	  }
	  rownames(pData(protocolData)) = sampleNames(cnSet)
	  protocolData(cnSet) = protocolData
	##featureData(cnSet) = featureData
	  featureNames(cnSet) = featureNames(featureData)
	  cnSet$gender <- initializeBigVector("gender", ncol(cnSet), vmode="integer")
	  cnSet$SNR = initializeBigVector("crlmmSNR-", ncol(cnSet), "double")
	  cnSet$SKW = initializeBigVector("crlmmSKW-", ncol(cnSet), "double")
	##sampleNames(cnSet) <- basename(sampleNames(cnSet))
        } else { # if XY specified, easier set-up of cnSet
          narrays = ncol(XY)
          if(verbose) message("Initializing container for genotyping and copy number estimation")           
          if(!is.null(batch)) {
              stopifnot(length(batch) == narrays)
          }
          if(is.null(batch)) {
              batch = rep("batch1", narrays) # assume only one batch stop("Must specify 'batch'")
          }
          if(is(batch, "factor")) batch = as.character(batch)
          if(cdfName!="nopackage") {
            pkgname <- getCrlmmAnnotationName(cdfName)
            path <- system.file("extdata", package=pkgname)
            genome <- getAvailableIlluminaGenomeBuild(path)
            featureData = getFeatureData(cdfName, copynumber=TRUE, genome=genome)
          } else {
                      anno = as(anno, "AnnotatedDataFrame")
 anno = as(anno, "GenomeAnnotatedDataFrame")
featureData=anno
           
          }
          nr <- nrow(featureData)
          nc <- narrays
          sns <-  sampleNames(XY)
          biga <- initializeBigMatrix(name="A", nr, nc)
          bigb <- initializeBigMatrix(name="B", nr, nc)
          bigc <- initializeBigMatrix(name="call", nr, nc)
          bigd <- initializeBigMatrix(name="callPr", nr,nc)
          colnames(biga) <- colnames(bigb) <- colnames(bigc) <- colnames(bigd) <- sns
          cnSet <- new("CNSet",
                       alleleA=biga,
                       alleleB=bigb,
                       call=bigc,
                       callProbability=bigd,
                       annotation=cdfName,
                       featureData=featureData,
                       batch=batch,
                       genome=genome)
          protocolData = annotatedDataFrameFrom(A(cnSet), byrow=FALSE)
          rownames(pData(protocolData)) = sampleNames(cnSet)
          protocolData(cnSet) = protocolData
          featureNames(cnSet) = featureNames(featureData)
          cnSet$gender = initializeBigVector("gender", ncol(cnSet), vmode="integer")
          cnSet$SNR = initializeBigVector("crlmmSNR-", ncol(cnSet), "double")
          cnSet$SKW = initializeBigVector("crlmmSKW-", ncol(cnSet), "double")
        }
	return(cnSet)
}
construct.Illumina <- constructInf

preprocessInf <- function(cnSet,
		       sampleSheet=NULL,
		       arrayNames=NULL,
		       ids=NULL,
		       path=".",
		       arrayInfoColNames=list(barcode="SentrixBarcode_A", position="SentrixPosition_A"),
		       highDensity=TRUE,
		       sep="_",
		       fileExt=list(green="Grn.idat", red="Red.idat"),
                       XY,
                       anno,
		       saveDate=TRUE,
		       stripNorm=TRUE,
		       useTarget=TRUE,
		       mixtureSampleSize=10^5,
		       fitMixture=TRUE,
                       quantile.method="between",
		       eps=0.1,
		       verbose=TRUE,
		       seed=1, cdfName){
	stopifnot(all(c("gender", "SNR", "SKW") %in% varLabels(cnSet)))
	open(A(cnSet))
	open(B(cnSet))
	sns <- sampleNames(cnSet)
	pkgname = getCrlmmAnnotationName(annotation(cnSet))
        is.snp = isSnp(cnSet)
	snp.index = which(is.snp)
	narrays = ncol(cnSet)
	sampleBatches <- splitIndicesByLength(seq_len(ncol(cnSet)), ocSamples()/getDoParWorkers())
	mixtureParams = initializeBigMatrix("crlmmMixt-", 4, narrays, "double")
	ocLapply(seq_along(sampleBatches),
		 processIDAT, # crlmm:::
		 sampleBatches=sampleBatches,
		 sampleSheet=sampleSheet,
		 arrayNames=arrayNames,
		 ids=ids,
		 path=path,
		 arrayInfoColNames=arrayInfoColNames,
		 highDensity=highDensity,
		 sep=sep,
		 fileExt=fileExt,
                 XY=XY,
                 anno=anno,
		 saveDate=saveDate,
		 verbose=verbose,
		 mixtureSampleSize=mixtureSampleSize,
		 fitMixture=fitMixture,
		 eps=eps,
		 seed=seed,
		 cdfName=cdfName,
		 sns=sns,
		 stripNorm=stripNorm,
		 useTarget=useTarget,
                 quantile.method=quantile.method,                 
		 A=A(cnSet),
		 B=B(cnSet),
		 GT=calls(cnSet),
		 GTP=snpCallProbability(cnSet),
		 SKW=cnSet$SKW,
		 SNR=cnSet$SNR,
		 mixtureParams=mixtureParams,
		 is.snp=is.snp,
		 neededPkgs=c("crlmm", pkgname)) # outdir=outdir,
	return(mixtureParams)
}
preprocess <- preprocessInf

genotypeInf <- function(cnSet, mixtureParams, probs=rep(1/3,3),
			SNRMin=5,
			recallMin=10,
			recallRegMin=1000,
			verbose=TRUE,
			returnParams=TRUE,
			badSNP=0.7,
			gender=NULL,
			DF=6,
			cdfName,
                        nopackage.norm="quantile",
                        call.method="crlmm",
                        trueCalls=NULL){
	is.snp = isSnp(cnSet)
	snp.index = which(is.snp)
##	narrays = ncol(cnSet)
##	open(A(cnSet))
##	open(B(cnSet))
##	open(snpCall(cnSet))
##	open(snpCallProbability(cnSet))
##	## crlmmGT2 overwrites the normalized intensities with calls and confidenceScores
##	message("Writing to call and callProbability slots")
##	for(j in 1:ncol(cnSet)){
##		snpCall(cnSet)[snp.index, j] <- as.integer(A(cnSet)[snp.index, j])
##		snpCallProbability(cnSet)[snp.index, j] <- as.integer(B(cnSet)[snp.index, j])
##	}
##	message("Writing complete.  Begin genotyping...")
##	close(A(cnSet))
##	close(B(cnSet))
        if(call.method=="crlmm") {
  	  tmp <- crlmmGT2(A=A(cnSet),
			B=B(cnSet),
			SNR=cnSet$SNR,
			mixtureParams=mixtureParams,
			cdfName=cdfName,
			col.names=sampleNames(cnSet),
			probs=probs,
			DF=DF,
			SNRMin=SNRMin,
			recallMin=recallMin,
			recallRegMin=recallRegMin,
			gender=gender,
			verbose=verbose,
			returnParams=returnParams,
			badSNP=badSNP,
			callsGt=calls(cnSet),
			callsPr=snpCallProbability(cnSet))#,
	##RS:  I changed the API for crlmmGT2 to be consistent with crlmmGT
           open(cnSet$gender)
           cnSet$gender[,] = tmp$gender
           close(cnSet$gender)
        }
        if(call.method=="krlmm") {
            if(cdfName=="nopackage") {
                tmp <- krlmmnopkg(cnSet, offset=16, gender=gender, normalize.method=nopackage.norm, annotation=anno, verbose=verbose)
            } else {
              tmp <- krlmm(cnSet, cdfName, gender=gender, trueCalls=trueCalls, verbose=verbose)
	## snp.names=featureNames(cnSet)[snp.index])
            }
        }
	if(verbose) message("Genotyping finished.") # Updating container with genotype calls and confidence scores.")
	TRUE
}


genotype.Illumina <- function(sampleSheet=NULL,
			      arrayNames=NULL,
			      ids=NULL,
			      path=".",
			      arrayInfoColNames=list(barcode="SentrixBarcode_A", position="SentrixPosition_A"),
			      highDensity=FALSE,
			      sep="_",
			      fileExt=list(green="Grn.idat", red="Red.idat"),
                              XY=NULL,
                              anno,
                              genome,
                              call.method="crlmm",
                              trueCalls=NULL,
			      cdfName,
			      copynumber=TRUE,
			      batch=NULL,
			      saveDate=FALSE,
			      stripNorm=TRUE,
			      useTarget=TRUE,
                              quantile.method="between",
                              nopackage.norm="quantile",
			      mixtureSampleSize=10^5,
			      fitMixture=TRUE,
			      eps=0.1,
			      verbose=TRUE,
			      seed=1,
			      sns,
			      probs=rep(1/3, 3),
			      DF=6,
			      SNRMin=5,
			      recallMin=10,
			      recallRegMin=1000,
			      gender=NULL,
			      returnParams=TRUE,
			      badSNP=0.7) {
        krlmm.supported = c("humanomni1quadv1b",      # Omni1 quad
	                    "humanomni25quadv1b",     # Omni2.5 quad
	                    "humanomni258v1a",        # Omni2.5 8 v1 A
                            "humanomni258v1p1b",      # Omni2.5 8 v1.1 B
                            "humanomni5quadv1b",      # Omni5 quad
			    "humanexome12v1p2a",      # Exome 12 v1.2 A
                            "humanomniexpexome8v1p1b", # Omni Express Exome 8 v1.1b
                            "nopackage")
        crlmm.supported = c("human1mv1c",             # 1M
                            "human370v1c",            # 370CNV
	                    "human650v3a",            # 650Y
	                    "human610quadv1b",        # 610 quad
	                    "human660quadv1a",        # 660 quad
	                    "human370quadv3c",        # 370CNV quad
	                    "human550v3b",            # 550K
	                    "human1mduov3b",          # 1M Duo
                            "humanomni1quadv1b",      # Omni1 quad
#			    "humanomni258v1a",        # Omni2.5 8 v1 A
#                           "humanomni258v1p1b",      # Omni2.5 8 v1.1 B
	                    "humanomniexpress12v1b",  # Omni express 12
	                    "humanimmuno12v1b",       # Immuno chip 12
                            "humancytosnp12v2p1h",    # CytoSNP 12
                            "humanomniexpexome8v1p1b") # Omni Express Exome 8 v1.1b
        if(call.method=="krlmm") {
          if(!any(cdfName==krlmm.supported))
             stop(cdfName, " platform not supported by krlmm.  Consider setting call.method=\'crlmm\'")           
          fitMixture=FALSE
          quantile.method="within"
        }
        if(call.method=="crlmm") {
          if(!any(cdfName==crlmm.supported))
             stop(cdfName, " platform not supported by crlmm.  Consider setting call.method=\'krlmm\'")
          fitMixture=TRUE
          # quantile.method="between"
        }
	is.lds = ifelse(isPackageLoaded("ff"), TRUE, FALSE)
        if (!(is.lds))
            stop("Package ff not loaded")
        if(!is.null(XY) && missing(cdfName))
          cdfName = getCrlmmAnnotationName(annotation(cnSet))
	if(missing(cdfName)) stop("must specify cdfName")
	if(!isValidCdfName(cdfName)) stop("cdfName not valid.  see validCdfNames")
#	stopifnot(!missing(batch))
#	if(is(batch, "factor")) batch <- as.character(batch)
        pkgname <- getCrlmmAnnotationName(cdfName)
#        if(is.null(cnSet)) {
  	  message("Instantiate CNSet container.")
	  cnSet <- constructInf(sampleSheet=sampleSheet,
				    arrayNames=arrayNames,
				    path=path,
				    arrayInfoColNames=arrayInfoColNames,
				    highDensity=highDensity,
				    sep=sep,
				    fileExt=fileExt,
                                    XY=XY,
                                    anno=anno,
                                    genome=genome,
				    cdfName=cdfName,
				    verbose=verbose,
				    batch=batch,
				    saveDate=saveDate)
#        }
        if(call.method=="krlmm" && ncol(cnSet)<8)
           stop(paste("To run krlmm you need at least 8 samples (in general, the more the better).  You currently have", ncol(cnSet)))
	mixtureParams <- preprocessInf(cnSet=cnSet,
				    sampleSheet=sampleSheet,
				    arrayNames=arrayNames,
				    ids=ids,
				    path=path,
				    arrayInfoColNames=arrayInfoColNames,
				    highDensity=highDensity,
				    sep=sep,
				    fileExt=fileExt,
				    saveDate=saveDate,
                                    XY=XY,
                                    anno=anno, 
				    stripNorm=stripNorm,
				    useTarget=useTarget,
				    mixtureSampleSize=mixtureSampleSize,
				    fitMixture=fitMixture,
                                    quantile.method=quantile.method,
				    eps=eps,
				    verbose=verbose,
				    seed=seed,
				    cdfName=cdfName)
	message("Begin genotyping...")
	genotypeInf(cnSet=cnSet, mixtureParams=mixtureParams,
		    probs=probs,
		    SNRMin=SNRMin,
		    recallMin=recallMin,
		    recallRegMin=recallRegMin,
		    verbose=verbose,
		    returnParams=returnParams,
		    badSNP=badSNP,
		    gender=gender,
		    DF=DF,
		    cdfName=cdfName,
                    nopackage.norm=nopackage.norm,
                    call.method=call.method,
                    trueCalls=trueCalls)
	return(cnSet)
}

crlmmIllumina <- genotype.Illumina



processIDAT <- function(stratum, sampleBatches, sampleSheet=NULL,
			arrayNames=NULL,
			ids=NULL,
			path=".",
			arrayInfoColNames=list(barcode="SentrixBarcode_A", position="SentrixPosition_A"),
			highDensity=FALSE,
			sep="_",
			fileExt=list(green="Grn.idat", red="Red.idat"),
                        XY,
                        anno,
			saveDate=FALSE,
			verbose=TRUE,
			mixtureSampleSize=10^5,
			fitMixture=TRUE,
			eps=0.1,
			seed=1,
			cdfName,
			sns,
			stripNorm=TRUE,
			useTarget=TRUE,
                        quantile.method=quantile.method,                        
			A, B,
			GT,
			GTP,
			SKW, SNR, mixtureParams, is.snp) { #, outdir=".") {
	message("Processing sample stratum ", stratum, " of ", length(sampleBatches))
	sel <- sampleBatches[[stratum]]
        if(length(path)>= length(sel)) path = path[sel]
        if(is.null(XY)) {
          RG = readIdatFiles(sampleSheet=sampleSheet[sel,], arrayNames=arrayNames[sel],
                       ids=ids, path=path, arrayInfoColNames=arrayInfoColNames,
                       highDensity=highDensity, sep=sep, fileExt=fileExt, saveDate=saveDate, verbose=verbose)
          XY = RGtoXY(RG, chipType=cdfName, anno=anno)
          rm(RG)
          gc(verbose=FALSE)
          if (missing(sns) || length(sns)!=ncol(XY)) sns = sampleNames(XY)
          if(cdfName!="nopackage") {
             res = preprocessInfinium2(XY, mixtureSampleSize=mixtureSampleSize, fitMixture=TRUE, verbose=verbose,
                               seed=seed, eps=eps, cdfName=cdfName, sns=sns, stripNorm=stripNorm, useTarget=useTarget,
                               quantile.method=quantile.method) #, outdir=outdir
             rm(XY)
          } else {
              res = list(A=XY@assayData$X, B=XY@assayData$Y, zero=XY@assayData$zero, sns=sns, gns=names(XY), cdfName=cdfName,
                  SNR=NA, SKW=NA, mixtureParams=NA)
              rm(XY)
          }
          } else{ #XY already available
          if(missing(sns) || length(sns)!=ncol(XY)) sns = sampleNames(XY)
          if(cdfName!="nopackage") {
            res = preprocessInfinium2(XY[,sel], mixtureSampleSize=mixtureSampleSize, fitMixture=TRUE, verbose=verbose,
                                           seed=seed, eps=eps, cdfName=cdfName, sns=sns[sel], stripNorm=stripNorm, useTarget=useTarget,
                                           quantile.method=quantile.method)
          } else {
              res = list(A=XY@assayData$X, B=XY@assayData$Y, zero=XY@assayData$zero, sns=sns, gns=names(XY), cdfName=cdfName,
                  SNR=NA, SKW=NA, mixtureParams=NA)
          }
        }
        gc(verbose=FALSE)
	if(verbose) message("Finished preprocessing.")
        snp.index = which(is.snp)
	np.index = which(!is.snp)
	open(A); open(B);
	open(GT); open(GTP)
	Amatrix <- res[["A"]]
	Bmatrix <- res[["B"]]

	## Amatrix and Bmatrix are ordinary matrices--not ff objects.
	## By writing columns of a ordinary matrix to GT and GTP, we
	## save one read step later on.  GT and GTP will be updated to
	## calls and call probabilities by the crlmmGT2 function. The A
	## and B slots will retain the normalized intensities for the
	## A and B alleles
        ## The loop below gives rise to warnings: In `[<-.ff_array`(`*tmp*`, snp.index, jj, value = structure(c(6648L,  ... :
        ## number of elements to replace is not multiple of values for replacement
	for(j in seq_along(sel)){
		jj <- sel[j]
		A[snp.index, jj] <- Amatrix[, j]
		GT[snp.index, jj] <- Amatrix[, j]
		B[snp.index, jj] <- Bmatrix[, j]
		GTP[snp.index, jj] <- Bmatrix[, j]
	}
	if(length(np.index)>0) {
		cnAmatrix <- res[["cnAB"]]$A
		cnBmatrix <- res[["cnAB"]]$B
		for(j in seq_along(sel)){
			jj <- sel[j]
			A[np.index, jj] <- cnAmatrix[, j]
			GT[np.index, jj] <- cnAmatrix[, j]
			B[np.index, jj] <- cnBmatrix[, j]
			GTP[np.index, jj] <- cnBmatrix[, j]
		}
        }
	open(SKW); open(SNR); open(mixtureParams)
	SKW[sel] = res[["SKW"]][]
	SNR[sel] = res[["SNR"]][]
	mixtureParams[, sel] = res[["mixtureParams"]][]
        close(A); close(B)
	close(GT); close(GTP)
        close(SNR); close(SKW)
        close(mixtureParams)
        rm(res)
	gc(verbose=FALSE)
        TRUE
      }

.test <- function(dir, pattern = "^test_.*\\.R$")
{
    .failure_details <- function(result) {
        res <- result[[1L]]
        if (res$nFail > 0 || res$nErr > 0) {
            Filter(function(x) length(x) > 0,
                   lapply(res$sourceFileResults,
                          function(fileRes) {
                              names(Filter(function(x) x$kind != "success",
                                           fileRes))
                          }))
        } else list()
    }

    if (missing(dir)) {
        dir <- system.file("unitTests", package="crlmm")
        if (!length(dir)) {
            dir <- system.file("UnitTests", package="crlmm")
            if (!length(dir))
                stop("unable to find unit tests, no 'unitTests' dir")
        }
    }
    require("RUnit", quietly=TRUE) || stop("RUnit package not found")
    RUnit_opts <- getOption("RUnit", list())
    RUnit_opts$verbose <- 0L
    RUnit_opts$silent <- TRUE
    RUnit_opts$verbose_fail_msg <- TRUE
    options(RUnit = RUnit_opts)
    suite <- defineTestSuite(name="crlmm RUnit Tests", dirs=dir,
                             testFileRegexp=pattern,
                             rngKind="default",
                             rngNormalKind="default")
    result <- runTestSuite(suite)
    cat("\n\n")
    printTextProtocol(result, showDetails=FALSE)
    if (length(details <- .failure_details(result)) >0) {
        cat("\nTest files with failing tests\n")
        for (i in seq_along(details)) {
            cat("\n  ", basename(names(details)[[i]]), "\n")
            for (j in seq_along(details[[i]])) {
                cat("    ", details[[i]][[j]], "\n")
            }
        }
        cat("\n\n")
        stop("unit tests failed for package crlmm")
    }
    result
}

## Method("initialize", "AlleleSet",
##        function(.Object,
##                 assayData = assayDataNew(alleleA=alleleA,
##                                          alleleB=alleleB, ...),
##                 phenoData = annotatedDataFrameFrom(assayData, byrow=FALSE),
##                 featureData = annotatedDataFrameFrom(assayData, byrow=TRUE),
##                 experimentData = new("MIAME"),
##                 annotation = character(),
##                 protocolData = phenoData[,integer(0)],
##                 alleleA = new("matrix"),
##                 alleleB = matrix(numeric(),
## 		                    nrow=nrow(alleleA), ncol=ncol(alleleA),
##                                  dimnames=dimnames(alleleA)),
## 		   chromosome=integer(),
## 		   position=integer(),
## 		   isSnp=integer(),
##                 ...) {
## 		  .Object <- callNextMethod(.Object,
## 					    assayData = assayData,
## 					    phenoData = phenoData,
## 					    featureData = featureData,
## 					    experimentData = experimentData,
## 					    annotation = annotation,
## 					    protocolData = protocolData)
## 		  if(length(annotation) < 1){
## 			  if((length(position) < 1 | length(chromosome) < 1| length(isSnp) < 1)){
## 				  stop("must specify annotation if 'chromosome', 'position', and 'isSnp' are missing")
## 			  } else {
## 				  pData(featureData)$chromosome <- chromosome
## 				  pData(featureData)$position <- position
## 				  pData(featureData)$isSnp <- isSnp
## 			  }
## 		  } else{
## 			  .Object@annotation <- annotation
## 			  if((length(position) < 1 | length(chromosome) < 1| length(isSnp) < 1)){
## 				  if(!isSupportedAnnotation(annotation)){
## 					  stop("The annotation is not supported. Arguments 'chromosome', 'position', and 'isSnp' can be omitted from the initialization only if the annotation is supported (see oligoClasses:::supportedAnnotation()).")
## 				  }
## 			  } else {
## 				  pData(featureData)$chromosome <- chromosome
## 				  pData(featureData)$position <- position
## 				  pData(featureData)$isSnp <- isSnp
## 			  }
## 			  .Object@featureData <- featureData
## 		  }
## 		  ## Do after annotation has been assigned
## 		  if(!(all(c("chromosome", "position", "isSnp") %in% varLabels(featureData))) & isSupportedAnnotation(annotation)){
## 			  ##update the featureData
## 			  .Object@featureData <- addFeatureAnnotation.crlmm(.Object)
## 		  }
## 		  .Object
##        })
## 
## ow to make the initialization platform-specific?
## Method("initialize", "SnpSuperSet",
##        function(.Object,
## 		   call=new("matrix"),		   
##                 callProbability=matrix(NA, nrow(call), ncol(call), dimnames=dimnames(call)),
##                 phenoData = annotatedDataFrameFrom(assayData, byrow=FALSE),		   
## 		   featureData=annotatedDataFrameFrom(assayData, byrow=TRUE),
## 		   experimentData=new("MIAME"),
## 		   annotation=character(),
## 		   protocolData=phenoData[, integer(0)],
## 		   position=integer(),
## 		   chromosome=integer(),
## 		   isSnp=integer(),...){
## 		  .Object <- callNextMethod(.Object,
## 					    call=call,
## 					    callProbability=callProbability,
## 					    phenoData=phenoData,
## 					    featureData=featureData,
## 					    experimentData=experimentData,
## 					    annotation=annotation,
## 					    protocolData=protocolData,
## 					    position=position,
## 					    chromosome=chromosome,
## 					    isSnp=isSnp, ...)
##        })
##

##setMethod("initialize", "SnpSuperSet",
##          function(.Object,
##		   call=new("matrix"),		   
##                   callProbability=matrix(NA, nrow(call), ncol(call), dimnames=dimnames(call)),
##                   alleleA = new("matrix"),
##                   alleleB = matrix(numeric(),
##		                    nrow=nrow(alleleA), ncol=ncol(alleleA),
##                                    dimnames=dimnames(alleleA)),		   
##                   phenoData = annotatedDataFrameFrom(call, byrow=FALSE),		   
##		   featureData=annotatedDataFrameFrom(call, byrow=TRUE),
##		   experimentData=new("MIAME"),
##		   protocolData=phenoData[, integer(0)],
##		   position=integer(),
##		   chromosome=integer(),
##		   isSnp=integer(),
##		   annotation=character(), ... ){
##		  ##browser()
##		  ##the ... should be additional assayDataElements, if any
##		  .Object <- callNextMethod(.Object,
##					    call=call,
##					    callProbability=callProbability,
##					    alleleA=alleleA,
##					    alleleB=alleleB,
##					    phenoData=phenoData,
##					    featureData=featureData,
##					    experimentData=experimentData,
##					    protocolData=protocolData,
##					    annotation=annotation, ...)
##		  annotation <- .Object@annotation
##		  ##add chromosome, position, isSnp to featureData
##		  if(length(annotation) < 1){
##			  if((length(position) < 1| length(chromosome) < 1 | length(isSnp) < 1)){
##				  stop("must specify annotation if 'chromosome', 'position', and 'isSnp' are missing")
##			  } else {
##				  pData(featureData)$chromosome <- chromosome
##				  pData(featureData)$position <- position
##				  pData(featureData)$isSnp <- isSnp
##			  }
##		  } else{
##			  if((length(position) < 1| length(chromosome) < 1 | length(isSnp) < 1)){
##				  if(!isSupportedAnnotation(annotation)){
##					  stop("The annotation is not supported. Arguments 'chromosome', 'position', and 'isSnp' can be omitted from the initialization only if the annotation is supported (see oligoClasses:::supportedAnnotation()).")
##				  }
##			  } else {
##				  pData(featureData)$chromosome <- chromosome
##				  pData(featureData)$position <- position
##				  pData(featureData)$isSnp <- isSnp
##			  }
##			  .Object@featureData <- featureData
##		  }
##		  ##Do after annotation has been assigned
##		  if(!(all(c("chromosome", "position", "isSnp") %in% varLabels(featureData))) & isSupportedAnnotation(annotation)){
##			  .Object@featureData <- addFeatureAnnotation.crlmm(.Object)
##		  }		  
##		  .Object
##          })



##setMethod("addFeatureAnnotation", "SnpSuperSet", function(object, ...){
##	addFeatureAnnotation.crlmm(object, ...)
##})

##getParam.SnpSuperSet <- function(object, name, batch){
##		  label <- paste(name, batch, sep="_")
##		  colindex <- grep(label, fvarLabels(object))
##		  if(length(colindex) == 1){
##			  param <- fData(object)[, colindex]
##		  }
##		  if(length(colindex) < 1){
##			  param <- NULL
##		  }
##		  if(is.na(colindex)){
##			  stop(paste(label, " not found in object"))
##		  }
##		  if(length(colindex) > 1){
##			  stop(paste(label, " not unique"))
##		  }
##		  return(param)
##	  }



##setMethod("splitByChromosome", "SnpSuperSet", function(object, cnOptions){
##	tmpdir <- cnOptions[["tmpdir"]]
##	outdir <- cnOptions[["outdir"]]	
##	save.it <- cnOptions[["save.it"]]
##	path <- system.file("extdata", package=paste(annotation(object), "Crlmm", sep=""))	
##	load(file.path(path, "snpProbes.rda"))
##	snpProbes <- get("snpProbes")
##	load(file.path(path, "cnProbes.rda"))
##	cnProbes <- get("cnProbes")	
##	k <- grep("chr", colnames(snpProbes))
##	if(length(k) < 1) stop("chr or chromosome not in colnames(snpProbes)")
##	for(CHR in 1:24){
##		cat("Chromosome ", CHR, "\n")
##		snps <- rownames(snpProbes)[snpProbes[, k] == CHR]
##		cnps <- rownames(cnProbes)[cnProbes[, k] == CHR]
##		index <- c(match(snps, featureNames(object)),
##			   match(cnps, featureNames(object)))
##		index <- index[!is.na(index)]
##		callSetPlus <- object[index, ]
##		if(CHR != 24){
##			cnSet <- computeCopynumber(callSetPlus, cnOptions)
##			
##		} else{
##			message("Copy number estimates not available for chromosome Y.  Saving only the 'callSetPlus' object for this chromosome")
##			save(callSetPlus, file=file.path(outdir, paste("callSetPlus_", CHR, ".rda", sep="")))
##		}
##		if(cnOptions[["hiddenMarkovModel"]] & CHR != 24){
##			cnSet <- computeHmm(cnSet, cnOptions)
##		}
##		save(cnSet, file=file.path(outdir, paste("cnSet_", CHR, ".rda", sep="")))
##		saved.objects <- list.files(outdir, pattern="cnSet", full.names=TRUE)
####		} else{ ## save crlmmSet to outdir
####			save(cnSet, file=file.path(outdir, paste("cnSet_", CHR, ".rda", sep="")))
####			saved.objects <- list.files(outdir, pattern="cnSet", full.names=TRUE)			
####		}		
##	}
##	saved.objects
##})

##setMethod("computeCopynumber", "SnpSuperSet",
##	  function(object, cnOptions){
##		  computeCopynumber.SnpSuperSet(object, cnOptions)
##	  })

	

crlmm <- function(filenames, row.names=TRUE, col.names=TRUE,
                  probs=c(1/3, 1/3, 1/3), DF=6, SNRMin=5, gender=NULL,
                  save.it=FALSE, load.it=FALSE, intensityFile,
                  mixtureSampleSize=10^5, eps=0.1, verbose=TRUE,
                  cdfName, sns, recallMin=10, recallRegMin=1000,
                  returnParams=FALSE, badSNP=.7){
  if ((load.it | save.it) & missing(intensityFile))
	  stop("'intensityFile' is missing, and you chose either load.it or save.it")
  if (missing(sns)) sns <- basename(filenames)
  if (any(duplicated(sns))) stop("sample identifiers are not unique")
  if (!missing(intensityFile))
    if (load.it & !file.exists(intensityFile)){
      load.it <- FALSE
      message("File ", intensityFile, " does not exist.")
      message("Not loading it, but running SNPRMA from scratch.")
    }
  if (!load.it){
    res <- snprma(filenames, fitMixture=TRUE,
                  mixtureSampleSize=mixtureSampleSize, verbose=verbose,
                  eps=eps, cdfName=cdfName, sns=sns)
    if(save.it){
      t0 <- proc.time()
      save(res, file=intensityFile)
      t0 <- proc.time()-t0
      if (verbose) message("Used ", t0[3], " seconds to save ", intensityFile, ".")
    }
  }else{
    if (verbose) message("Loading ", intensityFile, ".")
    obj <- load(intensityFile)
    if (verbose) message("Done.")
    if (obj != "res")
      stop("Object in ", intensityFile, " seems to be invalid.")
  }
  if(row.names) row.names=res$gns else row.names=NULL
  if(col.names) col.names=res$sns else col.names=NULL
  res2 <- crlmmGT(res[["A"]], res[["B"]], res[["SNR"]],
                  res[["mixtureParams"]], res[["cdfName"]],
                  gender=gender, row.names=row.names,
                  col.names=col.names, recallMin=recallMin,
                  recallRegMin=1000, SNRMin=SNRMin,
                  returnParams=returnParams, badSNP=badSNP,
                  verbose=verbose)

  res2[["SNR"]] <- res[["SNR"]]
  res2[["SKW"]] <- res[["SKW"]]
  return(list2SnpSet(res2, returnParams=returnParams))
}

crlmmGT <- function(A, B, SNR, mixtureParams, cdfName, row.names=NULL,
                    col.names=NULL, probs=c(1/3, 1/3, 1/3), DF=6,
                    SNRMin=5, recallMin=10, recallRegMin=1000,
                    gender=NULL, desctrucitve=FALSE, verbose=TRUE,
                    returnParams=FALSE, badSNP=.7){

  keepIndex <- which(SNR>SNRMin)
  if(length(keepIndex)==0) stop("No arrays above quality threshold!")

  NC <- ncol(A)
  NR <- nrow(A)

  pkgname <- getCrlmmAnnotationName(cdfName)
  if(!require(pkgname, character.only=TRUE, quietly=!verbose)){
    suggCall <- paste("library(", pkgname, ", lib.loc='/Altern/Lib/Loc')", sep="")
    msg <- paste("If", pkgname,
                 "is installed on an alternative location, please load it manually by using",
                 suggCall)
    message(strwrap(msg))
    stop("Package ", pkgname, " could not be found.")
    rm(suggCall, msg)
  }

  if(verbose) message("Loading annotations.")
  loader("genotypeStuff.rda", .crlmmPkgEnv, pkgname)
  loader("mixtureStuff.rda", .crlmmPkgEnv, pkgname)

  ## this is toget rid of the 'no visible binding' notes
  ## variable definitions
  XIndex <- getVarInEnv("XIndex")
  autosomeIndex <- getVarInEnv("autosomeIndex")
  YIndex <- getVarInEnv("YIndex")
  SMEDIAN <- getVarInEnv("SMEDIAN")
  theKnots <- getVarInEnv("theKnots")
  regionInfo <- getVarInEnv("regionInfo")
  params <- getVarInEnv("params")

  if(is.null(gender)){
	  if(verbose) message("Determining gender.")
	  gender <- imputeGender(A, B, XIndex, YIndex, SNR, SNRMin)
  }

  Indexes <- list(autosomeIndex, XIndex, YIndex)
  cIndexes <- list(keepIndex,
                   keepIndex[which(gender[keepIndex]==2)],
                   keepIndex[which(gender[keepIndex]==1)])

  if(verbose) cat("Calling", NR, "SNPs for recalibration... ")

  ## call C
  fIndex <- which(gender==2)
  mIndex <- which(gender==1)
  newparams <- gtypeCallerR(A, B, fIndex, mIndex,
                            params[["centers"]], params[["scales"]], params[["N"]],
                            Indexes, cIndexes,
                            sapply(Indexes, length), sapply(cIndexes, length),
                            SMEDIAN, theKnots,
                            mixtureParams, DF, probs, 0.025)
  gc(verbose=FALSE)
  names(newparams) <- c("centers", "scales", "N")

  if(verbose) message("Done.")
  if(verbose) message("Estimating recalibration parameters.")
  d <- newparams[["centers"]] - params$centers

  ##regression
  Index <- intersect(which(pmin(newparams[["N"]][, 1],
                                newparams[["N"]][, 2],
                                newparams[["N"]][, 3]) > recallMin &
                                !apply(regionInfo, 1, any)),
                                autosomeIndex)
  if(length(Index) < recallRegMin){
    warning("Recalibration not possible. Possible cause: small sample size.")
    newparams <- params
    dev <- vector("numeric", nrow(newparams[["centers"]]))
    SS <- matrix(Inf, 3, 3)
    DD <- 0
  }else{
    data4reg <- as.data.frame(newparams[["centers"]][Index,])
    names(data4reg) <- c("AA", "AB", "BB")
    regParams <- cbind(  coef(lm(AA~AB*BB, data=data4reg)),
                       c(coef(lm(AB~AA+BB, data=data4reg)), 0),
                       coef(lm(BB~AA*AB, data=data4reg)))
    rownames(regParams) <- c("intercept", "X", "Y", "XY")
    rm(data4reg)

    minN <- 3
    newparams[["centers"]][newparams[["N"]] < minN] <- NA
    Index <- setdiff(which(rowSums(is.na(newparams[["centers"]]))==1), YIndex)
    if(verbose) cat("Filling out empty centers")
    for(i in Index){
      if(verbose) if(i%%10000==0)cat(".")
      mu <- newparams[["centers"]][i, ]
      j <- which(is.na(mu))
      newparams[["centers"]][i, j] <- c(1, mu[-j], prod(mu[-j]))%*%regParams[, j]
    }

    ##remaing NAs are made like originals
    if(length(YIndex)>0){
      noMoveIndex <- union(setdiff(which(rowSums(is.na(newparams[["centers"]]))>0), YIndex),
                           YIndex[rowSums(is.na(newparams[["centers"]][YIndex, ])>1)])
    }
    snps2ignore <- which(rowSums(is.na(newparams[["centers"]])) > 0)
    snps2keep <- setdiff(autosomeIndex, snps2ignore)
    rm(snps2ignore)
    newparams[["centers"]][is.na(newparams[["centers"]])] <- params[["centers"]][is.na(newparams[["centers"]])]
    if(verbose) cat("\n")

    if(verbose) message("Calculating and standardizing size of shift.")
    GG <- DD <- newparams[["centers"]] - params[["centers"]]
    DD <- sweep(DD, 2, colMeans(DD[autosomeIndex, ]))
    SS <- cov(DD[autosomeIndex, ])
    SSI <- solve(SS)
    dev <- vector("numeric", nrow(DD))
    if(length(YIndex)){
      dev[-YIndex] <- apply(DD[-YIndex, ], 1, function(x) x%*%SSI%*%x)
      dev[-YIndex] <- 1/sqrt( (2*pi)^3*det(SS))*exp(-0.5*dev[-YIndex])
      ##Now Y (only two params)
      SSY <- SS[c(1, 3), c(1, 3)]
      SSI <- solve(SSY)
      dev[YIndex] <- apply(DD[YIndex, c(1, 3)], 1, function(x) x%*%SSI%*%x)
      dev[YIndex] <- 1/sqrt( (2*pi)^2*det(SSY))*exp(-0.5*dev[YIndex])
    } else {
      dev=apply(DD,1,function(x) x%*%SSI%*%x)
      dev=1/sqrt( (2*pi)^3*det(SS))*exp(-0.5*dev)
    }
  }

  ## BC: must keep SD
  params[-2] <- newparams[-2]

  rm(newparams);gc(verbose=FALSE)
  if(verbose) cat("Calling", NR, "SNPs... ")
  ## ###################
  ## ## MOVE TO C#######
  ImNull <- gtypeCallerR2(A, B, fIndex, mIndex, params[["centers"]],
                          params[["scales"]], params[["N"]], Indexes,
                          cIndexes, sapply(Indexes, length),
                          sapply(cIndexes, length), SMEDIAN, theKnots,
                          mixtureParams, DF, probs, 0.025,
                          which(regionInfo[,2]),
                          which(regionInfo[,1]))
  gc(verbose=FALSE)
  ##  END MOVE TO C#######
  ## ##################

  dev <- dev/(dev+1/383)
  if(!is.null(row.names)){ rownames(A) <- rownames(B) <- names(dev) <- row.names}
  if(!is.null(col.names)){ colnames(A) <- colnames(B) <- col.names}

  if(length(Index) >= recallRegMin){
   tmp4batchQC <- DD[autosomeIndex,]/(params[["N"]][autosomeIndex,]+1)
   tmpSnpQc <- dev[autosomeIndex]
   SS <- cov(tmp4batchQC[tmpSnpQc < badSNP,])
   batchQC <- mean(diag(SS))
  }else{
    SS <- matrix(0, 3, 3)
    batchQC <- Inf
  }

  if(verbose) message("Done.")
  if (returnParams){
    return(list(calls=A, confs=B, SNPQC=dev, batchQC=batchQC, params=params, DD=DD, covDD=SS, gender=gender, pkgname=pkgname))
  }else{
    return(list(calls=A, confs=B, SNPQC=dev, batchQC=batchQC, DD=DD, covDD=SS, gender=gender, pkgname=pkgname))
  }
}


gtypeCallerR <- function(A, B, fIndex, mIndex, theCenters, theScales,
                         theNs, Indexes, cIndexes, nIndexes,
                         ncIndexes, SMEDIAN, knots, params, dft,
                         probs, trim){

  stopifnot(!missing(A), !missing(B), dim(A)==dim(B),
            nrow(A)==nrow(theCenters), nrow(A)==nrow(theScales),
            nrow(A) == nrow(theNs), length(knots)==3, nrow(params)==4,
            ncol(params)==ncol(A), length(trim)==1, length(probs)==3)

  ## make code robust
  ## check types before passing to C

  .Call("gtypeCallerPart1", A, B, as.integer(fIndex),
        as.integer(mIndex), as.numeric(theCenters),
        as.numeric(theScales), as.integer(theNs),
        lapply(Indexes, as.integer), lapply(cIndexes, as.integer),
        as.integer(nIndexes), as.integer(ncIndexes),
        as.numeric(SMEDIAN), as.numeric(knots),
        as.numeric(params), as.integer(dft), as.numeric(probs),
        as.numeric(trim), PACKAGE="crlmm")

}

gtypeCallerR2 <- function(A, B, fIndex, mIndex, theCenters, theScales,
                         theNs, Indexes, cIndexes, nIndexes,
                         ncIndexes, SMEDIAN, knots, params, dft,
                         probs, trim, noTraining, noInfo){

  stopifnot(!missing(A), !missing(B), dim(A)==dim(B),
            nrow(A)==nrow(theCenters), nrow(A)==nrow(theScales),
            nrow(A) == nrow(theNs), length(knots)==3, nrow(params)==4,
            ncol(params)==ncol(A), length(trim)==1, length(probs)==3)

  .Call("gtypeCallerPart2", A, B,
        as.integer(fIndex), as.integer(mIndex),
        as.numeric(theCenters), as.numeric(theScales),
        as.integer(theNs), Indexes, cIndexes, nIndexes, ncIndexes,
        as.numeric(SMEDIAN), as.numeric(knots), as.numeric(params),
        as.integer(dft), as.numeric(probs), as.numeric(trim),
        as.integer(noTraining), as.integer(noInfo), PACKAGE="crlmm")

}

### parallel version
crlmm2 <- function(filenames, row.names=TRUE, col.names=TRUE,
                   probs=c(1/3, 1/3, 1/3), DF=6, SNRMin=5, gender=NULL,
                   save.it=FALSE, load.it=FALSE, intensityFile,
                   mixtureSampleSize=10^5, eps=0.1, verbose=TRUE,
                   cdfName, sns, recallMin=10, recallRegMin=1000,
                   returnParams=FALSE, badSNP=.7){
  if ((load.it || save.it) && missing(intensityFile))
    stop("'intensityFile' is missing, and you chose either load.it or save.it")
  if (missing(sns)) sns <- basename(filenames)
  if (!missing(intensityFile))
    if (load.it & !file.exists(intensityFile)){
      load.it <- FALSE
      message("File ", intensityFile, " does not exist.")
      message("Not loading it, but running SNPRMA from scratch.")
    }
  if (!load.it){
    res <- snprma2(filenames, fitMixture=TRUE,
                   mixtureSampleSize=mixtureSampleSize, verbose=verbose,
                   eps=eps, cdfName=cdfName, sns=sns)
    open(res[["A"]])
    open(res[["B"]])
    open(res[["SNR"]])
    open(res[["mixtureParams"]])
    if(save.it){
      t0 <- proc.time()
      save(res, file=intensityFile)
      t0 <- proc.time()-t0
      if (verbose) message("Used ", t0[3], " seconds to save ", intensityFile, ".")
    }
  }else{
    if (verbose) message("Loading ", intensityFile, ".")
    obj <- load(intensityFile)
    if (verbose) message("Done.")
    if (obj != "res")
      stop("Object in ", intensityFile, " seems to be invalid.")
  }
  if(row.names) row.names=res$gns else row.names=NULL
  if(col.names) col.names=res$sns else col.names=NULL
  res2 <- crlmmGT2(res[["A"]], res[["B"]], res[["SNR"]],
                   res[["mixtureParams"]], res[["cdfName"]],
                   gender=gender, row.names=row.names,
                   col.names=col.names, recallMin=recallMin,
                   recallRegMin=1000, SNRMin=SNRMin,
                   returnParams=returnParams, badSNP=badSNP,
                   verbose=verbose)

  res2[["SNR"]] <- res[["SNR"]]
  res2[["SKW"]] <- res[["SKW"]]
  return(list2SnpSet(res2, returnParams=returnParams))
}

imputeGender <- function(A, B, XIndex, YIndex, SNR, SNRMin){
	if(length(YIndex) > 0){
		a <- log2(as.matrix(A[XIndex,,drop=FALSE]))
		b <- log2(as.matrix(B[XIndex,,drop=FALSE]))
		meds.X <- (apply(a+b, 2, median))/2
		a <- log2(as.matrix(A[YIndex,,drop=FALSE]))
		b <- log2(as.matrix(B[YIndex,,drop=FALSE]))
		meds.Y <- (apply(a+b, 2, median))/2
		R <- meds.X - meds.Y
		if(sum(SNR[] > SNRMin) == 1){
			gender <- ifelse(R[SNR[] > SNRMin] > 0.5, 2L, 1L)
		} else{
			gender <- kmeans(R, c(min(R[SNR[]>SNRMin]), max(R[SNR[]>SNRMin])))[["cluster"]]
		}
	} else {
		XMedian <- apply(log2(as.matrix(A[XIndex,,drop=FALSE]))+log2(as.matrix(B[XIndex,, drop=FALSE])), 2, median)/2
		if(sum(SNR > SNRMin) == 1){
			gender <- which.min(c(abs(XMedian-8.9), abs(XMedian-9.5)))
		} else{
			gender <- kmeans(XMedian, c(min(XMedian[SNR[]>SNRMin]), max(XMedian[SNR[]>SNRMin])))[["cluster"]]
		}
	}
	return(gender)
}

snprma <- function(filenames, mixtureSampleSize=10^5, fitMixture=FALSE,
                   eps=0.1, verbose=TRUE, seed=1, cdfName, sns){
  if (missing(sns)) sns <- basename(filenames)
  if (missing(cdfName))
    cdfName <- read.celfile.header(filenames[1])$cdfName
  pkgname <- getCrlmmAnnotationName(cdfName)

  if(!require(pkgname, character.only=TRUE, quietly=!verbose)){
    suggCall <- paste("library(", pkgname,
                      ", lib.loc='/Altern/Lib/Loc')",
                      sep="")
    msg <- paste("If", pkgname,
                 "is installed on an alternative location,",
                 "please load it manually by using", suggCall)
    message(strwrap(msg))
    stop("Package ", pkgname, " could not be found.")
  }

  if(verbose) message("Loading annotations and mixture model parameters.")
  loader("preprocStuff.rda", .crlmmPkgEnv, pkgname)
  loader("genotypeStuff.rda", .crlmmPkgEnv, pkgname)
  loader("mixtureStuff.rda", .crlmmPkgEnv, pkgname)
  autosomeIndex <- getVarInEnv("autosomeIndex")
  pnsa <- getVarInEnv("pnsa")
  pnsb <- getVarInEnv("pnsb")
  fid <- getVarInEnv("fid")
  reference <- getVarInEnv("reference")
  aIndex <- getVarInEnv("aIndex")
  bIndex <- getVarInEnv("bIndex")
  SMEDIAN <- getVarInEnv("SMEDIAN")
  theKnots <- getVarInEnv("theKnots")
  gns <- getVarInEnv("gns")

  ##We will read each cel file, summarize, and run EM one by one
  ##We will save parameters of EM to use later
  mixtureParams <- matrix(0, 4, length(filenames))
  SNR <- vector("numeric", length(filenames))
  SKW <- vector("numeric", length(filenames))
##   mixtureParams <- initializeBigMatrix("crlmmMixt-", 4, length(filenames))
##   SNR <- initializeBigVector("crlmmSNR-", length(filenames), "numeric")
##   SKW <- initializeBigVector("crlmmSKW-", length(filenames), "numeric")

  ## This is the sample for the fitting of splines
  ## BC: I like better the idea of the user passing the seed,
  ##     because this might intefere with other analyses
  ##     (like what happened to GCRMA)
  set.seed(seed)
  idx <- sort(sample(autosomeIndex, mixtureSampleSize))
  ##S will hold (A+B)/2 and M will hold A-B
  ##NOTE: We actually dont need to save S. Only for pics etc...
  ##f is the correction. we save to avoid recomputing
  A <- matrix(as.integer(0), length(pnsa), length(filenames))
  B <- matrix(as.integer(0), length(pnsb), length(filenames))

  if(verbose){
    message("Processing ", length(filenames), " files.")
    pb <- txtProgressBar(min=0, max=length(filenames), style=3)
  }

  ##for skewness. no need to do everything
  idx2 <- sample(length(fid), 10^5)

  ##We start looping throug cel files
  for(i in seq(along=filenames)){
    y <- as.matrix(read.celfile(filenames[i], intensity.means.only=TRUE)[["INTENSITY"]][["MEAN"]][fid])
    x <- log2(y[idx2])
    SKW[i] <- mean((x-mean(x))^3)/(sd(x)^3)
    rm(x)
    y <- normalize.quantiles.use.target(y, target=reference)
    A[, i] <- intMedianSummaries(y[aIndex, 1, drop=FALSE], pnsa)
    B[, i] <- intMedianSummaries(y[bIndex, 1, drop=FALSE], pnsb)
    ##Now to fit the EM
    if(fitMixture){
      S <- (log2(A[idx, i])+log2(B[idx, i]))/2 - SMEDIAN
      M <- log2(A[idx, i])-log2(B[idx, i])

      ##we need to test the choice of eps.. it is not the max diff between funcs
      tmp <- fitAffySnpMixture56(S, M, theKnots, eps=eps)

      mixtureParams[, i] <- tmp[["coef"]]
      SNR[i] <- tmp[["medF1"]]^2/(tmp[["sigma1"]]^2+tmp[["sigma2"]]^2)
    }
    if (verbose) setTxtProgressBar(pb, i)
  }
  if (verbose) close(pb)
  if (!fitMixture) SNR <- mixtureParams <- NA
  ## gns comes from preprocStuff.rda
  list(A=A, B=B, sns=sns, gns=gns, SNR=SNR, SKW=SKW, mixtureParams=mixtureParams, cdfName=cdfName)
}

fitAffySnpMixture56 <- function(S, M, knots, probs=rep(1/3, 3), eps=.01, maxit=10, verbose=FALSE){
  ##56 stands for 5 and 6 arrays but will also work for Illumina
  ##Note the unfortunate choice of numbering:
  ##1 is BB, 2 AB, and 3 AA. Opposite to everything else!
  ##this is legacy code I decided not to change.
  ## this why at the end we report -coefs: F1 is the negative f
  mus <- append(quantile(M, c(1, 5)/6, names=FALSE), 0, 1)
  sigmas <- rep(mad(c(M[M<mus[1]]-mus[1], M[M>mus[3]]-mus[3])), 3)
  sigmas[2] <- sigmas[2]/2

  weights <- apply(cbind(mus, sigmas), 1, function(p) dnorm(M, p[1], p[2]))
  previousF1 <- -Inf
  change <- eps+1
  it <- 0

  if(verbose) message("Max change must be under ", eps, ".")
  matS <- stupidSplineBasis(S, knots)
  while (change > eps & it < maxit){
    it <- it+1
    ## E
    z <- sweep(weights, 2, probs, "*")
    LogLik <- rowSums(z)
    z <- sweep(z, 1, LogLik, "/")
    probs <- colMeans(z)

    ## M
    fit1 <- crossprod(chol2inv(chol(crossprod(sweep(matS, 1, z[, 1], FUN="*"), matS))), crossprod(matS, z[, 1]*M))

    fit2 <- sum(z[, 2]*M)/sum(z[, 2])
    F1 <- matS%*%fit1
    sigmas[c(1, 3)] <- sqrt(sum(z[, 1]*(M-F1)^2)/sum(z[, 1]))
    sigmas[2] <- sqrt(sum(z[, 2]*(M-fit2)^2)/sum(z[, 2]))

    weights[, 1] <- dnorm(M, F1, sigmas[1])
    weights[, 2] <- dnorm(M, fit2, sigmas[2])
    weights[, 3] <- dnorm(M, -F1, sigmas[3])

    change <- max(abs(F1-previousF1))
    previousF1 <- F1
    if(verbose) message("Iter ", it, ": ", change, ".")
  }
  medF1 <- median(-F1)
 return(list(coef= -fit1, medF1=medF1, sigma1=sigmas[1], sigma2=sigmas[2]))
}

snprma2 <- function(filenames, mixtureSampleSize=10^5, fitMixture=FALSE,
                    eps=0.1, verbose=TRUE, seed=1, cdfName, sns){
  if (missing(sns)) sns <- basename(filenames)
  if (missing(cdfName))
    cdfName <- read.celfile.header(filenames[1])[["cdfName"]]
  pkgname <- getCrlmmAnnotationName(cdfName)
  stopifnot(require(pkgname, character.only=TRUE, quietly=!verbose))

  if(verbose) message("Loading annotations and mixture model parameters.")
  obj <- loader("preprocStuff.rda", .crlmmPkgEnv, pkgname)
  pnsa <- getVarInEnv("pnsa")
  pnsb <- getVarInEnv("pnsb")
  gns <- getVarInEnv("gns")
  rm(list=obj, envir=.crlmmPkgEnv)
  rm(obj)

  ##We will read each cel file, summarize, and run EM one by one
  ##We will save parameters of EM to use later
  if(verbose) message("Initializing objects.")
  mixtureParams <- initializeBigMatrix("crlmmMixt-", 4, length(filenames), "double")
  SNR <- initializeBigVector("crlmmSNR-", length(filenames), "double")
  SKW <- initializeBigVector("crlmmSKW-", length(filenames), "double")

  ## This is the sample for the fitting of splines
  ## BC: I like better the idea of the user passing the seed,
  ##     because this might intefere with other analyses
  ##     (like what happened to GCRMA)
  ##S will hold (A+B)/2 and M will hold A-B
  ##NOTE: We actually dont need to save S. Only for pics etc...
  ##f is the correction. we save to avoid recomputing
  A <- initializeBigMatrix("crlmmA-", length(pnsa), length(filenames), "integer")
  B <- initializeBigMatrix("crlmmB-", length(pnsb), length(filenames), "integer")

  sampleBatches <- splitIndicesByNode(seq(along=filenames))

  if(verbose) message("Processing ", length(filenames), " files.")

  ocLapply(sampleBatches, processCEL, filenames=filenames,
           fitMixture=fitMixture, A=A, B=B, SKW=SKW, SNR=SNR,
           mixtureParams=mixtureParams, eps=eps, seed=seed,
           mixtureSampleSize=mixtureSampleSize, pkgname=pkgname,
           neededPkgs=c("crlmm", pkgname))
  close(mixtureParams)
  close(SNR)
  close(SKW)
  close(A)
  close(B)

  list(A=A, B=B, sns=sns, gns=gns, SNR=SNR, SKW=SKW, mixtureParams=mixtureParams, cdfName=cdfName)
}


processCEL <- function(i, filenames, fitMixture, A, B, SKW, SNR,
                       mixtureParams, eps, seed, mixtureSampleSize,
                       pkgname){
  obj1 <- loader("preprocStuff.rda", .crlmmPkgEnv, pkgname)
  obj2 <- loader("genotypeStuff.rda", .crlmmPkgEnv, pkgname)
  obj3 <- loader("mixtureStuff.rda", .crlmmPkgEnv, pkgname)
  autosomeIndex <- getVarInEnv("autosomeIndex")
  pnsa <- getVarInEnv("pnsa")
  pnsb <- getVarInEnv("pnsb")
  fid <- getVarInEnv("fid")
  reference <- getVarInEnv("reference")
  aIndex <- getVarInEnv("aIndex")
  bIndex <- getVarInEnv("bIndex")
  SMEDIAN <- getVarInEnv("SMEDIAN")
  theKnots <- getVarInEnv("theKnots")
  gns <- getVarInEnv("gns")
  rm(list=c(obj1, obj2, obj3), envir=.crlmmPkgEnv)
  rm(obj1, obj2, obj3)

  ## for mixture
  set.seed(seed)
  idx <- sort(sample(autosomeIndex, mixtureSampleSize))
  ##for skewness. no need to do everything
  idx2 <- sample(length(fid), 10^5)

  open(A)
  open(B)
  open(SKW)
  open(mixtureParams)
  open(SNR)

  for (k in i){
    y <- as.matrix(read.celfile(filenames[k], intensity.means.only=TRUE)[["INTENSITY"]][["MEAN"]][fid])
    x <- log2(y[idx2])
    SKW[k] <- mean((x-mean(x))^3)/(sd(x)^3)
    rm(x)
    y <- normalize.quantiles.use.target(y, target=reference)
    A[, k] <- intMedianSummaries(y[aIndex, 1, drop=FALSE], pnsa)
    B[, k] <- intMedianSummaries(y[bIndex, 1, drop=FALSE], pnsb)
    rm(y)

    if(fitMixture){
      S <- (log2(A[idx,k])+log2(B[idx, k]))/2 - SMEDIAN
      M <- log2(A[idx, k])-log2(B[idx, k])
      tmp <- fitAffySnpMixture56(S, M, theKnots, eps=eps)
      rm(S, M)
      mixtureParams[, k] <- tmp[["coef"]]
      SNR[k] <- tmp[["medF1"]]^2/(tmp[["sigma1"]]^2+tmp[["sigma2"]]^2)
      rm(tmp)
    } else {
      mixtureParams[, k] <- NA
      SNR[k] <- NA
    }
  }
  close(A)
  close(B)
  close(SKW)
  close(mixtureParams)
  close(SNR)
  rm(list=ls())
  gc(verbose=FALSE)
  TRUE
}

##setGeneric("batch", function(object) standardGeneric("batch"))
##setGeneric("getParam", function(object, name, batch) standardGeneric("getParam"))
setGeneric("cnIndex", function(object) standardGeneric("cnIndex"))
setGeneric("cnNames", function(object) standardGeneric("cnNames"))
##setGeneric("computeCopynumber", function(object, ...) standardGeneric("computeCopynumber"))
##setGeneric("pr", function(object, name, batch, value) standardGeneric("pr"))
setGeneric("snpIndex", function(object) standardGeneric("snpIndex"))
##setGeneric("snpNames", function(object) standardGeneric("snpNames"))

setGeneric("CA", function(object, ...) standardGeneric("CA"))
setGeneric("CB", function(object, ...) standardGeneric("CB"))
setGeneric("totalCopynumber", function(object, ...) standardGeneric("totalCopynumber"))


setGeneric("Ns", function(object, ...) standardGeneric("Ns"))
setGeneric("corr", function(object, ...) standardGeneric("corr"))
setGeneric("mads", function(object, ...) standardGeneric("mads"))
setGeneric("medians", function(object, ...) standardGeneric("medians"))
setGeneric("tau2", function(object, ...) standardGeneric("tau2"))

## The generics below are for internal use with copy number methods
## If we keep them in oligoClasses, we need to export and document
setGeneric("N.AA", function(object) standardGeneric("N.AA"))
setGeneric("N.AB", function(object) standardGeneric("N.AB"))
setGeneric("N.BB", function(object) standardGeneric("N.BB"))
setGeneric("N.AA<-", function(object, value) standardGeneric("N.AA<-"))
setGeneric("N.AB<-", function(object, value) standardGeneric("N.AB<-"))
setGeneric("N.BB<-", function(object, value) standardGeneric("N.BB<-"))

setGeneric("medians", function(object, ...) standardGeneric("medians"))
setGeneric("medianA.AA", function(object) standardGeneric("medianA.AA"))
setGeneric("medianA.AB", function(object) standardGeneric("medianA.AB"))
setGeneric("medianA.BB", function(object) standardGeneric("medianA.BB"))
setGeneric("medianB.AA", function(object) standardGeneric("medianB.AA"))
setGeneric("medianB.AB", function(object) standardGeneric("medianB.AB"))
setGeneric("medianB.BB", function(object) standardGeneric("medianB.BB"))
setGeneric("medianA.AA<-", function(object, value) standardGeneric("medianA.AA<-"))
setGeneric("medianA.AB<-", function(object, value) standardGeneric("medianA.AB<-"))
setGeneric("medianA.BB<-", function(object, value) standardGeneric("medianA.BB<-"))
setGeneric("medianB.AA<-", function(object, value) standardGeneric("medianB.AA<-"))
setGeneric("medianB.AB<-", function(object, value) standardGeneric("medianB.AB<-"))
setGeneric("medianB.BB<-", function(object, value) standardGeneric("medianB.BB<-"))



setGeneric("madA.AA", function(object) standardGeneric("madA.AA"))
setGeneric("madA.AB", function(object) standardGeneric("madA.AB"))
setGeneric("madA.BB", function(object) standardGeneric("madA.BB"))
setGeneric("madB.AA", function(object) standardGeneric("madB.AA"))
setGeneric("madB.AB", function(object) standardGeneric("madB.AB"))
setGeneric("madB.BB", function(object) standardGeneric("madB.BB"))
setGeneric("madA.AA<-", function(object, value) standardGeneric("madA.AA<-"))
setGeneric("madA.AB<-", function(object, value) standardGeneric("madA.AB<-"))
setGeneric("madA.BB<-", function(object, value) standardGeneric("madA.BB<-"))
setGeneric("madB.AA<-", function(object, value) standardGeneric("madB.AA<-"))
setGeneric("madB.AB<-", function(object, value) standardGeneric("madB.AB<-"))
setGeneric("madB.BB<-", function(object, value) standardGeneric("madB.BB<-"))

setGeneric("tau2A.AA", function(object) standardGeneric("tau2A.AA"))
setGeneric("tau2A.BB", function(object) standardGeneric("tau2A.BB"))
setGeneric("tau2B.AA", function(object) standardGeneric("tau2B.AA"))
setGeneric("tau2B.BB", function(object) standardGeneric("tau2B.BB"))
setGeneric("tau2A.AA<-", function(object, value) standardGeneric("tau2A.AA<-"))
setGeneric("tau2A.BB<-", function(object, value) standardGeneric("tau2A.BB<-"))
setGeneric("tau2B.AA<-", function(object, value) standardGeneric("tau2B.AA<-"))
setGeneric("tau2B.BB<-", function(object, value) standardGeneric("tau2B.BB<-"))

setGeneric("corrAA", function(object) standardGeneric("corrAA"))
setGeneric("corrAB", function(object) standardGeneric("corrAB"))
setGeneric("corrBB", function(object) standardGeneric("corrBB"))
setGeneric("corrAA<-", function(object, value) standardGeneric("corrAA<-"))
setGeneric("corrAB<-", function(object, value) standardGeneric("corrAB<-"))
setGeneric("corrBB<-", function(object, value) standardGeneric("corrBB<-"))



setGeneric("nuA", function(object) standardGeneric("nuA"))
setGeneric("nuB", function(object) standardGeneric("nuB"))
setGeneric("phiA", function(object) standardGeneric("phiA"))
setGeneric("phiB", function(object) standardGeneric("phiB"))
setGeneric("phiPrimeA", function(object) standardGeneric("phiPrimeA"))
setGeneric("phiPrimeB", function(object) standardGeneric("phiPrimeB"))
setGeneric("phiPrimeA<-", function(object, value) standardGeneric("phiPrimeA<-"))
setGeneric("phiPrimeB<-", function(object, value) standardGeneric("phiPrimeB<-"))
setGeneric("tau2A", function(object) standardGeneric("tau2A"))
setGeneric("tau2B", function(object) standardGeneric("tau2B"))
setGeneric("nuA<-", function(object, value) standardGeneric("nuA<-"))
setGeneric("nuB<-", function(object, value) standardGeneric("nuB<-"))
setGeneric("phiA<-", function(object, value) standardGeneric("phiA<-"))
setGeneric("phiB<-", function(object, value) standardGeneric("phiB<-"))
setGeneric("tau2A<-", function(object, value) standardGeneric("tau2A<-"))
setGeneric("tau2B<-", function(object, value) standardGeneric("tau2B<-"))
setGeneric("flags<-", function(object, value) standardGeneric("flags<-"))

setGeneric("posteriorMean", function(object) standardGeneric("posteriorMean"))
setGeneric("posteriorMean<-", function(object, value) standardGeneric("posteriorMean<-"))

setGeneric("posteriorProbability", function(object, predictRegion, copyNumber=0:4, w)
	   standardGeneric("posteriorProbability"))
setGeneric("calculatePosteriorMean", function(object, posteriorProb, copyNumber=0:4, ...)
	   standardGeneric("calculatePosteriorMean"))

setGeneric("predictionRegion", function(object, copyNumber=0:4)
	   standardGeneric("predictionRegion"))
setGeneric("xyplot", useAsDefault=function(x, data, ...) lattice::xyplot(x, data,...))
setGeneric("xyplotcrlmm", function(x, data, predictRegion,...) standardGeneric("xyplotcrlmm"))
setGeneric("calculateRBaf", function(object, batch.name, chrom) standardGeneric("calculateRBaf"))

##setGeneric("mixtureParams<-", function(object,value) standardGeneric("mixtureParams<-"))
setGeneric("OligoSetList", function(object,...) standardGeneric("OligoSetList"))
setGeneric("BafLrrSetList", function(object,...) standardGeneric("BafLrrSetList"))

plotSamples = function(cnSet, col=1, offset=0, xlim=c(9,16), ylim=c(-5,5), verbose=FALSE, sample=100000, seed=1, type="smoothScatter"){
    if(missing(col)){
        stop("col is missing, please specify which sample/(s) to plot")
    }
    set.seed(seed)
    row = sample(nrow(cnSet), sample)
    if(verbose)  
      message("compute M values")        
    M = computeLogRatio(cnSet, offset=offset, verbose=verbose, row=row, col=col)
    if(type=="beanplot") {
      sel = rowSums(is.finite(M[,]) & !is.na(M[,]))==ncol(M)
      beanplot(as.vector(M[sel,])~factor(rep(col, each=sum(sel))), ylab="M", xlab="Sample", ylim=ylim,
               beanlines="median", what=c(0,1,1,0), col="grey", border="black") #c("#CAB2D6", "#33A02C", "#B2DF8A"), border = "#CAB2D6", 
      abline(h=0, col="gray", lty=2)
      delete(M); rm(M)
    }
    if(type=="smoothScatter") {
      if(verbose)
        message("compute S values")
      S = computeAverageLogIntensity(cnSet, offset=offset, verbose=verbose, row=row, col=col)
      for(i in 1:length(col)) {
        smoothScatter(S[,i], M[,i], main=colnames(M)[i], ylab="M", xlab="S", xlim=xlim, ylim=ylim)
        abline(h=0, col="gray", lty=2)
      }
    delete(S, M); rm(S, M)
    }
}


plotSNPs = function(cnSet, row=1, offset=0, xlim=c(9,16), ylim=c(-5,5), verbose=FALSE){
    if(missing(row)){
        stop("row is missing, please specify which SNP/(s) to plot")
    }
    if(verbose)  
      message("compute M values")        
    M = computeLogRatio(cnSet, offset=offset, verbose=verbose, row=row)
    if(verbose)
      message("compute S values")   
    S = computeAverageLogIntensity(cnSet, offset=offset, verbose=verbose, row=row)
    if(verbose)
      message("get genotype calls")
    for(i in 1:length(row)){
       plot(S[i,], M[i,], xlab="S", ylab="M", xlim=xlim, ylim=ylim,
            main=rownames(M)[i], pch=19, col=calls(cnSet)[row[i],])
       abline(h=0, col="gray", lty=2)
    }
    rm(S, M)
}

setMethod("cnIndex", "eSet", function(object){
	index <- match(cnNames(object), featureNames(object), nomatch=0)
	index[index != 0]
  })

setMethod("cnNames", "eSet", function(object) {
	path <- system.file("extdata", package=paste(annotation(object), "Crlmm", sep=""))
	load(file.path(path, "cnProbes.rda"))
	cnProbes <- get("cnProbes")
	cnps <- rownames(cnProbes)
	cnps <- cnps[cnps %in% featureNames(object)]
	index <- match(cnps, featureNames(object), nomatch=0)
	index <- index[index != 0]
	featureNames(object)[index]
  })


##setMethod("snpIndex", "eSet", function(object){
##	index <- match(snpNames(object), featureNames(object), nomatch=0)
##	index[index != 0]
##})
##
##setMethod("snpNames", "eSet", function(object){
##	path <- system.file("extdata", package=paste(annotation(object), "Crlmm", sep=""))
##	load(file.path(path, "snpProbes.rda"))
##	snpProbes <- get("snpProbes")
##	snps <- rownames(snpProbes)
##	snps <- snps[snps %in% featureNames(object)]
##	index <- match(snps, featureNames(object), nomatch=0)
##	index <- index[index != 0]
##	featureNames(object)[index]
##})

##setMethod("combine", signature=signature(x="eSet", y="eSet"),
##	  function(x, y, ...){
##		  ##Check that both x and y are valid objects
##		  if(!validObject(x)) stop("x is not a valid object")
##		  if(!validObject(y)) stop("y is not a valid object")
##
##		  if(annotation(x) != annotation(y)) stop("must have same value for annotation slot")
##		  if(class(x) != class(y)) stop("objects must have the same class")
##		  if(storageMode(assayData(x)) != storageMode(assayData(y))){
##			  stop("objects must have same storage mode for assayData")
##		  }
##		  fd <- combine(featureData(x), featureData(y))
##		  pd <- combine(phenoData(x), phenoData(y))
##		  ad.x <- as.list(assayData(x))
##		  ad.y <- as.list(assayData(y))
##		  ad.xy <- mapply(rbind, ad.x, ad.y, SIMPLIFY=FALSE)
##		  id.x <- match(rownames(ad.xy[[1]]), featureNames(fd))
##		  ee <- combine(experimentData(x), experimentData(y))
##		  assayData(x) <- ad.xy
##		  storageMode(assayData(x)) <- storageMode(assayData(y))
##		  experimentData(x) <- ee
##		  featureData(x) <- fd
##		  phenoData(x) <- pd
##		  x
##          })

krlmm <- function(cnSet, cdfName, gender=NULL, trueCalls=NULL, verbose=TRUE) {
    pkgname <- getCrlmmAnnotationName(cdfName)
    
### pre-processing, output M    
    M = computeLogRatio(cnSet, verbose)
    message("Leaving out non-variant SNPs")
    
    # For SNPs with less than 3 distinct data point, exclude them from downstream analysis
    uniqueCount = apply(M[,], 1, function(x){length(unique(x))})
    SNPtoProcessIndex = uniqueCount >= 3
    noVariantSNPIndex = uniqueCount < 3
    M = M[SNPtoProcessIndex, ]

    numSNP = nrow(M)
    numSample = ncol(M)
    
    calls = oligoClasses::initializeBigMatrix(name="calls", numSNP, numSample, vmode = "integer")
    scores = oligoClasses::initializeBigMatrix(name="scores", numSNP, numSample, vmode = "double")
    open(calls)
    open(scores)

    rownames(calls) = rownames(M)
    rownames(scores) = rownames(M)
    colnames(calls) = colnames(M)
    colnames(scores) = colnames(M)  
    
    priormeans = calculatePriorValues(M, numSNP, verbose)
      
### retrieve or calculate coefficients
    krlmmCoefficients = getKrlmmVGLMCoefficients(pkgname, trueCalls, VGLMparameters, verbose, numSample, colnames(M))
    
### do VGLM fit, to predict k for each SNP
    kPrediction <- predictKwithVGLM(VGLMparameters, krlmmCoefficients, verbose)
    rm(VGLMparameters)
    
### assign calls
    assignCalls(calls, M, kPrediction, priormeans, numSNP, numSample, verbose)
    rm(kPrediction)

### assign confidence scores
    computeCallPr(scores, M, calls, numSNP, numSample, verbose)
    
### add back SNPs excluded before
    AddBackNoVarianceSNPs(cnSet, calls, scores, numSNP, numSample, SNPtoProcessIndex, noVariantSNPIndex)

    loader("genotypeStuff.rda", .crlmmPkgEnv, pkgname)
    YIndex <- getVarInEnv("YIndex")
    loader("annotation.rda", .crlmmPkgEnv, pkgname)
    annotation <- getVarInEnv("annot")
    
### impute gender if gender information not provided
    if (is.null(gender)) {
        gender = krlmmImputeGender(cnSet, annotation, YIndex, verbose)
    }

### double-check ChrY SNP cluster, impute gender if gender information not provided
    if (!(is.null(gender))) {
        verifyChrYSNPs(cnSet, M, calls, gender, annotation, YIndex, priormeans, verbose)
    }

    close(calls)
    close(scores)
    rm(calls)
    rm(scores)
    rm(M)
    rm(annotation)
    rm(YIndex)
    TRUE
}

krlmmnopkg <- function(cnSet, offset, gender=NULL, normalize.method=NULL, annotation=NULL, verbose=TRUE) {
    if(is.null(normalize.method)){
        message("No normalization method specified. Quantile normalization applied")
        M=quantilenormalization(cnSet,verbose,offset=offset)
    }else{
        if(normalize.method=="quantile"){
            M=quantilenormalization(cnSet,verbose,offset=offset)
        }
        if(normalize.method=="loess"){
            M=loessnormalization(cnSet,verbose,offset=offset)
	}
    }

    message("Leaving out non-variant SNPs")
    
    # For SNPs with less than 3 distinct data point, exclude them from downstream analysis
    uniqueCount = apply(M[,], 1, function(x){length(unique(x))})
    SNPtoProcessIndex = uniqueCount >= 3
    noVariantSNPIndex = uniqueCount < 3
    M = M[SNPtoProcessIndex, ]

    numSNP = nrow(M)
    numSample = ncol(M)

    calls = oligoClasses::initializeBigMatrix(name="calls", numSNP, numSample, vmode = "integer")
    scores = oligoClasses::initializeBigMatrix(name="scores", numSNP, numSample, vmode = "double")
    open(calls)
    open(scores)

    rownames(calls) = rownames(M)
    rownames(scores) = rownames(M)
    colnames(calls) = colnames(M)
    colnames(scores) = colnames(M)

    prepriormeans=calculatePriorcentersC(M,numSample)

    trueCalls=matrix(NA,nrow(M),ncol(M))

    for(i in 1:ncol(M)){
      tmp=kmeans(M[,i],centers=prepriormeans,nstart=45)
      trueCalls[,i]=tmp$cluster
    }
    colnames(trueCalls)=colnames(M)
    rownames(trueCalls)=rownames(M)
if(numSample<=30){
  priormeans=prepriormeans
}else{
  priormeans = calculatePriorValues(M, numSNP, verbose)
    if((abs(priormeans[1]-prepriormeans[1])+abs(priormeans[2]-prepriormeans[2])+abs(priormeans[3]-prepriormeans[3]))>=3){
priormeans=prepriormeans
}else{
      priormeans=priormeans
  }
 }
 
    
    VGLMparameters = calculateParameters(M, priormeans, numSNP, verbose)

### retrieve or calculate coefficients

    krlmmCoefficients = getKrlmmVGLMCoefficients(trueCalls=trueCalls,params=VGLMparameters,numSample=ncol(M),samplenames=colnames(M),verbose=T)
### do VGLM fit, to predict k for each SNP
    kPrediction <- predictKwithVGLM(VGLMparameters, krlmmCoefficients, verbose)
    rm(VGLMparameters)

### assign calls
    assignCalls(calls, M, kPrediction, priormeans, numSNP, numSample, verbose=T)
    rm(kPrediction)

### assign confidence scores
    computeCallPr(scores, M, calls, numSNP, numSample, verbose)

   YIndex = seq(1:nrow(cnSet))[chromosome(cnSet)==24 & isSnp(cnSet)]
    XIndex = seq(1:nrow(cnSet))[chromosome(cnSet)==23 & isSnp(cnSet)]
### impute gender if gender information not provided
    if (is.null(gender)) {
       if(sum(is.na(YIndex))==length(YIndex)){
		gender=NULL
       }else{
	gender = krlmmImputeGender(cnSet, annotation, YIndex, verbose, offset=offset)
       }
    }

### double-check ChrY ChrX SNP cluster, impute gender if gender information not provided
    if (!(is.null(gender))) {
        verifyChrYSNPs(cnSet, M, calls, gender, annotation, YIndex, priormeans, verbose)
        verifyChrXSNPs(cnSet, M, calls, gender, annotation, XIndex, priormeans, verbose)
    }

#    if (length(YIndex)>0  && !(is.null(gender))) {
#      verifyChrYSNPs(cnSet, M, calls, gender, annotation, YIndex, priormeans, verbose)
#    }
### add back SNPs excluded before
   AddBackNoVarianceSNPs(cnSet, calls, scores, numSNP, numSample, SNPtoProcessIndex, noVariantSNPIndex)

    close(calls)
    close(scores)
    rm(calls)
    rm(scores)
    rm(M)
    TRUE
}


#######################################################################################################

loessnormalization<- function(cnSet, verbose, offset=16, blockSize = 300000){
    # compute log-ratio in blocksize of 300,000 by default
    message("Start computing log-ratios")
    A <- A(cnSet)
    B <- B(cnSet)
    open(A)
    open(B)

    numSNP <- nrow(A)
    numSample <- ncol(A)
   

    M <- oligoClasses::initializeBigMatrix(name="M", numSNP, numSample, vmode = "double")
    S <- oligoClasses::initializeBigMatrix(name="S", numSNP, numSample, vmode = "double")

    rownames(M) = rownames(A)
    colnames(M) = colnames(A)

    numBlock = ceiling(numSNP / blockSize)
    for (i in 1:numBlock){
        if (verbose) message(" -- Processing segment ", i, " out of ", numBlock)
        subsetA = as.matrix(A[((i-1) * (blockSize) + 1):min(i * blockSize, numSNP), ])+offset
        subsetB = as.matrix(B[((i-1) * (blockSize) + 1):min(i * blockSize, numSNP), ])+offset
        subsetM = .Call("krlmmComputeM", subsetA, subsetB, PACKAGE="crlmm")
        M[((i-1) * (blockSize) + 1):min(i * blockSize, numSNP),] = subsetM[1:nrow(subsetM), ]
        subsetS = .Call("krlmmComputeS", subsetA, subsetB, PACKAGE="crlmm")
        S[((i-1) * (blockSize) + 1):min(i * blockSize, numSNP),] = subsetS[1:nrow(subsetS), ]

        rm(subsetA, subsetB, subsetM,subsetS)

    }
    close(A)
    close(B)
    gc()

    isna=is.na(M[,1])
    prepriormeansL=calculatePriorcentersC(M[,][!isna,],numSample)
    F01=abs(prepriormeansL[1])-prepriormeansL[2]   
    F02=abs(prepriormeansL[3])+prepriormeansL[2]

    for(i in 1:ncol(M)) {
        isna=is.na(M[,i])
        Msub = M[!isna,i]
        Mna=M[isna,i]
        Ssub = S[!isna,i]
        Sna=S[isna,i]

        km = kmeans(Msub, prepriormeansL)
        ind1v2 = km$cluster==1
        ind2v2 = km$cluster==2
        ind3v2 = km$cluster==3

        l1v2 = loessFit(Msub[ind1v2], Ssub[ind1v2])
        l2v2 = loessFit(Msub[ind2v2], Ssub[ind2v2])
        l3v2 = loessFit(Msub[ind3v2], Ssub[ind3v2])
        Msub[ind1v2] = l1v2$residuals+F01
        Msub[ind2v2] = l2v2$residuals
        Msub[ind3v2] = l3v2$residuals-F02
        Mnew=rbind(as.matrix(Msub),as.matrix(Mna))
        rownames(Mnew)=c(rownames(as.matrix(Msub)),rownames(as.matrix(Mna)))
        m=match(rownames(M),rownames(Mnew))
        Mnew=Mnew[m]
        M[,i] = Mnew
        rm(Msub, Ssub, ind1v2, ind2v2, ind3v2, l1v2, l2v2, l3v2)
    }
    return(M)
}


#######################################################################################################

quantilenormalization <- function(cnSet, verbose, offset=16, blockSize = 300000){
    A <- A(cnSet)
    B <- B(cnSet)
    open(A)
    open(B)

    numSNP <- nrow(A)
    numSample <- ncol(A)

    M <- oligoClasses::initializeBigMatrix(name="M", numSNP, numSample, vmode = "double")
    X <- oligoClasses::initializeBigMatrix(name="X", numSNP, numSample, vmode = "integer")
    Y <- oligoClasses::initializeBigMatrix(name="Y", numSNP, numSample, vmode = "integer")
    rownames(M) = rownames(X) = rownames(Y) = rownames(A)
    colnames(M) = colnames(X) = colnames(Y) = colnames(A)

    # This step may chew up a lot of memory...
    X =  normalize.quantiles(as.matrix(A[,]))+offset
    Y =  normalize.quantiles(as.matrix(B[,]))+offset

    numBlock = ceiling(numSNP / blockSize)
    for (i in 1:numBlock){
        if (verbose) message(" -- Processing segment ", i, " out of ", numBlock)
        subsetXqws = as.matrix(X[((i-1) * (blockSize) + 1):min(i * blockSize, numSNP), ])
        subsetYqws = as.matrix(Y[((i-1) * (blockSize) + 1):min(i * blockSize, numSNP), ])
        subsetM = .Call("krlmmComputeM", subsetXqws, subsetYqws, PACKAGE="crlmm")
        M[((i-1) * (blockSize) + 1):min(i * blockSize, numSNP),] = subsetM[1:nrow(subsetM), ]
        rm(subsetYqws, subsetXqws, subsetM)
    }

   close(A)
   close(B)
   return(M)
   delete(X)
   delete(Y)
   rm(X)
   rm(Y)
   gc()
}


#######################################################################################################

getNumberOfCores <- function(){
    defaultCores = min(detectCores(), 8)
    return(getOption("krlmm.cores", defaultCores))    
}


#######################################################################################################

computeLogRatio <- function(cnSet, verbose=FALSE, offset=0, blockSize = 500000, col=NULL, row=NULL){
    # compute log-ratio in blocksize of 500,000 by default
    if(verbose)
        message("Start computing log-ratios")
    if(!is.null(col) | !is.null(row)) {
      if(is.null(row)) {
       A <- A(cnSet)[,col]
       B <- B(cnSet)[,col]
      }
      if(is.null(col)) {
       A <- A(cnSet)[row,]
       B <- B(cnSet)[row,]
      }
      if(!is.null(col) & !is.null(row)) {
       A <- A(cnSet)[row,col]
       B <- B(cnSet)[row,col]
      }
    } else {
       A <- A(cnSet)
       B <- B(cnSet)
    }
    open(A)
    open(B)
    
    numSNP <- nrow(A)
    numSample <- ncol(A)
    
    M <- oligoClasses::initializeBigMatrix(name="M", numSNP, numSample, vmode = "double")
    rownames(M) = rownames(A)
    colnames(M) = colnames(A)
  
    numBlock = ceiling(numSNP / blockSize)
    for (i in 1:numBlock){
        if(verbose) message(" -- Processing segment ", i, " out of ", numBlock)
        subsetA = as.matrix(A[((i-1) * (blockSize) + 1):min(i * blockSize, numSNP), ])+offset
        subsetB = as.matrix(B[((i-1) * (blockSize) + 1):min(i * blockSize, numSNP), ])+offset              
        subsetM = .Call("krlmmComputeM", subsetA, subsetB, PACKAGE="crlmm")
        M[((i-1) * (blockSize) + 1):min(i * blockSize, numSNP),] = subsetM[1:nrow(subsetM), ]
        rm(subsetA, subsetB, subsetM)
    }
    
    close(A)
    close(B)
    if(verbose)
        message("Done computing log-ratios")
    return(M)    
}

computeAverageLogIntensity <- function(cnSet, verbose=FALSE, offset=0, blockSize = 500000, col=NULL, row=NULL){
    # compute average log intensity in blocksize of 500,000 by default
    if(verbose)
        message("Start computing average log-intensities")
    if(!is.null(col) | !is.null(row)) {
      if(is.null(row)) {
       A <- A(cnSet)[,col]
       B <- B(cnSet)[,col]
      }
      if(is.null(col)) {
       A <- A(cnSet)[row,]
       B <- B(cnSet)[row,]
      }
      if(!is.null(col) & !is.null(row)) {
       A <- A(cnSet)[row,col]
       B <- B(cnSet)[row,col]
      }
    } else {
       A <- A(cnSet)
       B <- B(cnSet)
    }
    open(A)
    open(B)
    
    numSNP <- nrow(A)
    numSample <- ncol(A)
    
    S <- oligoClasses::initializeBigMatrix(name="S", numSNP, numSample, vmode = "double")
    rownames(S) = rownames(A)
    colnames(S) = colnames(A)
      
    numBlock = ceiling(numSNP / blockSize)
    for (i in 1:numBlock){
        if(verbose) message(" -- Processing segment ", i, " out of ", numBlock)
        subsetA = as.matrix(A[((i-1) * (blockSize) + 1):min(i * blockSize, numSNP), ])+offset
        subsetB = as.matrix(B[((i-1) * (blockSize) + 1):min(i * blockSize, numSNP), ])+offset                     
        subsetS = .Call("krlmmComputeS", subsetA, subsetB, PACKAGE="crlmm")
        S[((i-1) * (blockSize) + 1):min(i * blockSize, numSNP),] = subsetS[1:nrow(subsetS), ]
        rm(subsetA, subsetB, subsetS)
    }
    
    close(A)
    close(B)
    if(verbose)
        message("Done computing average log-intensities")
    return(S)  
}


#######################################################################################################

calculatePriorValues <- function(M, numSNP, verbose) {

    calculateOneKmeans <- function(x) {
        tmp = kmeans(x, 3, nstart=45)
        return(sort(tmp$centers, decreasing = T))
    }

    if (verbose) message("Start calculating prior means")
    cl <- makeCluster(getNumberOfCores())
    centers <- parRapply(cl, M, calculateOneKmeans)
    stopCluster(cl) 
    centers <- matrix(centers, numSNP, 3, byrow = TRUE)
    priormeans = apply(centers, 2, FUN="median", na.rm=TRUE)
    if(abs(sum(priormeans))>1) {
      checksymmetric= apply(centers,1,function(x){abs(sum(x))})<1
      priormeans=apply(centers[checksymmetric,],2, FUN="median", na.rm=TRUE)
    }
    if (verbose) message("Done calculating prior means")
    return(priormeans)
}

calculatePriorcentersC <- function(M, numSample) {

    calculateOneKmeans <- function(x) {
        tmp = kmeans(x, 3, nstart=45)
        return(sort(tmp$centers, decreasing = T))
    }

    cl <- makeCluster(getNumberOfCores())
    centers <- parCapply(cl, M, calculateOneKmeans)
    stopCluster(cl) 
    centers <- matrix(centers, numSample, 3, byrow = TRUE)
    prepriormeans = apply(centers, 2, FUN="median", na.rm=TRUE)
    
    return(prepriormeans)
}


#######################################################################################################

calculateKrlmmCoefficients <- function(trueCalls, params, numSample, samplenames){
#    if (!require(VGAM)) {
#        message("VGAM package not found, fall back to use defined platform-specific coefficients")
#        return(NULL)
#    }
    amatch = match(rownames(params), rownames(trueCalls))
    amatch = amatch[!is.na(amatch)]
    trueCalls = trueCalls[amatch,]

    amatch = match(rownames(trueCalls), rownames(params))
    params = params[amatch, ]
    
    amatch = match(samplenames, colnames(trueCalls))
    amatch = amatch[!is.na(amatch)]    
    trueCalls = trueCalls[, amatch]
        
    if ((numSample <= 40) && (ncol(trueCalls) < round(numSample/2))){
        message("Sample size is ", numSample, ", KRLMM requires at least trueCalls of ", round(numSample/2), " samples to calculate coefficients")
        return(NULL)
    }
    if ((numSample > 40) && (ncol(trueCalls) < 20)){
        message("At least trueCalls of 20 samples are required to calculate coefficients")
        return(NULL)
    }
    if (nrow(trueCalls) < 1000){
        message("At lease trueCalls of 1000 SNPs are required to calculate coefficients")
        return(NULL)
    }

    getna = apply(trueCalls, 1, function(x){sum(is.na(x))>=10})
    truek = apply(trueCalls, 1, function(x){length(unique(x[!is.na(x)]))})
   
    params1 = params[!getna,]
    truek1 = truek[!getna]

    xx = data.frame(params1)
    t = as.factor(as.numeric(truek1)) 
    xx = data.frame(xx, t)
    fit = suppressWarnings(vglm(t~., multinomial(refLevel=1), xx))
    coe = coefficients(fit) # VGAM::
    return(coe)    
}

getKrlmmVGLMCoefficients <- function(pkgname, trueCalls, params, verbose, numSample, samplenames){
    if (!is.null(trueCalls)) {
        coe = calculateKrlmmCoefficients(trueCalls, params, numSample, samplenames)
        if (!is.null(coe)) {
            if (verbose) message ("Done calculating platform-specific coefficients to predict number of clusters")
            return(coe)
        }
    }
    if (!is.null(trueCalls)) message("Fall back to use defined platform-specific coefficients")
    if (verbose) message ("Retrieving defined platform-specific coefficients to predict number of clusters")
    loader("krlmmVGLMCoefficients.rda", .crlmmPkgEnv, pkgname)
    return(getVarInEnv("krlmmCoefficients"))      
}


predictKwithVGLM <- function(data, coe, verbose){
    if (verbose) message("Start predicting number of clusters")
    logit1 <- rep(0, nrow(data))
    logit2 <- coe[1]+coe[3]*data[,1]+coe[5]*data[,2]+coe[7]*data[,3]+coe[9]*data[,4]+coe[11]*data[,5]+coe[13]*data[,6]+coe[15]*data[,7]+coe[17]*data[,8]
    logit23 <- coe[2]+coe[4]*data[,1]+coe[6]*data[,2]+coe[8]*data[,3]+coe[10]*data[,4]+coe[12]*data[,5]+coe[14]*data[,6]+coe[16]*data[,7]+coe[18]*data[,8]
   
    logits <- cbind(logit1, logit2, logit23)
    rm(logit1)
    rm(logit2)
    rm(logit23)
    p.unscaled <- exp(logits)
    rm(logits)   
    p <- p.unscaled / rowSums(p.unscaled)
    clusterPrediction = apply(p, 1, function(x){which.max(x)})
    rm(p.unscaled)
    rm(p)
    if (verbose) message("Done predicting number of clusters")
    return(clusterPrediction)
}

#######################################################################################################

assignCalls3Cluster <- function(intensities, priormeans){
    prior31 = c(priormeans[1]/2,priormeans[2],priormeans[3])
    prior32 = c(priormeans[1],priormeans[2],priormeans[3]/2)	

    emp <- rep(NA, length(intensities))
	ansCalls <- emp
    tmp <- try(kmeans(intensities, priormeans, nstart=1),TRUE)
    if(class(tmp) == "try-error") {
        tmp1 <- try(kmeans(intensities, prior31, nstart=1),TRUE)
        if(class(tmp1) == "try-error"){
            tmp2 <- try(kmeans(intensities, prior32, nstart=1),TRUE)
            if(class(tmp2) == "try-error"){
                ansCalls = emp
            }else{
                ansCalls = tmp2$cluster
            }
        }else{
            ansCalls = tmp1$cluster
        }
    }else{
        ansCalls = tmp$cluster
    }
    rm(prior31, prior32)
    return(ansCalls)
 
}

#######################################################################################################

assignCalls2Cluster <- function(intensities, priormeans){
    closest <- rep(NA, 3)	
    for(j in 1:3){
        distance <- as.vector(abs(priormeans[j]-intensities))
        closest[j] <- intensities[which.min(distance)]
    }
    prior2 <- priormeans[priormeans!=priormeans[which.max(abs(closest-priormeans))]]
	
    emp <- rep(NA, length(intensities))
    ansCalls <- emp    
    tmp <- try(kmeans(intensities, prior2, nstart=1), TRUE)
    if(class(tmp) == "try-error") {
        ansCalls <- emp
    }else{
        if(prior2[1]==priormeans[2] && prior2[2]==priormeans[3]){
            mp <- abs(tmp$centers[1]-tmp$centers[2])
            pp <- abs(priormeans[2]-priormeans[3])
            if((mp/pp)<=0.25){
                ansCalls <- emp
            }else{
                ansCalls <- tmp$cluster+1
            }
        }
        if(prior2[1]==priormeans[1] && prior2[2]==priormeans[2]){
            mp=abs(tmp$centers[1]-tmp$centers[2])
            pp=abs(priormeans[1]-priormeans[2])
            if((mp/pp)<=0.25){
                ansCalls = emp
            }else{
                ansCalls <- tmp$cluster
            }
        }
        if(prior2[1]==priormeans[1] && prior2[2]==priormeans[3]){
            mp <- abs(tmp$centers[1]-tmp$centers[2])
            pp <- abs(priormeans[1]-priormeans[3])
            if ((mp/pp) <=0.25){
                ansCalls <- emp
            }else{
                ansCalls[tmp$cluster==1] <- 1
                ansCalls[tmp$cluster==2] <- 3
            }
        }
    }
    rm(tmp)
    return(ansCalls)
}

#######################################################################################################

assignCalls1Cluster <- function(intensities, priormeans){
    closest <- rep(NA, 3)
    for(j in 1:3){
        distance <- as.vector(abs(priormeans[j]-intensities))
        closest[j]=intensities[which.min(distance)]
    }
    clusterindex <- which.min(abs(closest-priormeans))
    return(rep(clusterindex, length(intensities)))
}
  
#######################################################################################################

assignCallsOneSNP <- function(x, priormeans, numSample){
    tolerance = 1e-3
       
    k = x[numSample + 1]
    values = x[1:numSample]  
    
    if (abs(k - 2) <= tolerance) {
        tmp <- try(kmeans(values, priormeans, nstart=1),TRUE)
        if (!(class(tmp)=="try-error")) {
            k <- 3;
        }
    }

    successful <- FALSE;
    if (abs(k - 3) <= tolerance){
        SNPCalls <- assignCalls3Cluster(values, priormeans)
        successful <- !is.na(SNPCalls[1])
        if (!successful) { 
            k <- 2
        }
    }
		
    if ( (abs(k - 2) <= tolerance) && (!successful)){
        SNPCalls <- assignCalls2Cluster(values, priormeans)
        successful <- !is.na(SNPCalls[1])
        if (!successful) { 
            k <- 1
        }			
    }

    if ( (abs(k - 1) <= tolerance) && (!successful)){
        SNPCalls <- assignCalls1Cluster(values, priormeans)
    }
    return(SNPCalls)
}


assignCalls <- function(callsGt, M, a, priormeans, numSNP, numSample, verbose, blockSize=500000){
    # process by block size of 500,000 by default
    message("Start assigning calls")
       
    numBlock = ceiling(numSNP / blockSize)

    for (i in 1:numBlock){
        if (verbose) message(" -- Processing segment ", i, " out of ", numBlock)
        subsetM = as.matrix(M[((i-1) * (blockSize) + 1):min(i * blockSize, numSNP), ])
        thisnumSNP = nrow(subsetM)
        subseta = a[((i-1) * (blockSize) + 1):min(i * blockSize, numSNP)]
        subseta = matrix(subseta, thisnumSNP, 1)

        testValues = cbind(subsetM, subseta)

        cl <- makeCluster(getNumberOfCores())
        callAns <- parRapply(cl, testValues, assignCallsOneSNP, priormeans, numSample)
        stopCluster(cl)

        callAnsAllValues = unlist(callAns)

        subsetcalls <- matrix(callAnsAllValues, thisnumSNP, numSample, byrow = TRUE)
        
        callsGt[((i-1) * (blockSize) + 1):min(i * blockSize, numSNP),] = subsetcalls[1:thisnumSNP, ]
        rm(subsetM, subseta, subsetcalls)
    }
    message("Done assigning calls")
}

#######################################################################################################


calculateParameters <- function(M, priormeans, numSNP, verbose) {
    if (verbose) message("Start calculating 3-cluster parameters")
    params3cluster <- calculateParameters3Cluster(M, priormeans, numSNP, verbose);
    if (verbose) message("Done calculating 3-cluster parameters")

    if (verbose) message("Start calculating 2-cluster parameters")
    params2cluster <- calculateParameters2Cluster(M, priormeans, numSNP, verbose);
    if (verbose) message("Done calculating 2-cluster parameters")

    if (verbose) message("Start calculating 1-cluster parameters")
    params1cluster <- calculateParameters1Cluster(M, priormeans, numSNP, verbose);    
    if (verbose) message("Done calculating 1-cluster parameters")

    parameters <- cbind(as.matrix(params1cluster$elemh), as.matrix(params1cluster$eless), as.matrix(params2cluster$elemh), as.matrix(params2cluster$eless),
                        as.matrix(params3cluster$elemh), as.matrix(params3cluster$eless), as.matrix(params2cluster$elehw), as.matrix(params3cluster$elehw));
    
    rownames(parameters) = rownames(M)
    rm(params3cluster)
    rm(params2cluster)
    rm(params1cluster)
    return(parameters)
}

hardyweinberg <- function(x, minN = 8){
   if (length(x) < minN){
       return(NA) 
    } else {
       result = .Call("krlmmHardyweinberg", x)
       return(result)
    }
}

calculateOneParams3Cluster <- function(x, priors, priorDown, priorUp){

    tmp = try(kmeans(x, priors, nstart=1), TRUE)
    if(class(tmp) == "kmeans") {
        flag = 1
     } else {
        tmp = try(kmeans(x, priorDown, nstart=1), TRUE)
        if(class(tmp) == "kmeans") {
            flag = 2
        } else {
            tmp = try(kmeans(x, priorUp, nstart=1), TRUE)
            if (class(tmp) == "kmeans") {
                flag = 3
            } else {
                tmp = kmeans(x, 3, nstart=35)
                flag = 4
            }
        }  
    }
    ss = sum(tmp$withinss)
    hw = hardyweinberg(tmp$cluster) 
    centers = sort(tmp$centers, decreasing = TRUE)

    return(c(centers, ss, hw, flag))
}

calculateParameters3Cluster <- function(M, priormeans, numSNP, verbose) {
    Apriors = priormeans
    ApriorDown = c(Apriors[1]/2, Apriors[2], Apriors[3]) # shift-down
    ApriorUp = c(Apriors[1], Apriors[2], Apriors[3]/2) # shift-up

    cl <- makeCluster(getNumberOfCores())
    clusterEvalQ(cl, library(crlmm))
    parAns <- parRapply(cl, M, calculateOneParams3Cluster, Apriors, ApriorDown, ApriorUp)

    stopCluster(cl)
    parAnsAllValues = unlist(parAns)
    parameters <- matrix(parAnsAllValues, numSNP, 6, byrow = TRUE)
   
    centers <- parameters[, 1:3]
    ss <- parameters[, 4]
    hw <- parameters[, 5]
    flag <- parameters[, 6]

    rm(parAns)
    rm(parameters)
   
    sigma=solve(cov(centers, use="complete.obs"))
    mh = calculateMahalDist3Cluster(centers, sigma, flag, Apriors, ApriorDown, ApriorUp, numSNP)

    rm(sigma)
    rm(centers)

    gc()
    return(list(eless = ss, elemh = mh, elehw = hw))
}


calculateMahalDist3Cluster <- function(centers, sigma, flag, priors, priorDown, priorUp, numSNP){
    mahaldist = rep(NA, numSNP)

    tolerance = 1e-3
    for (i in 1:numSNP) {
        if ((abs(flag[i] - 1) <= tolerance) || (abs(flag[i]- 4) <= tolerance)) difference = centers[i, ] - priors
        if (abs(flag[i] - 2) <= tolerance) difference = centers[i, ] - priorDown
        if (abs(flag[i] - 3) <= tolerance) difference = centers[i, ] - priorUp         
        mahaldist[i] = as.vector(difference)%*%sigma%*%as.vector(difference)
    }
    return(mahaldist)
}


calculateOneParams2Cluster <- function(x, priors){
    aa = rep(NA, 3)   
    for(j in 1:3){
        dist = as.vector(abs(priors[j]-x))
        aa[j]=x[which.min(dist)]
    } 
    prior2 = priors[priors!=priors[which.max(abs(aa-priors))]]
     

    tmp = try(kmeans(x, prior2, nstart=1), TRUE)
    if (class(tmp)=="kmeans") {
        centers = tmp$centers
        hw = hardyweinberg(tmp$cluster)
    }
    rm(tmp)
    tmp = kmeans(x, 2, nstart = 35)
    if (tmp$centers[1] < tmp$centers[2]){
        centers = c(tmp$centers[2], tmp$centers[1])
        hw = hardyweinberg(3 - tmp$cluster)   
    } else {
        centers = tmp$centers
        hw = hardyweinberg(tmp$cluster)
    }
    ss = sum(tmp$withinss)     
    return(c(centers, prior2, ss, hw))
}


calculateParameters2Cluster <- function(M, priormeans, numSNP, verbose) {
    Apriors = priormeans
   
    cl <- makeCluster(getNumberOfCores())
    clusterEvalQ(cl, library(crlmm))
    parAns <- parRapply(cl, M, calculateOneParams2Cluster, Apriors)
    stopCluster(cl)

    parAnsAllValues = unlist(parAns)
    parameters <- matrix(parAnsAllValues, numSNP, 6, byrow = TRUE)

    centers <- parameters[, 1:2]
    priors2cluster <- parameters[, 3:4]
    ss <- parameters[, 5]
    hw <- parameters[, 6]
    
    rm(parAns)
    rm(parameters)
    sigma=solve(cov(centers, use="complete.obs"))

    mh = calculateMahalDist2Cluster(centers, sigma, priors2cluster, numSNP)

    rm(sigma)
    rm(centers)

    gc()
    return(list(eless = ss, elemh = mh, elehw = hw))
}


calculateMahalDist2Cluster <- function(centers, sigma, priors, numSNP){
    mahaldist = rep(NA, numSNP)
    
    for (i in 1:numSNP) {
        difference <- centers[i,] - priors[i,]
        mahaldist[i] = as.vector(difference)%*%sigma%*%as.vector(difference)        
    }
    return(mahaldist)
}

calculateOneParams1Cluster <- function(x, priors){
    center = mean(x)
    diff <- x - center
    diffsquare <- diff^2
    ss = sum(diffsquare)

    closestPrior = priors[which.min(abs(priors - center))]

    return(c(center, closestPrior, ss))
}


calculateParameters1Cluster <- function(M, priormeans, numSNP, verbose) {
    Apriors = priormeans
   
    cl <- makeCluster(getNumberOfCores())
    clusterEvalQ(cl, library(crlmm))
    parAns <- parRapply(cl, M, calculateOneParams1Cluster, Apriors)
    stopCluster(cl)

    parAnsAllValues = unlist(parAns)
    parameters <- matrix(parAnsAllValues, numSNP, 3, byrow = TRUE)

    centers <- matrix(parameters[, 1], numSNP, 1)
    prior1cluster <- matrix(parameters[, 2], numSNP, 1)
    ss <- parameters[, 3]
    
    rm(parAns)
    rm(parameters)
    sigma=solve(cov(centers, use="complete.obs"))

    mh = calculateMahalDist1Cluster(centers, sigma, prior1cluster, numSNP)

    rm(sigma)
    rm(centers)

    gc()
    return(list(eless = ss, elemh = mh))
}


calculateMahalDist1Cluster <- function(centers, sigma, priors, numSNP){
    mahaldist = rep(NA, numSNP)
    
    for(i in 1:numSNP) {
        difference <- as.vector(centers[i, 1] - priors[i, 1])
	mahaldist[i]=difference%*%sigma%*%difference
    }
    return(mahaldist)
}
   
#############################################


computeCallPr <- function(callsPr, M, calls, numSNP, numSample, verbose, blockSize = 500000){
    # compute log-ratio in blocksize of 500,000 by default
    if (verbose) message("Start computing confidence scores")
    
    numBlock = ceiling(numSNP / blockSize)
    for (i in 1:numBlock){
        if (verbose) message(" -- Processing segment ", i, " out of ", numBlock)
        subsetM = as.matrix(M[((i-1) * (blockSize) + 1):min(i * blockSize, numSNP), ])
        subsetCalls = as.matrix(calls[((i-1) * (blockSize) + 1):min(i * blockSize, numSNP), ])                     
        subsetCallProb = .Call("krlmmConfidenceScore", subsetM, subsetCalls, PACKAGE="crlmm");
        callsPr[((i-1) * (blockSize) + 1):min(i * blockSize, numSNP),] = subsetCallProb[1:nrow(subsetM), ]
        rm(subsetM, subsetCalls, subsetCallProb)
    }

    if (verbose) message("Done computing confidence scores")
}

#############################################

AddBackNoVarianceSNPs <- function(cnSet, calls, scores, numSNP, numSample, variantSNPIndex, noVariantSNPIndex){
    callsGt = calls(cnSet)
    callsPr = snpCallProbability(cnSet)
    open(callsGt)
    open(callsPr)

    callsGt[variantSNPIndex, ] = calls[,]
    callsPr[variantSNPIndex, ] = scores[,]  
    
    callsGt[noVariantSNPIndex, ] = NA
    callsPr[noVariantSNPIndex, ] = 0     

    close(callsGt)
    close(callsPr)    
}

#############################################

krlmmImputeGender <- function(cnSet, annotation, YIndex, verbose, offset=0){
    if (verbose) message("Start imputing gender")    
    S = computeAverageLogIntensity(cnSet, verbose, offset=offset) 

    # S is calculated and saved in original SNP order. 
    matchy = match(rownames(annotation)[YIndex], rownames(S))
    matchy = matchy[!is.na(matchy)]
    if (length(matchy) <= 10){
        predictedGender = rep(NA, ncol(S))
    }
    Sy = S[matchy,]

    uniqueDataPoint = apply(Sy, 1, function(x){length(unique(x))})
    validYSNPs = uniqueDataPoint >= 2
    SyValid = Sy[validYSNPs, ]
    
    if (nrow(SyValid) < 20){
        message("Not enough ChrY SNPs, skipping gender prediction step");
        predictedGender = rep(NA, ncol(Sy))
    }
   
    rm(S)
    numYChrSNP = nrow(SyValid)

    allS = matrix(NA, numYChrSNP, 2)

    for (i in 1:numYChrSNP) {
        tmp = kmeans(SyValid[i,] ,2, nstart=45)
        allS[i,] = sort(tmp$centers, decreasing=F)
    }
    priorS = apply(allS, 2, FUN="median", na.rm=TRUE)

    if (abs(priorS[1] - priorS[2]) <= 1.6) {
        message("Separation between clusters too small (samples probabaly all the same gender): ");
      
    }
    
    meanmatrix = apply(Sy, 2, median)

    Sy1 = abs(meanmatrix - priorS[1])
    Sy2 = abs(meanmatrix - priorS[2])

    # output male - 1, female - 2, female S-values are smaller than male S-values. 
    test = cbind(Sy2, Sy1)
    predictedGender = apply(test, 1, which.min)

    open(cnSet$gender)
    cnSet$gender[,] = predictedGender
    close(cnSet$gender)
    
    if (verbose) message("Done imputing gender")
    return(predictedGender)   
}


#############################################

verifyChrYSNPs <- function(cnSet, M, calls, gender, annotation, YIndex, priormeans, verbose){
    if (verbose) message("Start verifying SNPs on Chromosome Y")
    callsGt = calls(cnSet)
    callsPr = snpCallProbability(cnSet)
    open(callsGt)
    open(callsPr)
       
    matchy = match(rownames(annotation)[YIndex], rownames(M))
    matchy = matchy[!is.na(matchy)]
   
    MChrY = M[matchy,]
    callsChrY = calls[matchy,]
  
    male = gender == 1
    female = gender == 2    
    
    checkK = apply(callsChrY[, male], 1, function(x){ length(unique(x[!is.na(x)])) } )
    
    for(i in 1:nrow(MChrY)){
        # Chromosome Y SNPs, no calls for female, k = 1 or 2 permitted for male samples
        thisChrYSNPorigPosition = match(rownames(callsChrY)[i], rownames(callsGt))
        callsGt[thisChrYSNPorigPosition, female] = NA
        callsPr[thisChrYSNPorigPosition, female] = 0

        # early exit for k = 1 or 2 cases. Only re-assign calls to male samples if we previouly assigned
        # male samples to 3 clusters by mistake. 
        if (checkK[i] < 3) next;
          
        if (class(try(kmeans(MChrY[i, male], c(priormeans[1], priormeans[3]), nstart=1), TRUE)) != "try-error"){
           
            maleSampleCalls = kmeans(MChrY[i, male],c(priormeans[1], priormeans[3]), nstart=45)$cluster
            callsGt[thisChrYSNPorigPosition, male][maleSampleCalls == 1] = 1
            callsGt[thisChrYSNPorigPosition, male][maleSampleCalls == 2] = 3
        } else {
                    
            distanceToPrior1 = mean(abs(MChrY[i, male] - priormeans[1]))
            distanceToPrior3 = mean(abs(MChrY[i, male] - priormeans[3]))                
            callsGt[thisChrYSNPorigPosition, male] = ifelse(distanceToPrior1 < distanceToPrior3, 1, 3)
        }

        MMaleSamples = MChrY[i, male]
        callsMaleSample = callsGt[thisChrYSNPorigPosition, male]
        scoresMaleSample = .Call("krlmmConfidenceScore", t(as.matrix(MMaleSamples)), t(as.matrix(callsMaleSample)), PACKAGE="crlmm");

        callsPr[thisChrYSNPorigPosition, male] = scoresMaleSample       
    }

    close(callsGt)
    close(callsPr)
    
    if (verbose) message("Done verifying SNPs on Chromosome Y")
}




#######################################
verifyChrXSNPs <- function(cnSet, M, calls, gender, annotation, XIndex, priormeans, verbose){
    if (verbose) message("Start verifying SNPs on Chromosome X")
    callsGt = calls(cnSet)
    callsPr = snpCallProbability(cnSet)
    open(callsGt)
    open(callsPr)
      
    matchx = match(rownames(annotation)[XIndex], rownames(M))
    matchx = matchx[!is.na(matchx)]
   
    MChrX= M[matchx,]
    callsChrX = callsGt[matchx,]
  
    male = gender == 1
    female = gender == 2    
    
    checkK = apply(callsChrX[, male], 1, function(x){ length(unique(x[!is.na(x)])) } )
    
    for(i in 1:nrow(MChrX)){
      
        # Chromosome X SNPs for male, k = 1 or 2 permitted for male samples
thisChrXSNPorigPosition = match(rownames(callsChrX)[i], rownames(callsGt))
        # early exit for k = 1 or 2 cases. Only re-assign calls to male samples if we previouly assigned
        # male samples to 3 clusters by mistake. 
        if (checkK[i] < 3) next;
          
        if (class(try(kmeans(MChrX[i, male], c(priormeans[1], priormeans[3]), nstart=1), TRUE)) != "try-error"){
           
            maleSampleCalls = kmeans(MChrX[i, male],c(priormeans[1], priormeans[3]), nstart=45)$cluster
            callsGt[thisChrXSNPorigPosition, male][maleSampleCalls == 1] = 1
            callsGt[thisChrXSNPorigPosition, male][maleSampleCalls == 2] = 3
        } else {
                    
            distanceToPrior1 = mean(abs(MChrX[i, male] - priormeans[1]))
            distanceToPrior3 = mean(abs(MChrX[i, male] - priormeans[3]))                
            callsGt[thisChrXSNPorigPosition, male] = ifelse(distanceToPrior1 < distanceToPrior3, 1, 3)
        }

        MMaleSamples = MChrX[i, male]
        callsMaleSample = callsGt[thisChrXSNPorigPosition, male]
        scoresMaleSample = .Call("krlmmConfidenceScore", t(as.matrix(MMaleSamples)), t(as.matrix(callsMaleSample)), PACKAGE="crlmm");

       callsPr[thisChrXSNPorigPosition, male] = scoresMaleSample       
    }

    close(callsGt)
    close(callsPr)
    
    if (verbose) message("Done verifying SNPs on Chromosome X")
}

%\VignetteIndexEntry{crlmm Vignette - Genotyping}
%\VignetteDepends{crlmm, hapmapsnp6, genomewidesnp6Crlmm}
%\VignetteKeywords{genotype, crlmm, SNP 5, SNP 6}
%\VignettePackage{crlmm}
\documentclass{article}

\newcommand{\Rfunction}[1]{{\texttt{#1}}}
\newcommand{\Rmethod}[1]{{\texttt{#1}}}
\newcommand{\Rcode}[1]{{\texttt{#1}}}
\newcommand{\Robject}[1]{{\texttt{#1}}}
\newcommand{\Rpackage}[1]{{\textsf{#1}}}
\newcommand{\Rclass}[1]{{\textit{#1}}}
\newcommand{\oligo}{\Rpackage{oligo }}

\begin{document}
\title{Genotyping with the \Rpackage{crlmm} Package}
\date{March, 2009}
\author{Benilton Carvalho}
\maketitle

<<setup, echo=FALSE, results=hide>>=
options(width=60)
options(continue=" ")
options(prompt="R> ")
@ 

\section{Quick intro to \Rpackage{crlmm}}

The \Rpackage{crlmm} package contains a new implementation for the
CRLMM algorithm (Carvalho et. al. 2007). Our focus is on efficient
genotyping of SNP 5.0 and 6.0 Affymetrix arrays, although extensions
of the method are under development for similar platforms.

This implementation, compared to the previous one (in
\Rpackage{oligo}), offers improved confidence scores, quality scores
for SNP's and batches, higher accuracy on different datasets and
better performance.

Additionally, this package does not use the pd.genomewidesnp packages
created via pdInfoBuilder for \Rpackage{oligo}. Instead, it uses
different annotation packages (\Rpackage{genomewidesnp.5} and
\Rpackage{genomewidesnp.6}), which use simple R objects to store only
the information needed for genotyping. This allowed us to improve the
speed of the method, as SQL queries are no longer performed here.

It is also our priority to make the package simple to use. Below we
demonstrate how to get genotype calls with the 'new' CRLMM. We use 3
samples on SNP 5.0 made available via the \Rpackage{hapmapsnp5}
package.

<<crlmm>>=
require(oligoClasses)
library(crlmm)
library(hapmapsnp6)
path <- system.file("celFiles", package="hapmapsnp6")
celFiles <- list.celfiles(path, full.names=TRUE)
system.time(crlmmResult <- crlmm(celFiles, verbose=FALSE))
@ 

The \Robject{crlmmResult} is a \Rclass{SnpSet} (see Biobase) object.
\begin{itemize}
\item \Robject{calls}: genotype calls (1 - AA; 2 - AB; 3 - BB);
\item \Robject{confs}: confidence scores, which can be translated to probabilities by using:
  \[ 1-2^-(\mbox{confs}/1000), \] although we prefer this
  representation as it saves a significant amount of memory;
\item \Robject{SNPQC}: SNP quality score;
%%\item \Robject{batchQC}: Batch quality score;
\item \Robject{SNR}: Signal-to-noise ratio.
\end{itemize}

<<out>>=
calls(crlmmResult)[1:10,]
confs(crlmmResult)[1:10,]
crlmmResult[["SNR"]]
@ 

\section{Details}

This document was written using:

<<>>=
sessionInfo()
@ 


\end{document}

%\VignetteIndexEntry{IlluminaPreprocessCN Vignette for Illumina}
%\VignetteDepends{crlmm, ff, cacheSweave}
%\VignetteKeywords{crlmm, illumina, copy number, SNP}
%\VignettePackage{crlmm}
\documentclass{article}
\usepackage{graphicx}
\usepackage{natbib}
\usepackage{amsmath}
\usepackage{url}
\usepackage[margin=1in]{geometry}
\newcommand{\crlmm}{\Rpackage{crlmm}}
\newcommand{\Rfunction}[1]{{\texttt{#1}}}
\newcommand{\Rmethod}[1]{{\texttt{#1}}}
\newcommand{\Rcode}[1]{{\texttt{#1}}}
\newcommand{\Robject}[1]{{\texttt{#1}}}
\newcommand{\Rpackage}[1]{{\textsf{#1}}}
\newcommand{\Rclass}[1]{{\textit{#1}}}
\newcommand{\oligo}{\Rpackage{oligo }}
\newcommand{\R}{\textsf{R}}
\newcommand{\ff}{\Rpackage{ff}}

\begin{document}
\title{Preprocessing and Genotyping Illumina Arrays for Copy Number Analysis}

\date{\today}

\author{Rob Scharpf}
\maketitle


\begin{abstract}

  This vignette illustrates the steps required prior to copy number
  analysis for Infinium platforms.  Specifically, we require
  construction of a container to store processed forms of the raw
  data, preprocessing to normalize the arrays, and genotyping using
  the CRLMM algorithm.  After completing these steps, users can refer
  to the \verb+copynumber+ vignette.

\end{abstract}


\section{Set up}

<<crlmm, results=hide, echo=FALSE>>=
library(Biobase)
library(crlmm)
options(width=70)
options(continue=" ")
@

%\textbf{Supported platforms:} The supported Infinium platforms are
%those for which a corresponding annotation package is available.  The
%annotation packages contain information on the markers, such as
%physical position and chromosome, as well as pre-computed parameters
%estimated from HapMap used during the preprocessing and genotyping
%steps. Currently supported Infinium platforms are listed in the
%following code chunk.
The following codechunk declares a directory for saving \Robject{ff}
files that will contain the normalized intensities and the genotype
calls.

<<ldpath,results=hide>>=
library(ff)
options(ffcaching="ffeachflush")
outdir <- paste("/local_data/r00/crlmm/", getRversion(), "/illumina_vignette", sep="")
ldPath(outdir)
dir.create(outdir, recursive=TRUE, showWarnings=FALSE)
@

We will also store cached computations in the directory \verb+outdir+.

<<cacheSweave, echo=FALSE, results=hide>>=
library(cacheSweave)
setCacheDir(outdir)
@

We declare that \crlmm{} should process 150,000 markers at a time
and/or 500 samples at a time (when possible) to reduce the memory
footprint.  As our example dataset in this vignette contains fewer
than 500 samples, all samples will be processed simultaneously.

<<ram>>=
ocProbesets(150e3)
ocSamples(500)
@

%\section{Initializing a container for storing processed data}
%
%This section will initialize a container for storing processed forms
%of the data, including the normalized intensities for the A and B
%alleles and the CRLMM genotype calls and confidence scores.  In
%addition, the container will store information on the markers
%(physical position, chromosome, and a SNP indicator), the batch, and
%the samples (e.g., gender).  To construct this container for Infinium
%platforms, several steps are required.

We begin by specifying the path containing the raw IDAT files for a
set of samples from the Infinium 370k platform.

<<datadir>>=
datadir <- "/thumper/ctsa/snpmicroarray/illumina/IDATS/370k"
@

For Infinium platforms, an Illumina sample sheet containing
information for reading the raw IDAT files is required. Please refer
to the BeadStudio Genotyping guide, Appendix A, for additional
information.  The following code reads in the samplesheet for the IDAT
files on our local server.

<<samplesheet>>=
samplesheet = read.csv(file.path(datadir, "HumanHap370Duo_Sample_Map.csv"), header=TRUE, as.is=TRUE)
@

For the purposes of this vignette, we indicate that we only wish to
process a subset of the arrays.  For our dataset, the file extensions
are `Grn.dat' and `Red.idat'.  We store the complete path to the
filename without the file extension in the \Robject{arrayNames} and
check that all of the green and red IDAT files exists.

<<subsetArrays>>=
samplesheet <- samplesheet[-c(28:46,61:75,78:79), ]
arrayNames <- file.path(datadir, unique(samplesheet[, "SentrixPosition"]))
all(file.exists(paste(arrayNames, "_Grn.idat", sep="")))
all(file.exists(paste(arrayNames, "_Red.idat", sep="")))
arrayInfo <- list(barcode=NULL, position="SentrixPosition")
@

All supported platforms have a corresponding annotation package.  The
appropriate annotation package is specified by the platform identifier
without the \verb+Crlmm+ postfix.

<<cdfname>>=
cdfName <- "human370v1c"
@

Next, we construct a character vector that specifies the batch for
each of the \Sexpr{length(arrayNames)} arrays.  Here, we have a small
dataset and process the samples in a single batch. Processing the
samples as a single batch is generally reasonable if the samples were
processed at similar times (e.g., within a few weeks).

<<batch>>=
batch <- rep("1", nrow(samplesheet))
@

%Finally, we initialize an object of class \Robject{CNSet} using the
%function \Rfunction{constructInf}.
%
%<<container,cache=TRUE>>=
%cnSet <- constructInf(sampleSheet=samplesheet,
%		      arrayNames=arrayNames,
%		      batch=batch,
%		      arrayInfoColNames=arrayInfo,
%		      cdfName=cdfName,
%		      verbose=TRUE,
%		      saveDate=TRUE)
%@
%
%A concise summary of the object's contents can be viewed with the
%\Rfunction{print} function.
%
%<<cnset>>=
%print(cnSet)
%@
%
%Note that the above object does not yet contain any processed data
%(only \verb+NA+'s).  As the elements of the \verb+assayData+ slot are
%\Robject{ff} objects (not matrices), several \verb+.ff+ files now
%appear in the \verb+outdir+. The \verb+.ff+ files should not be
%removed and can be listed using the \R{} function
%\Rfunction{list.files}.  %For the most part, the \emph{appearance} that
%the data is stored in memory is preserved.
%
%<<listff>>=
%sapply(assayData(cnSet), function(x) class(x)[1])
%list.files(outdir, pattern=".ff")[1:5]
%@
%
\section{Preprocessing and genotyping}

The raw intensities from the Infinium IDAT files are read and
normalized using the function \Rfunction{preprocessInf}.  The function
\Rfunction{preprocessInf} returns a \Robject{ff} object containing the
parameters for the mixture model used by the CRLMM genotyping
algorithm.  Following preprocessing, the \Rfunction{genotypeInf}
genotypes the samples. The function \Rfunction{genotype.Illumina} is a
wrapper to the above functions and returns an object of class
\Rclass{CNSet}.

%<<preprocess,cache=TRUE, results=hide>>=
%mixtureParams <- preprocessInf(cnSet=cnSet, sampleSheet=samplesheet,
%			       arrayNames=arrayNames,
%			       arrayInfoColNames=arrayInfo,
%			       cdfName=cdfName)
%@
%
%<<showMixtureParams>>=
%invisible(open(mixtureParams))
%str(mixtureParams[])
%invisible(close(mixtureParams))
%@
%
%Note that the normalized intensities for the A and B alleles are no
%longer \verb+NA+s and can be inspected using the methods \Rfunction{A}
%and \Rfunction{B}, respectively.
%
%<<intensities>>=
%invisible(open(A(cnSet)))
%invisible(open(B(cnSet)))
%as.matrix(A(cnSet)[1:5, 1:5])
%as.matrix(B(cnSet)[1:5, 1:5])
%invisible(close(A(cnSet)))
%invisible(close(B(cnSet)))
%@
%
%\section{Genotyping}
%
%CRLMM genotype calls and confidence scores are estimated using the
%function \Rfunction{genotypeInf}.
%
%<<genotype,cache=TRUE>>=
%updated <- genotypeInf(cnSet, mixtureParams=mixtureParams, cdfName=cdfName)
%@
%\vspace{-0.5em}
%<<>>=
%updated
%@
%
%\textbf{Wrapper:} As an alternative to calling the functions
%\Rfunction{constructInf}, \Rfunction{preprocessInf} and
%\Rfunction{genotypeInf} in sequence, a convenience function called
%\Rfunction{genotype.Illumina} is a wrapper for the above functions and
%produces identical results.

<<genotype.Illumina,cache=TRUE,results=hide>>=
cnSet <- genotype.Illumina(sampleSheet=samplesheet,
			   arrayNames=arrayNames,
			   arrayInfoColNames=arrayInfo,
			   cdfName="human370v1c",
			   batch=batch)
@


Note, to fully remove the data associated with the \Robject{cnSet2}
object, one should use the \Rfunction{delete} function in the
\Rpackage{ff} package followed by the \Rfunction{rm} function.  The
following code is not evaluated is it would change the results of the
cached computations in the previous code chunk.

<<delete, eval=FALSE>>=
lapply(assayData(cnSet), delete)
lapply(batchStatistics(cnSet), delete)
delete(cnSet$gender)
delete(cnSet$SNR)
delete(cnSet$SKW)
rm(cnSet)
@

%\section{Copy number estimation}

\SweaveInput{copynumber}

\section{Session information}
<<sessionInfo, results=tex>>=
toLatex(sessionInfo())
@

\begin{figure}[f]
  \begin{center}
  \includegraphics[width=0.6\textwidth]{IlluminaPreprocessCN-snr.pdf}
  \caption{The signal to noise ratio (SNR) for 180 HapMap samples. For
    Affymetrix platforms, SNR values below 5 can indicate possible
    problems with sample quality.  In some circumstances, it may be
    more helpful to exclude samples with poor DNA quality.}
\end{center}
\end{figure}


\bibliography{refs}
\bibliographystyle{plain}

\end{document}

  Copy number routines in the \crlmm{} package are available for
  Affymetrix 5.0 and 6.0 platforms, as well as several Illumina
  platforms.  This vignette assumes that the arrays have already been
  successfully preprocessed and genotyped as per the instructions in
  the \verb+AffyGW+ and \verb+IlluminaPreprocessCN+
  vignettes for the Affymetrix and Illumina platforms,
  respectively. While this vignette uses Affymetrix 6.0 arrays for
  illustration, the steps at this point are identical for both
  platforms.  See \citep{Scharpf2011} for details regarding the
  methodology implemented in \crlmm{} for copy number analysis.  In
  addition, a compendium describing copy number analysis using the
  \crlmm{} package is available from the author's website:
  \url{http://www.biostat.jhsph.edu/~rscharpf/crlmmCompendium/index.html}.


\textbf{Limitations:} While a minimum number of samples is not
required for preprocessing and genotyping, copy number estimation in
the \crlmm{} package currently requires at least 10 samples per batch.
The parameter estimates for copy number and the corresponding
estimates of raw copy number will tend to be more noisy for batches
with small sample sizes (e.g., $<$ 50).  Chemistry plate or scan date
are often useful surrogates for batch.  Samples that were processed at
similar times (e.g., in the same month) can be grouped together in the
same batch.

\section{Quality control}

The signal to noise ratio (SNR) estimated by the CRLMM genotyping
algorithm is an overall measure of the separation of the diallelic
genotype clusters at polymorphic loci and can be a useful measure of
array quality.  Small SNR values can indicate possible problems with
the DNA.  Depending on the size of the dataset and the number of
samples with low SNR, users may wish to rerun the preprocessing and
genotyping steps after excluding samples with low SNR.  The SNR is
stored in the \verb+phenoData+ slot of the \Robject{CNSet} object and
is available after preprocessing and genotyping. SNR values below 5
for Affymetrix or below 25 for Illumina may indicate poor sample
quality.  The following code chunk makes a histogram of the SNR values
for the HapMap samples.

<<snr,fig=TRUE,include=FALSE,width=6, height=4>>=
library(lattice)
invisible(open(cnSet$SNR))
snr <- cnSet$SNR[]
close(cnSet$SNR)
print(histogram(~snr,
		panel=function(...){
			panel.histogram(...)},
		breaks=25, xlim=range(snr), xlab="SNR"))
@



\section{Copy number estimation}

There are two ways to obtain marker-level estimates of copy number
that are supported by \Rpackage{crlmm}. One approach is to fit a
linear model to the normalized intensities stratified by the diallelic
genotype call at each SNP, as described in \cite{Scharpf2011}.
Another alternative is to compute log R ratio and B allele
frequencies. The latter is often better supported by downstream hidden
Markov models such as those in the \Rpackage{VanillaICE} package. We
describe each approach in the following two sections.

\subsection{Linear model for normalized intensities}

In this section, we fit a linear model to the normalized intensities
stratified by the diallic genotype call.  The intercept and slope from
the linear model are both SNP- and batch-specific.  The implementation
in the \crlmm{} package is encapsulated by the function
\Rfunction{crlmmCopynumber} that, using the default settings, can be
called by passing a single object of class \Rclass{CNSet}.

<<LDS_copynumber,cache=TRUE>>=
crlmmCopynumber(cnSet)
@

The following steps were performed by the \Rfunction{crlmmCopynumber}
function:
\begin{itemize}
\item sufficient statistics for the genotype clusters for
  each batch
\item unobserved genotype centers imputed
\item posterior summaries of sufficient statistics
\item intercept and slope for linear model
\end{itemize}
Depending on the value of \verb+ocProbesets()+, these summaries are
computed for subsets of the markers to reduce the required RAM. Note
that the value returned by the \Rfunction{crlmmCopynumber} function in
the above example is \verb+TRUE+.  The reason the function returns
\verb+TRUE+ in the above example is that the elements of the
\verb+batchStatistics+ slot have the class \Rclass{ff\_matrix}.
Rather than keep the statistical summaries in memory, the summaries
are written to files on disk using protocols described in the
\Rpackage{ff} package. Hence, while the \Robject{cnSet} object itself
is unchanged as a result of the \Rfunction{crlmmCopynumber} function,
the data on disk is updated accordingly.  Users that are interested in
accessing these low-level summaries can refer to the
\verb+Infrastructure+ vignette.  Note that the data structure depends
on whether the elements of the \verb+batchStatistics+ slot are
\Robject{ff} objects or ordinary matrices.  In this example, the
elements of \verb+batchStatistics+ have the class \Rclass{ff\_matrix}.

<<classes>>=
nms <- ls(batchStatistics(cnSet))
cls <- rep(NA, length(nms))
for(i in seq_along(nms)) cls[i] <- class(batchStatistics(cnSet)[[nms[i]]])[1]
all(cls == "ff_matrix")
@

The batch-specific statistical summaries computed by
\Robject{crlmmCopynumber} are written to files on disk using protocols
described in the \R{} package \Rpackage{ff}. The value returned by
\Rfunction{crlmmCopynumber} is \verb+TRUE+, indicating that the files
on disk have been successfully updated.  Note that while the
\Robject{cnSet} object is unchanged, the values on disk are different.


<<chr1index>>=
chr1.index <- which(chromosome(cnSet) == 1)
open(cnSet)
@

<<subset,cache=TRUE>>=
cnSet2 <- cnSet[chr1.index, ]
@

<<valuematrix>>=
close(cnSet)
for(i in seq_along(nms)) cls[i] <- class(batchStatistics(cnSet2)[[nms[i]]])[1]
all(cls == "matrix")
@

<<clean, echo=FALSE, results=hide>>=
rm(cnSet2); gc()
@

\paragraph{Raw total copy number.}
Several functions are available that will compute relatively quickly
the allele-specific, \emph{raw} copy number estimates. At allele $k$,
marker $i$, sample $j$, and batch $p$, the estimate of allele-specific
copy number is computed by subtracting the estimated background from
the normalized intensity and scaling by the slope coefficient. More
formally, \newcommand{\A}{A} \newcommand{\B}{B}
\begin{eqnarray}
  \label{eq:cnK}
{\hat c}_{k,ijp} &=& \mbox{max}\left\{\frac{1}{{\hat
    \phi}_{k,ip}}\left(I_{k,ijp}-{\hat \nu}_{k,ip}\right), ~0\right\}
\mbox{~for~} k \in \{\A, \B\}.
\end{eqnarray}
\noindent See \cite{Scharpf2011} for details.

The function \Rfunction{totalCopynumber} translates the normalized
intensities to an estimate of raw copy number by adding the
allele-specific summaries in Equation \eqref{eq:cnK}. For large
datasets, the calculation will not be instantaneous as the I/O can be
substantial.  Users should specify either a subset of the markers or a
subset of the samples to avoid using all of the available RAM.  For
example, in the following code chunk we compute the total copy number
at all markers for the first 2 samples, and the total copy number for
chromosome 20 for the first 50 samples.

<<totalCopynumber>>=
tmp <- totalCopynumber(cnSet, i=seq_len(nrow(cnSet)), j=1:2)
dim(tmp)
tmp2 <- totalCopynumber(cnSet, i=which(chromosome(cnSet) == 20), j=seq_len(ncol(cnSet)))
dim(tmp2)
@

Alternatively, the functions \Rfunction{CA} and \Rfunction{CB} compute
the allele-specific copy number.  For instance, the following code
chunk computes the allele-specific summaries at all polymorphic loci
for the first 2 samples.

<<ca>>=
snp.index <- which(isSnp(cnSet) & !is.na(chromosome(cnSet)))
ca <- CA(cnSet, i=snp.index, j=1:2)
cb <- CB(cnSet, i=snp.index, j=1:2)
@


\subsection{Container for log R ratios and B allele frequencies}
\label{sec:lrrBaf}

A useful container for storing the \crlmm{} genotypes, genotype
confidence scores, and the total or relative copy number at each
marker is the \Rclass{BafLrrSetList} class.  Coercion of a
\Rclass{CNSet} object to a \Rfunction{BafLrrSetList} object can be
acheived by the function \Rfunction{BafLrrSetList} as
illustrated below. Users should note that if the \verb+assayData+
elements in the \Rclass{CNSet} instance are \Rpackage{ff} objects, the
\verb+assayData+ elements of each element in the \Rclass{BafLrrSetList}
object will be \Rpackage{ff}-dervied objects (a new
\verb+total_cn*.ff+ file will be created in the \verb+ldPath()+
directory). The following code-chunk is not evalutated.

Estimation of log R ratios and B allele frequencies from an object of class \Rclass{CNSet} instantiated from genotyping Affymetrix CEL files or Illumina iDat files requires running the \R{} function \Rfunction{crlmmCopynumber} as a preliminary step.  Specifying \texttt{fit.linearModel=FALSE} will speed up computation as this function will by default compute summary statistics unnecessary for computing BAFs and log R ratios.

<<callingCrlmmCopynumber,eval=FALSE>>=
crlmmCopynumber(cnSet, fit.linearModel=FALSE)
@

<<oligoSnpSet>>=
open(cnSet)
oligoSetList <- BafLrrSetList(cnSet)
close(cnSet)
show(oligoSetList)
class(oligoSetList)
## oligoSnpSet of first chromosome
oligoSetList[[1]]
@

\noindent Log R ratios and B allele frequences can be retrieved by the
accessors \Rfunction{lrr} and \Rfunction{baf}, respectively.

<<testEqual>>=
lrrList <- lrr(oligoSetList)
class(lrrList)
dim(lrrList[[1]]) ## log R ratios for chromosome 1.
bafList <- baf(oligoSetList)
dim(bafList[[1]]) ## B allele frequencies for chromosome 1
@

If the \Rfunction{crlmmCopynumber} function is not run prior to the
\Rfunction{BafLrrSetList} construction, the log R ratios and BAFs will
be initialized as NAs.

%\VignetteIndexEntry{Preprocessing and genotyping Affymetrix arrays
%for copy number analysis}
%\VignetteDepends{crlmm, genomewidesnp6Crlmm, cacheSweave, ff}
%\VignetteKeywords{crlmm, SNP 6, copy number, SNP}
%\VignettePackage{crlmm}
\documentclass{article}
\usepackage{graphicx}
\usepackage{natbib}
\usepackage{amsmath}
\usepackage{url}
\newcommand{\Rfunction}[1]{{\texttt{#1}}}
\newcommand{\Rmethod}[1]{{\texttt{#1}}}
\newcommand{\Rcode}[1]{{\texttt{#1}}}
\newcommand{\Robject}[1]{{\texttt{#1}}}
\newcommand{\Rpackage}[1]{{\textsf{#1}}}
\newcommand{\Rclass}[1]{{\textit{#1}}}
\newcommand{\oligo}{\Rpackage{oligo }}
\newcommand{\R}{\textsf{R}}
\newcommand{\crlmm}{\Rpackage{crlmm}}
\usepackage[margin=1in]{geometry}

\begin{document}
\title{Preprocessing \& Genotyping Affymetrix Arrays for Copy Number Analysis}
\date{\today}
\author{Rob Scharpf}
\maketitle

<<setup, echo=FALSE, results=hide>>=
options(continue=" ", width=70)
@

%\section{Estimating copy number}

%At present, software for copy number estimation is provided only for the
%Affymetrix 6.0 platform.

\begin{abstract}

  This vignette describes the setup needed to analyze Affymetrix 6.0
  (or 5.0) CEL files and the steps for preprocessing and
  genotyping. These steps must be completed prior to copy number
  analyses in \crlmm{}.  After completing these steps, users can refer
  to the \verb+copynumber+ vignette.

\end{abstract}

\section{Set up}

<<cdfname, results=hide>>=
library(oligoClasses)
library2(crlmm)
library2(ff)
if(!exists("useCache")) useCache <- TRUE
if(useCache) library2(cacheSweave)
@

This vignette analyzes HapMap samples assayed on the Affymetrix 6.0
platform.  The annotation package for this platform is
\Rpackage{genomewidesnp6Crlmm}.  We assign the name of the annotation
package without the \verb+Crlmm+ postfix to the name \verb+cdfName+.
We use the \R{} package \Rpackage{cacheSweave} to cache long
computations in this vignette.  Users should refer to the
\Rpackage{cacheSweave} package for additional details regarding
cacheing.

<<cdfname>>=
cdfName <- "genomewidesnp6"
@

The HapMap CEL files are stored in a local directory assigned to
\verb+pathToCels+ in the following code. The genotyping step will
create several files with \verb+ff+ extensions. The ff objects contain
the low-level, normalized intensities as well as parameters used to
subsequently estimate copy number and B allele frequencies. These
files should not be deleted or moved.  We will store these files to
the path indicated by \verb+outdir+.

<<setup>>=
pathToCels <- "/thumper/ctsa/snpmicroarray/hapmap/raw/affy/1m"
v <- paste0("crlmm_v", gsub("\\.", "_", packageDescription("crlmm")$Version))
outdir <- file.path("/thumper/ctsa/snpmicroarray/rs/ProcessedData", v)
dir.create(outdir, recursive=TRUE, showWarnings=FALSE)
@

By providing the path in \verb+outdir+ as an argument to the \R{}
function \Rfunction{ldPath}, all of the \verb+ff+ files created during
the genotyping step will be stored in \verb+outdir+.

<<ldpath>>=
ldPath(outdir)
@

% only needed if cacheing
<<cachedir, echo=FALSE>>=
if(useCache) setCacheDir(outdir)
@

The \R{} functions \Rfunction{ocProbesets} and \Rfunction{ocSamples}
manage the RAM required for our analysis. See the documentation for
these functions and the \verb+CopyNumberOverview+ vignette for
additional details.

<<ram>>=
ocProbesets(100000)
ocSamples(200)
@


Next we indicate the local directory that contains the CEL files. For
the purposes of this vignette, we only analyze the CEPH ('C') and
Yoruban ('Y') samples.

<<celfiles>>=
celFiles <- list.celfiles(pathToCels, full.names=TRUE, pattern=".CEL")[1:10]
celFiles <- celFiles[substr(basename(celFiles), 13, 13) %in% c("C", "Y")]
if(exists("file.index")){
	celFiles <- celFiles[file.index]
}
@

Finally, copy number analyses using \crlmm{} require specification of
a batch variable that is used to indicate which samples were processed
together.  For example, if some of the samples were processed in April
and another set of samples were processed in June, we could name the
batches 'April' and 'June', respectively.  A useful surrogate for
batch is often the chemistry plate or the scan date of the array. For
the HapMap CEL files analyzed in this vignette, the CEPH (C) and
Yoruban (Y) samples were prepared on separate chemistry plates.  In
the following code chunk, we extract the population identifier from
the CEL file names and assign these identifiers to the variable
\Robject{plate}.

<<plates>>=
plates <- substr(basename(celFiles), 13, 13)
@

\section{Preprocessing and genotyping.}

The preprocessing steps for copy number estimation includes quantile
normalization of the raw intensities for each probe and a step that
summarizes the intensities of multiple probes at a single locus.  For
example, the Affymetrix 6.0 platform has 3 or 4 identical probes at
each polymorphic locus and the normalized intensities are summarized
by a median.  For the nonpolymorphic markers on Affymetrix 6.0, only
one probe per locus is available and the summarization step is not
needed.  After preprocessing the arrays, the \crlmm{} package
estimates the genotype using the CRLMM algorithm and provides a
confidence score for the genotype calls.   To begin, we initialize a
container for the normalized intensities:

<<constructCNSet, cache=TRUE>>=
cnSet <- constructAffyCNSet(celFiles, batch=plates,
			    cdfName="genomewidesnp6",
			    genome="hg19")
@


We quantile normalize the SNPs and nonpolymorphic markers separately.
Since the normalized intensities are ff objects, the functions
\Rfunction{cnrmaAffy} and \Rfunction{snprmaAffy} write the normalized
intensities to disk and nothing is returned.

<<cnrmaAffy, cache=TRUE>>=
cnrmaAffy(cnSet)
@

Any segment fault that occurs during the normalization can often be
traced to a corrupt cel file. To check if any of the files are
corrupt, one can use the function \Rfunction{validCEL} that tries to
read each files as in the following unevaluated codechunk:

<<checkcorrupt,eval=FALSE>>=
validCEL(celFiles)
@

<<snprmaAffy, cache=TRUE>>=
snprmaAffy(cnSet)
@

The function \Rfunction{genotypeAffy} performs performs the genotyping.

<<genotypeAffy, cache=TRUE>>=
genotypeAffy(cnSet, gender=NULL)
@

The above function also imputes the gender from the chromosome X and Y
intensities when the argument gender is \texttt{NULL}. The imputed
genders are

<<gender>>=
table(c("male", "female")[cnSet$gender[]])
@

<<genderCheck,echo=FALSE>>=
if(any(is.na(cnSet$gender[]))) stop("missing genders")
@

The normalized intensities, genotype calls, and confidence scores are
stored as \verb+ff+ objects in the \verb+assayData+ slot.  A concise
summary of this object can be obtained throught the \Rfunction{print}
or \Rfunction{show} methods.

<<show>>=
print(cnSet)
@

Note that the object is relatively small as the intensities and
genotype calls are stored on disk rather than in active memory.

<<objectsize>>=
print(object.size(cnSet), units="Mb")
@

\SweaveInput{copynumber}

A sample-specific estimate of the signal to noise ratio (SNR)
measuring the overall separation of the genotypes provides a measure
of sample quality.  Samples with SNRs below 5 typically indicate poor
quality, and typically have genotypes with lower confidence scores and
noisier copy number estimates.  The SNR is stored in the
\Robject{phenoData} slot of the \Rclass{CNSet} class and can be
accessed using the ``\$" operator.


\section{Session information}
<<sessionInfo, results=tex>>=
toLatex(sessionInfo())
@


\begin{figure}[f]
  \begin{center}
  \includegraphics[width=0.6\textwidth]{AffyGW-snr.pdf}
  \caption{The signal to noise ratio (SNR) for 180 HapMap samples. For
    Affymetrix platforms, SNR values below 5 can indicate possible
    problems with sample quality.  }
\end{center}
\end{figure}


%\begin{bibliography}
  \bibliographystyle{plain}
  \bibliography{refs}
%\end{bibliography}

\end{document}

%\VignetteIndexEntry{Infrastructure for copy number analysis}
%\VignetteDepends{crlmm, genomewidesnp6Crlmm}
%\VignetteKeywords{crlmm, SNP 6}
%\VignettePackage{crlmm}
\documentclass{article}
\usepackage{graphicx}
\usepackage{natbib}
\usepackage{url}
\newcommand{\Rfunction}[1]{{\texttt{#1}}}
\newcommand{\Rmethod}[1]{{\texttt{#1}}}
\newcommand{\Rcode}[1]{{\texttt{#1}}}
\newcommand{\Robject}[1]{{\texttt{#1}}}
\newcommand{\Rpackage}[1]{{\textsf{#1}}}
\newcommand{\Rclass}[1]{{\textit{#1}}}
\newcommand{\oligo}{\Rpackage{oligo }}
\newcommand{\R}{\textsf{R}}
\newcommand{\crlmm}{\Rpackage{crlmm}}
\newcommand{\ff}{\Rpackage{ff}}
\usepackage[margin=1in]{geometry}

\begin{document}
\title{Infrastructure for copy number analysis in \crlmm{}}
\date{\today}
\author{Rob Scharpf}

\maketitle

\begin{abstract}

  This vignette provides an overview of the \Rclass{CNSet} class and a
  brief discussion of the underlying infrastructure for large data
  support with the \Rpackage{ff} package.  This vignette instantiates
  an object of class \Rclass{CNSet} using a trivial dataset with 3
  files.  As this sample size is too small for estimating copy number
  with the \crlmm{} package, the final section of this vignette loads
  an object created by the analysis of 180 HapMap CEL files
  (Affymetrix 6.0 platform).  This object was instantiated by running
  the \verb+AffyGW+ vignette.

\end{abstract}

<<libraries,results=hide>>=
library(ff)
library(crlmm)
@

\section{Supported platforms}

The supported Affymetrix and Infinium platforms are those for which a
corresponding annotation package is available.  The annotation
packages contain information on the markers, such as physical position
and chromosome, as well as pre-computed parameters estimated from
HapMap used during the preprocessing and genotyping steps. For
Affymetrix, the 5.0 and 6.0 platforms are supported and the
corresponding annotation packages are \Rpackage{genomewidesnp5Crlmm}
and \Rpackage{genomewidesnp6Crlmm}.  Supported Infinium platforms are
listed in the following code chunk.

<<supportedPlatforms>>=
pkgs <- annotationPackages()
crlmm.pkgs <- pkgs[grep("Crlmm", pkgs)]
crlmm.pkgs[grep("human", crlmm.pkgs)]
@

\textbf{Large data:} In order to reduce \crlmm{}'s memory footprint
for copy number estimation, we require the \Rpackage{ff}.  The
\Rpackage{ff} package provides infrastructure for accessing and
writing data to disk instead of keeping data in memory.  As the
functions for preprocessing, genotyping, and copy number estimation do
not simultaneously require all samples and all probes in memory,
memory-usage by \crlmm{} can be fine-tuned by reading in and
processing subsets of the markers and/or samples. The functions
\Rfunction{ocSamples} and \Rfunction{ocProbesets} in the
\Rpackage{oligoClasses} package can be used to declare how many
markers and samples to read at once.  In general, specifying smaller
values should reduce the RAM required for a particular job.  In
general, smaller values will increase the run-time. In the following
code-chunk, we declare that \crlmm{} should process 100,000 markers at
a time (when possible) and 200 samples at a time.  If our dataset
contained fewer than 200 samples, the \Rfunction{ocSamples} option
would not have any effect.  One can view the current settings for
these commands, by typing the functions without an argument.

<<ram>>=
ocProbesets(100e3)
ocSamples(200)
@

\section{The \Rclass{CNSet} container}

\subsection{Instantiating an object of class \Rclass{CNSet}}

An object of class \Rclass{CNSet} can be instantiated by one of two
methods:

\begin{itemize}

\item[Approach 1:] during the preprocessing of the raw intensities for
  Illumina and Affymetrix arrays by the the functions
  \Rfunction{constructInf} and \Rfunction{genotype},
  respectively. (The \Rfunction{genotype} calls the function
  \Rfunction{constructAffy} to initialize a \Rclass{CNSet} object for
  Affymetrix platforms.)

\item[Approach 2:] by subsetting an existing \Robject{CNSet} object.
  As per usual, the `[' method can be used to extract a subset of
  markers $i$ as in `[i, ]', a subset of samples $j$ as in `[, j]', or
  a subset of markers $i$ and samples $j$ as in `[i, j]'.

\end{itemize}

There are important differences in the underlying data representation
depending on how the \Rclass{CNSet} object was instantiated.  In
particular, objects generated by the functions
\Rfunction{constructInf} and \Rfunction{genotype} store
high-dimensional data on disk rather than in memory through protocols
defined in the \R{} package \ff{}.  For instance, the normalized
intensities and genotype calls in a \Rclass{CNSet}-instance from
approach (1) are \ff{}-derived objects.  By contrast, when such an
objected generated by approach (1) is subset by the `[' method, an
object of the same class is returned but the \Rpackage{ff}-derived
objects are coerced to ordinary matrices. Note, therefore, that both
approaches (1) and (2) may involve substantial I/O and that (2) should
be performed judiciously.

\subsubsection{Approach 1}

To illustrate the first approach, we begin by specifying a local
directory to store output files and setting the \Rfunction{ldPath}
function with this path.

<<path>>=
outdir <- paste("/local_data/r00/crlmm/", getRversion(), "/infrastructure", sep="")
ldPath(outdir)
if(!file.exists(outdir)) dir.create(outdir)
@


Next we load the annotation package required, as well as the \R{}
package \Rpackage{hapmapsnp6} that contains 3 example CEL files.  We
use the \Rfunction{system.file} function to find the path to the CEL
files.

<<annotation_metadata>>=
require(genomewidesnp6Crlmm) & require(hapmapsnp6)
path <- system.file("celFiles", package="hapmapsnp6")
celfiles <- list.celfiles(path, full.names=TRUE)
@

Typically, an object of class \Rclass{CNSet} is instantiated as part
of the preprocessing and genotyping by calling the function
\Rfunction{genotype}, as illustrated in the
\verb+AffyGW+ vignette.

<<instantiateToyExample,cache=TRUE>>=
exampleSet <- genotype(celfiles, batch=rep("1", 3), cdfName="genomewidesnp6")
@

Several files with \verb+.ff+ extensions now appear in the directory
indicated by the \Rfunction{ldPath} function.

<<ldpath>>=
ldPath()
@

One could also instantiate an object of class \Rclass{CNSet} without
preprocessing/genotyping by calling the exported function
\Rfunction{constructAffy} directly using the \Rfunction{:::} operator.

<<constructAffy,eval=FALSE>>=
tmp <- constructAffy(celfiles, batch=rep("1", 3), cdfName="genomewidesnp6")
@

The \Rfunction{show} method provides a concise summary of the
\Robject{exampleSet} object. Note the class of the elements in the
\verb+batchStatistics+ and \verb+assayData+ slots is indicated in the
first line of the summary.

<<show>>=
invisible(open(exampleSet))
exampleSet
@

% We briefly outline some of the unique aspects of the
% \Rclass{CNSet}-class using in \crlmm{} that may differ from the more
% standard extensions of the virtual class \Rclass{eSet} defined in
% the \Rpackage{Biobase} package.

As the \verb+assayData+ elements of the \Robject{exampleSet} object
are stored on disk rather than in memory, we inspect attributes of the
elements by first opening the file connection using the
\Rfunction{open}. For example, in the following code we extract the
normalized intensities for the A allele by first opening the object
returned by the \Rfunction{A} function.  The name of the file where
the data is stored on disk is provided by the
\Rfunction{filename}. Finally, the normalized intensities can be
pulled from disk to memory by the `[' method.  It can be useful to
wrap the `[' method by the \Rfunction{as.matrix} to ensure that the
output is the desired class.

<<approach1>>=
invisible(open(exampleSet))
class(A(exampleSet))
filename(A(exampleSet))
as.matrix(A(exampleSet)[1:5, ])
@

\paragraph{Moving \texttt{*.ff} files} Ideally, the files with
\verb+.ff+ extension should not be moved.  However, if this is not
possible, the safest way to move these files is to clone all of the
\ff{} objects using the \Rfunction{clone}, followed by the
\Rfunction{delete} function to remove the original files on disk. An
example of the \Rfunction{delete} function is included at the end of
the \verb+IlluminaPreprocessCN+ vignette. See the documentation for
the \Rfunction{clone} and \Rfunction{delete} functions in the \ff{}
package for additional details.


\paragraph{Order of operations:} For \Rclass{CNSet}-instances derived
by approach (1), users should be mindful of the substantial I/O when
using accessors to extract data from the class.  For example, the
following 2 methods would extract identical results, with the latter
being much more efficient (extra parentheses are added to the second
operation to emphasize the order of operations):

<<orderOfOperations>>=
##inefficient
##invisible(open(cnSet))
A(exampleSet[1:5, ])
## preferred
(A(exampleSet))[1:5, ]
@


\subsubsection{Approach 2: using `['}

Here we instantiate a new object of class \Rclass{CNSet} by applying the
`[' method to an existing object of class \Rclass{CNSet}.

<<approach2>>=
cnset.subset <- exampleSet[1:5, ]
@

Note the class of the \verb+batchStatistics+ and \verb+assayData+
elements of the \Robject{cnset.subset} object printed in the first
line of the summary.

<<show.subset>>=
show(cnset.subset)
@

\subsection{Slots of class \Rclass{CNSet}}
\subsubsection{\texttt{featureData}}

Information on physical position, chromosome, and whether the marker
is a SNP can be accessed through accessors defined for the
\verb+featureData+.

<<featuredataAccessors>>=
library(Biobase)
fvarLabels(exampleSet)
position(exampleSet)[1:10]
chromosome(exampleSet)[1:10]
is.snp <- isSnp(exampleSet)
table(is.snp)
snp.index <- which(is.snp)
np.index <- which(!is.snp)
chr1.index <- which(chromosome(exampleSet) == 1)
@

\subsubsection{\texttt{assayData}}

The \verb+assayData+ elements are of the class
\Rclass{ff\_matrix}/\Rclass{ffdf} or \Rclass{matrix}, depending on how
the \Rclass{CNSet} object was instantiated.  Elements in the
\verb+assayData+ environment can be listed using the \Rfunction{ls}
function.

<<assayData>>=
ls(assayData(exampleSet))
@

The normalized intensities for the A and B alleles have names
\verb+alleleA+ and \verb+alleleB+ and can be accessed by the methods
\Rfunction{A} and \Rfunction{B}, respectively.  Genotype calls and
confidence scores can be accessed by \Rfunction{snpCall} and
\Rfunction{snpCallProbability}, respectively.  Note that the
confidence scores are represented as integers to reduce the filesize,
but can be tranlated to the probability scale using the \R{} function
\Rfunction{i2p}. For example,

<<i2p>>=
scores <- as.matrix(snpCallProbability(exampleSet)[1:5, 1:2])
i2p(scores)
@

Note that for the Affymetrix 6.0 platform the assay data elements each
have a row dimension corresponding to the total number of polymorphic
and nonpolymorphic markers interrogated by the Affymetrix 6.0
platform.  A consequence of keeping the rows of the assay data
elements the same for all of the statistical summaries is that the
matrix used to store genotype calls is larger than necessary.

\subsubsection{\texttt{batch} and \texttt{batchStatistics}}

As defined in Leek \textit{et al.} 2010, \textit{Batch effects are
  sub-groups of measurements that have qualitatively different
  behaviour across conditions and are unrelated to the biological or
  scientific variables in a study}.  The \verb+batchStatistics+ slot
is an environment used to store SNP- and batch-specific summaries,
such as the sufficient statistics for the genotype clusters and the
linear model parameters used for copy number estimation.  The
\verb+batch+ slot is used to store the 'batch name' for each
array. For small studies in which the samples were processed at
similar times (e.g., within a month), all the samples can be
considered to be in the same batch.  For large studies in which the
samples were processed over several months, users should the scan date
of the array or the chemistry plate are useful surrogates. The only
constraint on the \verb+batch+ variable is that it must be a character
vector that is the same length as the number of samples to be
processed.  The \verb+batch+ is specified as an argument to the \R{}
functions \Rfunction{constructInf} and \Rfunction{genotype} that
instantiate \Rclass{CNSet} objects for the Illumina and Affymetrix
platforms, respectively.  The \Rfunction{batch} function can be used
to access the \verb+batch+ information on the samples as in the
following example.

<<batchAccessor>>=
batch(exampleSet)
@

For the \verb+batchStatistics+ slot, the elements in the environment
have the class \Rclass{ff\_matrix}/\Rclass{ffdf} or \Rclass{matrix},
depending on how the \Rclass{CNSet} object was instantiated.  The
dimension of each element is the number of markers (SNPs +
nonpolymorphic markers) $\times$ the number of batches. The names of
the elements in the environment can be list using the \R{} function
\Rfunction{ls}.

<<batchStatistics>>=
ls(batchStatistics(exampleSet))
@

Currently, the batch-specific summaries are stored to allow some
flexibility in the choice of downstream analyses of copy number and
visual assessments of model fit.  Documentation for such applications
will be expanded in future versions of \crlmm{}, and are currently not
intended to be accessed directly by the user.

%The \Robject{CNSet} class also contains the slot
%\Robject{batchStatistics} that contains batch-specific summaries
%needed for copy number estimation. In particular, each element is a
%matrix (or an ff object) with R rows and C columns, correspoinding to
%R markers and C batches.  The summaries includes the within genotype
%cluster medians and median absolute deviations (mads), but also
%parameters estimated from the linear model.  (For unobserved
%genotypes, the medians are imputed and the variance is obtained the
%median variance (across markers) within a batch. ) The elements of the
%slot can be listed as follows.

\subsubsection{\texttt{phenoData}}

Sample-level summaries obtained during the preprocessing/genotyping
steps include skew (SKW), the signal to noise ratio (SNR), and gender
(1=male, 2=female) are stored in the \verb+phenoData+ slot.  As for
other \Rclass{eSet} extensions, the \verb+$+ method can be used to
extract these summaries.  For \Rclass{CNSet} objects generated by
approach (1), these elements are of the class
\Rclass{ff\_vector}.

<<phenodataAccessors>>=
varLabels(exampleSet)
class(exampleSet$gender)
invisible(open(exampleSet$gender))
exampleSet$gender
@

The `[' methods without arguments can be used to coerce to a vector.

<<vector>>=
c("male", "female")[exampleSet$gender[]]
invisible(close(exampleSet$gender))
@




\subsubsection{\texttt{protocolData}}

The scan date of the arrays are stored in the \verb+protocolData+.

<<protocolData>>=
varLabels(protocolData(exampleSet))
protocolData(exampleSet)$ScanDate
@

\section{Trouble shooting with a HapMap example}

This section uses an object of class \Rclass{CNSet} instantiated by
the \verb+AffyGW+ vignette and saved to a local path
on our computing cluster indicated by the object \Robject{outdir}
below.  The \verb+copynumber+ vignette was used to fill out the
\verb+batchStatistics+ slot of the \Robject{cnSet} object.

<<path>>=
if(getRversion() < "2.13.0"){
	rpath <- getRversion()
} else rpath <- "trunk"
outdir <- paste("/thumper/ctsa/snpmicroarray/rs/ProcessedData/crlmm/", rpath, "/copynumber_vignette", sep="")
@

Next, we load the \Robject{cnSet} object.

<<loadcnset>>=
if(!exists("cnSet")) load(file.path(outdir, "cnSet.rda"))
invisible(open(cnSet))
@


\subsection{Missing values}

Most often, missing values occur when the genotype confidence scores
for a SNP were below the threshold used by the
\Robject{crlmmCopynumber} function. For the HapMap analysis, we used a
confidence threshold of 0.80 (the default).  In the following code, we
assess NA's appearing for the raw copy number estimates for the first
10 samples.

<<NA>>=
GT.CONF.THR <- 0.80
autosome.index <- which(isSnp(cnSet) & chromosome(cnSet) < 23)
sample.index <- 1:10
ct <- totalCopynumber(cnSet, i=autosome.index, j=sample.index)
ca <- CA(cnSet, i=autosome.index, j=sample.index)
cb <- CB(cnSet, i=autosome.index, j=sample.index)
missing.ca <- is.na(ca)
missing.cb <- is.na(cb)
(nmissing.ca <- sum(missing.ca))
(nmissing.cb <- sum(missing.cb))
identical(nmissing.ca, nmissing.cb)
@

If \Robject{nmissing.ca} is nonzero, check the genotype confidence
scores provided by the crlmm genotyping algorithm against the
threshold specified in \Robject{crlmmCopynumber}.

<<NAconfidenceScores>>=
if(nmissing.ca > 0){
	##invisible(open(snpCallProbability(cnSet)))
	gt.conf <- i2p(snpCallProbability(cnSet)[autosome.index, sample.index])
	##invisible(close(snpCallProbability(cnSet)))
	below.thr <- gt.conf < GT.CONF.THR
	index.allbelow <- as.integer(which(rowSums(below.thr) == length(sample.index)))
	nmissingBecauseOfGtThr <- length(index.allbelow) * length(sample.index)
	stopifnot(identical(nmissingBecauseOfGtThr, nmissing.ca))
	## or calculate the proportion of missing effected by low crlmm confidence
	length(index.allbelow) * length(sample.index)/nmissing.ca
}
@

One could inspect the cluster plots for the 'low confidence' calls.

<<clusterPlotsLowConfidence>>=
## TODO
@

We repeat the above check for missing values at polymorphic loci on
chromosome X.  In this case, we compare the \Robject{rowSums} of the
missing values to the number of samples to check whether all of the
estimates are missing for a given SNP.

<<NAchromosomeX>>=
## start with first batch
sample.index <- which(batch(cnSet)==batch(cnSet)[1])
X.index <- which(isSnp(cnSet) & chromosome(cnSet) == 23)
ca.X <- CA(cnSet, i=X.index, j=sample.index)
missing.caX <- is.na(ca.X)
(nmissing.caX <- sum(missing.caX))
missing.snp.index <- which(rowSums(missing.caX) == length(sample.index))
index <- which(rowSums(missing.caX) == length(sample.index))
##length(index)*length(sample.index)/nmissing.caX
@

From the above codechunk, we see that \Sexpr{length(index)} SNPs have
NAs for all the samples. Next, we tally the number of NAs for
polymorphic markers on chromosome X that are below the confidence
threshold.  For the HapMap analysis, all of the missing values arose
from SNPs in which either the men or the women had confidence scores
that were all below the threshold.

<<NAcrlmmConfidenceScore>>=
invisible(open(cnSet$gender))
F <- which(cnSet$gender[sample.index] == 2)
M <- which(cnSet$gender[sample.index] == 1)
gt.conf <- i2p(snpCallProbability(cnSet)[X.index, sample.index])
below.thr <- gt.conf < GT.CONF.THR
index.allbelowF <- as.integer(which(rowSums(below.thr[, F]) == length(F)))
index.allbelowM <- as.integer(which(rowSums(below.thr[, M]) == length(M)))
all.equal(index.allbelowF, index.allbelowM)
all.equal(as.integer(index.allbelowF), as.integer(index))
@

For nonpolymorphic loci, the genotype confidence scores are irrelevant
and estimates are available at most markers.

<<NAnonpolymorphic.autosome>>=
np.index <- which(!isSnp(cnSet) & chromosome(cnSet)==23)
ca.F <- CA(cnSet, i=np.index, j=F)
ca.M <- CA(cnSet, i=np.index, j=M)
## NAs for one marker
ca.F <- ca.F[-match("CN_974939", rownames(ca.F)), ]
ca.M <- ca.M[-match("CN_974939", rownames(ca.M)), ]
sum(is.na(ca.F))
sum(is.na(ca.M))
@

%TODO: marker CN\_974939 has NAs for the normalized intensities.  This
%is because CN\_974939 is not in the \Robject{npProbesFid} file in
%\Rpackage{genomewidesnp6Crlmm}.  The \Robject{npProbesFid} file should
%be updated in the next \Rpackage{genomewidesnp6Crlmm} release.

In total, there were \Sexpr{length(missing.snp.index)} polymorphic
markers on chromosome X for which copy number estimates are not
available.  Lowering the confidence threshold would permit estimation
of copy number at most of these loci.  A confidence threshold is
included as a parameter for the copy number estimation as an approach
to reduce the sensitivity of genotype-specific summary statistics,
such as the within-genotype median, to intensities from samples that
do not clearly fall into one of the biallelic genotype clusters. There
are drawbacks to this approach, including variance estimates that can
be a bit optimistic at some loci.  More direct approaches for outlier
detection and removal may be explored in the future.

Copy number estimates for other chromosomes, such as mitochondrial and
chromosome Y, are not currently available in \crlmm{}.

<<close>>=
invisible(close(cnSet))
invisible(close(cnSet$gender))
@


\section{Session information}
<<sessionInfo, results=tex>>=
toLatex(sessionInfo())
@

%\section*{References}
%
%%\begin{bibliography}
%  \bibliographystyle{plain}
%  \bibliography{refs}
%\end{bibliography}


\end{document}

%\VignetteIndexEntry{From Genotypes to Association}
%\VignetteKeywords{genotype, crlmm, SNP 5, SNP 6}
%\VignettePackage{crlmm}

\documentclass[12pt]{article}

\usepackage{amsmath}
\usepackage[authoryear,round]{natbib}
\usepackage{hyperref}


\textwidth=6.2in
\textheight=8.5in
%\parskip=.3cm
\oddsidemargin=.1in
\evensidemargin=.1in
\headheight=-.3in

\newcommand{\scscst}{\scriptscriptstyle}
\newcommand{\scst}{\scriptstyle}


\newcommand{\Rfunction}[1]{{\texttt{#1}}}
\newcommand{\Robject}[1]{{\texttt{#1}}}
\newcommand{\Rpackage}[1]{{\textit{#1}}}
\newcommand{\Rmethod}[1]{{\texttt{#1}}}
\newcommand{\Rfunarg}[1]{{\texttt{#1}}}
\newcommand{\Rclass}[1]{{\textit{#1}}}

\textwidth=6.2in

\bibliographystyle{plainnat}

\begin{document}
%\setkeys{Gin}{width=0.55\textwidth}

\title{crlmm to downstream data analysis}
\author{VJ Carey, B Carvalho}
\date{March, 2012}
\maketitle

\section{Running CRLMM on a nontrivial set of CEL files}
To use the \Rmethod{crlmm} algorithm, the user must load the
\Rpackage{crlmm} package, as described below:
<<loadPkg>>=
library(crlmm)
@

We work with the 90 CEU samples hybridized to Affy 6.0 chips. When CEL
files are available, they must be identified and passed to
\Rmethod{crlmm}, as shown below. In this example, we assume that the
results are stored in a variable called \Robject{crlmmResult}.
<<lkd, eval=FALSE>>=
celFiles <- list.celfiles()
crlmmResult <- crlmm(celFiles)
@

Alternatively, the data aforementioned are available through the
\Rpackage{hapmapsnp6} package (required minimum version $1.3.6$) and can
be loaded by using:

<<loadFromPkg>>=
suppressPackageStartupMessages(library(hapmapsnp6))
data(crlmmResult)
@

This is currently a \Rclass{SnpSet} object.
<<lkj21>>=
  class(crlmmResult)
@

%% In order to reduce the memory requirements for this task, we will use
%% only results for chromosome 20.
%%
%% <<getSubset>>=
%% @

\section{Adding information to a \Rclass{SnpSet}}

We will use the \Rpackage{GGdata} package to obtain extra information
on the samples. This will be later used when building an \Rclass{eSet}
extension to store the genotyping results.
<<getpd>>=
  suppressPackageStartupMessages(library(GGdata))
  hmceuB36 <- getSS('GGdata', as.character(1:22))
  pd <- phenoData(hmceuB36)
  ggn <- sampleNames(pd)
  preSN <- sampleNames(crlmmResult)
  simpSN <- gsub("_.*", "", preSN)
  if (!all.equal(simpSN, ggn)) stop("align GGdata phenoData with crlmmResult read")
@

The additional information obtained from \Rpackage{GGdata} can be
easily combined to what is already available on \Robject{crlmmResult}.
<<docl>>=
  sampleNames(crlmmResult) <- simpSN
  phenoData(crlmmResult) <- combine(pd, phenoData(crlmmResult))
  dim(calls(crlmmResult))
  dim(confs(crlmmResult, FALSE))
  calls(crlmmResult)[1:10, 1:2]
  confs(crlmmResult, FALSE)[1:10, 1:2]
@


\section{Coercing to SnpMatrix as a prelude to a GWAS}

From this point on, we will use only the genotype calls. Therefore, to
reduce memory requirements, we will recode the \Rpackage{crlmm} genotype
calls, so the \Rpackage{snpStats} package can be used, and delete the
remaining \Rmethod{crlmm} results.
<<clean>>=
theCalls <- t(calls(crlmmResult))-1L
rm(crlmmResult)
@

<<morecleaning, echo=FALSE>>=
gc()
@

SNP's for which all the samples have the same genotype are not
informative for association studies. Therefore, we remove such SNP's
prior to fitting the models.

<<rmNonInformative>>=
gtypeCounts <- rbind(AA=colSums(theCalls == 0L),
                     AB=colSums(theCalls == 1L),
                     BB=colSums(theCalls == 2L))
gtypeCounts[, 1:5]
toRemove <- which(colSums(gtypeCounts == 0) == 2L)
gtypeCounts[, toRemove[1:4]]
theCalls <- theCalls[, -toRemove]
@

The \Rpackage{snpStats} provides tools to simplify the analysis of
GWAS. The snippet below shows how to load the package and convert the
genotype calls to a format that \Rpackage{snpStats} is able to handle.
<<lksnm>>=
suppressPackageStartupMessages(library(snpStats))
crlmmSM <- new("SnpMatrix", theCalls)
crlmmSM
@

\section{Conducting a GWAS}

We want to find SNP for which genotype is predictive of expression of CPNE1.
We will use expression data available from GGdata, using a naive analysis.
<<doa>>=
suppressPackageStartupMessages(library(illuminaHumanv1.db))
rmm <- revmap(illuminaHumanv1SYMBOL)
mypr <- get("CPNE1", rmm)
ex <- as.numeric(exprs(hmceuB36)[mypr[1],])
subjdata <- pData(hmceuB36)
subjdata[["ex"]] <- ex
head(subjdata)
@

With the expression data now available in \Robject{subjdata}, we can use
the tools from \Rpackage{SnpMatrix} to fit models that will be used to
evaluate the association between the genotypes of each available SNP and
the expression levels of CPNE1.
<<model>>=
gwas <- snp.rhs.tests(ex~male, data=subjdata, snp.data=crlmmSM, family="gaussian")
ok <- which(p.value(gwas) < 1e-10)
gwas[ok,]
@

<<dopl,fig=TRUE>>=
snp <- names(gwas[ok,])[1]
gtypes <- theCalls[,snp]+1L
boxplot(ex~gtypes, xlab=paste("Genotype Call for", snp),
        ylab="CPNE1 Expression", xaxt="n", range=0)
points(ex~jitter(gtypes), col=gtypes, pch=19)
axis(1, at=1:3, labels=c("AA", "AB", "BB"))
@

\section{Session Info}

This vignette was created using the following packages:
<<lksess>>=
sessionInfo()
@

\end{document}

%\VignetteIndexEntry{crlmm Vignette - Illumina 370k chip}
%\VignetteKeywords{genotype, crlmm, Illumina}
%\VignettePackage{crlmm}

\documentclass[12pt]{article}

\usepackage{amsmath,pstricks}
\usepackage[authoryear,round]{natbib}
\usepackage{hyperref}
\usepackage{Sweave}

\textwidth=6.2in
\textheight=8.5in
%\parskip=.3cm
\oddsidemargin=.1in
\evensidemargin=.1in
\headheight=-.3in

\newcommand{\scscst}{\scriptscriptstyle}
\newcommand{\scst}{\scriptstyle}


\newcommand{\Rfunction}[1]{{\texttt{#1}}}
\newcommand{\Robject}[1]{{\texttt{#1}}}
\newcommand{\Rpackage}[1]{{\textit{#1}}}
\newcommand{\Rmethod}[1]{{\texttt{#1}}}
\newcommand{\Rfunarg}[1]{{\texttt{#1}}}
\newcommand{\Rclass}[1]{{\textit{#1}}}

\textwidth=6.2in

\bibliographystyle{plainnat}

\begin{document}
%\setkeys{Gin}{width=0.55\textwidth}

\title{Using \Rpackage{crlmm} to genotype data from Illumina's Infinium BeadChips}
\author{Matt Ritchie}
\maketitle

\section{Getting started}

In this user guide we read in and genotype data from 40 HapMap samples
which have been analyzed using Illumina's 370k Duo BeadChips.
This data is available in the \Rpackage{hapmap370k} package.
Additional chip-specific model parameters and basic SNP annotation
information used by CRLMM is stored in the \Rpackage{human370v1cCrlmm} package.
The required packages can be installed in the usual way using the \Rfunction{biocLite} function.

<<echo=TRUE, results=hide, eval=FALSE>>=
if (!requireNamespace("BiocManager", quietly=TRUE))
    install.packages("BiocManager")
BiocManager::install(c("crlmm", "hapmap370k", "human370v1cCrlmm"))
@

\section{Reading in data}
The function \Rfunction{readIdatFiles} extracts the Red and Green intensities
from the binary {\tt idat} files output by Illumina's scanning device.
The file {\tt samples370k.csv} contains information about each sample.

<<echo=FALSE, results=hide, eval=TRUE>>=
options(width=70)
@

<<read, results=hide, eval=TRUE>>=
library(crlmm)
library(ff)
library(hapmap370k)

data.dir = system.file("idatFiles", package="hapmap370k")

# Read in sample annotation info
samples = read.csv(file.path(data.dir, "samples370k.csv"), as.is=TRUE)
samples[1:5,]
@

\section{Genotyping}

Next we use the function \Rfunction{crlmmIllumina} which performs preprocessing followed by genotyping using the CRLMM algorithm.
It reads in data from idat files and stores results in a \Rclass{CNSet} object.

<<genotype, results=hide, cache=TRUE>>=
crlmmResult = crlmmIllumina(samples, path=data.dir,
				arrayInfoColNames=list(barcode=NULL,position="SentrixPosition"),
				saveDate=TRUE, cdfName="human370v1c", verbose=FALSE)

class(crlmmResult)
dim(crlmmResult)
slotNames(crlmmResult)
calls(crlmmResult)[1:10, 1:5]
i2p(confs(crlmmResult)[1:10,1:5])
@

If GenCall output is available instead of idat files, the function \Rfunction{readGenCallOutput} can be
used to read in the data.
This function assumes the GenCall output is formatted to have samples listed one below the other,
and that the columns 'X Raw' and 'Y Raw' are available in the file.
The resulting \Robject{NChannelSet} from this function can be used as input to 
\Rfunction{crlmmIllumina} (or equivalently \code{genotype.Illumina}) via the \Robject{XY} argument 
(instead of the usual \Rfunction{RG} argument used when the data has been read in from idat files).

Plotting the {\it SNR} reveals no obvious batch effects in this data set (different symbols are used for
arrays scanned on different days).

<<snr,  fig=TRUE, width=8, height=6>>=
plot(crlmmResult[["SNR"]], pch=scanbatch, xlab="Array", ylab="SNR",
     main="Signal-to-noise ratio per array",las=2)
@

<<plotsamples, fig=TRUE, width=8, height=8>>=
par(mfrow=c(2,2))
plotSamples(crlmmResult, col=1:4)
@

<<plotsnps, fig=TRUE, width=8, height=8>>=
par(mfrow=c(2,2))
plotSNPs(crlmmResult, row=1:4)
@ 

\section{System information}

This analysis was carried out on a linux machine with 32GB of RAM
using the following packages:

<<session>>=
sessionInfo()
@

\end{document}

\name{genotype}
\alias{genotype}
\alias{genotype2}
\alias{genotypeLD}
\title{
	Preprocessing and genotyping of Affymetrix arrays.
}
\description{
	Preprocessing and genotyping of Affymetrix arrays.
}
\usage{
genotype(filenames, cdfName, batch, mixtureSampleSize = 10^5, eps =0.1,
         verbose = TRUE, seed = 1, sns, probs = rep(1/3, 3),
         DF = 6, SNRMin = 5, recallMin = 10, recallRegMin = 1000,
         gender = NULL, returnParams = TRUE, badSNP = 0.7,  genome=c("hg19", "hg18"))
}
\arguments{
  \item{filenames}{ complete path to CEL files}
  \item{cdfName}{  annotation package  (see also
  \code{validCdfNames})}
  \item{batch}{ vector of class \code{character} denoting the batch for
    each sample in \code{filenames}. The
    \code{batch} vector must be the same length as the number of
    samples. See details. }
  \item{mixtureSampleSize}{    Sample size to be use when fitting the mixture model.}
 \item{eps}{   Stop criteria.}
  \item{verbose}{  Logical.  Whether to print descriptive messages during processing.}
  \item{seed}{ Seed to be used when sampling. Useful for reproducibility}
  \item{sns}{The sample identifiers.  If missing, the default sample names are \code{basename(filenames)}}
  \item{probs}{'numeric' vector with priors for AA, AB and BB.}
  \item{DF}{'integer' with number of degrees of freedom to use with t-distribution.}
  \item{SNRMin}{'numeric' scalar defining the minimum SNR used to filter
  out samples.}
  \item{recallMin}{Minimum number of samples for recalibration. }
  \item{recallRegMin}{Minimum number of SNP's for regression.}
  \item{gender}{  integer vector (  male = 1, female =2 ) or missing,
  with same length as filenames.  If missing, the gender is predicted.}
  \item{returnParams}{'logical'. Return recalibrated parameters from
    crlmm.}
  \item{badSNP}{'numeric'. Threshold to flag as bad SNP (affects
    batchQC)}
  \item{genome}{character string indicating the UCSC genome build for
    the SNP annotation}
}

\details{

	For large datasets it is important to utilize the large data
	support by installing and loading the ff package before calling
	the \code{genotype} function. In previous versions of the
	\code{crlmm} package, we useed different functions for
	genotyping depending on whether the ff package is loaded, namely
	\code{genotype} and \code{genotype2}.  The \code{genotype}
	function now handles both instances.

	\code{genotype} is essentially a wrapper of the \code{crlmm}
	function for genotyping.  Differences include (1) that the copy
	number probes (if present) are also quantile-normalized and (2)
	the class of object returned by this function, \code{CNSet}, is
	needed for subsequent copy number estimation.  Note that the
	batch variable that must be passed to this function has no
	effect on the normalization or genotyping steps.  Rather,
	\code{batch} is required in order to initialize a \code{CNSet}
	container with the appropriate dimensions and is used directly
	when estimating copy number.
}

\value{	A \code{SnpSuperSet} instance.}
\references{

  Carvalho B, Bengtsson H, Speed TP, Irizarry RA. Exploration,
  normalization, and genotype calls of high-density oligonucleotide SNP
  array data. Biostatistics. 2007 Apr;8(2):485-99. Epub 2006 Dec
  22. PMID: 17189563.

  Carvalho BS, Louis TA, Irizarry RA.
  Quantifying uncertainty in genotype calls.
  Bioinformatics. 2010 Jan 15;26(2):242-9.

}
\author{R. Scharpf}
\note{For large datasets, load the 'ff' package prior to genotyping --
this will greatly reduce the RAM required for big jobs.  See
\code{ldPath} and \code{ocSamples}.}

\seealso{
	\code{\link{snprma}}, \code{\link{crlmm}},
	\code{\link[oligoClasses]{ocSamples}},
	\code{\link[oligoClasses]{ldOpts}},
	\code{\link{batch}},
	\code{\link{crlmmCopynumber}}
}
\examples{
if (require(ff) & require(genomewidesnp6Crlmm) & require(hapmapsnp6)){
  ldPath(tempdir())
  path <- system.file("celFiles", package="hapmapsnp6")
  ## the filenames with full path...
  ## very useful when genotyping samples not in the working directory
  cels <- list.celfiles(path, full.names=TRUE)
  ## Note: one would need at least 10 CEL files for copy number estimation
  ## To use less RAM, specify a smaller argument to ocProbesets
  ocProbesets(50e3)
  batch <- rep("A", length(cels))
  (cnSet <- genotype(cels, cdfName="genomewidesnp6", batch=batch))

##Segment faults that occur with the above step can often be traced to a
##corrupt cel file. To check if any of the files are corrupt, try
##reading the files in one at a time:

\dontrun{
require(affyio)
validCEL(cels)
}

  ## when gender is not specified (as in the above example), crlmm tries
  ## to predict the gender from SNPs on chromosome X
  cnSet$gender

  ## If gender is known, one should check that the assigned gender is
  ## correct. Alternatively, one can pass gender as an argument to the
  ## genotype function.
  gender <- c("female", "female", "male")
  gender[gender == "female"] <- 2
  gender[gender == "male"] <- 1
  dim(cnSet)
  table(isSnp(cnSet))
}
}
\keyword{ classif }

\name{genotypes}
\alias{genotypes}

\title{
  The possible genotypes for an integer copy number.
}
\description{
  The possible genotypes for an integer copy number (0-4).
}
\usage{
genotypes(copyNumber, is.snp=TRUE)
}

\arguments{
  \item{copyNumber}{
    Integer (0-4 allowed).}

  \item{is.snp}{Logical. If TRUE, possible genotypes for a polymorphic
    SNP is returned. If FALSE, only monomorphic genotypes returned.}
}

\value{
  Character vector.
}

\author{
R. Scharpf
}

\examples{

for(i in 0:4) print(genotypes(i))
for(i in 0:4) print(genotypes(i, FALSE))

}

\keyword{manip}

\name{preprocessInf}
\alias{preprocessInf}
\title{

	Preprocessing of Illumina Infinium II arrays.

}
\description{

	This function normalizes the intensities for the 'A' and 'B'
	alleles for a \code{CNSet} object and estimates mixture
	parameters used for subsequent genotyping.  See details for
	how the normalized intensities are written to file.  This step
	is required for subsequent genotyping and copy number
	estimation.

}
\usage{
preprocessInf(cnSet, sampleSheet=NULL, arrayNames = NULL, ids = NULL,
path = ".", arrayInfoColNames = list(barcode = "SentrixBarcode_A",
position = "SentrixPosition_A"), highDensity = TRUE, sep = "_", fileExt
= list(green = "Grn.idat", red = "Red.idat"), XY, anno, saveDate = TRUE, stripNorm
= TRUE, useTarget = TRUE, mixtureSampleSize = 10^5, fitMixture = TRUE, 
quantile.method="between", eps = 0.1, verbose = TRUE, seed = 1, cdfName)
}

\arguments{
  \item{cnSet}{
  object of class \code{CNSet}
}

  \item{sampleSheet}{\code{data.frame} containing Illumina sample sheet
    information (for required columns, refer to BeadStudio Genotyping
    guide - Appendix A).}

  \item{arrayNames}{character vector containing names of arrays to be
    read in.  If \code{NULL}, all arrays that can be found in the
    specified working directory will be read in.}

  \item{ids}{vector containing ids of probes to be read in.  If
    \code{NULL} all probes found on the first array are read in.}

  \item{path}{character string specifying the location of files to be
    read by the function}

  \item{arrayInfoColNames}{(used when \code{sampleSheet} is specified)
    list containing elements 'barcode' which indicates column names in
    the \code{sampleSheet} which contains the arrayNumber/barcode number
    and 'position' which indicates the strip number.  In older style
    sample sheets, this information is combined (usually in a column
    named 'SentrixPosition') and this should be specified as
    \code{list(barcode=NULL, position="SentrixPosition")}}

  \item{highDensity}{logical (used when \code{sampleSheet} is
    specified). If \code{TRUE}, array extensions '\_A', '\_B' in
    sampleSheet are replaced with 'R01C01', 'R01C02' etc.}
  \item{sep}{character string specifying separator used in .idat file
    names.}
  \item{fileExt}{list containing elements 'Green' and 'Red' which
    specify the .idat file extension for the Cy3 and Cy5 channels.}
  \item{XY}{an \code{NChannelSet} object containing X and Y intensities.}
  \item{anno}{data.frame containing SNP annotation information from 
    manifest and additional columns 'isSnp', 'position', 'chromosome' 
    and 'featureNames'. For use when \code{cdfName}='nopackage'}
  \item{saveDate}{'logical'.  Should the dates from each .idat be saved
    with sample information?}
  \item{stripNorm}{'logical'.  Should the data be strip-level normalized?}
  \item{useTarget}{'logical' (only used when \code{stripNorm=TRUE}).
    Should the reference HapMap intensities be used in strip-level normalization?}
  \item{mixtureSampleSize}{ Sample size to be use when fitting the mixture model.}
  \item{fitMixture}{ 'logical.' Whether to fit per-array mixture
    model.}
  \item{quantile.method}{character string specifying the quantile normalization method to use ('within' or 'between' channels).}
  \item{eps}{   Stop criteria.}
  \item{verbose}{  'logical.'  Whether to print descriptive messages during processing.}
  \item{seed}{ Seed to be used when sampling. Useful for
    reproducibility}
  \item{cdfName}{ \code{character} string indicating which annotation
    package to load.}
}
\details{

	The normalized intensities are written to disk using package
	\code{ff} protocols for writing/reading to disk. Note that the
	object \code{CNSet} containing the \code{ff} objects in the
	\code{assayData} slot will be updated after applying this
	function.

}
\value{

	A \code{ff_matrix} object containing parameters for fitting the
	mixture model.  Note that while the \code{CNSet} object is not
	returned by this function, the object will be updated as the
	normalized intensities are written to disk.  In particular,
	after applying this function the normalized intensities in the
	\code{alleleA} and \code{alleleB} elements of \code{assayData}
	are now available.

}
\author{
R. Scharpf
}
\seealso{
	\code{\link{CNSet-class}}, \code{\link{A}}, \code{\link{B}},
	\code{\link{constructInf}}, \code{\link{genotypeInf}}, \code{\link{annotationPackages}}
}
\examples{
	## See the 'illumina_copynumber' vignette in inst/scripts of
	## the source package
}
\keyword{manip}

\name{cnSetExample}
\alias{cnSetExample}
\alias{cnSetExample2}
\docType{data}
\title{
	Object of class 'CNSet'
}
\description{

	The data for the first 16 polymorphic markers in the HapMap analysis.

}
\details{
  This object was created from the copynumber vignette in
  inst/scripts. A subset of markers was selected to keep the package
  size small.
}
\usage{

data(cnSetExample)
data(cnSetExample2)

}
\format{
  The data illustrates the \code{CNSet-class}, with
	\code{assayData} containing the quantile-normalized
	intensities for the A and B alleles, genotype calls and
	confidence scores.

}
\examples{
     data(cnSetExample)
     data(cnSetExample2)
}
\keyword{datasets}

\name{PredictionRegion-class}
\Rdversion{1.1}
\docType{class}
\alias{PredictionRegion-class}
\alias{[,PredictionRegion,ANY,ANY,ANY-method}

\title{Class \code{"PredictionRegion"}}
\description{
  A container for bivariate normal prediction regions for SNP data and
  univarite prediction regions for nonpolymorphic markers.
}
\section{Objects from the Class}{
  Objects from the class are created from the \code{predictionRegion} function.
}
\section{Slots}{
  \describe{
    \item{\code{.Data}:}{Object of class \code{"list"} ~~ }
  }
}
\section{Extends}{
Class \code{"\linkS4class{list}"}, from data part.
Class \code{"\linkS4class{vector}"}, by class "list", distance 2.
Class \code{"\linkS4class{AssayData}"}, by class "list", distance 2.
Class \code{"\linkS4class{list_or_ffdf}"}, by class "list", distance 2.
Class \code{vectorORfactor}, by class "list", distance 3.
}
\section{Methods}{
  \describe{
    \item{[}{\code{signature(x = "PredictionRegion")}: ... } Prediction
    regions can be subset by markers.
	 }
}

\author{
R. Scharpf
}


\seealso{
\code{\link{predictionRegion}}
}
\examples{
showClass("PredictionRegion")
}
\keyword{classes}

\name{crlmm-package}
\alias{crlmm-package}
\docType{package}
\title{
Genotype Calling via CRLMM Algorithm
}
\description{
Faster implementation of CRLMM specific to SNP 5.0 and 6.0 arrays.
}
\details{
Index:
\preformatted{
crlmm-package           New implementation of the CRLMM Algorithm.
crlmm                   Genotype SNP 5.0 or 6.0 samples.
calls                   Accessor for genotype calls.
confs                   Accessor for confidences.
}
The 'crlmm' package reimplements the CRLMM algorithm present in the
'oligo' package. This implementation primes for efficient genotyping of
samples on SNP 5.0 and SNP 6.0 Affymetrix arrays.

To use this package, the user must have additional data packages:
'genomewidesnp5Crlmm' - SNP 5.0 arrays
'genomewidesnp6Crlmm' - SNP 6.0 arrays
}
\author{
Rafael A Irizarry
Maintainer: Benilton S Carvalho <carvalho@bclab.org>
}
\references{
  Carvalho BS, Louis TA, Irizarry RA. Quantifying uncertainty in genotype
  calls. Bioinformatics. 2010 Jan 15;26(2):242-9. Epub 2009 Nov 11.
}
\keyword{ package }

\name{genotypeAffy}
\alias{genotypeAffy}
\title{Genotype Affymetrix CEL files}
\description{Assign diallelic genotypes at polymorphic markers}
\usage{
genotypeAffy(cnSet, SNRMin = 5, recallMin = 10, recallRegMin = 1000, gender = NULL, badSNP = 0.7, returnParams = TRUE, verbose = TRUE)
}

\arguments{
  \item{cnSet}{
    An object of class \code{CNSet}
  }
  \item{SNRMin}{
    See \code{\link{crlmm}}
  }
  \item{recallMin}{
    See \code{\link{crlmm}}
}
  \item{recallRegMin}{
    See \code{\link{crlmm}}
}
  \item{gender}{
    See \code{\link{crlmm}}
}
  \item{badSNP}{
    See \code{\link{crlmm}}
}
  \item{returnParams}{
    See \code{\link{crlmm}}
}
  \item{verbose}{
    Logical.
}
}
\details{
  Wrapper for crlmm genotyping.
}
\value{
  Returns logical.  SNP genotypes and confidence scores are written to
  \code{ff_matrix} objects.
}

\author{
R.Scharpf
}
\seealso{
  \code{\link{crlmm}}, \code{\link{calls}}, \code{\link{confs}}
}
\keyword{classif}

\name{crlmm}
\alias{crlmm}
\alias{crlmm2}

\title{Genotype oligonucleotide arrays with CRLMM}
\description{
  This is a faster and more efficient implementation of the CRLMM
  algorithm, especially designed for Affymetrix SNP 5 and 6 arrays (to
  be soon extended to other platforms).
}
\usage{
crlmm(filenames, row.names=TRUE, col.names=TRUE,
      probs=c(1/3, 1/3, 1/3), DF=6, SNRMin=5,
      gender=NULL, save.it=FALSE, load.it=FALSE,
      intensityFile, mixtureSampleSize=10^5,
      eps=0.1, verbose=TRUE, cdfName, sns, recallMin=10,
      recallRegMin=1000, returnParams=FALSE, badSNP=0.7)
crlmm2(filenames, row.names=TRUE, col.names=TRUE,
      probs=c(1/3, 1/3, 1/3), DF=6, SNRMin=5,
      gender=NULL, save.it=FALSE, load.it=FALSE,
      intensityFile, mixtureSampleSize=10^5,
      eps=0.1, verbose=TRUE, cdfName, sns, recallMin=10,
      recallRegMin=1000, returnParams=FALSE, badSNP=0.7)
}

\arguments{
  \item{filenames}{'character' vector with CEL files to be genotyped.}
  \item{row.names}{'logical'. Use rownames - SNP names?}
  \item{col.names}{'logical'. Use colnames - Sample names?}
  \item{probs}{'numeric' vector with priors for AA, AB and BB.}
  \item{DF}{'integer' with number of degrees of freedom to use with t-distribution.}
  \item{SNRMin}{'numeric' scalar defining the minimum SNR used to filter
  out samples.}
  \item{gender}{'integer' vector, with same length as 'filenames',
    defining sex. (1 - male; 2 - female)}
  \item{save.it}{'logical'. Save preprocessed data?}
  \item{load.it}{'logical'. Load preprocessed data to speed up analysis?}
  \item{intensityFile}{'character' with filename to be saved/loaded -
    preprocessed data.}
  \item{mixtureSampleSize}{Number of SNP's to be used with the mixture model.}
  \item{eps}{Minimum change for mixture model.}
  \item{verbose}{'logical'.}
  \item{cdfName}{'character' defining the CDF name to use
    ('GenomeWideSnp5', 'GenomeWideSnp6')}
  \item{sns}{'character' vector with sample names to be used.}
  \item{recallMin}{Minimum number of samples for recalibration.}
  \item{recallRegMin}{Minimum number of SNP's for regression.}
  \item{returnParams}{'logical'. Return recalibrated parameters.}
  \item{badSNP}{'numeric'. Threshold to flag as bad SNP (affects batchQC)}
}
\value{
  A \code{SnpSet} object.
  \item{calls}{Genotype calls (1 - AA, 2 - AB, 3 - BB)}
  \item{confs}{Confidence scores 'round(-1000*log2(1-p))'}
  \item{SNPQC}{SNP Quality Scores}
  \item{batchQC}{Batch Quality Score}
  \item{params}{Recalibrated parameters}
}
\details{
  'crlmm2' allows one to genotype very large datasets (via ff package) and also permits
  the use of clusters or multiple cores (via snow package) to speed up
  genotyping.

  As noted above, the call probabilities are stored using an integer
  representation to reduce file size using the transformation
  'round(-1000*log2(1-p))', where p is the probability.  The function
  \code{i2P} can be used to convert the integers back to the scale of
  probabilities.
}


\seealso{ \code{\link{i2p}}, \code{\link{snpCall}}, \code{\link{snpCallProbability}}}

\references{ Carvalho B, Bengtsson H, Speed TP,
  Irizarry RA. Exploration, normalization, and genotype calls of
  high-density oligonucleotide SNP array data. Biostatistics. 2007
  Apr;8(2):485-99. Epub 2006 Dec 22. PMID: 17189563.

  Carvalho BS, Louis TA, Irizarry RA.
  Quantifying uncertainty in genotype calls.
  Bioinformatics. 2010 Jan 15;26(2):242-9.
}
\examples{
## this can be slow
library(oligoClasses)
if (require(genomewidesnp6Crlmm) & require(hapmapsnp6)){
  path <- system.file("celFiles", package="hapmapsnp6")

  ## the filenames with full path...
  ## very useful when genotyping samples not in the working directory
  cels <- list.celfiles(path, full.names=TRUE)
  (crlmmOutput <- crlmm(cels))
  ## If gender is known, one should check that the assigned gender is
  ## correct, or pass the integer coding of gender as an argument to the
  ## crlmm function as done below
}

\dontrun{
## HPC Example
library(ff)
library(snow)
library(crlmm)
## genotype 50K SNPs at a time
ocProbesets(50000)
## setup cluster - 8 cores on the machine
library(doSNOW)
cl <- makeCluster(8, "SOCK")
registerDoSNOW(cl)
##setCluster(8, "SOCK")

path <- system.file("celFiles", package="hapmapsnp6")
cels <- list.celfiles(path, full.names=TRUE)
crlmmOutput <- crlmm2(cels)
}

}
\keyword{classif}

\name{genotypeInf}
\alias{genotypeInf}

\title{

	Genotyping of Illumina Infinium II arrays.

}
\description{

	Genotyping of Illumina Infinium II arrays. This function
	provides CRLMM/KRLMM genotypes and confidence scores for the the
	polymorphic markers and is a required step prior to copy
	number estimation.

}
\usage{
genotypeInf(cnSet, mixtureParams, probs = rep(1/3, 3), SNRMin = 5,
             recallMin = 10, recallRegMin = 1000, verbose = TRUE, returnParams = TRUE, 
             badSNP = 0.7, gender = NULL, DF = 6, cdfName, nopackage.norm="quantile",
             call.method="crlmm", trueCalls = NULL)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{cnSet}{An object of class \code{CNSet}}
  \item{mixtureParams}{
    \code{data.frame} containing mixture model parameters needed for
    genotyping.  The mixture model parameters are estimated from the
    \code{preprocessInf} function.
  }
  \item{probs}{'numeric' vector with priors for AA, AB and BB.}
  \item{SNRMin}{'numeric' scalar defining the minimum SNR used to filter
    out samples.}
  \item{recallMin}{Minimum number of samples for recalibration. }
  \item{recallRegMin}{Minimum number of SNP's for regression.}
  \item{verbose}{  'logical.'  Whether to print descriptive messages during processing.}
  \item{returnParams}{'logical'. Return recalibrated parameters from crlmm.}
  \item{badSNP}{'numeric'. Threshold to flag as bad SNP (affects
  batchQC)}
  \item{gender}{  integer vector (  male = 1, female =2 ) or missing,
  with same length as filenames.  If missing, the gender is
  predicted.}
  \item{DF}{'integer' with number of degrees of freedom to use with
    t-distribution.}
  \item{cdfName}{\code{character} string indicating which annotation
    package to load.}
  \item{nopackage.norm}{character string specifying normalization to be used when \code{cdfName}='nopackage'.
    Options are 'none', 'quantile' (within channel, between array) and 'quantileloess'.}
  \item{call.method}{character string specifying the genotype calling algorithm to use ('crlmm' or 'krlmm').}
  \item{trueCalls}{ matrix specifying known Genotype calls for a subset of samples and features(1 - AA, 2 - AB, 3 - BB).}
}

\details{

	The genotype calls and confidence scores are written to
	file using \code{ff} protocols for I/O.  For the most part,
	the calls and confidence scores can be accessed as though the
	data is in memory through the methods \code{snpCall} and
	\code{snpCallProbability}, respectively.

	The genotype calls are stored using an integer representation: 1
	- AA, 2 - AB, 3 - BB.  Similarly, the call probabilities are
	stored using an integer representation to reduce file size using
	the transformation 'round(-1000*log2(1-p))', where p is the
	probability.  The function \code{i2P} can be used to convert the
	integers back to the scale of probabilities.

	An optional \code{trueCalls} argument can be provided to KRLMM method
	which contains known genotype calls(can contain some NAs) for some 
	samples and SNPs. This will used to compute KRLMM parameters by calling
	\code{vglm} function from \code{VGAM} package.

	The KRLMM method makes use of functions provided in \code{parallel} 
	package to speed up the process. It by default initialises up to 
	8 clusters. This is configurable by setting up an option named 
	"krlmm.cores", e.g. options("krlmm.cores" = 16). 
      }

\value{

Logical.  If the genotyping is completed, the value 'TRUE' is
returned.  Note that \code{assayData} elements 'call' and
'callProbability' are updated on disk. Therefore, the genotypes and
confidence scores can be retrieved using accessors for the
\code{CNSet}  class.

}
\author{
R. Scharpf
}
\seealso{
	\code{\link{crlmm}}, \code{\link{snpCall}},
	\code{\link{snpCallProbability}}, \code{\link{annotationPackages}}
}
\examples{
	## See the 'illumina_copynumber' vignette in inst/scripts of
	## the source package
}
\keyword{classif}

\name{AssayData-methods}
\Rdversion{1.1}
\docType{methods}
\alias{Ns,AssayData-method}
\alias{corr,AssayData-method}
\alias{mads,AssayData-method}
\alias{medians,AssayData-method}
\alias{tau2,AssayData-method}

\title{Methods for class "AssayData" in crlmm}

\description{

	The \code{batchStatistics} slot in a \code{CNSet} object is an
	instance of the \code{AssayData} slot. In general, the
	accessors for \code{AssayData} are called indirectly by the
	corresponding method for the \code{CNSet} class and not called
	directly by the user.

}

\section{Methods}{
  \describe{
    \item{Ns}{\code{signature(object="AssayData")}:  Accessor for genotype
  frequencies}
    \item{corr}{\code{signature(object="AssayData")}: Accessor for the correlation of the
  log-transformed normalized intensities within the diallelic genotype clusters}
    \item{mads}{\code{signature(x="AssayData")}: Accessor for the median
  absolute deviation of the  normalized intensities within the
  diallelic genotype clusters }
    \item{medians}{\code{signature(object="AssayData")}: Accessor for
  the posterior mean of the normalized intensity within the diallelic genotype
  clusters.}
    \item{tau2}{\code{signature(object="AssayData")}: Accessor for the
  median absolute deviation of the log-transformed intensities within
  the diallelic genotype clusters}
  }
}

\seealso{
  \code{\link{CNSet-class}}, \code{\link{Ns}}, \code{\link{tau2}}, \code{\link{corr}}, \code{\link{mads}}, \code{\link{medians}}
}
\keyword{manip}

\name{ListClassConstructors}
\alias{BafLrrSetList}
\alias{OligoSetList}
\title{ Methods for Function \code{BafLrrSetList} in Package \pkg{crlmm} ~~}
\description{
  Constructors for \code{BafLrrSetList} and \code{OligoSetList} objects.
}

\usage{
BafLrrSetList(object, ...)
OligoSetList(object, ...)
}

\arguments{
  \item{object}{ A \code{CNSet} object.}
  \item{...}{Additional arguments \code{batch.name} and \code{chrom} can
  be used to specify specific batches or chromosomes in the \code{CNSet}
  object.
}
}

\details{

  Constructs a \code{BafLrrSetList} object or a \code{OligoSetList}
  object from an object of class \code{CNSet}.

}

\value{A \code{BafLrrSetList} or \code{OligoSetList}}

\examples{
data(cnSetExample)
oligoList <- OligoSetList(cnSetExample)
## only contains 1 chromosome, so list only has one element
dims(oligoList)
brList <- BafLrrSetList(cnSetExample)
dims(brList)
}

\seealso{\code{\linkS4class{BeadStudioSetList}}}


\keyword{manip}

\name{crlmmCopynumber}
\alias{crlmmCopynumber}
\alias{crlmmCopynumber2}
\alias{crlmmCopynumberLD}
\title{Locus- and allele-specific estimation of copy number}
\description{
  Locus- and allele-specific estimation of copy number.
}
\usage{
crlmmCopynumber(object, MIN.SAMPLES=10, SNRMin = 5, MIN.OBS = 1,
	        DF.PRIOR = 50, bias.adj = FALSE,
                prior.prob = rep(1/4, 4), seed = 1, verbose = TRUE,
                GT.CONF.THR = 0.80, MIN.NU = 2^3, MIN.PHI = 2^3,
                THR.NU.PHI = TRUE, type=c("SNP", "NP", "X.SNP", "X.NP"),
                fit.linearModel=TRUE)
}

\arguments{

  \item{object}{object of class \code{CNSet}.}

 \item{MIN.SAMPLES}{ 'Integer'.  The minimum number of samples in a
  batch.  Bathes with fewer than MIN.SAMPLES are skipped.  Therefore,
  samples in batches with fewer than MIN.SAMPLES have NA's for the
  allele-specific copy number and NA's for the linear model
  parameters.

}

  \item{SNRMin}{ Samples with low signal to noise ratios are  excluded.  }

  \item{MIN.OBS}{

  For a SNP with with fewer than \code{MIN.OBS} of a genotype in a given
  batch, the within-genotype median is imputed.  The imputation is based
  on a regression using SNPs for which all three biallelic genotypes are
  observed.  For example, assume at at a given SNP genotypes AA and AB
  were observed and BB is an unobserved genotype.  For SNPs in which all
  3 genotypes were observed, we fit the model E(mean_BB) = beta0 +
  beta1*mean_AA + beta2*mean_AB, obtaining estimates; of beta0, beta1,
  and beta2.  The imputed mean at the SNP with unobserved BB is then
  beta0hat + beta1hat * mean_AA of beta2hat * mean_AB.

}

  \item{DF.PRIOR}{

  The 2 x 2 covariance matrix of the background and signal variances
  is estimated from the data at each locus.  This matrix is then
  smoothed towards a common matrix estimated from all of the loci.
  DF.PRIOR controls the amount of smoothing towards the common matrix,
  with higher values corresponding to greater smoothing.  Currently,
  DF.PRIOR is not estimated from the data.  Future versions may
  estimate DF.PRIOR empirically.

}

\item{bias.adj}{

  \code{bias.adj} is currently ignored (as well as the prior.prob
  argument).  We plan to add this feature back to the crlmm package in
  the near future. This feature, when \code{TRUE}, updated initial
  estimates from the linear model after excluding samples with a low
  posterior probability of normal copy number.  Excluding samples that
  have a low posterior probability can be helpful at loci in which a
  substantial fraction of the samples have a copy number alteration.
  For additional information, see Scharpf et al., 2010.

}
  \item{prior.prob}{

    This argument is currently ignored.  A numerical vector providing
  prior probabilities for copy number states corresponding to homozygous
  deletion, hemizygous deletion, normal copy number, and amplification,
  respectively.

}
  \item{seed}{ Seed for random number generation.}

  \item{verbose}{ Logical. }

  \item{GT.CONF.THR}{

    Confidence threshold for genotype calls (0, 1).  Calls with
    confidence scores below this theshold are not used to estimate the
    within-genotype medians. See Carvalho et al., 2007 for information
    regarding confidence scores of biallelic genotypes.

}


  \item{MIN.NU}{ numeric. Minimum value for background
    intensity. Ignored if \code{THR.NU.PHI} is \code{FALSE}. }

  \item{MIN.PHI}{numeric. Minimum value for slope. Ignored if
    \code{THR.NU.PHI} is \code{FALSE}.}

  \item{THR.NU.PHI}{  If \code{THR.NU.PHI} is \code{FALSE},
    \code{MIN.NU} and \code{MIN.PHI} are ignored. When TRUE, background
    (nu) and slope (phi) coefficients below MIN.NU and MIN.PHI are set
    to MIN.NU and MIN.PHI, respectively.}

  \item{type}{ Character string vector that must be one or more of
    "SNP", "NP", "X.SNP", or "X.NP". Type refers to a set of
    markers. See details below}

  \item{fit.linearModel}{Logical.  If TRUE, a linear model is fit to
  estimate the parameters for computing the absolute copy number.  If
  FALSE, we compute the batch-specific, within-genotype median and MAD
  at polymorphic loci and the median and MAD at nonpolymorphic loci.}

 }

 \references{

  Carvalho B, Bengtsson H, Speed TP, Irizarry RA. Exploration,
  normalization, and genotype calls of high-density oligonucleotide SNP
  array data. Biostatistics. 2007 Apr;8(2):485-99. Epub 2006 Dec
  22. PMID: 17189563.

  Carvalho BS, Louis TA, Irizarry RA.
  Quantifying uncertainty in genotype calls.
  Bioinformatics. 2010 Jan 15;26(2):242-9.

  Scharpf RB, Ruczinski I, Carvalho B, Doan B, Chakravarti A, and
  Irizarry RA, Biostatistics.  Biostatistics, Epub July 2010.

}

\details{

    We suggest a minimum of 10 samples per batch for using
    \code{crlmmCopynumber}.  50 or more samples per batch is preferred
    and will improve the estimates.

    The functions \code{crlmmCopynumberLD} and
    \code{crlmmCopynumber2} have been deprecated.

	The argument \code{type} can be used to specify a subset of
	markers for which the copy number estimation algorithm is run.
	One or more of the following possible entries are valid: 'SNP',
	'NP', 'X.SNP', and 'X.NP'.

	'SNP' referers to autosomal SNPs.

	'NP' refers to autosomal nonpolymorphic markers.

	'X.SNP' refers to SNPs on chromosome X.

	'X.NP' refers to autosomes on chromosome X.

	However, users must run 'SNP' prior to running 'NP' and 'X.NP',
	or specify \code{type = c('SNP', 'X.NP')}.

      }

      \value{

	The value returned by the \code{crlmmCopynumber} function
	depends on whether the data is stored in RAM or whether the data
	is stored on disk using the R package \code{ff} for reading /
	writing.  If uncertain, the first line of the \code{show} method
	defined for \code{CNSet} objects prints whether the
	\code{assayData} elements are derived from the \code{ff} package
	in the first line.  Specifically,

	- if the elements of the \code{batchStaticts} slot in the
	\code{CNSet} object have the class "ff_matrix" or "ffdf", then
	the \code{crlmmCopynumber} function updates the data stored on
	disk and returns the value \code{TRUE}.

	- if the elements of the \code{batchStatistics} slot in the
	\code{CNSet} object have the class 'matrix', then the
	\code{crlmmCopynumber} function returns an object of class
	\code{CNSet} with the elements of \code{batchStatistics}
	updated.

      }

\author{R. Scharpf}
\keyword{manip}

\name{copynumberAccessors}
\alias{CA}
\alias{CB}
\alias{nuA}
\alias{nuB}
\alias{phiA}
\alias{phiB}
\alias{totalCopynumber}
\alias{rawCopynumber}

\title{

	Accessors for allele-specific or total copy number

}


\description{

These methods can be applied after an object of class \code{CNSet} has
been generated by the \code{crlmmCopynumber} function.

}
\usage{
CA(object, ...)
CB(object, ...)
nuA(object)
nuB(object)
phiA(object)
phiB(object)
totalCopynumber(object,...)
rawCopynumber(object,...)
}

\arguments{
  \item{object}{  An object of class \code{CNSet}.}
  \item{\dots}{
  An additional argument named 'i' can be passed to subset the markers
  and an argument 'j' can be passed to subset the samples.  Other
  arguments are ignored.
  }
}

\details{

	At polymorphic markers, nuA and nuB provide the intercept
	coefficient (the estimated background intensity) for the A and
	B alleles, respectively. phiA and phiB provide the slope
	coefficients for the A and B alleles, respectively.

	At nonpolymorphic markers, nuB and phiB are 'NA'.

	These functions can be used to tranlate the normalized
	intensities to the copy number scale.  Plotting the copy
	number estimates as a function of physical position can be
	used to guide downstream algorithms that smooth, as well as to
	assess possible mosaicism.


}

\value{

	nu[A/B] and phi[A/B] return matrices of the intercept and
	slope coefficients, respectively.

	CA and CB return matrices of allele-specific copy number.

	totalCopynumber (or rawCopynumber) returns a matrix of CA+CB.

      }

\note{

Subsetting the \code{CNSet} object before extracting copy number can be
very inefficient when the data set is very large, particularly if using
ff objects.  The \code{[} method will subset all of the assay data
elements and all of the elements in the LinearModelParameter slot.

}


\seealso{

	\code{\link{crlmmCopynumber}}, \code{\link{CNSet-class}}

}

\examples{
\dontrun{
data(cnSetExample)
all(isCurrent(cnSetExample)) ## is the cnSet object current?

## --------------------------------------------------
## calculating allele-specific copy number
## --------------------------------------------------
## copy number for allele A, first 5 markers, first 2 samples
(ca <- CA(cnSetExample, i=1:5, j=1:2))
## copy number for allele B, first 5 markers, first 2 samples
(cb <- CB(cnSetExample, i=1:5, j=1:2))
## total copy number for first 5 markers, first 2 samples
(cn1 <- ca+cb)

## total copy number at first 5 nonpolymorphic loci
index <- which(!isSnp(cnSetExample))[1:5]
cn2 <- CA(cnSetExample, i=index, j=1:2)
## note, cb is NA at nonpolymorphic loci
(cb <- CB(cnSetExample, i=index, j=1:2))
## note, ca+cb will give NAs at nonpolymorphic loci
CA(cnSetExample, i=index, j=1:2) + cb
## A shortcut for total copy number
cn3 <- totalCopynumber(cnSetExample, i=1:5, j=1:2)
all.equal(cn3, cn1)
cn4 <- totalCopynumber(cnSetExample, i=index, j=1:2)
all.equal(cn4, cn2)

## markers 1-5, all samples
cn5 <- totalCopynumber(cnSetExample, i=1:5)
## all markers, samples 1-5
cn6 <- totalCopynumber(cnSetExample, j=1:2)
}
}
\keyword{manip}

\name{constructAffyCNSet}
\alias{constructAffyCNSet}

\title{ Construct an object of class CNSet from Affymetrix cel files
}
\description{
Construct a container for normalized intensities for Affymetrix cel
files, referred to as a \code{CNSet}
}
\usage{
constructAffyCNSet(filenames, sns, cdfName, batch, verbose = TRUE, genome)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{filenames}{
    Vector of cel file names.
}
  \item{sns}{
    Sample identifiers. Defaults to \code{basename(filenames)}.
}
  \item{cdfName}{
    Character string indicating annotation package (e.g., "genomewidesnp6Crlmm")
}
  \item{batch}{
    Vector of same length as filenames indicating batch.
}
  \item{verbose}{
    Logical.
}
  \item{genome}{
    Character string indicating UCSC genome build (hg18 or hg19 supported)
}
}

\value{
  An object of class \code{CNSet}
}

\author{
R. Scharpf
}
% Add one or more standard keywords, see file 'KEYWORDS' in the
% R documentation directory.
\keyword{manip}

\name{xyplot}
\alias{xyplot}
\alias{xyplot,formula,CNSet-method}
\title{Plot prediction regions and normalized intensities.}
\description{
  Plot prediction regions for integer copy number and normalized intensities.
}
\usage{
xyplot(x, data, ...)
}

\arguments{
  \item{x}{
    A \code{formula}.
  }
  \item{data}{
    A \code{CNSet} object.
  }
  \item{\dots}{
    Additional arguments passed to \code{xyplot} function in lattice.
  }
}

\value{
  A \code{trellis} object.
}

\author{
R. Scharpf
}

\seealso{
  \code{\link[lattice]{xyplot}}, \code{\link{ABpanel}}
}

\examples{
library(oligoClasses)
data(cnSetExample2)
table(batch(cnSetExample2))
sample.index <- which(batch(cnSetExample2) == "CUPID")
## A single SNP
pr <- predictionRegion(cnSetExample2[1:4, sample.index], copyNumber=0:4)
gt <- calls(cnSetExample2[1:4, sample.index])
lim <- c(6,13)
xyplot(B~A|snpid, data=cnSetExample2[1:4, sample.index],
       predictRegion=pr,
       panel=ABpanel,
       pch=21,
       fill=c("red", "blue", "green3")[gt],
       xlim=lim, ylim=lim)

## multiple SNPs, prediction regions for 3 batches
\dontrun{
	tab <- table(batch(cnSetExample2))
	bns <- names(tab)[tab > 50]
	sample.index <- which(batch(cnSetExample2) %in% bns[1:3])
	pr <- predictionRegion(cnSetExample2[1:10, sample.index], copyNumber=0:4)
	gt <- as.integer(calls(cnSetExample2[1:10, sample.index]))
	xyplot(B~A|snpid, data=cnSetExample2[1:10, sample.index],
	       predictRegion=pr,
	       panel=ABpanel,
	       pch=21,
	       fill=c("red", "blue", "green3")[gt],
	       xlim=c(6,12), ylim=c(6,12))

	## nonpolymorphic markers
	data(cnSetExample2)
	tab <- table(batch(cnSetExample2))
	bns <- names(tab)[tab > 50]
	sample.index <- which(batch(cnSetExample2)%in%bns[1:3])
	np.index <- which(!isSnp(cnSetExample2))[1:10]
	taus <- tau2(cnSetExample)[np.index, , , ]
	pr <- predictionRegion(cnSetExample2[np.index, sample.index],
			       copyNumber=0:4)
	pp <- posteriorProbability(cnSetExample2[np.index, sample.index],
				   predictRegion=pr,
				   copyNumber=0:4)
}
}
\keyword{dplot}
\keyword{hplot}

\name{posteriorProbability}
\alias{posteriorProbability}
\alias{posteriorProbability,CNSet-method}
%- Also NEED an '\alias' for EACH other topic documented here.
\title{Calculate the posterior probability for integer copy numbers.}

\description{
  Calculate the posterior probability for integer copy numbers using the
  bivariate normal prediction regions.
}

\usage{
posteriorProbability(object, predictRegion, copyNumber = 0:4, w)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{object}{
    A \code{CNSet} object.
}
  \item{predictRegion}{
    A list containing the bivariate normal prediction region for each of
    the possible genotypes.
}
  \item{copyNumber}{
    Integer vector.
  }

  \item{w}{ numeric vector of prior probabilities for each of the copy
    number states.  Must be the same length as \code{copyNumber} and sum
    to 1.}

}

\details{
This is currently under development.
}

\value{
An array (features x samples x copy number)
}

\author{
R. Scharpf
}

\note{
  This is under development. Use at your own risk.
}

%% ~Make other sections like Warning with \section{Warning }{....} ~

\seealso{
  \code{\link{predictionRegion}}, \code{\link{genotypes}} %\code{\link{calculatePosteriorMean}}
}
\examples{
data(cnSetExample)
pr <- predictionRegion(cnSetExample, copyNumber=0:4)
pp <- posteriorProbability(cnSetExample, predictRegion=pr)
dim(pp)

## multiple batches
data(cnSetExample2)
pr <- predictionRegion(cnSetExample2, copyNumber=0:4)
pp <- posteriorProbability(cnSetExample2, predictRegion=pr)
}
% Add one or more standard keywords, see file 'KEYWORDS' in the
% R documentation directory.
\keyword{array}

\name{constructInf}
\alias{constructInf}
\title{
	Instantiate an object of class CNSet for the Infinium platforms.
}
\description{
	Instantiates an object of class CNSet for the Infinium
	platforms.  Elements of \code{assayData} and
	\code{batchStatistics} will be \code{ff} objects. See details.
}
\usage{
constructInf(sampleSheet = NULL, arrayNames = NULL, path = ".",
       arrayInfoColNames = list(barcode="SentrixBarcode_A",position="SentrixPosition_A"), highDensity = FALSE, sep = "_",
       fileExt = list(green = "Grn.idat", red = "Red.idat"), XY, cdfName, anno, genome, verbose = FALSE, batch=NULL, saveDate = TRUE)
}
\arguments{
  \item{sampleSheet}{\code{data.frame} containing Illumina sample sheet
    information (for required columns, refer to BeadStudio Genotyping
    guide - Appendix A).
}
  \item{arrayNames}{character vector containing names of arrays to be
    read in.  If \code{NULL}, all arrays that can be found in the
    specified working directory will be read in.}
  \item{path}{character string specifying the location of files to be
    read by the function}
  \item{arrayInfoColNames}{(used when \code{sampleSheet} is specified)
    list containing elements 'barcode' which indicates column names in
    the \code{sampleSheet} which contains the arrayNumber/barcode number
    and 'position' which indicates the strip number.  In older style
    sample sheets, this information is combined (usually in a column
    named 'SentrixPosition') and this should be specified as
    \code{list(barcode=NULL, position="SentrixPosition")}}
  \item{highDensity}{logical (used when \code{sampleSheet} is
    specified). If \code{TRUE}, array extensions '\_A', '\_B' in
    sampleSheet are replaced with 'R01C01', 'R01C02' etc.}
  \item{sep}{character string specifying separator used in .idat file
    names.}
  \item{fileExt}{list containing elements 'Green' and 'Red' which
    specify the .idat file extension for the Cy3 and Cy5 channels.}
  \item{XY}{an \code{NChannelSet} containing X and Y intensities.}
  \item{cdfName}{annotation package (see also \code{validCdfNames}) or 
    'nopackage' when an \code{anno} data.frame and \code{genome} supplied}
  \item{anno}{data.frame containing SNP annotation information from 
    manifest and additional columns 'isSnp', 'position', 'chromosome' 
    and 'featureNames'. For use when \code{cdfName}='nopackage'}
  \item{genome}{character string specifying which genome is used in annotation}
  \item{verbose}{  'logical.'  Whether to print descriptive messages
  during processing.}
 \item{batch}{ batch variable. See details.}
  \item{saveDate}{'logical'.  Should the dates from each .idat be saved
    with sample information?}
}
\details{

	This function initializes a container for storing the normalized
	intensities for the A and B alleles at polymorphic loci and the
	normalized intensities for the 'A' allele at nonpolymorphic
	loci. CRLMM genotype calls and confidence scores are also stored
	in the assayData. This function does not do any preprocessing or
	genotyping -- it only creates an object of the appropriate size.
	The initialized values will all be 'NA'.

	The ff package provides infrastructure for accessing and writing
	data to disk instead of keeping data in memory.  Each element of
	the \code{assayData} and \code{batchStatistics} slot are ff
	objects.  ff objects in the R workspace contain pointers to
	several files with the '.ff' extension on disk.  The location of
	where the data is stored on disk can be specified by use of the
	\code{ldPath} function.  Users should not move or rename this
	directory.  If only output files are stored in \code{ldPath},
	one can either remove the entire directory prior to rerunning
	the analysis or all of the '.ff' files.  Otherwise, one would
	accumulate a large number of '.ff' files on disk that are no
	longer in use.

	We have adopted the \code{ff} package in order to reduce crlmm's
	memory footprint. The memory usage can be fine-tuned by the
	utilities \code{ocSamples} and \code{ocProbesets} provided in
	the \code{oligoClasses} package. In most instances, the
	user-level interface will be no different than accessing data
	from ordinary matrices in R.  However, the differences in the
	underlying representation can become more noticeable for very
	large datasets in which the I/O for accessing data from the disk
	can be substantial.

}
\value{

	A \code{CNSet} object

}
\author{
R. Scharpf
}
\seealso{
	\code{\link{ldPath}}, \code{\link{ocSamples}}, \code{\link{ocProbesets}}, \code{\link{CNSet-class}}, \code{\link{preprocessInf}}, \code{\link{genotypeInf}}
}
\examples{
## See the Illumina vignettes in inst/scripts of the
## source package for an example
}
\keyword{manip}

\name{snprma}
\Rdversion{1.1}
\alias{snprma}
\alias{snprma2}

\title{
  Preprocessing tool for SNP arrays.
}
\description{
  SNPRMA will preprocess SNP chips. The preprocessing consists of
  quantile normalization to a known target distribution and
  summarization to the SNP-Allele level.
}
\usage{
snprma(filenames, mixtureSampleSize = 10^5, fitMixture = FALSE, eps = 0.1, verbose = TRUE, seed = 1, cdfName, sns)
snprma2(filenames, mixtureSampleSize = 10^5, fitMixture = FALSE, eps = 0.1, verbose = TRUE, seed = 1, cdfName, sns)
}
\arguments{
  \item{filenames}{
    'character' vector with file names.
}
  \item{mixtureSampleSize}{
    Sample size to be use when fitting the mixture model.
}
  \item{fitMixture}{
    'logical'. Fit the mixture model?
}
  \item{eps}{
    Stop criteria.
}
  \item{verbose}{
    'logical'.
}
  \item{seed}{
    Seed to be used when sampling.
}
  \item{cdfName}{
    cdfName: 'GenomeWideSnp\_5', 'GenomeWideSnp\_6'
}
  \item{sns}{
    Sample names.
}
}
\value{
  \item{A}{Summarized intensities for Allele A}
  \item{B}{Summarized intensities for Allele B}
  \item{sns}{Sample names}
  \item{gns}{SNP names}
  \item{SNR}{Signal-to-noise ratio}
  \item{SKW}{Skewness}
  \item{mixtureParams}{Parameters from mixture model}
  \item{cdfName}{Name of the CDF}
}
\details{
  'snprma2' allows one to genotype very large datasets (via ff package) and also permits
  the use of clusters or multiple cores (via snow package) to speed up preprocessing.
  }
\examples{
if (require(genomewidesnp6Crlmm) & require(hapmapsnp6) & require(oligoClasses)){
  path <- system.file("celFiles", package="hapmapsnp6")

  ## the filenames with full path...
  ## very useful when genotyping samples not in the working directory
  cels <- list.celfiles(path, full.names=TRUE)
  snprmaOutput <- snprma(cels)
  snprmaOutput[["A"]][1:10,]
  snprmaOutput[["B"]][1:10,]
}
\dontrun{
## HPC Example
library(ff)
library(snow)
library(crlmm)
## genotype 50K SNPs at a time
ocProbesets(50000)
## setup cluster - 8 cores on the machine
setCluster(8, "SOCK")

path <- system.file("celFiles", package="hapmapsnp6")
cels <- list.celfiles(path, full.names=TRUE)
snprmaOutput <- snprma2(cels)

}

}
\keyword{manip}
\keyword{classif}

\name{predictionRegion}
\alias{predictionRegion}
\alias{predictionRegion,CNSet,integer-method}
\title{Prediction regions for integer copy number}
\description{Bivariate normal prediction regions for integer copy
  number.  Copy numbers 0-4 allowed.}
\usage{
predictionRegion(object, copyNumber)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{object}{A \code{CNSet} object.}
  \item{copyNumber}{Integer vector.  0-4 allowed.}
}
\details{

  We fit a linear regression for each allele to the diallic genotype
  cluster medians.  Denoting the background and slope by nu and phi,
  respectively, the mean for the bivariate normal prediction region is
  given by

  mu_A = nu_A + CA * phi_A

  and

  mu_B nu_B + CB * phi_B

  The variance and correlation of the normalized intensities is
  estimated from the diallelic genotype clusters AA, AB, and BB on the
  log-scale.  For copy number not equal to two, we assume that the
  variance is approximately the same for copy number not equal to 2.

}

\value{

  A list named by the genotype.  `NULL' refers to copy number zero, `A'
  is a hemizygous deletion, etc. Each element is a list of the means
  (mu) and covariance (cov) for each marker stored as an array.  For
  `mu', the dimensions of the array are marker x allele (A or B) x
  batch.  For `cov', the dimensions of the array are marker x 3
  (varA, cor, and varB) x batch.

}


\references{
  Scharpf et al., 2011, Biostatistics.
}
\author{
  R. Scharpf
}

%% ~Make other sections like Warning with \section{Warning }{....} ~

\seealso{
%  \code{\link{calculatePosteriorMean}},
  \code{\link{posteriorProbability}}, \code{\link{genotypes}}
}
\examples{
data(cnSetExample)
pr <- predictionRegion(cnSetExample, copyNumber=0:4)
names(pr)
## bivariate normal prediction region for NULL genotype (homozygous deletion)
str(pr[["NULL"]])
}
% Add one or more standard keywords, see file 'KEYWORDS' in the
% R documentation directory.
\keyword{distribution}
\keyword{list}

\name{cnrmaAffy}
\alias{cnrmaAffy}
\title{quantile normalize nonpolymorphic markers}
\description{Quantile normalize nonpolymorphic markers to hapmap reference distribution}
\usage{
cnrmaAffy(cnSet, seed = 1, verbose = TRUE)
}

\arguments{
  \item{cnSet}{
    Object of class \code{CNSet}
  }
  \item{seed}{
    Random number seed
  }
  \item{verbose}{
    Logical.
  }
}
\value{
  Returns logical.  Normalized intensities are written to the alleleA
  \code{ff_matrix} stored in the \code{CNSet} assayData.
}
\author{ R. Scharpf}

\seealso{\code{\link{snprmaAffy}}}
% Add one or more standard keywords, see file 'KEYWORDS' in the
% R documentation directory.
\keyword{robust}

\name{batchStatisticAccessors}
\alias{Ns}
\alias{corr}
\alias{tau2}
\alias{mads}
\alias{medians}

\title{

	Accessors for batch-specific summary statistics.

}

\description{

	The summary statistics stored here are used by the tools for
	copy number estimation.

}

\usage{
corr(object, ...)
tau2(object, ...)
mads(object,...)
medians(object,...)
Ns(object,...)
}

\arguments{
  \item{object}{  An object of class \code{CNSet}.}
  \item{\dots}{Ignored}
}

\value{

An array with dimension R x A x G x C, or R x G x C.

R: number of markers
A: number of alleles (2)
G: number of biallelic genotypes (3)
C: number of batches

\code{Ns} returns an array of genotype frequencies stratified by
batch.  Dimension R x G x C.

\code{corr} returns an array of within-genotype correlations
(log2-scale) stratified by batch. Dimension R x G x C.

\code{medians} returns an array of the within-genotype medians
(intensity-scale) stratified by batch and allele. Dimension R x A x G
x C.

\code{mads} returns an array of the within-genotype median absolute
deviations (intensity-scale) stratified by batch and allele. Dimension
is the same as for \code{medians}.

\code{tau2} returns an array of the squared within-genotype median
absolute deviation on the log-scale.  Only the mads for AA and BB
genotypes are stored.  Dimension is R x A x G x C, where G is AA or
BB.  Note that the mad for allele A/B for subjects with genotype BB/AA
is a robust estimate of the background variance, whereas the the mad
for allele A/B for subjects with genotype AA/BB is a robust estimate
of the variance for copy number greater than 0 (we assume that on the
log-scale the variance is rougly constant for CA, CB > 0).

}


\seealso{

	\code{\link{batchStatistics}}

}

\examples{
data(cnSetExample)
Ns(cnSetExample)[1:5, , ]
corr(cnSetExample)[1:5, , ]
meds <- medians(cnSetExample)
mads(cnSetExample)[1:5, , ,]
tau2(cnSetExample)[1:5, , ,]
}
\keyword{manip}

\name{ABpanel}
\alias{ABpanel}
\title{A panel function for plotting prediction regions and
  log-normalized intensities
}
\description{
  A panel function for plotting prediction regions and
  log-normalized intensities
}
\usage{
ABpanel(x, y, predictRegion, copyNumber = 0:4, fill, ..., subscripts)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{x}{
    log-normalized intensities for the A or B allele
}
\item{y}{
      log-normalized intensities for the A or B allele
}
  \item{predictRegion}{
    A \code{list}. See \code{predictionRegion}.
}
  \item{copyNumber}{
    Integer vector.  Indicates which prediction regions are drawn.
}
  \item{fill}{
    Character or integer vector for coloring the points.  Only valid for
    certain point symbols.  See \code{points}.
}
  \item{\dots}{
    Additional arguments to \code{panel.xyplot} and \code{lpolygon}.
}
  \item{subscripts}{
    See \code{xyplot} in the \pkg{lattice} package.
}
}

\value{
  Not applicable
}

\note{

  \code{ABpanel} can be passed as the argument to panel in the
  \code{xyplot} method for \code{CNSet} objects.  See the examples in
  \code{xyplot}.

}

\author{
R. Scharpf
}

\seealso{
  \code{\link[lattice]{xyplot}},   \code{\link[lattice]{panel.xyplot}}
  \code{\link[lattice]{lpolygon}}
}

% Add one or more standard keywords, see file 'KEYWORDS' in the
% R documentation directory.
\keyword{aplot}

\name{CNSet-methods}
\Rdversion{1.1}
\docType{methods}
\alias{coerce,CNSet,oligoSnpSet-method}
\alias{CA,CNSet-method}
\alias{CB,CNSet-method}
\alias{lines,CNSet-method}
\alias{totalCopynumber,CNSet-method}
\alias{rawCopynumber,CNSet-method}
\alias{nuA,CNSet-method}
\alias{nuB,CNSet-method}
\alias{Ns,CNSet-method}
\alias{corr,CNSet-method}
\alias{mads,CNSet-method}
\alias{medians,CNSet-method}
\alias{phiA,CNSet-method}
\alias{phiB,CNSet-method}
\alias{tau2,CNSet-method}
\alias{OligoSetList,CNSet-method}
\alias{BafLrrSetList,CNSet-method}

\title{crlmm methods for class "CNSet"}
\description{

	CNSet is a container defined in the oligoClasses package for
	 storing normalized intensities for genotyping platforms,
	 genotype calls, and parameters estimated for copy
	 number. Accessors for data that an object of this class
	 contains are largely defined in the package
	 oligoClasses. CNSet methods that involve more complex
	 calculations that are specific to the crlmm package, such as
	 computing allele-specific copy number, are included in crlmm
	 and described here.

}

\section{Methods}{
  \describe{
    \item{}{\code{as(from, "oligoSnpSet")}: Method for coercing object
      \code{from} (class \code{CNSet}) to an object of class
      \code{oligoSnpSet}.}
    \item{CA}{\code{signature(object="CNSet")}: calculates raw copy number
      for allele A}
    \item{CB}{\code{signature(object="CNSet")}: calculates raw copy
  number for allele B}
    \item{lines}{\code{signature(x="CNSet")}: plot ellipses (95th
  percentile) for prediction regions}
    \item{totalCopynumber}{\code{signature(object="CNSet")}: calculates
  total raw copy number}
    \item{rawCopynumber}{\code{signature(object="CNSet")}: same as totalCopynumber}
    \item{nuA}{\code{signature(object="CNSet")}: estimate of mean
  background (intensity-scale) for allele A}
    \item{nuB}{\code{signature(object="CNSet")}: estimate of mean
  background (intensity-scale) for allele A}
    \item{phiA}{\code{signature(object="CNSet")}: estimate of slope coefficient (intensity-scale) for allele A}
    \item{phiB}{\code{signature(object="CNSet")}: estimate of slope coefficient (intensity-scale) for allele B}
    \item{Ns}{\code{signature(object="CNSet")}: genotype frequencies }
    \item{corr}{\code{signature(object="CNSet")}: correlation of
  log-transformed normalized intensities within the genotype clusters}
    \item{mads}{\code{signature(x="CNSet")}: ...}
    \item{medians}{\code{signature(object="CNSet")}: ...}
    \item{tau2}{\code{signature(object="CNSet")}: ...}


  \item{}{\code{OligoSetList(object)}: constructs an object of class
    \code{OligoSetList} from \code{object} having class \code{CNSet}.
  }

  \item{}{\code{BafLrrSetList(object)}: constructs an object of class
    \code{BafLrrSetList} from \code{object} having class \code{CNSet}.
  }

  }
}

\seealso{
  \code{\link{CNSet-class}}, \code{\link{CA}}, \code{\link{CB}},
  \code{\link{totalCopynumber}}, \code{\link{rawCopynumber}}
}
\keyword{methods}

\name{readGenCallOutput}
\alias{readGenCallOutput}

\title{Read X and Y intensities from GenCall output}

\description{
  This function reads the raw X and Y intensities output by GenomeStudio's GenCall genotyping module in preparation for genotyping with \code{crlmm}.
}

\usage{
readGenCallOutput(filenames, path=".", cdfName, colnames=list("SampleID"="Sample ID", "SNPID"="SNP Name", "XRaw"="X Raw", "YRaw"="Y Raw"),
                  type=list("SampleID"="character", "SNPID"="character", "XRaw"="integer", "YRaw"="integer"), verbose=FALSE)
}

\arguments{
  \item{filenames}{'character' string, or a vector of character string specifying the name of the file(s) to read in}
  \item{path}{'character' string specifying the location of file to be read by the function}
  \item{cdfName}{'character' defining the chip annotation (manifest) to use
    ('human370v1c', human550v3b', 'human650v3a', 'human1mv1c',
    'human370quadv3c', 'human610quadv1b', 'human660quadv1a',
    'human1mduov3b', 'humanomni1quadv1b', 'humanomniexpress12v1b', 'humancytosnp12v2p1h')}
  \item{colnames}{list containing elements 'SampleID', 'SNPID', 'XRaw' and 'YRaw', which specify the column names from
     in 'file' that pertain to these variables.  The default should suffice in most situations.}
  \item{type}{list containing data types for the columns to be read in.  The default should be fine in most situations.}
  \item{verbose}{'logical'.  Should processing information be displayed as data is read in?}
}

\details{
This function returns an \code{NChannelSet} containing raw intensity data (X and Y) from GenCall final report file. 
It assumes the GenCall output is formatted to have samples listed one below the other, and that the columns 'X Raw' and 'Y Raw' are available in the file.
The function crlmmillumina() can be run on the output of the \code{readGenCallOutput} function. 
}

\value{
\code{NChannelSet} containing X and Y bead intensities.
}

\references{
  Ritchie ME, Carvalho BS, Hetrick KN, Tavar\'{e} S, Irizarry RA.
  R/Bioconductor software for Illumina's Infinium whole-genome 
  genotyping BeadChips. Bioinformatics. 2009 Oct 1;25(19):2621-3.
}

\author{Cynthia Liu, Matt Ritchie, Zhiyin Dai}

\examples{
#XY = readGenCallOutput(file="Hap650Yv3_Final_Report.txt", cdfName="human650v3a")
#crlmmOut = crlmmIllumina(XY=XY)
}

\keyword{IO}

\name{readIdatFiles}
\alias{readIdatFiles}
\alias{readIdatFiles2}

\title{Reads Idat Files from Infinium II Illumina BeadChips}

\description{
  Reads intensity information for each bead type from
  .idat files of Infinium II genotyping BeadChips}

\usage{
readIdatFiles(sampleSheet=NULL, arrayNames=NULL, ids=NULL, path="",
              arrayInfoColNames=list(barcode="SentrixBarcode_A",
                                     position="SentrixPosition_A"),
              highDensity=FALSE, sep="_",
              fileExt=list(green="Grn.idat", red="Red.idat"),
              saveDate=FALSE, verbose=FALSE)
}

\arguments{
  \item{sampleSheet}{\code{data.frame} containing Illumina sample sheet
    information (for required columns, refer to BeadStudio Genotyping
    guide - Appendix A).}
  \item{arrayNames}{character vector containing names of arrays to be
    read in.  If \code{NULL}, all arrays that can be found in the
    specified working directory will be read in.}
  \item{ids}{vector containing ids of probes to be read in.  If
    \code{NULL} all probes found on the first array are read in.}
  \item{path}{character string specifying the location of files to be
    read by the function}
  \item{arrayInfoColNames}{(used when \code{sampleSheet} is specified)
    list containing elements 'barcode' which indicates column names in
    the \code{sampleSheet} which contains the arrayNumber/barcode number
    and 'position' which indicates the strip number.  In older style
    sample sheets, this information is combined (usually in a column
    named 'SentrixPosition') and this should be specified as
    \code{list(barcode=NULL, position="SentrixPosition")}}
  \item{highDensity}{logical (used when \code{sampleSheet} is
    specified). If \code{TRUE}, array extensions '\_A', '\_B' in
    sampleSheet are replaced with 'R01C01', 'R01C02' etc.}
  \item{sep}{character string specifying separator used in .idat file
    names.}
  \item{fileExt}{list containing elements 'Green' and 'Red' which
    specify the .idat file extension for the Cy3 and Cy5 channels.}
  \item{saveDate}{logical.  Should the dates from each .idat be saved
    with sample information?}
  \item{verbose}{logical.  Should processing information be displayed as data is read in?}
}

\details{
The summarised Cy3 (G) and Cy5 (R) intensities (on the orginal scale)
are read in from the .idat files.

Where available, a \code{sampleSheet} data.frame, in the same format
as used by BeadStudio (columns 'Sample\_ID', 'SentrixBarcode\_A' and
'SentrixPosition\_A' are required) which keeps track of sample
information can be specified.

Thanks to Keith Baggerly who provided the code to read in the binary .idat files.
}

\value{
  NChannelSet with intensity data (\code{R}, \code{G}), and indicator 
  for SNPs with 0 beads (\code{zero}) for each bead type.
}

\references{
  Ritchie ME, Carvalho BS, Hetrick KN, Tavar\'{e} S, Irizarry RA.
  R/Bioconductor software for Illumina's Infinium whole-genome 
  genotyping BeadChips. Bioinformatics. 2009 Oct 1;25(19):2621-3.
}

\author{Matt Ritchie}

\examples{

#RG = readIdatFiles()

}
\keyword{IO}

\name{calculateRBaf}
\alias{calculateRBaf}
\alias{calculateRBaf,CNSet-method}
\title{Calculate log R ratios and B allele frequencies.}
\description{Calculate log R ratios and B allele frequencies from
  a \code{CNSet} object}
\usage{
calculateRBaf(object, batch.name, chrom)
}

\arguments{

  \item{object}{A \code{CNSet} object.}

  \item{batch.name}{A character string indicating the batch.  If
  missing, log R ratios and B allele frequencies are calculated for all
  batches in the \code{object}.}

  \item{chrom}{Integer indicating which chromosome to process. If
  missing, B allele frequencies and log R ratios are calculated for all
  autosomal chromosomes and chromosome X that are included in
  \code{object}.}

}

\details{

  \code{batch.name} must be a value in \code{batch(object)}. Currently,
  one must specify a single \code{batch.name}. If a character vector for
  \code{batch.name} is supplied, only the first is evaluated.

  TODO: A description of how these values are calculated.

}

\value{
  A named list.

  \code{baf}: Each element in the baf list is a matrix of B allele
  frequencies (one matrix for each chromosome).

  \code{lrr}: Each element in the lrr list is a matrix of log R ratios
  (one matrix for each chromosome).

  The log R ratios were scaled by a factor of 100 and stored as an
  integer.  B allele frequencies were scaled by a factor of 1000 and
  stored as an integer.

}

\references{

  Peiffer et al., High-resolution genomic profiling of chromosomal aberrations using
  Infinium whole-genome genotyping (2006), Genome Research

}

\author{Lynn Mireless}

\examples{
	data(cnSetExample)
	baf.lrr <- suppressWarnings(calculateRBaf(cnSetExample, "SHELF"))
	hist(baf.lrr[["baf"]][[1]]/1000, breaks=100)
	hist(baf.lrr[["lrr"]][[1]]/100, breaks=100)
	\dontrun{
		library(ff)
		baf.lrr <- suppressWarnings(calculateRBaf(cnSetExample, "SHELF"))
		class(baf.lrr[["baf"]][[1]]) ## ff_matrix
		class(baf.lrr[["lrr"]][[1]]) ## ff_matrix
	}
}
\keyword{list}
\keyword{methods}

\name{snprmaAffy}
\alias{snprmaAffy}
\title{Quantile normalize intensities for SNPs}
\description{Quantile normalize intensities for SNPs to a HapMap target
  reference distribution}
\usage{
snprmaAffy(cnSet, mixtureSampleSize = 10^5, eps = 0.1, seed = 1, verbose = TRUE)
}

\arguments{
  \item{cnSet}{Object of class \code{CNSet}}
  \item{mixtureSampleSize}{Sample size to be use when fitting the
  mixture model. }
  \item{eps}{Stop criteria.}
  \item{seed}{Seed to be used when sampling. }
  \item{verbose}{Logical.}
}

\value{
  Returns nothing.  Normalized intensities are written to files.
}
\author{
R.Scharpf
}

\seealso{
  \code{\link{snprma}}
}
% Add one or more standard keywords, see file 'KEYWORDS' in the
% R documentation directory.
\keyword{robust}

\name{genotype.Illumina}
\alias{genotype.Illumina}
\alias{crlmmIllumina}

\title{
	Preprocessing and genotyping of Illumina Infinium II arrays.
}
\description{
	Preprocessing and genotyping of Illumina Infinium II arrays.
}
\usage{
genotype.Illumina(sampleSheet=NULL, arrayNames=NULL, ids=NULL, path=".",
      arrayInfoColNames=list(barcode="SentrixBarcode_A", position="SentrixPosition_A"),
      highDensity=FALSE, sep="_", fileExt=list(green="Grn.idat", red="Red.idat"), XY=NULL, anno, genome, 
      call.method="crlmm", trueCalls=NULL, cdfName, copynumber=TRUE, batch=NULL, saveDate=FALSE, stripNorm=TRUE, 
      useTarget=TRUE, quantile.method="between", nopackage.norm="quantile", mixtureSampleSize=10^5, fitMixture=TRUE,
      eps=0.1, verbose = TRUE, seed = 1, sns, probs = rep(1/3, 3), DF = 6, SNRMin = 5, 
      recallMin = 10, recallRegMin = 1000, gender = NULL, returnParams = TRUE, badSNP = 0.7)

crlmmIllumina(sampleSheet=NULL, arrayNames=NULL, ids=NULL, path=".",
      arrayInfoColNames=list(barcode="SentrixBarcode_A", position="SentrixPosition_A"),
      highDensity=FALSE, sep="_", fileExt=list(green="Grn.idat", red="Red.idat"), XY=NULL, anno, genome, 
      call.method="crlmm", trueCalls=NULL, cdfName, copynumber=TRUE, batch=NULL, saveDate=FALSE, stripNorm=TRUE, 
      useTarget=TRUE, quantile.method="between", nopackage.norm="quantile", mixtureSampleSize=10^5, fitMixture=TRUE,
      eps=0.1, verbose = TRUE, seed = 1, sns, probs = rep(1/3, 3), DF = 6, SNRMin = 5, 
      recallMin = 10, recallRegMin = 1000, gender = NULL, returnParams = TRUE, badSNP = 0.7)

}
\arguments{
  \item{sampleSheet}{\code{data.frame} containing Illumina sample sheet
    information (for required columns, refer to BeadStudio Genotyping
    guide - Appendix A).}
  \item{arrayNames}{character vector containing names of arrays to be
    read in.  If \code{NULL}, all arrays that can be found in the
    specified working directory will be read in.}
  \item{ids}{vector containing ids of probes to be read in.  If
    \code{NULL} all probes found on the first array are read in.}
  \item{path}{character string specifying the location of files to be
    read by the function}
  \item{arrayInfoColNames}{(used when \code{sampleSheet} is specified)
    list containing elements 'barcode' which indicates column names in
    the \code{sampleSheet} which contains the arrayNumber/barcode number
    and 'position' which indicates the strip number.  In older style
    sample sheets, this information is combined (usually in a column
    named 'SentrixPosition') and this should be specified as
    \code{list(barcode=NULL, position="SentrixPosition")}}
  \item{highDensity}{logical (used when \code{sampleSheet} is
    specified). If \code{TRUE}, array extensions '\_A', '\_B' in
    sampleSheet are replaced with 'R01C01', 'R01C02' etc.}
  \item{sep}{character string specifying separator used in .idat file
    names.}
  \item{fileExt}{list containing elements 'Green' and 'Red' which
    specify the .idat file extension for the Cy3 and Cy5 channels.}
  \item{XY}{\code{NChannelSet} containing X and Y intensities.}
  \item{anno}{data.frame containing SNP annotation information from 
    manifest and additional columns 'isSnp', 'position', 'chromosome' 
    and 'featureNames'. For use when \code{cdfName}='nopackage'}
  \item{genome}{character string specifying which genome is used in annotation}
  \item{call.method}{character string specifying the genotype calling algorithm to use ('crlmm' or 'krlmm').}
  \item{trueCalls}{matrix specifying known Genotype calls(can contain some NAs) for a subset of samples and features (1 - AA, 2 - AB, 3 - BB).}
  \item{cdfName}{annotation package (see also \code{validCdfNames}) or 'nopackage' when combined with 'krlmm', an \code{anno} data.frame and \code{genome}.}
  \item{copynumber}{ 'logical.' Whether to store copy number intensities with SNP output.}
  \item{batch}{ character vector indicating the batch variable. Must be
	the same length as the number of samples. See details.}
  \item{saveDate}{'logical'.  Should the dates from each .idat be saved
    with sample information?}
  \item{stripNorm}{'logical'.  Should the data be strip-level normalized?}
  \item{useTarget}{'logical' (only used when \code{stripNorm=TRUE}).
    Should the reference HapMap intensities be used in strip-level normalization?}
  \item{quantile.method}{character string specifying the quantile normalization method to use ('within' or 'between' channels).}
  \item{nopackage.norm}{character string specifying normalization to be used when \code{cdfName}='nopackage'.
    Options are 'none', 'quantile' (within channel, between array) and 'loess'.}
  \item{mixtureSampleSize}{ Sample size to be use when fitting the mixture model.}
  \item{fitMixture}{ 'logical.' Whether to fit per-array mixture model.}
  \item{eps}{   Stop criteria.}
  \item{verbose}{  'logical.'  Whether to print descriptive messages during processing.}
  \item{seed}{ Seed to be used when sampling. Useful for reproducibility}
  \item{sns}{The sample identifiers.  If missing, the default sample names are \code{basename(filenames)}}
  \item{probs}{'numeric' vector with priors for AA, AB and BB.}
  \item{DF}{'integer' with number of degrees of freedom to use with t-distribution.}
  \item{SNRMin}{'numeric' scalar defining the minimum SNR used to filter
  out samples.}
  \item{recallMin}{Minimum number of samples for recalibration. }
  \item{recallRegMin}{Minimum number of SNP's for regression.}
  \item{gender}{  integer vector (  male = 1, female = 2 ) or missing,
  with same length as filenames.  If missing, the gender is predicted.}
  \item{returnParams}{'logical'. Return recalibrated parameters from crlmm.}
  \item{badSNP}{'numeric'. Threshold to flag as bad SNP (affects batchQC)}
}

\details{

	\code{genotype.Illumina} (or equivalently \code{crlmmIllumina}) 
        is a wrapper of the \code{crlmm} function for genotyping.  
        Differences include (1) that the copy number probes (if present) 
        are also quantile-normalized and (2) the class of object returned 
        by this function, \code{CNSet}, is needed for subsequent copy number 
        estimation.  Note that the batch variable (a character string) has 
        no effect on the normalization or genotyping steps. Rather, \code{batch} 
        is required in order to initialize a \code{CNSet} container with the 
        appropriate dimensions.

        The new 'krlmm' option is available for certain chip types. Optional 
	argument \code{trueCalls} matrix contains known Genotype calls 
	(1 - AA, 2 - AB, 3 - BB) for a subset of samples and features. This 
	will used to compute KRLMM coefficients by calling \code{vglm} function 
	from \code{VGAM} package.

	The 'krlmm' method makes use of functions provided in \code{parallel} 
	package to speed up the process. It by default initialises up to 
	8 clusters. This is configurable by setting up an option named 
	"krlmm.cores", e.g. options("krlmm.cores" = 16). 

        In general, a chip specific annotation package is required to use the 
        \code{genotype.Illumina} function. If this is not available (newer chip 
        types or custom chips often don't have a chip-specific package available
        on Bioconductor), consider using \code{cdfName}='nopackage' and specifying 
        \code{anno} and \code{genome}, which runs 'krlmm' on the samples available.
        Here \code{anno} is a data.frame read in from the relevant chip-specific
        manifest, which must have additional columns 'isSnp' which is a logical that
        indicates whether a probe is polymorphic or not, 'position', 'chromosome' and
        'featureNames' that give the location on the chromosome and SNP name.
      }

\value{	A \code{SnpSuperSet} instance.}
\references{
  Ritchie ME, Carvalho BS, Hetrick KN, Tavar\'{e} S, Irizarry RA.
  R/Bioconductor software for Illumina's Infinium whole-genome
  genotyping BeadChips. Bioinformatics. 2009 Oct 1;25(19):2621-3.

  Liu R, Dai Z, Yeager M, Irizarry RA1, Ritchie ME.
  KRLMM: an adaptive genotype calling method for common and low frequency variants.
  BMC Bioinformatics. 2014 May 23;15:158.

  Carvalho B, Bengtsson H, Speed TP, Irizarry RA. Exploration,
  normalization, and genotype calls of high-density oligonucleotide SNP
  array data. Biostatistics. 2007 Apr;8(2):485-99. Epub 2006 Dec
  22. PMID: 17189563.

  Carvalho BS, Louis TA, Irizarry RA.
  Quantifying uncertainty in genotype calls.
  Bioinformatics. 2010 Jan 15;26(2):242-9.
}

\author{Matt Ritchie, Cynthia Liu, Zhiyin Dai}

\seealso{
	\code{\link[oligoClasses]{ocSamples}},
	\code{\link[oligoClasses]{ldOpts}}
}
\examples{
\dontrun{
	# example for 'crlmm' option
	library(ff)
	library(crlmm)
	## to enable paralellization, set to TRUE
	if(FALSE){
		library(snow)
		library(doSNOW)
		## with 10 workers
		cl <- makeCluster(10, type="SOCK")
		registerDoSNOW(cl)
	}
	## path to idat files
	datadir <- "/thumper/ctsa/snpmicroarray/illumina/IDATS/370k"
	## read in your samplesheet
	samplesheet = read.csv(file.path(datadir, "HumanHap370Duo_Sample_Map.csv"), header=TRUE, as.is=TRUE)
	samplesheet <- samplesheet[-c(28:46,61:75,78:79), ]
	arrayNames <- file.path(datadir, unique(samplesheet[, "SentrixPosition"]))
	arrayInfo <- list(barcode=NULL, position="SentrixPosition")
	cnSet <- genotype.Illumina(sampleSheet=samplesheet,
				   arrayNames=arrayNames,
				   arrayInfoColNames=arrayInfo,
				   cdfName="human370v1c",
				   batch=rep("1", nrow(samplesheet)))

}
\dontrun{
	# example for 'krlmm' option
	library(crlmm)
	library(ff)
	# line below is an optional step for krlmm to initialise 16 workers 
	# options("krlmm.cores" = 16)
	# read in raw X and Y intensities output by GenomeStudio's GenCall genotyping module
	XY = readGenCallOutput(c("HumanOmni2-5_4v1_FinalReport_83TUSCAN.csv","HumanOmni2-5_4v1_FinalReport_88CHB-JPT.csv"),
				cdfName="humanomni25quadv1b",
				verbose=TRUE)
	krlmmResult = genotype.Illumina(XY=XY, 
		      			cdfName=ThiscdfName, 
					call.method="krlmm", 
					verbose=TRUE)

	# example for 'krlmm' option with known genotype call for some SNPs and samples
	library(VGAM)
	hapmapCalls = load("hapmapCalls.rda")
	# hapmapCalls should have rownames and colnames corresponding to XY featureNames and sampleNames
	krlmmResult = genotype.Illumina(XY=XY,
					cdfName=ThiscdfName, 
					call.method="krlmm", 
					trueCalls=hapmapCalls, 
					verbose=TRUE)		
}
}
\keyword{classif}

\name{plotSNPs}
\alias{plotSNPs}
\alias{plotSamples}

\title{
	Make M vs S plot for SNPs or samples.
}

\description{
	These functions plot the M-values (log-ratios) versus S-values (average intensities)
        for given SNP/(s) or sample/(s) or beanplots for M-values from different samples.
}

\usage{
	plotSNPs(cnSet, row=1, offset=0, xlim=c(9,16), ylim=c(-5,5), verbose=FALSE)
	plotSamples(cnSet, col=1, offset=0, xlim=c(9,16), ylim=c(-5,5), verbose=FALSE, sample=100000, seed=1, type="smoothScatter")
}

\arguments{
  \item{cnSet}{An object of class \code{CNSet}}
  \item{row}{scalar/vector of SNP indexes to plot}
  \item{col}{scalar/vector of sample indexes to plot}
  \item{offset}{numeric, offset to add to intensities in \code{cnSet} 
        before log2-transforming to make log-ratios or average log-intensities}
  \item{xlim}{the x limits of the plot}
  \item{ylim}{the y limits of the plot}
  \item{verbose}{'logical.'  Whether to print descriptive messages during processing}
  \item{sample}{integer indicating the number of SNPs to sample for the plot}
  \item{seed}{integer seed for the random number generator to sample the SNPs}
  \item{type}{character vector specifying the type of sample plot (either 'smoothScatter' or 'beanplot')}
}

\details{
	The \code{plotSNPs} and \code{plotSamples} functions plot the M and S
        values derived from the \code{cnSet} object.
      }

\value{
      One or more M vs S plot for \code{plotSNPs} for a given SNP(/s)
      or either a smoothed scatter plot of M vs S or a beanplot of the M-values 
      for a selected sample(/s) for \code{plotSamples}.
}

\author{
      Matt Ritchie and Cynthia Liu
}

\seealso{
	\code{\link{genotype.Illumina}}
}
\examples{
\dontrun{
   crlmmResult <- genotype.Illumina(sampleSheet=samples[1:10,], path=path,
                                        arrayInfoColNames=list(barcode=NULL,
                                        position="SentrixPosition"),
                                        saveDate=TRUE, cdfName="human370v1c")
   par(mfrow=c(2,2))
   plotSamples(crlmmResult, col=1:4)
   plotSNPs(crlmmResult, row=1:4)
}
}
\keyword{hplot}

\name{validCdfNames}
\alias{validCdfNames}

\title{ Supported annotation packages for crlmm genotyping}
\description{
Supported annotation packages for crlmm genotyping
}
\usage{
validCdfNames()
}
\details{
  List of available annotation packages
}
\value{
  character vector
}

\author{
R.Scharpf
}


\examples{
validCdfNames()
}
\keyword{manip}

\name{validCEL}
\alias{celDates}
\alias{validCEL}
\title{
  Reads cel files and return an error if a file is not read
}
\description{
  Reads cel files and return an error if a file is not read
}
\usage{
validCEL(celfiles)
celDates(celfiles)
}
\arguments{
  \item{celfiles}{
    vector of cel file names to read
}
}
\value{
  Returns a message that cel files were successfully read, or an error
  if there were problems reading the cel files.
}

\author{
R. Scharpf
}

\seealso{ \cite{\link[affyio]{read.celfile.header}},
  \cite{\link{POSIXt}},
  \code{\link{read.celfile}}
}

\examples{
library(oligoClasses)
if(require(hapmapsnp6)){
  path <- system.file("celFiles", package="hapmapsnp6")
  cels <- list.celfiles(path, full.names=TRUE)
  validCEL(cels)
  celDates(cels)
}
}
\keyword{IO}
